/* addEffect.h
 */
#ifndef _CATEGORY_ADDEFFECT_H
#define _CATEGORY_ADDEFFECT_H

#include "osl/category/categoryEnv.h"
#include "osl/category/categoryTraits.h"
#include "osl/category/ptypePtypeProbTable.h"
#include "osl/misc/carray.h"
#include "osl/container/moveVector.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/container/moveStack.h"
#include "osl/effect_util/effectUtil.h"
#include <utility>

namespace osl
{
  namespace category
  {
    /**
     * ŨζĤ
     */
    struct AddEffect : public TableProbabilityCategory
    {
      static Interpolator<PtypePtype2ProbTable> probTable;
      static const char *getName() { return "AddEffect"; }
      static std::pair<bool,size_t> 
      getIndexIf(const CategoryEnv& env, Move move, int pieceValue)
      {
	const Player opponent = alt(env.state->getTurn());
	CArray<Piece,2> highest; 
	EffectUtil::mostValuableEffectOfMoveNotKing
	  (opponent, *env.state, move, highest);
	if (highest[0] != Piece::EMPTY())
	{
	  CArray<bool,2> supported;
	  if(highest[0].position().isPieceStand())
	    supported[0]=false;
	  else
	    supported[0] = 
	      env.state->hasEffectBy(opponent, highest[0].position());
	  if(highest[1].position().isPieceStand())
	    supported[1]=false;
	  else
	    supported[1] = 
	      env.state->hasEffectBy(opponent, highest[1].position());
	  const size_t index = 
	    probTable.getIndex(highest[0].ptype(),  highest[1].ptype(),
			       supported, move.ptype(), pieceValue);
	    return std::make_pair(true,index);
	}
	return std::make_pair(false,0u);
      }

    };

    /**
     * ľưŨζĤ
     *
     * Ǥ϶»ʤΤΤߤоݤˤ (»ޤ)
     * (ϸ AddEffect )
     * 
     * FIXME: isInCategory Ǥ hasEffectFromTo ȤäƤ뤿
     * generate  ̵
     */
    struct AddEffectToLastMove : public TableProbabilityCategory
    {
      static Interpolator<PtypePtypeProbTable> probTable;
      static const char *getName();
      static const int minProb() { return 381; }
      static const int pieceValueLimit;
      static std::pair<bool,size_t> 
      getIndexIf(const CategoryEnv& env, Move move, int pieceValue)
      {
	const Move last_move = env.history->lastMove();
	if ((! last_move.isNormal()) || (pieceValue < pieceValueLimit))
	  return std::make_pair(false,0u);
	if (env.state->hasEffectFromTo(move.ptypeO(),move.to(),
				       env.history->lastMove().to()))
	{
	  const Ptype target = last_move.ptype();
	  const bool supported 
	    = env.state->hasEffectBy(alt(last_move.player()),last_move.to());
	  const size_t index 
	    = probTable.getIndex(target, supported,move.ptype(), pieceValue);
	  return std::make_pair(true,index);
	}
	return std::make_pair(false,0u);
      }
      static void generate(const CategoryEnv& env, MoveLogProbVector& out);
      static void generate(const CategoryEnv& env, MoveLogProbVector& out,
			   int *indices);
    };

  } // namespace category
} // namespace osl

#endif /* _CATEGORY_ADDEFFECT_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
