/* probOfEachCategory.h
 */
#ifndef CATEGORY_PROBOFEACHCATEGORY_H
#define CATEGORY_PROBOFEACHCATEGORY_H

#include "osl/category/categoryEnv.h"
#include "osl/category/categoryList.h"
#include "osl/category/categoryTraits.h"
#include "osl/container/moveLogProbVector.h"
#include <iosfwd>
namespace osl
{
  namespace category
  {
    template <class T> class Classify;
    template <class T> class CategoryUtil;
    namespace analyzer
    {
      /**
       * ؼФƳƥƥ꤬ɤγΨդĴ٤
       */
      class ProbOfEachCategory
      {
      public:
	struct CallBack
	{
	  virtual ~CallBack() {}
	  virtual void found(const char *name, int prob){}
	  virtual void notFound(const char *name) {}
	};
      private:
	const CategoryEnv& env;
	Move target;
	CallBack *function;
	int classify_piece_value;
      public:
	ProbOfEachCategory(const CategoryEnv& e, Move t, CallBack *f);
	template <typename Head, typename Tail>
	void analyze(CategoryList<Head, Tail>);
      private:
	/**
	 * @return match
	 */
	template <typename Category>
	bool analyze(Category)
	{
	  if (MinProb<Category>::minProb() > env.limit)
	    return false;
	  MoveLogProbVector moves;
	  CategoryUtil<Category>::generate(env,moves);
	  for (size_t i=0; i<moves.size(); ++i)
	  {
	    if (moves[i].getMove() == target)
	    {
	      function->found(Category::getName(), moves[i].getLogProb());
	      return true;
	    }
	  }
	  function->notFound(Category::getName());
	  return false;
	}
	bool analyze(NullCategory);
	template <typename SubCategories>
	bool analyze(Classify<SubCategories>)
	{
	  if (MinProb<Classify<SubCategories> >::minProb() > env.limit)
	    return false;
	  analyzeClassify(SubCategories());
	  return true;
	}
	template <typename Head, typename Tail>
	void analyzeClassify(CategoryList<Head,Tail>);
	void analyzeClassify(NullCategory);
      public:
	struct Reporter : public ProbOfEachCategory::CallBack
	{
	  std::ostream& os;
	  bool verbose;
	  explicit Reporter(std::ostream&, bool verbose=false);
	  void found(const char *name, int prob);
	  void notFound(const char *name);
	};
      };
    } // namespace analyzer
  } // namespace category
} // namespace osl

template <typename Head, typename Tail>
void osl::category::analyzer::
ProbOfEachCategory::analyze(CategoryList<Head, Tail>)
{
  const bool match = analyze(Head());
  if (Head::IsKingEscape && match)
    return;		// λƨ
  analyze(Tail());
}

template <typename Head, typename Tail>
void osl::category::analyzer::
ProbOfEachCategory::analyzeClassify(CategoryList<Head,Tail>)
{
  const std::pair<bool,int> r = 
    CategoryUtil<Head>::getProbIf(env, target, classify_piece_value);
  if (r.first)
    function->found(Head::getName(), r.second);
  else
    function->notFound(Head::getName());
  return analyzeClassify(Tail());
}

#endif /* CATEGORY_PROBOFEACHCATEGORY_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
