/* categoryListUtil.h
 */
#ifndef _CATEGORYLISTUTIL_H
#define _CATEGORYLISTUTIL_H
#include "osl/category/categoryConcept.h"
#include "osl/category/categoryList.h"
#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/category/categoryFlags.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/effect_util/effectUtil.h"

namespace osl
{
  namespace container
  {
    class MoveLogProbVector;
    class MoveLogProbSet;
    class MoveStack;
  }
  namespace category
  {
    using namespace container;
    struct CategoryListUtil
    {
    public:
      /**
       * ʣΥƥ˻ dispatcher
       * @return cut äɤ
       */
      template <typename Head, typename Tail, typename Function>
      static bool forEachCategory(CategoryList<Head,Tail>, 
				  const CategoryEnv& env, Function f);
      /**
       * CategoryList κǸξ
       */
      template <typename Function>
      static bool forEachCategory(NullCategory, const CategoryEnv&, Function)
      {
	return false;
      }
      /**
       * ʣΥƥ˻ dispatcher
       * ȤäɤФƤС
       */
      template <typename Head, typename Tail, typename Function>
      static bool forEachCategory(CategoryList<Head,Tail>, 
				  const CategoryEnv& env, Function f,
				  CategoryFlags& flags, size_t cur=0);      
      template <typename Function>
      static bool forEachCategory(NullCategory, 
				  const CategoryEnv&, Function,
				  CategoryFlags&, size_t)
      {
	return false;
      }

      /**
       * ƤΥƥμ out 
       */
      template <typename Categories, typename MoveSet>
      static void gatherAllMoves(const CategoryEnv& env, MoveSet& out);
    private:
      /**
       * ĤΥƥμ
       * @return Υƥɬפ뤫ɤ
       */
      template <typename Category, typename Function>
      static bool useOneCategory(const CategoryEnv& env, Function f)
      {
	if (MinProb<Category>::minProb() > env.limit)
	  return true;

	MoveLogProbVector moves;
	generate<Category>(env, moves);
      
	const bool try_next = f.template operator()<Category::IsKingEscape>
	  (Category::getName(), moves);

	return try_next;
      }
      /** 
       * Generator Ȥä out ˻ؼ롣
       */
      template <class Generator>
      static void generate(const CategoryEnv&, MoveLogProbVector& out);
    };

    template <typename Head, typename Tail, typename Function>
    bool CategoryListUtil::forEachCategory(CategoryList<Head,Tail>, 
					   const CategoryEnv& env, Function f)
    {
      const bool try_next = 
	useOneCategory<Head,Function>(env, f);
      if (! try_next)
	return true;
      // ⤷꤬äƤ¾Υƥ꡼ϻƤ⤷礦ʤ
      if (Head::IsKingEscape 
	  && EffectUtil::isKingInCheck(env.state->getTurn(), *env.state))
	return false;
      
      return forEachCategory(Tail(), env, f);
    }

    template <typename Head, typename Tail, typename Function>
    bool CategoryListUtil::forEachCategory(CategoryList<Head,Tail>, 
					   const CategoryEnv& env, Function f,
					   CategoryFlags& flags, size_t cur)
    {
      if (! flags.isSet(cur))
      {
	flags.set(cur);
	const bool try_next = 
	  useOneCategory<Head,Function>(env, f);
	if (! try_next)
	  return true;
      }

      // ⤷꤬äƤ¾Υƥ꡼ϻƤ⤷礦ʤ
      if (Head::IsKingEscape
	  && EffectUtil::isKingInCheck(env.state->getTurn(), *env.state))
	return false;

      return forEachCategory(Tail(), env, f, flags, cur+1);
    }
    
  } // namespace category
  using category::CategoryListUtil;
} // namespace osl

#endif /* _CATEGORYLISTUTIL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
