/* categoryUtil.h
 */
#ifndef _CATEGORYUTIL_H
#define _CATEGORYUTIL_H

#include "osl/category/categoryEnv.h"
#include "osl/category/categoryTraits.h"
#include "osl/category/moveEvaluation.h"

#include "osl/search/simpleHashRecord.h"
#include "osl/effect_util/sendOffPosition.h"

#include "osl/move_classifier/moveAdaptor.h"
#include "osl/move.h"
#include "osl/container/moveVector.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/container/position8.h"
#include "osl/misc/loki.h"
#include <utility>

// ǲȤ
#include "osl/eval/pieceEval.h"

namespace osl
{
  namespace category
  {
    struct CategoryUtilBase
    {
      static void findSendOff(const CategoryEnv& env, Position8& out)
      {
	const Player turn = env.state->getTurn();
	if (env.record)
	{
	  out = env.record->qrecord.sendOffPosition(turn, *env.state);
	}
	else
	{
	  const Position king_position = env.state->getKingPosition(alt(turn));
	  effect_util::SendOffPosition::find(turn, *env.state, king_position, 
					     out);
	}
      }

      // Ŭ
      enum {
	SendOffBonus = eval::PtypeEvalTraits<PAWN>::val,
	CaptureIfCheckmateBonus = eval::PtypeEvalTraits<PAWN>::val*2
      };
    };
    
    template <class Category> 
    struct CategoryUtil : public CategoryUtilBase
    {
    private:
      static void doNotCompile();	// not implemented;
    private:
      static std::pair<bool,size_t>
      getIndexIf(const CategoryEnv& env, Move m, int, Int2Type<FIXED>)
      {
	if (Category::isInCategory(env, m))
	  return std::make_pair(true, 0u);
	return std::make_pair(false,0u);
      }
      static std::pair<bool,size_t> 
      getIndexIf(const CategoryEnv& env, Move m, int pieceValue, 
		 Int2Type<STANDARD_TABLE>)
      {
	if (Category::isInCategory(env, m))
	  return std::make_pair(true, Category::getIndex(env,m,pieceValue));
	return std::make_pair(false, 0u);
      }
      static std::pair<bool,size_t> 
      getIndexIf(const CategoryEnv& env, Move m, int pieceValue, 
		 Int2Type<TABLE>)
      {
	return Category::getIndexIf(env, m, pieceValue);
      }
      static std::pair<bool,size_t> 
      getIndexIf(const CategoryEnv& env, Move m, int pieceValue, 
		 Int2Type<SPECIAL>)
      {
	doNotCompile();
	return std::make_pair(false, 0u);
      }
    public:
      /**
       * ؼ m  Category ˴ޤޤƤ뤫ɤȡޤޤƤ
       * ɽΥǥåƱĴ٤롥
       * ƱĴ٤ΤϰΥƥǤη׻Ψ뤿
       *
       * ʬ෿ƥ˻Ѳǽ
       * @param pieceValue ֤
       */
      static std::pair<bool,size_t> getIndexIf(const CategoryEnv& env, 
					       Move m, int pieceValue)
      {
	return getIndexIf(env, m, pieceValue, Int2Type<Category::probType>());
      }
    private:
      static std::pair<bool,int>
      getProbIf(const CategoryEnv& env, Move m, int pieceValue,
		Int2Type<false>) // FIXED, STANDARD_TABLE, TABLE
      {
	const std::pair<bool,size_t> r = 
	  getIndexIf(env, m, pieceValue);
	if (r.first)
	  return std::make_pair
	    (true, Category::probTable[r.second]);
	return std::make_pair(false,0);
      }
      static std::pair<bool,int> 
      getProbIf(const CategoryEnv& env, Move m, int pieceValue, 
		Int2Type<true>)	// SPECIAL
      {
	return Category::getProbIf(env, m, pieceValue);
      }
    public:
      /**
       * ؼ m  Category ˴ޤޤƤ뤫ɤȡޤޤƤ
       * ΨƱĴ٤롥
       * ƱĴ٤ΤϰΥƥǤη׻Ψ뤿
       *
       * ʬ෿ƥ˻Ѳǽ
       * @param pieceValue ֤
       */
      static std::pair<bool,int> getProbIf(const CategoryEnv& env, 
					   Move m,
					   int pieceValue)
      {
	return getProbIf(env, m, pieceValue, 
			 Int2Type<Category::probType==SPECIAL>());
      }
    private:
      static void generate(const CategoryEnv& env, MoveLogProbVector& out,
			   Int2Type<SPECIAL>)
      {
	Category::generate(env, out);
	out.sortByProbability();
      }
      static void generate(const CategoryEnv& env, MoveLogProbVector& out,
			   Int2Type<TABLE>)
      {
	// ΤȤ getIndex ͳ⼫ʬ餻®
	// see AddEffect, Escape
	Category::generate(env, out);
	out.sortByProbability();
      }
      static void generate(const CategoryEnv& env, MoveLogProbVector& out,
			   Int2Type<STANDARD_TABLE>)
      {
	MoveVector moves;
	Category::generate(env, moves);
	Position8 sendoffs;
	if (Category::AdjustSendOff)
	{
	  findSendOff(env, sendoffs);
	}
	const int limit = env.limit;
	for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	{
	  int diff = MoveEvaluation::evalLight(*env.state, *p);
	  if (Category::AdjustSendOff
	      && sendoffs.isMember(p->to()))
	    diff = std::max(diff, (const int)SendOffBonus);
	  
	  const size_t index = Category::getIndex(env, *p, diff);
	  const int moveProb = Category::probTable[index];
	  assert(p->isValid());
	  if (! Category::IsKingEscape)
	  {
	    if (moveProb <= limit) 
	      out.push_back(*p, moveProb);
	  }
	  else
	  {
	    // ̤ƨϺʤõޤưʤ
	    out.push_back(*p, std::min(moveProb, limit));
	  }
	}
	out.sortByProbability();
      }
      static void generate(const CategoryEnv& env, MoveLogProbVector& out,
			   Int2Type<FIXED>)
      {
	assert(env.limit >= Category::probTable[0]);
	MoveVector moves;
	Category::generate(env, moves);
	for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	{
	  assert(p->isValid());
	  out.push_back(*p, Category::probTable[0]);
	}
	// no need to sort here
      }
    public:
      /**
       * ƥ Category ˻ؼ
       * 
       * ƥ˻Ѳǽ
       */
      static void generate(const CategoryEnv& env, MoveLogProbVector& out)
      {
	generate(env, out, Int2Type<Category::probType>());
      }
    private:
      static void generateWithIndex(const CategoryEnv&, MoveLogProbVector&,
				    int *, Int2Type<SPECIAL>)
      {
	doNotCompile();
      }
      static void generateWithIndex(const CategoryEnv& env, 
				    MoveLogProbVector& out, int *indices,
				    Int2Type<TABLE>)
      {
	Category::generate(env, out, indices);
	out.sortByProbability();
      }
      static void generateWithIndex(const CategoryEnv& env, 
				    MoveLogProbVector& out, int *indices,
				    Int2Type<STANDARD_TABLE>)
      {
	MoveVector moves;
	Category::generate(env, moves);
	for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	{
	  const int diff = MoveEvaluation::evalLight(*env.state, *p);
	  const size_t index = Category::getIndex(env, *p, diff);
	  const int moveProb = Category::probTable[index];
	  if (moveProb <= env.limit) 
	  {
	    *indices++ = index;
	    out.push_back(*p, moveProb);
	  }
	}
	out.sortByProbability();
      }
      static void generateWithIndex(const CategoryEnv& env, 
				    MoveLogProbVector& out, int *indices,
				    Int2Type<FIXED>)
      {
	assert(env.limit >= Category::probTable[0]);
	MoveVector moves;
	Category::generate(env, moves);
	for (MoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
	{
	  out.push_back(*p, Category::probTable[0]);
	  *indices++ = 0;
	}
	// no need to sort here
      }
    public:
      /**
       * ƥ Category ˻ؼindex 
       * Ψ¬
       * 
       * ƥ˻Ѳǽ
       */
      static void generateWithIndex(const CategoryEnv& env, 
				    MoveLogProbVector& out, int *indices)
      {
	generateWithIndex(env, out, indices, Int2Type<Category::probType>());
      }
    };
  } // namespace category
} // namespace osl


#endif /* _CATEGORYUTIL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
