/* gold.h
 */
#ifndef OSL_CATEGARY_GOLD_H
#define OSL_CATEGARY_GOLD_H

#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/container/moveStack.h"
#include "osl/container/moveVector.h"

namespace osl
{
  namespace category
  {
    struct LureGoldDiagonally : public FixedProbabilityCategory
    {
      static FixedValueTable probTable; // 400
      static const char *getName() { return "LureGoldDiagonally"; }
      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (env.limit < 400)
	  return false;
	if (! move.isDrop())
	  return false;
	if (! (move.ptype() == SILVER || move.ptype() == GOLD ||
	       move.ptype() == BISHOP))
	  return false;
	const Position kingPosition = env.state->getKingPosition(alt(env.state->getTurn()));
	const int x_diff = kingPosition.x() - move.to().x();
	if (x_diff < -2 || x_diff > 2 || x_diff == 0)
	  return false;

	const Player player = env.state->getTurn();
	if (!(env.state->hasPieceOnStand(player, GOLD) ||
	      env.state->hasPieceOnStand(player, ROOK)))
	  return false;

	if ((move.ptype() == GOLD &&
	     env.state->countPiecesOnStand(player, GOLD) < 2)
	    || (move.ptype() == ROOK &&
		env.state->countPiecesOnStand(player, ROOK) < 2))
	  return false;

	const Position r = Board_Table.nextPosition(player, move.to(), R);
	if (!r.isOnBoard())
	  return false;
	const Position ru = Board_Table.nextPosition(player,
						     r, U);
	if (!ru.isOnBoard())
	  return false;
	const Piece ru_piece = env.state->getPieceAt(ru);
	if ((Board_Table.nextPosition(player, ru, U) == kingPosition
	     || Board_Table.nextPosition(player, ru, R) == kingPosition)
	    && !ru_piece.isEmpty() && ru_piece.ptype() == GOLD
	    && env.state->hasEffectBy(player, ru))
	  return true;

	const Position l = Board_Table.nextPosition(player, move.to(), L);
	if (!l.isOnBoard())
	  return false;
	const Position lu = Board_Table.nextPosition(player,
						     l, U);
	if (!lu.isOnBoard())
	  return false;
	const Piece lu_piece = env.state->getPieceAt(lu);
	if ((Board_Table.nextPosition(player, lu, U) == kingPosition
	     || Board_Table.nextPosition(player, lu, L) == kingPosition)
	    && !lu_piece.isEmpty() && lu_piece.ptype() == GOLD
	    && env.state->hasEffectBy(player, lu))
	  return true;
	return false;
      }
    };

    /**
     * ˼Τƶ
     */
    struct PrepareForkGold : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;
      static const char *getName() { return "PrepareForkGold"; }

      template <int EffectAllowed>
      static bool dropSuitable(const NumEffectState& state, Player player,
			       Position gold, Position to_drop)
      {
	if (! state.getPieceOnBoard(to_drop).isEmpty())
	  return false;
	return (state.countEffect(alt(player), to_drop) <= EffectAllowed)
	  && (EffectAllowed == 1 // 2ξ϶̤ʤΤǹ¦ɬ
	      || (state.hasEffectBy(player, to_drop)
		  // ϽĤϽ֤ΤﳲϾʤ
		  || state.hasEffectByPtype<ROOK>(player, gold)));
      }

      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	const Position to = move.to();
	const Player player = move.player();
	if (env.state->hasMultipleEffectBy(alt(player), to))
	  return false;
	
	const Position gold = Board_Table.nextPosition(player, to, U);
	if (env.state->getPieceAt(gold).ptype() != GOLD)
	  return false;
	if (env.state->getPieceAt(gold).owner() == player)
	  return false;
	int bishop = env.state->countPiecesOnStand(player, BISHOP);
	int silver = env.state->countPiecesOnStand(player, SILVER);
	if (move.isDrop())
	{
	  if (move.ptype() == BISHOP)
	    --bishop;
	  else if (move.ptype() == SILVER)
	    --silver;
	}
	if (bishop == 0 && silver == 0)
	  return false;
	// KING
	{
	  const Position king = env.state->getKingPosition(alt(player));
	  const Offset o = Board_Table.getShortOffset(Offset32(king, to));
	  const Position center = to + o;
	  if (center + o != king) // o.zero() ξfalse
	    goto test_rook;
	  if (o.dy() == 0)	// horizontal
	  {
	    if (env.state->getPieceOnBoard(center).isEmpty())
	      goto test_rook;
	    const Position to_drop
	      = Board_Table.nextPosition(player, center, U);
	    if (dropSuitable<2>(*env.state, player, gold, to_drop))
	      return true;
	  }
	  else if (o == Board_Table.getOffset(player, UL)
		   || o == Board_Table.getOffset(player, UR))
	  {
	    if (dropSuitable<2>(*env.state, player, gold, center))
	      return true;
	  }
	} 
      test_rook:
	// ROOK
	for (int i = PtypeTraits<ROOK>::indexMin;
	     i < PtypeTraits<ROOK>::indexLimit; ++i)
	{
	  const Piece rook_piece = env.state->getPieceOf(i);
	  if (! rook_piece.isOnBoardByOwner(alt(player)))
	    continue;

	  const Position rook = rook_piece.position();
	  if (rook == to)
	    continue;
	  const Offset o = Board_Table.getShortOffset(Offset32(rook, to));
	  const Position center = to + o;
	  if (center + o != rook) // o.zero() ξfalse
	    continue;
	  if (o.dy() == 0)	// horizontal
	  {
	    if (env.state->getPieceOnBoard(center).isEmpty())
	      continue;
	    const Position to_drop
	      = Board_Table.nextPosition(player, center, U);
	    if (dropSuitable<1>(*env.state, player, gold, to_drop))
	      return true;
	  }
	  else if (o == Board_Table.getOffset(player, UL)
		   || o == Board_Table.getOffset(player, UR))
	  {
	    if (dropSuitable<1>(*env.state, player, gold, center))
	      return true;
	  }
	}
	return false;
      }
    };
  } // namespace category
} // namespace osl

#endif /* OSL_CATEGARY_GOLD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
