/* knight.h
 */
#ifndef OSL_CATEGARY_KNIGHT_H
#define OSL_CATEGARY_KNIGHT_H

#include "osl/category/categoryTraits.h"
#include "osl/category/categoryEnv.h"
#include "osl/container/moveStack.h"
#include "osl/container/moveVector.h"

namespace osl
{
  namespace category
  {
    /**
     * Ϥξ
     */
    struct KnightDoubleAttack : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;	// 300
      static const char *getName() { return "KnightDoubleAttack"; }

      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (move.ptype() != KNIGHT)
	  return false;
	const Player P = move.player();
	const Position to = move.to();
	// TODO: relax the condition?
	if (env.state->hasEffectBy(alt(P), to))
	  return false;
	const Position front = Board_Table.nextPosition(P, to, U);
	const Piece left = 
	  env.state->getPieceAt(Board_Table.nextPosition(P, front, UL));
	const Piece right = 
	  env.state->getPieceAt(Board_Table.nextPosition(P, front, UR));

	return (left.isPiece() && right.isPiece() &&
		left.owner() != P && right.owner() != P &&
		left.ptype() != PAWN && right.ptype() != PAWN);
      }
    };

    /**
     * ˤϹǤ
     * ˲ˤʤΤ
     */
    struct KnightToAttack : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;	// 400
      static const char *getName() { return "KnightToAttack"; }

      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (move.ptype() != KNIGHT || !move.isDrop())
	  return false;
	const Player P = move.player();
	const Position to = move.to();
	// relax condition?
	if (env.state->hasEffectBy(alt(P), to))
	  return false;
	const Piece king = env.state->getKingPiece(alt(P));
	const int y = (P == BLACK ? (king.position().y() + 4) :
		       (king.position().y() - 4));
	const int x_diff = king.position().x() - to.x();

	return (y == to.y() &&
		(x_diff == -2 || x_diff == 2 || x_diff == 0));
      }
    };
    /**
     * Ҥ
     */
    struct JoiningKnights : public FixedProbabilityCategory
    {
      static FixedValueTable probTable;	// 400
      static const char *getName() { return "JoiningKnights"; }

      static bool isInCategory(const CategoryEnv& env, Move move)
      {
	if (move.ptype() != KNIGHT || !move.isDrop())
	  return false;
	const Player P = move.player();
	const Position to = move.to();
	const Piece king = env.state->getKingPiece(alt(P));
	if (!env.state->hasEffectFromTo(move.ptypeO(), to, king.position()))
	  return false;
	if (env.state->hasEffectByPtype<KNIGHT>(P, to))
	  return true;
	return false;
      }
    };
  } // namespace category
} // namespace osl

#endif /* OSL_CATEGARY_KNIGHT_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
