#ifndef _CHECKMATE_SEARHCER_H
#define _CHECKMATE_SEARHCER_H
#include "osl/checkmate/proofDisproof.h"
#include "osl/checkmate/checkHashTable.h"
#include "osl/checkmate/checkHashRecord.h"
#include "osl/state/numEffectState.h"
#include "osl/hash/hashKey.h"
#include "osl/pathEncoding.h"
#include "osl/move.h"
#include <cstddef>

#ifndef CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT
#define CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT 5000000
#endif

namespace osl
{
  namespace checkmate
  {
    class CheckMove;
    class CheckMoveList;
    class LibertyEstimator;
    class PieceCost;
    /**
     * dfpn^+ extension.
     *
     * - extension by Kishimoto
     *   - avoid GHI problem by path-encoding and smulation
     *   - decrease ratio of re-extension of interior nodes
     * - evaluation functions
     *   - more pieces -> less proof number (Kishimoto)
     * @param Table SimpleCheckHashTable Ȥ ArrayCheckHashTable ʤ
     */
    template<class Table=CheckHashTable, 
	     class HEstimator=LibertyEstimator, class CostEstimator=PieceCost>
    class CheckmateSearcher
    {
    public:
      typedef NumEffectState state_t;
      typedef Table table_t;
    private:
      state_t *state;		// acquaintance, delete ʤ
      Table& table;
      HashKey key;
      PathEncoding path;
      int depth;
      /** ߤõǾ񤷤Ρɿ */
      size_t node_count;
      /** ߤõǾǽʥΡɿ */
      size_t search_node_limit;
      /** objectƤ񤷤Ρɿ */
      size_t total_node_count;
      /** objectƤǽʥΡɿ */
      const size_t total_node_limit;
      bool verbose_destructor;
      const Player attacker;
      void clearNodeCount() { total_node_count+=node_count; node_count=0; }
    public:
      /**
       * @param attacker ¦ץ쥤䡼 (֥Ȥϥץ쥤)
       */
      CheckmateSearcher(Player attacker, Table& t,
			size_t limit=CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT);
      ~CheckmateSearcher();
      size_t totalNodeCount() const { return total_node_count; }
      bool verbose() const { return verbose_destructor; }
      void setVerbose(bool verbose=true) { verbose_destructor = verbose; }
      /**
       * stateP̤֤ͤ.
       * stateμ֤PȰפƤȤ
       * @param search_node_limit õnodeܰ
       * Successλbest_moveˤμ
       */
      template <Player P>
      const ProofDisproof
      hasCheckmateMove(state_t& state, 
		       const HashKey& key, const PathEncoding& path, 
		       size_t search_node_limit, Move& best_move);
      const ProofDisproof 
      hasCheckmateMove(state_t& state, const HashKey& key,
		       const PathEncoding& path, size_t limit, Move& best_move)
      {
	if (attacker == BLACK)
	  return hasCheckmateMove<BLACK>(state, key, path, limit, best_move);
	else
	  return hasCheckmateMove<WHITE>(state, key, path, limit, best_move);
      }
      /**
       * statePˤäƵͤǤ̤֤
       * ΤȤǤͤˤбƤʤ
       * ꤬äƤʤˤϸƤФʤ
       * stateμ֤alt(P)ȰפƤȤ
       * statePˤäƲ꤬äƤȤ
       * @param search_node_limit õnodeܰ
       * @param last_move ǤͤȽɬ
       */
      template <Player P>
      const ProofDisproof
      hasEscapeMove(state_t& state, const HashKey&, const PathEncoding& path, 
		    size_t search_node_limit, Move last_move);
      const ProofDisproof
      hasEscapeMove(state_t& state, 
		    const HashKey& key, const PathEncoding& path, 
		    size_t limit, Move last_move)
      {
	if (attacker == BLACK)
	  return hasEscapeMove<BLACK>(state, key, path, limit, last_move);
	else
	  return hasEscapeMove<WHITE>(state, key, path, limit, last_move);
      }
      /**
       * attack¦
       * @param record ΥΡɤ CheckHashRecord Υǡ
       *   0 ʳϤ attack ǥơ֥ҤѤ
       *   0 Ϥ attack ǿ˥쥳ɤä礽Υݥ󥿤
       *   ݥ󥿤λȤʤΤ
       *   ProofDisproof  record->proofDisproof() ˳ǼƵ
       * @param parent ƥΡɤΥݥ󥿡ήåѡ
       */
      template <Player P>
      void attack(unsigned int proofLimit, unsigned int disproofLimit, 
		  CheckHashRecord *parent, CheckHashRecord *record);
      /**
       * ͤߤƨ
       * alt(P)ˤ
       * @param record ΥΡɤ CheckHashRecord Υǡ
       *   0 ʳϤ attack ǥơ֥ҤѤ
       *   0 Ϥ attack ǿ˥쥳ɤä礽Υݥ󥿤
       *   ProofDisproof  record->proofDisproof() ˳ǼƵ
       */
      template <Player P>
      void defense(unsigned int proofLimit, unsigned int disproofLimit,
		   CheckHashRecord *parent, CheckHashRecord *record);

      const table_t& getTable() const { return table; }
    private:
      /**
       * attack ν: ˬ̤ǡήåǡݡʤɤԤ
       * @param record ƤӽФ0, ƤӽФϳݤǡؤΥݥ
       * @return true ʤlimit over
       */
      template <Player P>
      bool setUpAttackNode(CheckHashRecord *record);
      /**
       * defense ν: ˬ̤ǡήåǡݡʤɤԤ
       * @param record ƤӽФ0, ƤӽФϳݤǡؤΥݥ
       * @return true ʤlimit over
       */
      template <Player P>
      bool setUpDefenseNode(CheckHashRecord *record);

      bool exceedNodeCount(unsigned int futureCost) const;
      static bool exceedRootTolerance(unsigned int proofNumber, unsigned int disproofNumber,
				      unsigned int continuousNoExpandLoop);
    };
  } // namespace checkmate
  using checkmate::CheckmateSearcher;
} // namespace osl
#endif // _CHECKMATE_SEARHCER_H
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
