/* dualCheckmateSearcher.h
 */
#ifndef _DUALCHECKMATESEARCHER_H
#define _DUALCHECKMATESEARCHER_H

#include "osl/checkmate/checkmateSearcher.h"
#include "osl/checkmate/checkHashTable.h"
#include "osl/checkmate/oraclePool.h"
#include "osl/checkmate/oraclePoolLastMove.h"
#include "osl/checkmate/oracleProver.h"
#include "osl/checkmate/oracleProverLight.h"
#include "osl/checkmate/oracleDisprover.h"
#include "osl/checkmate/oracleAges.h"

#ifdef CHECKMATE_DEBUG
#  include "osl/checkmate/analyzer/checkTableAnalyzer.h"
#endif

#include "osl/misc/carray.h"
#include <cassert>
/** ORACLE ǽ˻ */
#define USE_ORACLE

namespace osl
{
  class RepetitionCounter;
  namespace container
  {
    class MoveStack;
  } // namespace container
  using container::MoveStack;
  namespace checkmate
  {
    inline size_t limitToCheckCount(int limit){
      assert(limit >= 0);
      extern CArray<size_t,32> limitToCheckCountTable;
      return limitToCheckCountTable[limit>>7];
    }

    /**
     * ξ Player εͤߤ򰷤;
     */
    template <class Table=CheckHashTable, class HEstimator=LibertyEstimator, class CostEstimator=PieceCost>
    class DualCheckmateSearcher
    {
    public:
      typedef Table table_t;
      typedef CheckmateSearcher<table_t, HEstimator, CostEstimator> checkmate_t;
      typedef OracleProver<table_t> prover_t;
      typedef OracleDisprover<table_t> disprover_t;
      struct CheckmatorWithOracle
      {
	table_t table;
	checkmate_t searcher;
	prover_t prover;
	disprover_t disprover;
	OraclePool oracles;
	OraclePoolLastMove oracles_last_move;
	OraclePool oracles_after_attack;
	CheckmatorWithOracle(Player attacker, size_t total_nodecount);
	void showOracleStatus() const;
      };
    private:
      CheckmatorWithOracle black, white;
      CheckmatorWithOracle& get(Player attacker)
      {
	return (attacker == BLACK) ? black : white;
      }
      const CheckmatorWithOracle& get(Player attacker) const
      {
	return (attacker == BLACK) ? black : white;
      }
      static const int disproof_oracle_record_limit = 500;
      // statistics
      size_t simulation_node_count;
      unsigned int proof_by_oracle, unknown_by_oracle, disproof_by_oracle;
      unsigned int proof_by_search, unknown_by_search, disproof_by_search;
    public:
      /**
       * @param total_node_count instanceͳɤΥΡɿ.
       * isWinningState ʤɤŬڤ node_limit ͿƤ̾
       * 礭ͤɤ.
       */
      explicit DualCheckmateSearcher(size_t total_node_count=CHECKMATE_DEFAULT_TOTAL_NODE_LIMIT);
      virtual ~DualCheckmateSearcher();
      void setVerbose(bool verbose=true) { 
	get(BLACK).searcher.setVerbose(verbose);
	get(WHITE).searcher.setVerbose(verbose);
      }
    private:
      template <Player P>
      bool isWinningStateByOracle(bool memorize, NumEffectState& state, 
				  const HashKey& key, const PathEncoding& path,
				  Move& best_move, ProofOracleAttack<P> oracle)
      {
	if (! oracle.isValid())
	  return false;

	if (memorize)
	{
	  if (get(P).prover.proofWin(state, key, path, oracle, best_move))
	  {
	    ++proof_by_oracle;
	    assert(best_move.isValid());
	    return true;
	  }
	}
	else 
	{
	  OracleProverLight<P> prover;
	  const bool result = prover.proofWin(state, oracle, best_move);
	  simulation_node_count += prover.nodeCount();
	  if (result)
	  {
	    ++proof_by_oracle;
	    assert(best_move.isValid());
	    return true;
	  }
	}
	++unknown_by_oracle;
	return false;
      }
      template <Player P>
      bool isNotWinningStateByOracle(NumEffectState& state, 
				     const HashKey& key, const PathEncoding& path,
				     const DisproofOracleAttack<P>& oracle)
      {
	if (oracle.isValid())
	{
	  if (get(P).disprover.proofNoCheckmate(state, key, path, oracle))
	  {
	    ++disproof_by_oracle;
	    return true;
	  }
	  ++unknown_by_oracle;
	  return false;
	}
	return false;
      }
    public:
      /**
       * @return ͤɤ
       */
      template <Player P>
      bool isWinningStateByOracle(NumEffectState& state, 
				  const HashKey& key, const PathEncoding& path,
				  Move& best_move, unsigned short& oracle_age,
				  bool memorize=false)
      {
	assert(state.getTurn() == P);
	const PieceStand black_stand = key.blackStand();
	while (true)
	{
#ifndef NDEBUG
	  const unsigned int old_age = oracle_age;
#endif
	  const CheckHashRecord *guide 
	    = get(P).oracles.findProofOracle(state, black_stand, oracle_age);
	  if (! guide)
	    break;
	  assert(old_age < oracle_age);
	  ProofOracleAttack<P> oracle(guide);
	  if (isWinningStateByOracle(memorize, state, key, path, best_move, oracle))
	    return true;
	  if (oracle_age == 1)
	    break;
	}
	return false;
      }
      /**
       * @param oracle_age Ʊoracle2ٻʤ
       * @return Եͤɤ
       */
      template <Player P>
      bool isNotWinningStateByOracle(NumEffectState& state, 
				     const HashKey& key, const PathEncoding& path,
				     unsigned short& oracle_age)
      {
	assert(state.getTurn() == P);
	const PieceStand black_stand = key.blackStand();
	const CheckHashRecord *guide 
	  = get(P).oracles.findDisproofOracle(state, black_stand, oracle_age);
	DisproofOracleAttack<P> oracle(guide);
	return isNotWinningStateByOracle(state, key, path, oracle);
      }
      /**
       * @param oracle_age Ʊoracle2ٻʤ
       * @return ͤɤ
       */
      template <Player P>
      bool isWinningStateByOracleLastMove(NumEffectState& state, 
					  const HashKey& key, const PathEncoding& path,
					  Move& best_move, Move last_move,
					  unsigned short& oracle_age,
					  bool memorize=false)
      {
	assert(state.getTurn() == P);
	if ((! last_move.isNormal()) || last_move.isDrop())
	  return false;

	const PieceStand black_stand = key.blackStand();
	const CheckHashRecord *guide 
	  = get(P).oracles_last_move.findOracle(state, last_move, black_stand,
						oracle_age);
	ProofOracleAttack<P> oracle(guide);
	return isWinningStateByOracle(memorize,
				      state, key, path, best_move, oracle);
      }
      bool isWinningStateByOracleLastMove(NumEffectState& state, 
					  const HashKey& key, const PathEncoding& path,
					  Move& best_move, Move last_move,
					  unsigned short& age)
      {
	if (state.getTurn() == BLACK)
	  return isWinningStateByOracleLastMove<BLACK>
	    (state, key, path, best_move, last_move, age);
	else
	  return isWinningStateByOracleLastMove<WHITE>
	    (state, key, path, best_move, last_move, age);
      }
      /**
       * @return ̤ͤ
       * @param oracle_age :˻oracleο:returnޤǤ˻oracleο. (oracleϽˤʤǤΤΤ)
       */
      template <Player P>
      bool isWinningState(int node_limit, NumEffectState& state, 
			  const HashKey& key, const PathEncoding& path,
			  Move& best_move, AttackOracleAges& oracle_age,
			  Move last_move=Move::INVALID())
      {
	assert(state.getTurn() == P);
	assert(path.turn() == P);
#ifdef USE_ORACLE
	const bool memorize = (node_limit > 10);
	if (isWinningStateByOracleLastMove<P>(state, key, path, best_move, last_move,
					      oracle_age.proof_last_move,
					      memorize))
	  return true;
	if (isWinningStateByOracle<P>(state, key, path, best_move,oracle_age.proof,
				      memorize))
	  return true;
	if (node_limit == 0)
	  return false;
	if (node_limit > disproof_oracle_record_limit)
	{
	  if (isNotWinningStateByOracle<P>(state, key, path, oracle_age.disproof))
	    return false;
	}
#endif
	const size_t node_count_before = searcher(P).totalNodeCount();
	const ProofDisproof proof_disproof 
	  = get(P).searcher.template hasCheckmateMove<P>
	  (state, key, path, node_limit, best_move);
	const size_t node_count_after = searcher(P).totalNodeCount();
	if (proof_disproof.isCheckmateSuccess())
	{
	  assert(best_move.isValid());
	  ++proof_by_search;
#ifdef USE_ORACLE
	  const CheckHashRecord *record 
	    = get(P).searcher.getTable().find(key);
	  assert(record);
	  get(P).oracles.addProofOracle(state, record);
	  if (last_move.isNormal() && (! last_move.isDrop()))
	  {
	    get(P).oracles_last_move.addOracle(state, last_move, record);
	  }
#endif
	  return true;
	}
	if (proof_disproof.isCheckmateFail())
	{
	  ++disproof_by_search;
#ifdef USE_ORACLE
	  if (static_cast<int>(node_count_after - node_count_before)
	      > disproof_oracle_record_limit)
	  {
	    const CheckHashRecord *record 
	      = get(P).searcher.getTable().find(key);
	    assert(record);
	    get(P).oracles.addDisproofOracle(state, record);
	  }
#endif
	}
	else
	{
	  ++unknown_by_search;
	}
	return false;
      }
      bool isWinningState(int node_limit, NumEffectState& state, 
			  const HashKey& key, const PathEncoding& path,
			  Move& best_move, AttackOracleAges& oracle_age,
			  Move last_move=Move::INVALID());
      bool isWinningState(int node_limit, NumEffectState& state, 
			  const HashKey& key, const PathEncoding& path,
			  Move& best_move, Move last_move=Move::INVALID())
      {
	AttackOracleAges dummy_age;
	return isWinningState(node_limit, state, key, path, best_move, 
			      dummy_age, last_move);
      }
      /**
       * @return ̤ͤ
       * : ̤˲꤬äƤ뤳
       * @param P  
       * @param oracle_age :˻oracleο:returnޤǤ˻oracleο. (oracleϽˤʤǤΤΤ)
       */
      template <Player P>
      bool isLosingState(int node_limit, NumEffectState& state, 
			 const HashKey& key, const PathEncoding& path,
			 Move last_move=Move::INVALID())
      {
	assert(state.getTurn() == P);
	assert(path.turn() == P);
	const Player attacker = PlayerTraits<P>::opponent;
	OraclePool& oracles = get(attacker).oracles_after_attack;
#ifdef USE_ORACLE
	unsigned short castle_oracle_age=0; // dummy
	const PieceStand black_stand = key.blackStand();
	const CheckHashRecord *guide 
	  = oracles.findProofOracle(state, black_stand, castle_oracle_age);
	ProofOracleDefense<attacker> oracle(guide);
	if (oracle.isValid())
	{
	  if (get(attacker).prover.proofLose(state, key,
					     path, oracle, last_move))
	  {
	    ++proof_by_oracle;
	    return true;
	  }
	  ++unknown_by_oracle;
	}
#endif
	const ProofDisproof proof_disproof = get(attacker).searcher.template 
	  hasEscapeMove<attacker>(state, key, path, 
				  node_limit, last_move);
	if (proof_disproof.isCheckmateSuccess())
	{
	  ++proof_by_search;
#ifdef USE_ORACLE
	  const CheckHashRecord *record
	    = get(attacker).searcher.getTable().find(key);
	  assert(record);
	  oracles.addProofOracle(state, record);
#endif
	  return true;
	}
	++unknown_by_search;
	return false;
      }
      bool isLosingState(int node_limit, NumEffectState& state, 
			 const HashKey&, const PathEncoding& path, 
			 Move last_move=Move::INVALID());

      const checkmate_t& searcher(Player P) const { return get(P).searcher; }
      const table_t& getTable(Player P) const { return get(P).table; }
      table_t& getTable(Player P)  { return get(P).table; }
      size_t totalNodeCount() const
      {
	return (simulation_node_count
		+ searcher(BLACK).totalNodeCount() 
		+ searcher(WHITE).totalNodeCount()
		+ get(BLACK).prover.nodeCount()
		+ get(WHITE).prover.nodeCount()
		+ get(BLACK).disprover.nodeCount()
		+ get(WHITE).disprover.nodeCount());
      }

      void writeRootHistory(const RepetitionCounter& counter,
			    const MoveStack& moves,
			    const SimpleState& state, Player attack);
      void undoWriteRootHistory(const RepetitionCounter& counter,
				const MoveStack& moves,
				const SimpleState& state, Player attack);
    };

  } // namespace checkmate
  using checkmate::DualCheckmateSearcher;
} // namespace osl

#endif /* _DUALCHECKMATESEARCHER_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
