/* dualCheckmateSearcher.tcc
 */
#ifndef _DUALCHECKMATESEARCHER_TCC
#define _DUALCHECKMATESEARCHER_TCC

#include "osl/checkmate/dualCheckmateSearcher.h"
#include "osl/checkmate/checkHistoryToTable.h"
#include <iostream>
#include <iomanip>

template <class Table,class HEstimator,class CostEstimator>
osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
CheckmatorWithOracle::
CheckmatorWithOracle(Player attacker, size_t total_nodecount)
  : table(attacker), 
    searcher(attacker, table, total_nodecount),
    prover(table), disprover(table), oracles(attacker), oracles_last_move(attacker),
    oracles_after_attack(attacker)
{
}

template <class Table,class HEstimator,class CostEstimator>
void osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
CheckmatorWithOracle::showOracleStatus() const
{
  std::cerr << oracles.keySize() << ":"
	    << oracles.totalSize() << " "
	    << oracles_after_attack.keySize() << ":"
	    << oracles_after_attack.totalSize();
}

/* ------------------------------------------------------------------------- */
template <class Table,class HEstimator,class CostEstimator>
osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
DualCheckmateSearcher(size_t total_nodecount)
  : black(BLACK, total_nodecount), white(WHITE, total_nodecount), 
    simulation_node_count(0),
    proof_by_oracle(0), unknown_by_oracle(0), disproof_by_oracle(0),
    proof_by_search(0), unknown_by_search(0), disproof_by_search(0)
{
}
template <class Table,class HEstimator,class CostEstimator>
osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
~DualCheckmateSearcher()
{
  if (get(BLACK).searcher.verbose() && (get(BLACK).oracles.keySize() > 1))
  {
    std::cerr << "oracle p " << std::setw(5) << proof_by_oracle 
	      << " u " << std::setw(5) << unknown_by_oracle
	      << " d " << std::setw(5) << disproof_by_oracle << "\n";
    std::cerr << "search p " << std::setw(5) << proof_by_search
	      << " u " << std::setw(5) << unknown_by_search 
	      << " d " << std::setw(5) << disproof_by_search << "\n";
    std::cerr << "oracles ";
    get(BLACK).showOracleStatus();
    std::cerr << " ";
    get(WHITE).showOracleStatus();
    std::cerr << "\n";
  }
}

template <class Table,class HEstimator,class CostEstimator>
bool osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
isWinningState(int node_limit, NumEffectState& state, 
	       const HashKey& key, const PathEncoding& path,
	       Move& best_move, AttackOracleAges& age, Move last_move)
{
  if (state.getTurn() == BLACK)
    return isWinningState<BLACK>(node_limit, state, key, path, best_move, age, last_move);
  else
    return isWinningState<WHITE>(node_limit, state, key, path, best_move, age, last_move);
}

template <class Table,class HEstimator,class CostEstimator>
bool osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
isLosingState(int node_limit, NumEffectState& state, 
	      const HashKey& key, const PathEncoding& path,
	      Move last_move)
{
  if (state.getTurn() == BLACK)
    return isLosingState<BLACK>(node_limit, state, key, path, last_move);
  else
    return isLosingState<WHITE>(node_limit, state, key, path, last_move);
}

template <class Table,class HEstimator,class CostEstimator>
void osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
writeRootHistory(const RepetitionCounter& counter, const MoveStack& moves,
		 const SimpleState& state, Player attack)
{
  CheckHistoryToTable::write(getTable(attack), counter, moves,
			     state, attack);
}

template <class Table,class HEstimator,class CostEstimator>
void osl::checkmate::DualCheckmateSearcher<Table,HEstimator,CostEstimator>::
undoWriteRootHistory(const RepetitionCounter& counter,
		     const MoveStack&,
		     const SimpleState&, Player attack)
{
#ifdef CHECKMATE_DEBUG
  CheckHistoryToTable::undoWrite(getTable(attack), counter, attack);
#endif
}

#endif /* _DUALCHECKMATESEARCHER_TCC */
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
