/* libertyEstimator.h
 */
#ifndef _LIBERTYESTIMATOR_H
#define _LIBERTYESTIMATOR_H

#include "osl/piece.h"
#include "osl/move.h"
namespace osl
{
  namespace checkmate
  {
    struct PureLibertyEstimator
    {
      /** @return move ζ̤ΤޤΤʤޥ(ͽ) */
      template <class State>
      static void attackH(Player attacker, const State& state, Move move, 
			  unsigned int& proof_number, 
			  unsigned int& disproof_number)
      {
	const Player defender = alt(attacker);
	const Position king_position = state.getKingPosition(defender);
	proof_number = std::max(safePositionsAfterAttack(state, attacker, move, king_position),
			       1);
	disproof_number = 1;
      }
      /** @return move ζ̤ΤޤΤޥ(ͽ) */
      template <class State>
      static void defenseH(Player attacker, const State&, Move move, 
			   unsigned int& proof_number, 
			   unsigned int& disproof_number)
      {
	proof_number = 1;
	disproof_number = 1;
      }

      template <class State>
      static int safePositionsAfterAttack(const State& state, Player,
					  Move attack, Position king_position);
      template <class State>
      static int safePositionsAfterMove(const State& state, Player,
					Move attack, Position king_position,
					Piece attack_piece);
      template <class State>
      static int safePositionsAfterDrop(const State& state, Player,
					Move attack, Position king_position);
    private:
      /**
       * attack_piece ư
       */
      template <class State>
      static void countIfSafePosition(int& count, Player attacker, 
				      const State& state, 
				      Position target, Move attack, 
				      Piece attack_piece)
      {
	if (state.getPieceAt(target).isEmpty()
	    && (! state.hasEffectNotBy(attacker, attack_piece, target))
	    && (! state.hasEffectFromTo(attack.ptypeO(), attack.to(), target)))
	  ++count;
      }
      /**
       * dropξ
       */
      template <class State>
      static void countIfSafePosition(int& count, Player attacker, 
				      const State& state, 
				      Position target, Move attack)
      {
	if (state.getPieceAt(target).isEmpty()
	    && (! state.hasEffectBy(attacker, target))
	    && (! state.hasEffectFromTo(attack.ptypeO(), attack.to(), target)))
	  ++count;
      }
    };
    /**
     * ̤ưܤȤHο.
     * /ΤƤʤɤ¿θ.
     */
    struct LibertyEstimator
    {

      /** ¦ move Ф proof_number  disproof_number ͽۤ */
      template <class State>
      static void attackH(Player attacker, const State&, Move move, 
			  unsigned int& proof_number, unsigned int& disproof_number);
      /** ɸ¦ move Ф proof_number  disproof_number ͽۤ */
      template <class State>
      static void defenseH(Player attacker, const State&, Move move, 
			   unsigned int& proof_number, unsigned int& disproof_number);
    };

  } // namespace checkmate
} // namespace osl

template<typename State> inline
int osl::checkmate::PureLibertyEstimator::
safePositionsAfterAttack(const State& state, Player attacker,
			 Move attack, Position king_position)
{
  assert(attacker == attack.player());
  const Position from = attack.from();
  if (from.isPieceStand())
    return safePositionsAfterDrop(state, attacker, attack, king_position);
  const Piece attack_piece = state.getPieceOnBoard(from);
  return safePositionsAfterMove(state, attacker, attack, king_position,
				attack_piece);
}

template<typename State> inline
int osl::checkmate::PureLibertyEstimator::
safePositionsAfterMove(const State& state, Player attacker,
		       Move attack, Position king_position, Piece attack_piece)
{
  // Ψϸ
  int result = 0;
  Position target = king_position + DirectionPlayerTraits<UL,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<U,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<UR,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<L,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<R,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<DL,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<D,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  target = king_position + DirectionPlayerTraits<DR,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack, attack_piece);
  return result;
}

template<typename State> inline
int osl::checkmate::PureLibertyEstimator::
safePositionsAfterDrop(const State& state, Player attacker,
		       Move attack, Position king_position)
{
  // Ψϸ
  int result = 0;
  Position target = king_position + DirectionPlayerTraits<UL,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<U,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<UR,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<L,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<R,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<DL,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<D,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  target = king_position + DirectionPlayerTraits<DR,BLACK>::offset();
  countIfSafePosition(result, attacker, state, target, attack);
  return result;
}

template<typename State>
void osl::checkmate::LibertyEstimator::
attackH(Player attacker, const State& state, Move move, 
	unsigned int& proof_number, unsigned int& disproof_number)
{
  const Player defender = alt(attacker);
  const Position king_position = state.getKingPosition(defender);
  proof_number = std::max(PureLibertyEstimator::safePositionsAfterAttack
			  (state, attacker, move, king_position),
			 1);
  // Ϥ뤬®ʤ¿
  if (state.hasMultipleEffectBy(defender, move.to()))
    ++proof_number;

  const Position from=move.from();
  const Position to=move.to();
  const int attack_support = state.countEffect(attacker,to);
  const int defense_support = state.countEffect(defender,to);
  if ((attack_support + (from.isPieceStand() ? 1 : 0)) > defense_support)
  {
    /** äƤˤäƤߤ */
    disproof_number=2;
  }
  else if (move.capturePtype()!=PTYPE_EMPTY)
  {
    /**  */
    Ptype capturePtype=unpromote(move.capturePtype());
    if ((capturePtype == SILVER)
	|| (capturePtype == GOLD))
    {
      disproof_number=2;
    }
    else
    {
      proof_number+=1;
      disproof_number=1;
    }
  }
  else
  {
    proof_number+=1;
    disproof_number=1;
  }
}

template<typename State>
void osl::checkmate::LibertyEstimator::
defenseH(Player attacker, const State& state, Move move, 
		 unsigned int& proof_number, unsigned int& disproof_number)
{
  /** captureϲͤ⤤ */
  if (move.capturePtype()!=PTYPE_EMPTY)
  {
    proof_number=2;
    disproof_number=1;
    return;
  }
  if (move.ptype()==KING)
  {
    proof_number=1;
    disproof_number=1;
    return;
  }
  const Position to = move.to();
  if ((state.countEffect(attacker,to) + (move.isDrop() ? 1 : 0))
      <= state.countEffect(alt(attacker),to))
  {
    proof_number=2;
    disproof_number=1;
    return;
  }
  proof_number=1;
  disproof_number = 2;
}

#endif /* _LIBERTYESTIMATOR_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
