/* generalSimpleHashTable.cc
 */
#include "osl/container/generalSimpleHashTable.h"
#include "osl/hash/hashKey.h"
#include "osl/stl/hash_map.h"

template <typename Record>
struct osl::container::GeneralSimpleHashTable<Record>::Table 
  : public hash_map<HashKey, Record>
{
  /*
   * Object larger than 192 bytes will cause malloc at each allocation.
   */
  BOOST_STATIC_ASSERT(sizeof(Record) <= 192);
  /** icc hash_map capacity äƤʤäΤǼʬǻ */
  const unsigned int capacity;
  int numCacheHit, numRecordAfterFull;
  typedef hash_map<HashKey, Record> base_t;
  Table(unsigned int c) 
    : base_t(c), capacity(c), numCacheHit(0), numRecordAfterFull(0)
  {
  }
  ~Table()
  {
  }
  void clear()
  {
    base_t::clear();
    numCacheHit = 0;
    numRecordAfterFull = 0;
  }
  size_t size() const
  {
    return base_t::size();
  }
  Record *find(const HashKey& key)
  {
    typename base_t::iterator pos = base_t::find(key);
    if (pos == base_t::end())
      return 0;

    ++numCacheHit;
    return &pos->second;
  }
    
  Record *allocate(const HashKey& key)
  {
    const size_t currentSize = size();
    if (currentSize < capacity)
    {
      Record *record = &base_t::operator[](key);
      if (currentSize == size())
	++numCacheHit;
      return record;
    }
    // 䤵ʤ褦õ
    Record *result = find(key);
    if ((result == 0) && (numRecordAfterFull++ == 0) && capacity)
    {
      throw TableFull();
    }
    return result;
  }
};

  
template <typename Record>
osl::container::GeneralSimpleHashTable<Record>::
GeneralSimpleHashTable(unsigned int capacity) 
  : table(new Table(capacity))
{
}
  
template <typename Record>
osl::container::GeneralSimpleHashTable<Record>::
~GeneralSimpleHashTable() {
}

template <typename Record>
void osl::container::GeneralSimpleHashTable<Record>::
clear()
{
  table->clear();
}

template <typename Record>
Record * 
osl::container::GeneralSimpleHashTable<Record>::
allocate(const HashKey& key)
{
  return table->allocate(key);
}

template <typename Record>
Record * 
osl::container::GeneralSimpleHashTable<Record>::
find(const HashKey& key)
{
  return table->find(key);
}

template <typename Record>
const Record * 
osl::container::GeneralSimpleHashTable<Record>::
find(const HashKey& key) const
{
  return table->find(key);
}

template <typename Record>
unsigned int osl::container::GeneralSimpleHashTable<Record>::
size() const
{
  return table->size();
}

template <typename Record>
unsigned int osl::container::GeneralSimpleHashTable<Record>::
capacity() const
{
  return table->capacity;
}

template <typename Record>
int osl::container::GeneralSimpleHashTable<Record>::
numCacheHit() const
{
  return table->numCacheHit;
}

template <typename Record>
int osl::container::GeneralSimpleHashTable<Record>::
numRecordAfterFull() const
{
  return table->numRecordAfterFull;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
