/* progressEval.h
 */
#ifndef EVAL_PROGRESSEVAL_H
#define EVAL_PROGRESSEVAL_H

#include "osl/eval/pieceEval.h"
#include "osl/eval/ppair/piecePairPieceEval.h"
#include "osl/eval/endgame/attackDefense.h"
#include "osl/progress/effect5x3.h"
#include "osl/state/numEffectState.h"
namespace osl
{
  namespace eval
  {
    /**
     * פȽפʬȤɾؿ.
     * ɬפǤСפȽפɾؿtemplate parameterˤ롥
     */
    class ProgressEval
    {
    public:
      typedef PiecePairPieceEval opening_eval_t;
      typedef endgame::AttackDefense endgame_eval_t;
      typedef progress::Effect5x3 progress_t;
    private:
      opening_eval_t opening_eval;
      endgame_eval_t endgame_eval;
      progress_t current_progress;
      CArray<int,2> pawn_on_stand;
      CArray<int,2> pawns;
    public:
      ProgressEval(const NumEffectState& state) 
	: opening_eval(state), endgame_eval(state),
	  current_progress(state)
      {
	pawn_on_stand[playerToIndex(BLACK)] = state.countPiecesOnStand(BLACK, PAWN);
	pawn_on_stand[playerToIndex(WHITE)] = state.countPiecesOnStand(WHITE, PAWN);
	pawns[playerToIndex(BLACK)] = 0;
	pawns[playerToIndex(WHITE)] = 0;
	for (int i = PtypeTraits<PAWN>::indexMin;
	     i < PtypeTraits<PAWN>::indexLimit; i++)
	{
	  const Piece pawn = state.getPieceOf(i);
	  if (pawn.owner() == BLACK)
	    pawns[playerToIndex(BLACK)]++;
	  else
	    pawns[playerToIndex(WHITE)]++;
	}
      }

      static bool initialized() 
      { 
	return opening_eval_t::initialized();
      } 
      static bool setUp(const char *filename)
      {
	return opening_eval_t::setUp(filename);
      }
      static bool setUp()
      {
	return opening_eval_t::setUp();
      }

      /** roundup  2^n Ǥ뤳 */
      static const int ROUND_UP = 64;
      static int composeValue(int value_opening, int value_endgame, 
			      Progress16 progress16,
			      Progress16 black,
			      Progress16 white,
			      int black_pawn, int white_pawn,
			      int black_pawn_total = 0, int white_pawn_total = 0)
      {
	assert(progress16.isValid());
	assert(black.isValid());
	assert(white.isValid());
	assert(value_opening % 2 == 0);
	assert(value_endgame % 2 == 0);
	int result = value_opening*(16-progress16.value())
	  + value_endgame*progress16.value();
	const int black_progress = black.value();
	const int white_progress = white.value();
	if (black_progress != white_progress)
	{
	  const int bonus = Ptype_Eval_Table.value(newPtypeO(BLACK, BISHOP));
	  result += (bonus * 2 + Ptype_Eval_Table.value(newPtypeO(BLACK, PAWN))) * (white_progress - black_progress) * progress16.value() / 16;
	  result &= ~1;
	}
	if (black_pawn > 1)
	{
	  result -= Ptype_Eval_Table.value(newPtypeO(BLACK, PAWN)) *
	    (black_pawn - 1) * progress16.value() / 32;
	}
	if (white_pawn > 1)
	{
	  result -= Ptype_Eval_Table.value(newPtypeO(WHITE, PAWN)) *
	    (white_pawn - 1) * progress16.value() / 32;
	}
	if (black_pawn == 0 && black_pawn_total < white_pawn_total)
	  result -= Ptype_Eval_Table.value(newPtypeO(BLACK, PAWN)) / 2;
	if (white_pawn == 0 && white_pawn_total < black_pawn_total)
	  result -= Ptype_Eval_Table.value(newPtypeO(WHITE, PAWN)) / 2;
	result &= ~(ROUND_UP-1);
	assert(result % 2 == 0);
	return result;
      }
      const Progress16 progress16() const { return current_progress.progress16(); }
      int value() const
      {
	assert(pawn_on_stand[playerToIndex(BLACK)] >= 0);
	assert(pawn_on_stand[playerToIndex(WHITE)] >= 0);
	return composeValue(opening_eval.value(), endgame_eval.value(), 
			    progress16(), current_progress.progress16(BLACK),
			    current_progress.progress16(WHITE),
			    pawn_on_stand[playerToIndex(BLACK)], pawn_on_stand[playerToIndex(WHITE)],
			    pawns[playerToIndex(BLACK)], pawns[playerToIndex(WHITE)]);
      }
      static void setValues(const SimpleState&, Progress16 progress16,
			    container::PieceValues&);
      static void setValues(const SimpleState& s, container::PieceValues& o);

      void doSimpleMove(const NumEffectState& state,
			Position from, Position to,  int promote_mask)
      {
	opening_eval.doSimpleMove(state,from,to,promote_mask);
	endgame_eval.doSimpleMove(state,from,to,promote_mask);
	current_progress.doSimpleMove(state,from,to,promote_mask);
      }
      void doDropMove(const NumEffectState& state,Position to,Ptype ptype)
      {
	opening_eval.doDropMove(state,to,ptype);
	endgame_eval.doDropMove(state,to,ptype);
	current_progress.doDropMove(state,to,ptype);
	if (ptype == PAWN)
	{
	  pawn_on_stand[playerToIndex(state.getTurn())]--;
	  assert(pawn_on_stand[playerToIndex(BLACK)] >= 0);
	  assert(pawn_on_stand[playerToIndex(WHITE)] >= 0);
	}
      }
      void doCaptureMove(const NumEffectState& state,
			 Position from, Position to, 
			 Piece victim,int promote_mask)
      {
	opening_eval.doCaptureMove(state,from,to,victim,promote_mask);
	endgame_eval.doCaptureMove(state,from,to,victim,promote_mask);
	current_progress.doCaptureMove(state,from,to,victim,promote_mask);
	if (unpromote(victim.ptype()) == PAWN)
	{
	  pawn_on_stand[playerToIndex(state.getTurn())]++;
	  pawns[playerToIndex(state.getTurn())]++;
	  pawns[playerToIndex(alt(state.getTurn()))]--;
	  assert(pawns[playerToIndex(BLACK)] + pawns[playerToIndex(WHITE)] == 18);
	  assert(pawn_on_stand[playerToIndex(BLACK)] >= 0);
	  assert(pawn_on_stand[playerToIndex(WHITE)] >= 0);
	}
      }
    public:
      static int infty()
      {
	assert(endgame_eval_t::infty() <= opening_eval_t::infty());
	// פ
	return composeValue(opening_eval_t::infty(), 0, Progress16(0),
			    Progress16(0), Progress16(0), 0, 0);
      }

      static int captureValue(PtypeO ptypeO)
      {
	// פ
	return composeValue(opening_eval_t::captureValue(ptypeO), 0, 
			    Progress16(0), Progress16(0), Progress16(0),
			    0, 0);
      }
    };

  } // namespace eval
} // namespace osl

#endif /* EVAL_PROGRESSEVAL_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
