/* searchPlayer.tcc
 */
#ifndef GAMEPLAYING_SEARCHPLAYER_TCC
#define GAMEPLAYING_SEARCHPLAYER_TCC
#include "osl/game_playing/searchPlayer.h"
#include "osl/game_playing/gameState.h"
#include "osl/checkmate/dualCheckmateSearcher.h"
#include "osl/category/bigramPosition.h"
#include "osl/container/moveStack.h"
#include "osl/misc/realTime.h"

#ifdef USE_NTESUKI
#  include "osl/ntesuki/ntesukiSearcher.h"
#  include "osl/ntesuki/ntesukiMoveGenerator.h"
#endif

#include <boost/scoped_ptr.hpp>

struct osl::game_playing::SearchPlayer::CheckmateSearcher
  : public osl::checkmate::DualCheckmateSearcher<>
{
  explicit CheckmateSearcher(size_t node_limit) 
    : DualCheckmateSearcher<>(node_limit)
  {
  }
};

struct osl::game_playing::SearchPlayer::SearchResult
{
  HashKey root;
  Move best_move;
};

#ifdef USE_NTESUKI
struct
osl::game_playing::SearchPlayer::NtesukiThread
{
  explicit NtesukiThread(Move& next_move,
			 volatile bool *thread_finished,
			 volatile bool *stop_flag,
			 NumEffectState state);

  void operator()();

  Move& next_move;
  volatile bool *thread_finished;
  volatile bool *stop_flag;
  NumEffectState state;
};
#endif

template <class Searcher>
osl::game_playing::ComputerPlayer* osl::game_playing::
SearchPlayer::cloneIt(const Searcher& copy) const
{
  return new Searcher(copy);
}

template <class Searcher>
const osl::Move osl::game_playing::
SearchPlayer::searchWithSecondsForThisMove(const GameState& state, Searcher& searcher, 
					   int seconds_for_this_move)
{
  typedef typename Searcher::eval_t eval_t;
  if (! eval_t::initialized())
    throw std::runtime_error("evaluation function not initialized");
  
  const MoveStack& history = state.moveHistory();

#ifdef USE_NTESUKI
  volatile bool ntesuki_thread_finished;
  volatile bool stop_ntesuki;
  Move ntesuki_next_move = Move::INVALID();

  NtesukiThread thread(ntesuki_next_move, &ntesuki_thread_finished,
		       &stop_ntesuki, state.state());
  boost::thread ntesuki_thread(thread);
#endif
  category::BigramHistory::setupTable(state.getInitialState(),
				      state.state(), history);
  searcher.setHistory(history);
#ifdef GLOBAL_BIGRAMKILLERMOVE
  searcher.setBigramKillerMove(*bigram_ptr);
#endif
  int deepening_step_for_this_move = config.deepening_step;
  if (seconds_for_this_move > 0)
  {
    if (seconds_for_this_move < 10)
      deepening_step_for_this_move = std::min(100, config.deepening_step);
  }
  stop_flag = searcher.alarmSwitch();
  const Move best_move = 
    searcher.computeBestMoveIteratively(config.limit, deepening_step_for_this_move,
					config.initial_limit,
					config.node_limit,
					misc::RealTime(seconds_for_this_move));
  stop_flag = AlarmSwitch(0);
#ifdef GLOBAL_BIGRAMKILLERMOVE
  *bigram_ptr = searcher.bigramKillerMove();
#endif
  search_result_ptr->root = state.state().getHash();
  search_result_ptr->best_move = best_move;
#ifdef USE_NTESUKI
  if (ntesuki_thread_finished)
  {
    if (ntesuki_next_move.isNormal())
    {
      return ntesuki_next_move;
    }
    else
    {
      //ntesuki_finished
    }
  }
  else
  {
    //force_finish
    stop_ntesuki = true;
    ntesuki_thread.join();
  }
#endif
  if (copy_history_table)
    history_table = searcher.historyTable();
  return best_move;
}

template <class Searcher>
bool osl::game_playing::
SearchPlayer::isReasonableMove(Searcher& searcher, Move move, int pawn_sacrifice)
{
  return searcher.isReasonableMove(move, pawn_sacrifice);
}

#endif /* GAMEPLAYING_SEARCHPLAYER_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
