/* hashKey.h
 */
#ifndef _HASH_KEY_H
#define _HASH_KEY_H

#include "osl/config.h"
#include "osl/hash/boardKey.h"
#include "osl/piece.h"
#include "osl/move.h"
#include "osl/pieceStand.h"
#include "osl/state/simpleState.h"
#include "osl/misc/carray.h"

namespace osl
{
  namespace hash
  {
#if OSL_WORDSIZE == 64
    typedef HashKey64 HashKeyBase;
    typedef BoardKey64 BoardKey;
    typedef SignatureKey64 SignatureKey;
#elif OSL_WORDSIZE == 32
    typedef HashKey32 HashKeyBase;
    typedef BoardKey32 BoardKey;
    typedef SignatureKey32 SignatureKey;
#endif
    class HashKey : public HashKeyBase
    {
    public:
      HashKey() :HashKeyBase(){}
      const HashKey newHashWithMove(Move move) const;

      PtypeO doSimpleMove(const SimpleState& state,
			  Position from, Position to, int promote_mask);
      void doDropMove(const SimpleState&, Position to,Ptype ptype);
      void doCaptureMove(const SimpleState& state,
			 Position from, Position to, Piece p1, int promote_mask);
      static HashKey calcHash(const SimpleState& state);

    };
    std::ostream& operator<<(std::ostream& os,const HashKey& h);

    class HashGenTable
    {
      CArray2d<HashKey,Position::SIZE,PTYPEO_SIZE> hashKey;
    public:
      HashGenTable();
      void addHashKey(HashKey& hk,Position pos,PtypeO ptypeo) const{
	assert(pos.isValid() && isValidPtypeO(ptypeo));
	hk+=hashKey[pos.index()][ptypeo-PTYPEO_MIN];
      }
      void subHashKey(HashKey& hk,Position pos,PtypeO ptypeo) const{
	assert(pos.isValid() && isValidPtypeO(ptypeo));
	hk-=hashKey[pos.index()][ptypeo-PTYPEO_MIN];
      }
    };
    extern const HashGenTable Hash_Gen_Table;

    inline 
    PtypeO HashKey::doSimpleMove(const SimpleState& state,
				 Position from, Position to, int promote_mask) 
    {
      const PtypeO ptypeo=state.getPieceAt(from).ptypeO();
      const PtypeO newPtypeo=promoteWithMask(ptypeo,promote_mask);
      Hash_Gen_Table.subHashKey(*this,from,ptypeo);
      Hash_Gen_Table.addHashKey(*this,to,newPtypeo);
      return newPtypeo;
    }
    inline
    void HashKey::doDropMove(const SimpleState& state, Position to,Ptype ptype)
    {
      const Player p=state.getTurn();
      Hash_Gen_Table.subHashKey(*this,Position::STAND(),newPtypeO(p,ptype));
      Hash_Gen_Table.addHashKey(*this,to,newPtypeO(p,ptype));
    }
    inline
    void HashKey::doCaptureMove(const SimpleState& state, Position from, 
				Position to, Piece p1, int promote_mask)
    {
      const Ptype capturePtype=p1.ptype();
      const Piece p0=state.getPieceAt(from);
      const Piece newPiece=p0.promoteWithMask(promote_mask);
      const Player p=state.getTurn();
      // ä
      Hash_Gen_Table.subHashKey(*this,to,newPtypeO(alt(p),capturePtype));
      Hash_Gen_Table.addHashKey(*this,Position::STAND(),newPtypeO(p,unpromote(capturePtype)));
      // ư
      Hash_Gen_Table.subHashKey(*this,from,p0.ptypeO());
      Hash_Gen_Table.addHashKey(*this,to,newPiece.ptypeO());
    }

  } // namespace hash
  using hash::HashKey;
  using hash::SignatureKey;

  namespace stl
  {
    template <typename T> struct hash;
    template <>
    struct hash<hash::HashKey>{
      unsigned long operator()(const HashKey& h) const{
	return h.getSignature();
      }
    };
    template<>
    struct hash<hash::SignatureKey>
    {
      unsigned long operator()(const SignatureKey& h) const
      {
	return h.getSignature();
      }
    };
  } // namespace stl

} // namespace osl

#endif /* _HASH_KEY_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
