#ifndef MISC_PERFMON_H
#define MISC_PERFMON_H
#if defined(i386) || defined(__x86_64)
#  define HAVE_TSC 1
#endif
#include <sys/time.h>
#ifndef HAVE_TSC
#  include <sys/resource.h>
#endif
#include <iosfwd>
#include <cassert>
namespace osl
{
  namespace misc
  {
    class PerfMon
    {
#ifdef HAVE_TSC
      unsigned long long start_time;
#else
      rusage start_time;
#endif
    public:
      void restart()
      {
#ifdef HAVE_TSC
	unsigned int ax,dx;
	asm volatile("rdtsc\nmovl %%eax,%0\nmovl %%edx,%1":"=g"(ax),"=g"(dx): :"eax","edx");
	start_time =
	  (static_cast<unsigned long long>(dx)<<32)
	  + static_cast<unsigned long long>(ax);
#else
#ifdef NDEBUG
	getrusage(RUSAGE_SELF, &start_time);
#else
	int ret=getrusage(RUSAGE_SELF, &start_time);
	assert(ret==0);
#endif
#endif
      }
      PerfMon() {
	restart();
      }
      unsigned long long stop(){
#ifdef HAVE_TSC
	unsigned int ax,dx;
	asm volatile("rdtsc\nmovl %%eax,%0\nmovl %%edx,%1":"=g"(ax),"=g"(dx): :"eax","edx");
	const unsigned long long end_time
	  = ((static_cast<unsigned long long>(dx)<<32)
	     + static_cast<unsigned long long>(ax));
	return (end_time - PerfMon::start_time);
#else
	rusage end_time;
#ifdef NDEBUG
	getrusage(RUSAGE_SELF,&end_time);
#else
	int ret=getrusage(RUSAGE_SELF,&end_time);
	assert(ret==0);
#endif
	return (end_time.ru_utime.tv_sec - start_time.ru_utime.tv_sec)*1000000
	  +(end_time.ru_utime.tv_usec - start_time.ru_utime.tv_usec);
#endif
      }
      void stop(const char *message,int loop){
	const unsigned long long cycles=stop();
	PerfMon::message(cycles, message, loop);
      }
      static void message(unsigned long long cycles,
			  const char *message,int loop);
    };
  
    class MeasureTimeLock
    {
      timeval start;
      std::ostream& os;
      char const* message;
    public:
      MeasureTimeLock (std::ostream& os, char const* message)
	: os (os), message (message)
      {
#ifndef NDEBUG
	int ret =
#endif
	  gettimeofday(&start, NULL);
	assert(ret == 0);
      }

      ~MeasureTimeLock();
    };
  } // namespace misc
} // namespace osl


#endif /* MISC_PERFMON_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
