#ifndef _REAL_TIME_H
#define _REAL_TIME_H
#include <string>
#include <sys/time.h>
#include <stdexcept>
#include <cassert>

namespace osl
{
namespace misc
{
  struct NoMoreTime : std::runtime_error
  {
    NoMoreTime() : std::runtime_error("time limit over")
    {
    }
  };
  class RealTime
  {
    timeval start, end;
  public:
    /** 
     * @param timeval Ĥ()
     */
    explicit RealTime (long timeleft)
    {
      gettimeofday (&start, 0);
      end = start;
      end.tv_sec += timeleft;
    }

    explicit RealTime (timeval timeleft)
    {
      gettimeofday (&start, 0);
      end = start;
      end.tv_sec += timeleft.tv_sec;
      end.tv_usec += timeleft.tv_usec;
    }
    bool isNegative() const
    {
      return end.tv_sec - start.tv_sec < 0;
    }

    bool
    timeLeft () const
    {
      timeval now;
      gettimeofday (&now , 0);

      if (now.tv_sec < end.tv_sec)
	{
	  return true;
	}
      else if (now.tv_sec == end.tv_sec)
	{
	  if (now.tv_usec < end.tv_usec)
	    {
	      return true;
	    }
	}
      return false;
    }

    static timeval diff(timeval l, timeval r)
    {
      timeval ret;
      ret.tv_sec = l.tv_sec - r.tv_sec;
      ret.tv_usec = l.tv_usec - r.tv_usec;

      if (ret.tv_usec < 0)
	{
	  --ret.tv_sec;
	  ret.tv_usec += 1000000;
	}

      return ret;
    }
    timeval getTimeLeft () const
    {
      timeval now;
      gettimeofday (&now , 0);
      return diff(end, now);
    }
    int getTimeLeftInSeconds () const
    {
      timeval now;
      gettimeofday (&now , 0);
      return diff(end, now).tv_sec;
    }
    timeval getConsumed () const
    {
      timeval now;
      gettimeofday (&now , 0);
      return diff(now, start);
    }
    double getConsumedInDouble() const
    {
      timeval consumed = getConsumed();
      return consumed.tv_sec + consumed.tv_usec/1000000.0;
    }

    long getTimeLeftApprox () const
    {
      timeval now;

      gettimeofday (&now , 0);
      long ret = end.tv_sec - now.tv_sec;
      return ret;
    }

    const std::string getEnd() const;
  };
} // namespace misc
} // namespace osl


#endif //_REAL_TIME
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
