#ifndef _MOVE_H
#define _MOVE_H
#include "osl/player.h"
#include "osl/ptype.h"
#include "osl/position.h"
#include <iosfwd>

/** move طǤĤޤΤʤ顼ǤȤ */
// #define MOVE_DEBUG
#ifdef MOVE_DEBUG
#  include <cassert>
#  define move_assert(x) assert(x)
#else
#  define move_assert(x) 
#endif

namespace osl
{
  /**
   * ̤Ƥʤ moveɽ .
   * - invalid: isInvalid ʳα黻ϤǤʤ
   * - declare_win: isInvalid ʳα黻ϤǤʤ
   * - pass: from, to, ptype, oldPtype ϤȤ롥player()ϤȤʤ
   * 
   * Piecepromotep򤽤 
   * ̤ 
   * - from     : 8 bit 
   * - to       : 8 bit 
   * - dummy    : 3 bit 
   * - promote? : 1 bit  
   * - capture ptype    : 4 bit 
   * - ptype    : 4 bit - promote moveξpromoteΤ
   * - owner    : signed 
   */
  class Move
  {
    int move;
    explicit Move(int value) : move(value)
    {
    }
    enum { 
      INVALID_VALUE = 1, DECLARE_WIN = 2,
      BLACK_PASS = 0, WHITE_PASS = (-1)<<28, 
    };
  public:
    int intValue() const { return move; }
    /** դιˡκ 
     * ʣƼ뤳Ȥϡ600Ǥ­⤷ʤ
     */
    static const unsigned int MaxUniqMoves=600;
  private:
    void init(Position from, Position to, Ptype ptype,
	      Ptype capture_ptype, bool is_promote, Player player)
    {
      move = (from.uintValue()
	      | (to.uintValue() <<8)
	      | (static_cast<unsigned int>(is_promote)<<19)
	      | (static_cast<unsigned int>(capture_ptype)<<20)
	      | (static_cast<unsigned int>(ptype)<<24)
	      | (static_cast<int>(player)<<28));
    }
  public:
    Move() : move(INVALID_VALUE)
    {
    }
    /** INVALID Ǥ PASS Ǥʤ. isValid()ɤʬʤ*/
    bool isNormal() const { 
      // PASS  INVALID  to()  00
      return move & 0xff00; 
    }
    bool isPass() const { return (move & 0xffff) == 0; }
    static const Move makeDirect(int value) { return Move(value); }
    static const Move PASS(Player P) { return Move(P<<28); }
    static const Move INVALID() { return Move(INVALID_VALUE); }
    static const Move DeclareWin() { return Move(DECLARE_WIN); }
    /**
     * ư
     */
    Move(Position from, Position to, Ptype ptype,
	 Ptype capture_ptype, bool is_promote, Player player)
    {
      move_assert(from.isValid());
      move_assert(to.isOnBoard());
      move_assert(isValid(ptype));
      move_assert(isValid(capture_ptype));
      move_assert(isValid(player));
      init(from, to, ptype, capture_ptype, is_promote, player);
      move_assert(isValid());
    }
    /**
     * drop
     */
    Move(Position to, Ptype ptype, Player player)
    {
      move_assert(to.isOnBoard());
      move_assert(isValid(ptype));
      move_assert(isValid(player));
      init(Position::STAND(), to, ptype, PTYPE_EMPTY, false, player);
      move_assert(isValid());
    }

    const Position from() const 
    {
      assert(! isInvalid());
      move_assert(isValidOrPass());
      const Position result = Position::makeDirect(move & 0xff);
      return result;
    }
    const Position to() const {
      assert(! isInvalid());
      move_assert(isValidOrPass());
      const Position result = Position::makeDirect((move >> 8) & 0xff);
      return result;
    }
    int promoteMask() const {
      assert(isNormal());
      return (static_cast<int>(move)&(1<<19));
    }
    static int promoteMask(bool is_promote) {
      return is_promote<<19;
    }
    bool isPromote() const { assert(isNormal()); return (move & (1<<19)); }
    bool isDrop() const { assert(isNormal()); return from().isPieceStand(); }
      
    Ptype ptype() const {
      assert(! isInvalid());
      move_assert(isValidOrPass());
      const Ptype result = static_cast<Ptype>((move >> 24) & 0xf);
      return result;
    }
    /** ưPtype, i.e., ää */
    PtypeO ptypeO() const {
      assert(! isInvalid());
      const PtypeO result = static_cast<PtypeO>(move >> 24);
      return result;
    }
    /** ưPtypeO, i.e., ä */
    PtypeO oldPtypeO() const {
      assert(! isInvalid());
      const PtypeO result = static_cast<PtypeO>((move>>24)+((move >> (19-3))&8));
      return result;
    }
    /** ưPtype, i.e., ä */
    Ptype oldPtype() const { 
      assert(! isInvalid());
      move_assert(isValidOrPass());
      const PtypeO old_ptypeo = static_cast<PtypeO>((move>>24)+((move >> (19-3))&8));
      return getPtype(old_ptypeo); 
    }
    Ptype capturePtype() const {
      assert(isNormal());
      const Ptype result = static_cast<Ptype>((move>>20)&0xf);
      return result;
    }

    Player player() const {
      assert(! isInvalid());
      const Player result = static_cast<Player>(move>>28);
      return result;
    }

    const Move noPromote() const {
      assert(isNormal());
      move_assert(move.isPromote() && isPromoted(move.ptype()));
      return Move(move^((1<<19)|(1<<27)));
    }
    void setCapturePtype(Ptype capture_ptype)
    {
      assert(isNormal());
      unsigned int result = static_cast<unsigned int>(move);
      result &= ~(15 << 20);
      result += static_cast<unsigned int>(capture_ptype) << 20;
      move = result;
    }

    bool isValid() const;
    /** state  apply ǽǤʤtrue */
    bool isInvalid() const { 
      return static_cast<unsigned int>(move-1) < DECLARE_WIN; 
    }
    bool isValidOrPass() const { return isPass() || isValid(); }
  };

  inline bool operator<(Move lhs, Move rhs)
  {
    return lhs.intValue() < rhs.intValue();
  }
  inline bool operator==(Move lhs, Move rhs)
  {
    return lhs.intValue() == rhs.intValue();
  }
  inline bool operator!=(Move lhs, Move rhs)
  {
    return ! (lhs == rhs);
  }

  std::ostream& operator<<(std::ostream& os, Move move);
}
#endif /* _MOVE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
