#ifndef _GENERATE_DROP_MOVES_H
#define _GENERATE_DROP_MOVES_H

#include "osl/player.h"
#include "osl/ptype.h"
#include "osl/position.h"
#include "osl/direction.h"
#include "osl/piece.h"
#include "osl/move_action/concept.h"
#include "osl/container/bitXmask.h"
#include <boost/static_assert.hpp>

namespace osl
{
  namespace move_generator
  {
    /**
     * Ǥļ
     */
    template<class State,class Action>
    class Drop
    {
      BOOST_CLASS_REQUIRE(Action,osl::move_action,Concept);
      BOOST_STATIC_ASSERT(sizeof(typename State::simple_board_t));
    
    private:
      template<Player P,bool hasPawn,bool hasLance,bool hasKnight,bool hasSilver,bool hasGold,bool hasBishop,bool hasRook>
      static void generateDropMovesX(const State& state,Action& action,int x);

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance,bool hasKnight,bool hasSilver,bool hasGold,bool hasBishop,bool hasRook>
      static void generateDropMoves(const State& state,Action& action,PawnMaskT pawnMask);

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance,bool hasKnight,bool hasSilver,bool hasGold,bool hasBishop>
      static void generateDropMovesBishop(const State& state,Action& action,PawnMaskT pawnMask){
	if(state.template hasPieceOnStand<P,ROOK>())
	  generateDropMoves<P,PawnMaskT,hasPawn,hasLance,hasKnight,hasSilver,hasGold,hasBishop,true>(state,action,pawnMask);
	else
	  generateDropMoves<P,PawnMaskT,hasPawn,hasLance,hasKnight,hasSilver,hasGold,hasBishop,false>(state,action,pawnMask);
      }

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance,bool hasKnight,bool hasSilver,bool hasGold>
      static void generateDropMovesGold(const State& state,Action& action,PawnMaskT pawnMask);

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance,bool hasKnight,bool hasSilver>
      static void generateDropMovesSilver(const State& state,Action& action,PawnMaskT pawnMask){
	if(state.template hasPieceOnStand<P,GOLD>())
	  generateDropMovesGold<P,PawnMaskT,hasPawn,hasLance,hasKnight,hasSilver,true>(state,action,pawnMask);
	else
	  generateDropMovesGold<P,PawnMaskT,hasPawn,hasLance,hasKnight,hasSilver,false>(state,action,pawnMask);
      }

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance,bool hasKnight>
      static void generateDropMovesKnight(const State& state,Action& action,PawnMaskT pawnMask){
	if(state.template hasPieceOnStand<P,SILVER>())
	  generateDropMovesSilver<P,PawnMaskT,hasPawn,hasLance,hasKnight,true>(state,action,pawnMask);
	else
	  generateDropMovesSilver<P,PawnMaskT,hasPawn,hasLance,hasKnight,false>(state,action,pawnMask);
      }

      template<Player P,typename PawnMaskT,bool hasPawn,bool hasLance>
      static void generateDropMovesLance(const State& state,Action& action,PawnMaskT pawnMask){
	if(state.template hasPieceOnStand<P,KNIGHT>())
	  generateDropMovesKnight<P,PawnMaskT,hasPawn,hasLance,true>(state,action,pawnMask);
	else
	  generateDropMovesKnight<P,PawnMaskT,hasPawn,hasLance,false>(state,action,pawnMask);
      }

      template<Player P,typename PawnMaskT,bool hasPawn>
      static void generateDropMovesPawn(const State& state,Action& action,PawnMaskT pawnMask){
	if(state.template hasPieceOnStand<P,LANCE>())
	  generateDropMovesLance<P,PawnMaskT,hasPawn,true>(state,action,pawnMask);
	else
	  generateDropMovesLance<P,PawnMaskT,hasPawn,false>(state,action,pawnMask);
      }
    public:
      template<Player P>
      static void generateMoves(const State& state,Action& action);
      template<Player P>
      static void generateMovesWithPawnMask(const State& state,Action& action,BitXmask pawnMask);      
    };
  } // namespace move_generator
} // namespace osl

#endif /* _GENERATE_DROP_MOVES_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
