#ifndef _COMPACT_BOARD_H
#define _COMPACT_BOARD_H
#include "osl/state/simpleState.h"
#include "osl/stl/vector.h"

namespace osl
{
  namespace record
  {
    class OPiece
    {
    public:
      OPiece(Piece p)
      {
	const Position pos = p.position();
	const int bitPos = pos.isPieceStand() ? 0 : pos.x() << 4 | pos.y();
	value = (static_cast<int>(p.owner()) << 20 |
		 static_cast<int>(p.ptype()) << 16 | bitPos);
      }
      OPiece(int i)
      {
	value = i;
      }
      Position getPosition() const
      {
	if ((value & 0xff) == 0)
	  return Position::STAND();
	else
	  return Position((value >> 4) & 0xf, value & 0xf);
      }
      Ptype getPtype() const
      {
	return static_cast<Ptype>((value >> 16) & 0xf);
      }
      Player getOwner() const
      {
	return static_cast<Player>(value >> 20);
      }
      operator int() const { return value; }
    private:
      int value;
    };

    class CompactBoard
    {
    public:
      CompactBoard() {}
      explicit CompactBoard(const SimpleState& state);
      SimpleState getState();
      friend std::ostream& operator<<(std::ostream& os, const CompactBoard& c);
      friend std::istream& operator>>(std::istream& os, CompactBoard& c);
    private:
      osl::vector<OPiece> pieces;
      Player turn;
    };
  }
}

#endif // _COMPACT_BOARD_H
