/* alphaBeta.h
 */
#ifndef _ALPHA_BETA_H
#define _ALPHA_BETA_H

#include "osl/search/realizationProbability.h"
#include "osl/search/plainNullWindowSearch.h"
#include "osl/search/searchBase.h"
#include "osl/search/searchWindow.h"
#include "osl/search/killerMoveTable.h"
#include "osl/checkmate/dualCheckmateSearcher.h"
#include "osl/eval/evalTraits.h"
#include "osl/misc/realTime.h"
#include "osl/container/moveStack.h"

namespace osl
{
  namespace search
  {
    class SimpleHashRecord;
    class SimpleHashTable;
    /**
     * ¸ΨѤ alpha beta õ.
     *   ƤmoveĴ٤֤ͤؿ
     *   alphaBetaSearchAfterMove:
     *   - null window search (alpha ͤ򹹿뤫)
     *   - alpha beta search (alpha beta õ + õĹ)
     *   
     */
    template<typename Eval, typename MoveGenerator,
	     typename Table=SimpleHashTable, 
	     typename Recorder=CountRecorder, 
	     typename Probabilities=RealizationProbability>
    class AlphaBeta
      : public PlainNullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>
    {
      BOOST_CLASS_REQUIRE(Eval, osl::eval, Concept);
      typedef SearchFramework<Eval,MoveGenerator,Table,Recorder,
			      Probabilities> framework_t;
      typedef PlainNullWindowSearch<Eval,MoveGenerator,Table,Recorder,
				    Probabilities> base_t;
      misc::RealTime timer;
    public:
      typedef HashEffectState state_t;
      typedef typename state_t::effect_state_t effect_state_t;
      typedef Eval eval_t;
    
      typedef MoveGenerator MoveGenerator_t; // ̾ΤĤɤ?
      typedef Probabilities Probabilities_t;
      typedef AlphaBeta<Eval,MoveGenerator,Table,Recorder,Probabilities> 
      alpha_beta_t;
    public:
      explicit AlphaBeta(const HashEffectState& s, 
			 typename framework_t::checkmate_t& checker,
			 Table *t, Recorder& r);
      ~AlphaBeta();

      /** 
       * entrance of alpha beta window search.
       *
       * see http://www.logos.t.u-tokyo.ac.jp/~gekisashi/algorithm/re_search.html.
       * rootDepth, curLimit are initialized here.
       */
      template <Player P>
      int alphaBetaSearch(AlphaBetaWindow window, MoveLogProb& best_move,
			  int limit);
      static const AlphaBetaWindow fullWindow(Player P)
      {
	return AlphaBetaWindow(base_t::winThreshold(alt(P)), 
			       base_t::winThreshold(P));
      }
      int alphaBetaSearch(AlphaBetaWindow window, MoveLogProb& best_move, int limit)
      {
	const Player P = base_t::state().getTurn();
	if (P == BLACK)
	  return alphaBetaSearch<BLACK>(window, best_move, limit);
	else
	  return alphaBetaSearch<WHITE>(window, best_move, limit);
      }
      int alphaBetaSearch(MoveLogProb& best_move, int limit)
      {
	AlphaBetaWindow full_window = fullWindow(base_t::state().getTurn());
	return alphaBetaSearch(full_window, best_move, limit);
      }
    
      /**
       * entrance of alpha beta iterative search.
       */
      Move computeBestMoveIteratively(int limit,
				      const int step=100, 
				      const int initialLimit=600,
				      int nodeLimit=1600000,
				      const misc::RealTime& timer
				      =misc::RealTime(60));

      /** (private) */
      template <Player P>
      bool alphaBetaSearch(const SearchMove& move, AlphaBetaWindow& window, 
			   int& base_value, SearchMove& best_move);
      /** (private) 
       * : this->null_window ˲
       */
      template <Player P>
      int alphaBetaSearchAfterMove(const SearchMove& move, 
				   AlphaBetaWindow window, 
				   SearchMove& best_move);

    }; // class AlphaBeta
  } // namespace search
  using search::AlphaBeta;
} // namespace osl

#endif /* _ALPHA_BETA_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
