#ifndef _MTDF_TCC
#define _MTDF_TCC
#include "osl/search/mtdf.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/category/categoryListUtil.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/record/csa.h"
#include <stdexcept>
#include <iostream>

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
MTDF(const HashEffectState& s, checkmate_t& c, Table& t, Recorder& r) 
  : base_t(r, &t), checkmate_searcher(c),
    memory_tester(s, checkmate_searcher, &t, r),
    timer(60)
{
  setInitialGuess();
  checkmate_searcher.setVerbose(base_t::table->isVerbose());
}
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
~MTDF() 
{
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setInitialGuess(int guess) 
{
  threshold = (guess & (~1)); 
  threshold -= eval::delta(memory_tester.state().getTurn());
  assert(threshold % 2);
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setInitialGuess() 
{ 
  const Eval eval(memory_tester.state());
  setInitialGuess(eval.value());
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
bool osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
isReasonableMove(Move move, int pawn_sacrifice)
{
  return memory_tester.isReasonableMove(move, pawn_sacrifice);
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::misc::AlarmSwitch 
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
alarmSwitch()
{ 
  return memory_tester.alarmSwitch();
}
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::search::BigramKillerMove& 
osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
bigramKillerMove()
{
  return memory_tester.bigramKillerMove();
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
setBigramKillerMove(const BigramKillerMove& killers)
{
  memory_tester.setBigramKillerMove(killers);
}

/* ------------------------------------------------------------------------- */
/*	Methods for the toplevel loop					     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMove(int limit, const RealTime& timer)
{
  const bool use_timer = ! timer.isNegative();
  if (use_timer)
  {
    this->timer = timer;
    memory_tester.setTimeOut(timer.getTimeLeftInSeconds());
  }
  else
  {
    this->timer = RealTime(6000); // INFTY
  }
  Move best_move;
  if (memory_tester.state().getTurn()==BLACK)
    best_move = computeBestMoveOfPlayer<BLACK>(limit);
  else
    best_move = computeBestMoveOfPlayer<WHITE>(limit);
  if (use_timer)
    memory_tester.resetTimeOut();
  return best_move;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveIteratively(int limit, const int step, 
			   const int initial_limit, int node_limit, 
			   const RealTime& timer)
{
  int last_iteration_consumed = 0;
  int last_time_left = timer.getTimeLeftApprox();
  base_t::recorder.resetNodeCount();
  int limit_iterative=initial_limit;
  Move best_move = Move::INVALID();
  try
  {
    while (limit_iterative < limit)
    {
      if (base_t::table->isVerbose())
	std::cerr << "iteration " << limit_iterative 
		  << " " << timer.getTimeLeftApprox() << " left\n"
		  << std::flush;
      best_move = computeBestMove(limit_iterative, timer);
      if (base_t::table->isVerbose())
	std::cerr << "selected " << record::csa::show(best_move) << "\n";
      memory_tester.testTimeOut();
      if (best_move.isNormal())
      {
	bool time_over = false;
	if (! timer.isNegative())
	{
	  const int current_time_left = timer.getTimeLeftInSeconds();
	  last_iteration_consumed = (last_time_left - current_time_left);
	  last_time_left = current_time_left;
	  if (current_time_left < last_iteration_consumed)
	  {
	    time_over = true;
	    if (base_t::table->isVerbose())
	      std::cerr << "time limit " << current_time_left 
			<< " < " << last_iteration_consumed << "\n";
	  }
	}
	// depth  200 䤹8ܤ餤ˤʤȤɤ
	// פ20ܤˤʤ뤳Ȥ⤢
	if (time_over || (base_t::recorder.nodeCount() *8 > node_limit))
	{
	  if (base_t::table->isVerbose())
	    std::cerr << "iteration stop\n";
	  return best_move;
	}
      }
      limit_iterative += step;
    }
    if (base_t::table->isVerbose())
      std::cerr << "final iteration " << limit
		<< " " << timer.getTimeLeftApprox() << " left\n";
    while (true)
    {
      best_move = computeBestMove(limit, timer);
      if (best_move.isNormal())
	break;
      memory_tester.testTimeOut();

      // ۤäƤλ
      if (limit >= 2000)
	break;

      limit += 200;
      if (base_t::table->isVerbose())
	std::cerr << "extend limit to " << limit << " before resign\n";
    }
  }
  catch (std::exception& e)
  {
    std::cerr << "std exception " << e.what() << "\n";
  }
  catch (...)
  {
    std::cerr << "unkown exception\n";
  }
  if (base_t::table->isVerbose())
    std::cerr << "selected " << record::csa::show(best_move) << "\n";
  return best_move;
}

/**
 * ᥤõ.
 * ǽ low/high fail ȽꤷѴؿ˸Ǥ
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
computeBestMoveOfPlayer(int limit)
{
  if (base_t::isWinValue(P, threshold))
  {
    threshold = base_t::windowMax(P) - eval::delta(P);
  }
  assert(threshold % 2);
  base_t::recorder.startSearch(limit);
  if (base_t::table->isVerbose())
    std::cerr << "threshold " << threshold << "\n";

  SimpleHashRecord *record_in_table
    = base_t::table->allocate(memory_tester.currentHash(), limit);
  SimpleHashRecord *record = record_in_table;
  boost::scoped_ptr<SimpleHashRecord> record_if_not_allocated;
  if (! record_in_table)
  {
    record_if_not_allocated.reset(new SimpleHashRecord());
    record = record_if_not_allocated.get();
  }
  assert(record);
  assert(memory_tester.rootRecord() == 0
	 || memory_tester.rootRecord() == record);
  memory_tester.setRootRecord(record);

  // ޤơ֥򸣤
  SearchMove best_move_in_table;
  if (record_in_table)
  {
    best_move_in_table = record->bestMove();
    if (best_move_in_table.getMove().isNormal())
    {
      // ;Ǿ
      int table_value;
      if (record->template hasGreaterLowerBound<P>
	  (limit, base_t::winThreshold(P), table_value))
	return best_move_in_table.getMove();
    }
    if ((! best_move_in_table.getMove().isNormal())
	|| (best_move_in_table.record == 0))
    {
      best_move_in_table = SearchMove(); // invalidate
    }
  }

  // ƤΥƥμ򽸤
  SearchMoveSet& move_set = record->moves();
  while (true)
  {
    const category::CategoryEnv env(&memory_tester.state(), limit, 
				    &memory_tester.history(), record, 0,
				    &memory_tester.historyTable());
    CategoryListUtil::gatherAllMoves<MoveGenerator>(env, move_set);
    if ((move_set.size() > 1) || (limit > 2000))
      break;
    limit += 100;
  }
  if (move_set.size() == 1)
    return move_set.begin()->getMove();

#ifndef DONT_USE_CHECKMATE
  // ;ƤǤߤ root ǤƤǤʤ
  if (record_in_table)
  {
    const int check_limit 
      = ((limit < 500) 
	 ? 800 
	 : ((limit < 700) ? 3800 : 4000));
    const int checkmate_max = checkmate::limitToCheckCount(check_limit);
    const int checkmate_node = record->checkmateNodesLeft(checkmate_max);
    if (checkmate_node > 0)
    {
      PathEncoding path = memory_tester.path();
      const bool my_king_in_check
	= memory_tester.state().hasEffectBy(alt(P),memory_tester.state().template getKingPosition<P>());
      if (my_king_in_check)
      {
	// ꤫鲦꤬äƤ
	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool lose = memory_tester.template isLosingState<P>(checkmate_node);
	base_t::recorder.backFromCheckmateSearch();
	if (lose)
	{
	  base_t::recordLoseByCheckmate(P, record);
	  return Move::INVALID();
	}
      }
      // ͤޤʤСͤޤȤͤ 
      {
	Move checkmate_move;
	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool win = memory_tester.template isWinningState<P>
	  (checkmate_node, checkmate_move, record->qrecord.attack_oracle);
	base_t::recorder.backFromCheckmateSearch();
	if (win)
	{
	  base_t::recordWinByCheckmate(P, record, checkmate_move);
	  return checkmate_move;
	}
      }
      // ͤͤ
      if ((! my_king_in_check)
	  && (! record->threatmate().isThreatmate(P)))
      {
	Move threatmate_move;

	base_t::recorder.gotoCheckmateSearch(memory_tester.state(), checkmate_node);
	const bool threatmate 
	  = memory_tester.template isThreatmateState<P>
	  (checkmate_node, threatmate_move, record->qrecord.threatmate_oracle);
	base_t::recorder.backFromCheckmateSearch();
	if (threatmate)
	{
	  record->threatmate().setThreatmate(P, threatmate_move);
	  if (base_t::table->isVerbose())
	    std::cerr << "root threatmate " << threatmate_move << "\n";
	}
      }
    }
  }
#endif
  /**
   * threshold ۤȱۤʤä
   */
  SearchMoveSorter<P> lower_moves, higher_moves;

  /** 
   * ߤκɾ threshold ˤä upperbound ä
   * lower bound äꤹΤ threshold ѹ 
   */
  int currentVal = base_t::minusInfty(P);
  // ǽκ
  base_t::recorder.newCategory("table",limit);
  if (base_t::validTableMove(memory_tester.state(), best_move_in_table.moveLogProb(), limit))
  {
    SearchMove trying_move = best_move_in_table;
    trying_move.setLogProbAtMost(Probabilities::TableMove);
    int val = base_t::minusInfty(P);
    for (int i=0; i<2; ++i)	// 2ޤbestmove¤γ
    {
      if (i > 0)
      {
	if (base_t::isWinValue(alt(P), val))
	  break;
	threshold = EvalTraits<P>::max(currentVal - eval::delta(P)
				       + Eval::captureValue(newPtypeO(P,PAWN)),
				       base_t::winThreshold(alt(P)));
	currentVal = base_t::minusInfty(P);
      }
      if (testMove<P>(trying_move, val, currentVal, limit))
      {
	if (base_t::isWinValue(alt(P), currentVal))
	  return trying_move.getMove();
	threshold = currentVal + eval::delta(P);
	// record->bestMoveϥɥ쥹Ȥʤ
	SearchMove *stored_move = move_set.find(trying_move.getMove());
	higher_moves.insert(std::make_pair(currentVal, stored_move));
	break;
      }
    }
    if (higher_moves.empty() && (! base_t::isWinValue(alt(P), val)))
    {
      threshold = currentVal - eval::delta(P);
      currentVal = base_t::minusInfty(P);
    }
  }
  // 礤κϿ
  const int qvalue
    = memory_tester.template quiescenceValue<P>(base_t::winThreshold(alt(P)),
						base_t::winThreshold(P));
  if (higher_moves.empty())
  {
    const int safe_threshold 
      = qvalue + Eval::captureValue(newPtypeO(P,PAWN))*2 - eval::delta(P);
    if (EvalTraits<P>::betterThan(threshold, safe_threshold))
    {
      threshold = safe_threshold;
      currentVal = base_t::minusInfty(P);
      threshold = EvalTraits<P>::max(threshold, currentVal+eval::delta(P));
    }
  }
  if (record->qrecord.bestMove().isNormal())
  {
    const MoveLogProb quiescence_move(record->qrecord.bestMove(),
				      Probabilities::KillerMove-1);
    move_set.assignIfBetter(quiescence_move);
  }
  // ٤Ƥμ
  SearchMoveVector all_moves;
  for (SearchMoveSet::iterator p=move_set.begin(); p!=move_set.end(); ++p)
  {
    if (move_classifier::MoveAdaptor<move_classifier::PawnDropCheckmate<P> >
	::isMember(memory_tester.state(), p->getMove()))
      continue; // 

    if (higher_moves.empty() 	// ʣʤ
	|| (p->getMove() != higher_moves.begin()->second->getMove()))
    {
      p->setLogProbAtMost(1000);	// ϳ̵褦
      if (p->getMove().isNormal() && (p->getLogProb() <= limit))
	all_moves.push_back(&*p);
    }
  }
  all_moves.sortByValue<P>(threshold + Eval::captureValue(newPtypeO(P,PAWN))*2);
  examineMoves<P>("uniqued moves", currentVal, limit, all_moves,
		  lower_moves, higher_moves);

  Move best_move;
  if (higher_moves.empty()) // low fail
  {
    best_move 
      = retryByLowerF<P>(record_in_table, limit, lower_moves, higher_moves);
  } 
  else 
  {
    // high fail: loop ƬǤ threshold Ͱʾμʣǧ
    best_move = selectBestMoveByHigherF<P>(record_in_table, limit, 
					   lower_moves, higher_moves);
  }
  base_t::recorder.setCheckmateCount(checkmate_searcher.totalNodeCount());
  const double consumed = timer.getConsumedInDouble();
  base_t::recorder.finishSearch(best_move, consumed,base_t::table->isVerbose());
  return best_move;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *category_name, 
	     int& currentVal, int limit, const SearchMoveVector& moves, 
	     SearchMoveSorter<P>& lower_moves, SearchMoveSorter<P>& higher_moves) 
{
  base_t::recorder.newCategory(category_name, limit);
  for (SearchMoveVector::const_iterator p=moves.begin(); p!=moves.end(); ++p)
  {
    if (base_t::isWinValue(P, currentVal))
    {
      assert(! higher_moves.empty());
      return;
    }
    if ((*p)->getMove().isPass()
	|| ShouldPromoteCut::canIgnoreAndNotDrop((*p)->getMove()))
      continue;

    int val;
    if (testMove<P>(**p, val, currentVal, limit))
    {
      threshold = currentVal + eval::delta(P);
      higher_moves.insert(std::make_pair(currentVal, *p));
      if ((moves.size() > 30) && base_t::table->isVerbose())
      {
	std::cerr << currentVal << " ";
	csaShow(std::cerr, (*p)->getMove());
	std::cerr << "\n";
      }
    }
    if (higher_moves.empty() 
	&& (! base_t::isWinValue(alt(P), val)))
    {
      // low fail ʤΤǤƤ
      lower_moves.insert(std::make_pair(val, *p));
      threshold = val - eval::delta(P); // ɤɤ f 򲼤Ƹǻ
      threshold += Eval::captureValue(newPtypeO(P,PAWN)); // 2礵˲
      threshold = EvalTraits<P>::max(threshold, base_t::winThreshold(alt(P)));
      currentVal = threshold - eval::delta(P);
    }
  }
  assert(higher_moves.isValidAll());
}
  
/**
 * ȥåץ٥ 롼 ( low fail )
 * 
 * threshold 򲼤ʤõthreshold ꤬ۤǤ selectBestMoveByHigherF 
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
retryByLowerF(SimpleHashRecord *record, int limit, 
	      SearchMoveSorter<P>& lower_moves, SearchMoveSorter<P>& higher_moves)
{
  assert(higher_moves.empty());
  assert(lower_moves.isValidAll());
  SearchMoveVector moves;
  while (higher_moves.empty() && (lower_moves.size() > 1))
  {
    base_t::recorder.recordTopLevelLowFail(lower_moves.begin()->second->moveLogProb(),
					   threshold);
    if (base_t::table->isVerbose())
      std::cerr << "low fail " << threshold << "\n";
    const int old_threshold = threshold;
    for (typename SearchMoveSorter<P>::const_iterator p=lower_moves.begin(); p!=lower_moves.end(); ++p)
      moves.push_back(p->second);
    lower_moves.clear();
    // 餱λؤϿʤΤ lower moves Ϥ󸺤äƤ椯ǽ
    int currentVal = base_t::minusInfty(P);
    examineMoves<P>("low fail", currentVal, limit, 
		    moves, lower_moves, higher_moves);
    moves.clear();
    threshold += Eval::captureValue(newPtypeO(P,PAWN))*4; // 8礵˲
    threshold = EvalTraits<P>::max(threshold, base_t::winThreshold(alt(P)));
    if (old_threshold == threshold)
      return Move::INVALID();
  }

  if (! higher_moves.empty())
    return selectBestMoveByHigherF<P>(record, limit, 
				      lower_moves, higher_moves);

  if (lower_moves.size() == 0)
    return Move::INVALID();

  assert(lower_moves.size() == 1);
  const SearchMove *best_move = lower_moves.begin()->second;
  const int val = lower_moves.begin()->first;
  assert(limit >= 0);
  if (record)
    base_t::recordLowerBound(P, record, limit, *best_move, val);

  // thresholdϤΤޤ: 򿶤櫓Τ¸
  return best_move->getMove();
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
const osl::Move osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
selectBestMoveByHigherF(SimpleHashRecord *record,
			int limit, SearchMoveSorter<P>& lower_moves, 
			SearchMoveSorter<P>& higher_moves)
{
  assert(higher_moves.isValidAll());
  assert(lower_moves.isValidAll());
  base_t::recorder.recordTopLevelHighFail(higher_moves.begin()->second->moveLogProb(),
					  threshold);
  // lower_moves ǽΤ򽦤
  {
    assert(! higher_moves.empty());
    int currentVal = higher_moves.begin()->first;
    threshold = currentVal + eval::delta(P);
    SearchMoveVector moves_maybe_beyond_f;
    for (typename SearchMoveSorter<P>::const_iterator p=lower_moves.begin(); 
	 p!=lower_moves.end(); ++p)
    {
      // p->first ˤ ¤äƤ롤 lower_moves Ͼ¤ι⤤¤Ǥ
      if (! eval::betterThan(P, p->first, threshold))
	break;
      moves_maybe_beyond_f.push_back(p->second);
    }
    examineMoves<P>("high fail", currentVal, limit,
		    moves_maybe_beyond_f, lower_moves, higher_moves);
  }
  const int triage_value = Eval::captureValue(newPtypeO(alt(P),PAWN))/2;
  triageMovesSpeculatively<P>(higher_moves, limit, triage_value);
  triageMovesSpeculatively<P>(higher_moves, limit, triage_value/8);

  while (higher_moves.size() > 1)
  {
    if (base_t::table->isVerbose())
    {
      std::cerr << "higher " << higher_moves.size() << "\n";
      higher_moves.dump(3);
    }
    typename SearchMoveSorter<P>::iterator p = higher_moves.begin();
    int best_value = p->first;
    if (base_t::isWinValue(P, best_value))
    {
      // ʬ⤱Ф⤦ɤμǤɤ(ߤϵͤߤ򸫤ĤȤ)
      break;
    }

    SearchMoveSorter<P> next_jobs;
    next_jobs.insert(*p);
    while (++p != higher_moves.end())
    {
      threshold = best_value + eval::delta(P);

      const SearchMove *target = p->second;
      SearchMove move_to_try = *target;
      if (higher_moves.size() <= 3)
	move_to_try.setLogProbAtMost(Probabilities::ReSearch);

      int val,max_value=threshold-eval::delta(P);
      if (testMove<P>(move_to_try, val, max_value, limit))
      {
	best_value = val;
	next_jobs.insert(std::make_pair(val, target)); // ȥå׸
	if (base_t::isWinValue(P, best_value))
	{
	  // ʬ⤱Ф⤦ɤμǤɤ(ߤϵͤߤ򸫤ĤȤ)
	  break;
	}
      }
    }
    higher_moves.swap(next_jobs);
  }
  const SearchMove *best_move = higher_moves.begin()->second;
  const int best_value = higher_moves.begin()->first;
  assert(limit >= 0);
  if (record)
  {
    base_t::recordLowerBound(P, record, limit, *best_move, best_value);
    assert((record->bestMove().getMove() == best_move->getMove())
	   || (record->lowerLimit() == limit && record->lowerBound() == best_value)
	   || (limit <= std::max(record->lowerLimit(), record->upperLimit())));
  }

  // thresholdϤΤޤ: 򿶤櫓Τ¸
  return best_move->getMove();
}

/**
 * ͤͿ higher_moves ʤ.
 * ޤ̤٤Τ绨Ĥͤ򤢤Ƥ
 * upperbound ޤäƤ鿿ܤ
 */
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
void osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
triageMovesSpeculatively(SearchMoveSorter<P>& higher_moves, int limit,
			 int jump)
{
  while (higher_moves.size() > 1)
  {
    SearchMoveSorter<P> next_jobs;
    typename SearchMoveSorter<P>::iterator p = higher_moves.begin();
    const int best_value = p->first;
    threshold = best_value + jump + eval::delta(P);
    if (base_t::isWinValue(P, threshold))
    {
      // ʬ⤱Ф⤦ɤμǤɤ
      return;
    }

    if (base_t::table->isVerbose())
    {
      std::cerr << "triage " << higher_moves.size() << " " << threshold << "\n";
      higher_moves.dump(3);
    }
    next_jobs.insert(*p);	// ǽμϻˤ
    while (++p != higher_moves.end())
    {
      const SearchMove *target = p->second;
      SearchMove move_to_try = *target;
      if (higher_moves.size() <= 3)
	move_to_try.setLogProbAtMost(Probabilities::ReSearch);

      int val;
      int max_value=threshold-eval::delta(P);
      if (testMove<P>(move_to_try, val, max_value, limit))
      {
	next_jobs.insert(std::make_pair(val, target)); // ͹
	next_jobs.insert(++p, higher_moves.end());
	break;
      }
    }
    if (next_jobs.size() <= 1)
      return;
    
    higher_moves.swap(next_jobs);
  }
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
bool osl::search::MTDF<Eval,MoveGenerator,Table,Recorder,Probabilities>::
testMove(const SearchMove& move, int& val, int& max_value, int limit)
{
  assert(P == move.getMove().player());
  assert(P == memory_tester.state().getTurn());
  assert(eval::betterThan(P, threshold,max_value));
  assert(threshold % 2);
  assert(! base_t::isWinValue(BLACK, threshold));
  assert(! base_t::isWinValue(WHITE, threshold));
  assert(move.getMove().isNormal());

  if ((! memory_tester.timerAvailable()) && (! timer.isNegative()) && (! timer.timeLeft()))
    throw misc::NoMoreTime();

  // adjusted_move: logp  history table 
  SearchMove adjusted_move = move;
  const int history_logp = memory_tester.historyTable().logp(move.getMove());
  if (history_logp < move.getLogProb())
    adjusted_move = SearchMove(MoveLogProb(move.getMove(), history_logp),
			       move.record);
  val = memory_tester.template nullWindowSearch<P>(threshold, adjusted_move,
						   limit);
  move.record = adjusted_move.record;
  
  if (eval::betterThan(P, val,max_value))
  {
    max_value = val;
    if (eval::betterThan(P, val,threshold))
    {
      memory_tester.setKillerMove(move.getMove());
      return true;
    }
    // fall thru
  }
  return false;
}

#endif /* _MTDF_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
