#ifndef _NULLWINDOW_SEARCH_TCC
#define _NULLWINDOW_SEARCH_TCC

#include "osl/search/nullWindowSearch.h"
#include "osl/search/nullWindowSearchPolicy.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/searchFramework.tcc"

// /** ơõ򤵤ʤȤ */
// #define NO_BEST_MOVE_EXTENSION

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
NullWindowSearch(const HashEffectState& s, typename framework_t::checkmate_t& c, 
		 Table *t, Recorder& r)
  : framework_t(s, c, t, r), null_window(0)
{
}
  
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
~NullWindowSearch()
{
}

/* ------------------------------------------------------------------------- */
/*	Methods								     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
int osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearchOfTurn(int threshold, const SearchMove& move, int limit)
{
  if (framework_t::state().getTurn() == BLACK)
    return nullWindowSearch<BLACK>(threshold, move, limit);
  else
    return nullWindowSearch<WHITE>(threshold, move, limit);
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
int osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearch(int threshold, const SearchMove& move, int limit)
{
  assert(threshold % 2);
  assert(! framework_t::isWinValue(BLACK, threshold));
  assert(! framework_t::isWinValue(WHITE, threshold));

  null_window = NullWindow(threshold);
  framework_t::setRoot(limit);
  assert(framework_t::curDepth() == 0);
  int max_value = framework_t::minusInfty(P);
  SearchMove best_move;
  nullWindowSearch<P,false>(move, max_value, best_move);
  return max_value;
}


template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, bool best_move_extension>
bool osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearch(const SearchMove& move, int& max_val, SearchMove& best_move)
{
  assert(P == move.getMove().player());
  assert(P == framework_t::state().getTurn());
  assert(eval::betterThan(P, null_window.beta(), max_val));
    
  typedef null_window::DoUndoNullWindow<NullWindowSearch,P,best_move_extension> 
    helper_t;
  helper_t helper(this, move, max_val);
  const int value
    = framework_t::template searchWithMove<P,helper_t>(move, helper);
  if (eval::betterThan(P, value, max_val))
  {
    max_val = value;
    best_move = move;
    if (eval::betterThan(P, value, null_window.alpha()))
      return true;
  }
  return false;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P> inline
int osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
normalSearch(const SearchMove& moved, 
	     int logprob, int base_value, SearchMove& best_move)
{
  const Player Turn = PlayerTraits<P>::opponent;
  typedef null_window::NullWindowSearchPolicy<NullWindowSearch,Turn,false>
    normal_policy_t;

  framework_t::recorder.tryMove(moved.moveLogProb(),
				null_window.alpha(),framework_t::curLimit());
  framework_t::subLimit(logprob);

  normal_policy_t normalPolicy(this, nullWindow());
  const int val = framework_t::template searchAllMoves<Turn,normal_policy_t>
    (moved.record, normalPolicy, best_move);

  framework_t::addLimit(logprob);
  framework_t::recorder.recordValue(moved.moveLogProb(),
				    val,eval::betterThan(P, val,base_value),
				    framework_t::curLimit());
  return val;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, bool best_move_extension>
int osl::search::NullWindowSearch<Eval,MoveGenerator,Table,Recorder,Probabilities>::
nullWindowSearchAfterMove(const SearchMove& moved, int base_value)
{
  assert(alt(P) == framework_t::state().getTurn());
  assert(P == moved.getMove().player());
  assert(eval::betterThan(P, null_window.alpha(), base_value));
  assert(moved.getLogProb() <= framework_t::curLimit());

  const Player Turn = PlayerTraits<P>::opponent;
  typedef null_window::NullWindowSearchPolicy<NullWindowSearch,Turn,true>
    extend_policy_t;
  SearchMove best_move;

  // normal search
  int val = normalSearch<P>(moved, moved.getLogProb(), base_value, best_move);

  if (! eval::betterThan(P, val, base_value))
    return val;
  if ((! best_move_extension) 
      && (! eval::betterThan(P, val, null_window.alpha())))
    return val;

  // first extension
  // ¿ȿ
  if (moved.getLogProb() 
      > Probabilities::ReSearch*4 + framework_t::ReSearchLimitMargin) 
  {
    val = normalSearch<P>(moved, Probabilities::ReSearch*4, base_value, best_move);
    if (! eval::betterThan(P, val, base_value))
      return val;
    if ((! best_move_extension)
	&& (! eval::betterThan(P, val, null_window.alpha())))
      return val;
  }
  if (moved.getLogProb() 
      > Probabilities::ReSearch*2 + framework_t::ReSearchLimitMargin) 
  {
    val = normalSearch<P>(moved, Probabilities::ReSearch*2, base_value, best_move);
    if (! eval::betterThan(P, val, base_value))
      return val;
    if ((! best_move_extension) 
	&& (! eval::betterThan(P, val, null_window.alpha())))
      return val;
  }
  if (moved.getLogProb() 
      > Probabilities::ReSearch + framework_t::ReSearchLimitMargin) 
  {
    val = normalSearch<P>(moved, Probabilities::ReSearch, base_value, best_move);
  }
  
  if (best_move_extension)
    return val;
  if (! eval::betterThan(P, val, null_window.alpha()))
    return val;

  // second extension
#ifndef NO_BEST_MOVE_EXTENSION
  // ơõ κϥåȤʤƤĹ⡼ɤ
  if (Probabilities::DefaultProb != Probabilities::ReSearch)
  {
    extend_policy_t extendPolicy(this, nullWindow());

    framework_t::recorder.retryMove(moved.moveLogProb(), null_window.alpha(),
				    framework_t::curLimit(),2);
    const int log_prob = std::min(moved.getLogProb(), (int)Probabilities::ReSearch);
    framework_t::subLimit(log_prob);

    val = framework_t::template searchAllMoves<Turn,extend_policy_t>
      (moved.record, extendPolicy, best_move);

    framework_t::addLimit(log_prob);
    framework_t::recorder.recordValue(moved.moveLogProb(), val, 
				      eval::betterThan(P, val,base_value),
				      framework_t::curLimit());
  }
#endif
  return val;
}

/* ------------------------------------------------------------------------- */

#endif /* _NULLWINDOW_SEARCH_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
