/* quiescenceGenerator.h
 */
#ifndef QUIESCENCEGENERATOR_H
#define QUIESCENCEGENERATOR_H

#include "osl/search/shouldPromoteCut.h"
#include "osl/move_generator/addEffect.h"
#include "osl/move_generator/capture.h"
#include "osl/move_generator/escape.h"
#include "osl/move_generator/promote.h"
#include "osl/move_generator/safeDropMajorPiece.h"
#include "osl/move_generator/addEffect8.h"
#include "osl/move_generator/additionalLance.h"
#include "osl/move_action/store.h"
#include "osl/eval/pieceEval.h"
#include "osl/state/numEffectState.h"
#include "osl/container/position8.h"

namespace osl
{
  namespace search
  {
    /**
     * QuiescenceSearch ǻȤؼ
     */
    template <Player P>
    struct QuiescenceGenerator
    {
      /**
       * P Ũ PTYPE 롥
       * @param dont_capture Ͻ
       */
      template <Ptype PTYPE, bool has_dont_capture>
      static void capture(const NumEffectState&,
			  MoveVector& moves, Piece dont_capture);
      static void captureAll(const NumEffectState&, MoveVector& moves);
      static void captureAll(const NumEffectState&,
			     MoveVector& moves, Piece dont_capture);
      /**
       * P Ũ target 롥
       */
      static void capture(const NumEffectState&,
			  Position target, MoveVector& moves);
      static void promote(const NumEffectState&, MoveVector& moves);
      template <Ptype PTYPE>
      static void promote(const NumEffectState&, MoveVector& moves);
      static void check(const NumEffectState&, MoveVector& moves,
			bool no_liberty=false);
      static void check(const NumEffectState&, bool no_liberty,
			const Position8& sendoffs, MoveVector& moves);
      static void escapeKing(const NumEffectState& state, MoveVector& moves); 
      /**
       * @return existance of safe (in takeback) move
       */
      static bool escapeKingInTakeBack(const NumEffectState& state, MoveVector& moves, bool check_by_lance); 
      static void dropMajorPiece(const NumEffectState& state, MoveVector& moves); 

      static void attackMajorPieceSecondSelection(const NumEffectState& state,
						  bool target_has_support,
						  const MoveVector& src,
						  MoveVector& out);
      static void attackMajorPieceFirstSelection(const NumEffectState& state,
						 const MoveVector& all_moves,
						 MoveVector& moves,
						 MoveVector& expensive_drops);

      static void attackMajorPiece(const NumEffectState& state,
				   MoveVector& moves); 
      static void escapeAll(const NumEffectState& state, MoveVector& moves); 
      /**
       * @param escape KINGʳζ
       */
      static void escapeNormalPiece(const NumEffectState& state, 
				    Piece escape, MoveVector& moves,
				    bool add_support_only=false); 
      /**
       * ľ˻ؼ꤫ƨ
       */
      static void escapeFromLastMove(const NumEffectState& state, 
				     Move last_move, MoveVector& moves); 
      static void escapeFromLastMoveOtherThanPawn(const NumEffectState& state, 
				     Move last_move, MoveVector& moves); 
      /**
       * @return Τʤ˰ưǽ
       */
      static bool escapeByMoveOnly(const NumEffectState& state, 
				   Piece piece, MoveVector& moves); 
      static void attackGoldWithPawn(const NumEffectState& state,
				     MoveVector& moves); 
      static void attackSilverWithPawn(const NumEffectState& state,
				       MoveVector& moves); 
      static void attackKnightWithPawn(const NumEffectState& state,
				       MoveVector& moves); 
      /**
       * Ѥ˿ʤ. ˤɤȴɤ
       */
      static void advanceBishop(const NumEffectState& state, 
				MoveVector& moves); 
      template <Direction DIR>
      static void advanceBishop(const NumEffectState& state, 
				const Position from, MoveVector& moves); 
      static void attackKing8(const NumEffectState& state, MoveVector& moves); 
      static void attackToPinned(const NumEffectState& state, MoveVector& moves); 

      static void utilizePromoted(const NumEffectState& state, 
				  Piece target,
				  MoveVector& moves); 

      static void breakThreatmate(const NumEffectState& state, 
				  Move threatmate, MoveVector& moves);
      static void kingWalk(const NumEffectState& state, 
			   MoveVector& moves);

      struct CaptureGenerator
      {
	const NumEffectState *state;
	MoveVector& out;
	void operator()(Piece target_piece)
	{
	  QuiescenceGenerator<P>::capture(*state, target_piece.position(), out);
	}
      };
      struct CaptureGeneratorWithCheck
      {
	const NumEffectState *state;
	MoveVector& out;
	Piece dont_capture;
	void operator()(Piece target_piece)
	{
	  if (target_piece != dont_capture)
	    QuiescenceGenerator<P>::capture(*state, target_piece.position(), out);
	}
      };
    };
  } // namespace search
} // namespace osl




template <osl::Player P> 
inline
void osl::search::QuiescenceGenerator<P>::
capture(const NumEffectState& state, Position target, MoveVector& moves)
{
  using namespace move_action;
  MoveVector all_moves;
  Store store(all_moves);
  move_generator::Capture<P>::generate(state, target, store);
  for (MoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end(); ++p)
  {
    if (ShouldPromoteCut::canIgnoreMove<P>(*p))
      continue;
    moves.push_back(*p);
  }
}

template <osl::Player P> 
template <osl::Ptype PTYPE> 
inline
void osl::search::QuiescenceGenerator<P>::
promote(const NumEffectState& state, MoveVector& moves)
{
  assert(moves.empty());
  using namespace move_action;
  Store store(moves);
  move_generator::Promote<P,NumEffectState,Store>
    ::template generateMovesPtype<PTYPE>(state, store);
  return;
}

template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
escapeKing(const NumEffectState& state, MoveVector& moves) 
{
  assert(moves.empty());
  MoveVector all_moves;
  move_action::Store store(all_moves);
  GenerateEscape<P>::generate(state, state.template getKingPiece<P>(), store);
  for (MoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end();
       ++p)
  {
    const Position to = p->to();
    if (p->isDrop())
    {
      // ʳɤޤʤ
      if (! state.hasEffectBy<P>(to))
      {
	if (! ((p->ptype() == PAWN)
	       && state.hasMultipleEffectBy(alt(P), to)))
	  continue;
      }
    }
    else if (p->ptype() != KING)
    {
      // ʳɤޤʤ
      if (p->capturePtype() == PTYPE_EMPTY)
      {
	if (! state.hasMultipleEffectBy(P, to))
	{
	  if (! ((p->ptype() == PAWN)
		 && state.hasMultipleEffectBy(alt(P), to)))
	    continue;
	}
      }
    }
    moves.push_back(*p);
  }
}

template <osl::Player P>
inline
bool osl::search::QuiescenceGenerator<P>::
escapeKingInTakeBack(const NumEffectState& state, MoveVector& moves,
		     bool check_by_lance) 
{
  bool has_safe_move = false;
  assert(moves.empty());
  MoveVector all_moves;
  move_action::Store store(all_moves);
  GenerateEscape<P>::generate(state, state.template getKingPiece<P>(), store);
  Position last_drop_to = Position::STAND();
  MoveVector drops;
  for (MoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end();
       ++p)
  {
    const Position to = p->to();
    if (p->ptype() == KING)
    {
      // TODO: Ϥ
      has_safe_move = true; // KFEND ή: ƨ꤬ɤȤ
      continue;
    }
    if (p->capturePtype() != PTYPE_EMPTY)
    {
      moves.push_back(*p);
      continue;
    }
    // ̵̹ɤޤʤ
    const int attack_count = state.countEffect(alt(P),to);
    const int defense_count 
      = state.countEffect(P,to) + (p->isDrop() ? 1 : 0);
    if (defense_count <= attack_count)
      continue;
    if ((attack_count == 1) && (! check_by_lance))
    {
      // ιФƼʬʬʤɤȤ
      has_safe_move = true;
      continue;
    }

    // ʾdrop β򤫤ʤΤǹ1ǵͤȽˤϽʬ
    // TODO: ְ¤ˤ
    if (to != last_drop_to)
    {
      last_drop_to = to;
      drops.push_back(*p);
    }
  }
  if (! has_safe_move)
    moves.push_back(drops.begin(), drops.end());
  return has_safe_move;
}

template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
dropMajorPiece(const NumEffectState& state, MoveVector& moves) 
{
  assert(moves.empty());
  using namespace move_action;
  Store store(moves);
  move_generator::SafeDropMajorPiece<P>::generate(state, store);
}

template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
attackGoldWithPawn(const NumEffectState& state, MoveVector& moves) 
{
  assert(moves.empty());
  using namespace move_action;

  const bool has_pawn = state.hasPieceOnStand<P, PAWN>();
  for (int i = PtypeTraits<GOLD>::indexMin;
       i < PtypeTraits<GOLD>::indexLimit; ++i)
  {
    const Piece p = state.getPieceOf(i);
    if (! p.isOnBoardByOwner<PlayerTraits<P>::opponent>())
      continue;
    const Position head = (Board_Table.nextPosition(P, p.position(), D));
    if (state.getPieceAt(head).isEmpty())
    {
      // TODO: Ϥ˵
      const int defense = state.countEffect(alt(P), head);
      const int attack = state.countEffect(P, head)
	+ state.hasEffectByPtype<BISHOP>(P, p.position());

      // move
      if (! head.canPromote<P>()) // ϤǤƤϤ
      {
	const Position origin = (Board_Table.nextPosition(P, head, D));
	if (defense <= attack
	    || state.hasEffectByPtype<BISHOP>(P, origin)) // open attack
	{
	  const Piece candidate = state.getPieceAt(origin);
	  if (candidate.ptype() == PAWN && candidate.owner() == P)
	  {
	    const Move move(origin, head, PAWN, PTYPE_EMPTY, false, P);
	    moves.push_back(move);
	  }
	}
      }      
      // drop
      if (defense <= attack)
      {
	if (has_pawn && !state.template isPawnMaskSet<P>(head.x()))
	{
	  const Move move(head, PAWN, P);
	  moves.push_back(move);
	}
      }
    }
  }
}

template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
attackKnightWithPawn(const NumEffectState& state, MoveVector& moves) 
{
  assert(moves.empty());
  using namespace move_action;

  const bool has_pawn = state.hasPieceOnStand<P, PAWN>();
  for (int i = PtypeTraits<KNIGHT>::indexMin;
       i < PtypeTraits<KNIGHT>::indexLimit; ++i)
  {
    const Piece p = state.getPieceOf(i);
    if (p.isOnBoard() && p.ptype() == KNIGHT && p.owner() == alt(P))
    {
      const Position head = (Board_Table.nextPosition(P, p.position(), D));
      // ɤ...
      if (state.getPieceAt(head).isEmpty())
      {
	const int defense = state.countEffect(alt(P), head);
	const int offense = state.countEffect(P, head);
	if ((defense <= offense)
	    && (! head.canPromote<P>())) // ϤǤƤϤ
	{
	  const Position origin = (Board_Table.nextPosition(P, head, D));
	  const Piece candidate = state.getPieceAt(origin);
	  if (candidate.ptype() == PAWN && candidate.owner() == P)
	  {
	    const Move move(origin, head, PAWN, PTYPE_EMPTY, false, P);
	    moves.push_back(move);
	  }
	}
	if (has_pawn && !state.template isPawnMaskSet<P>(head.x())
	    && (defense <= offense+1))
	{
	  const Move move(head, PAWN, P);
	  moves.push_back(move);
	}
      }
    }
  }
}

template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
advanceBishop(const NumEffectState& state, MoveVector& moves) 
{
  for (int i = PtypeTraits<BISHOP>::indexMin;
       i < PtypeTraits<BISHOP>::indexLimit; ++i)
  {
    const Piece bishop = state.getPieceOf(i);
    if (! bishop.isOnBoardByOwner<P>())
      continue;
    if (bishop.ptype() != BISHOP)
      continue;
    const Position from = bishop.position();
    if (state.hasEffectBy<PlayerTraits<P>::opponent>(from))
      continue;			// escape Ͻʣ
    advanceBishop<UL>(state, from, moves);
    advanceBishop<UR>(state, from, moves);
  }
}

template <osl::Player P>
template <osl::Direction DIR>
inline
void osl::search::QuiescenceGenerator<P>::
advanceBishop(const NumEffectState& state, 
	      const Position from, MoveVector& moves)
{
  const Offset offset = DirectionPlayerTraits<DIR,P>::offset();
  for (Position to=from+offset;; to+=offset)
  {
    if (! state.getPieceAt(to).isEmpty())
      break;		// capture ϽʣĤǤ׳Ƚ
    if (to.canPromote<P>())
      break;		// promote Ͻʣ
    if (state.hasEffectBy<PlayerTraits<P>::opponent>(to))
      continue;
    const Move move(from, to, BISHOP, PTYPE_EMPTY, false, P);
    assert(state.isAlmostValidMove<false>(move));
    moves.push_back(move);
  }
}
      
template <osl::Player P>
inline
void osl::search::QuiescenceGenerator<P>::
escapeFromLastMove(const NumEffectState& state, Move last_move, MoveVector& moves)
{
  if (! last_move.isNormal())
    return;
  if (last_move.ptype() != PAWN)
  {
    escapeFromLastMoveOtherThanPawn(state, last_move, moves);
    return;
  }
  const Position attack_from = last_move.to();
  const Position attack_to = (Board_Table.nextPosition(alt(P),attack_from, U));
  const Piece target = state.getPieceOnBoard(attack_to);
  if (! target.isOnBoardByOwner<P>())
    return;

  using namespace move_action;
  MoveVector all_moves;
  Store store(all_moves);
  
  GenerateEscape<P>::generate(state, target, store);
  for (MoveVector::const_iterator p=all_moves.begin(); p!=all_moves.end();
	 ++p)
  {
    assert(! p->isDrop());
    const Position to = p->to();
    // ƨˤνʬʤȤ
    const int defense = state.countEffect(P, to);
    const int offense = state.countEffect(alt(P), to);
    if (offense > defense)
      continue;
    if (to == attack_from)
    {
      // ϽʣξϿˤΤƤ
      if (offense == defense)
	continue;
      assert(p->capturePtype() == PAWN);
      if (! ShouldPromoteCut::canIgnoreMove<P>(*p))
	moves.push_back(*p);
      continue;
    }
    assert(p->from() == attack_to);
    if ((offense == defense)
	&& (p->capturePtype() == PTYPE_EMPTY))
      continue;

    // TODO: KnightCaptureDepth
    if ((! p->isPromote())
	&& (! ShouldPromoteCut::canIgnoreMove<P>(*p)))
    {
      moves.push_back(*p);
    }
  }
}

#endif /* QUIESCENCEGENERATOR_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
