/* quiescenceRecord.h
 */
#ifndef _QUIESCENCERECORD_H
#define _QUIESCENCERECORD_H

#include "osl/search/dualThreatmateState.h"
#include "osl/checkmate/oracleAges.h"
#include "osl/checkmate/king8Info.h"
#include "osl/effect_util/sendOffPosition.h"
#include "osl/ptype.h"
#include "osl/move.h"
#include "osl/container/moveVector.h"
#include "osl/stl/vector.h"
#include <iosfwd>
namespace osl
{
  namespace search
  {
    struct QSearchTraits
    {
      enum { 
	/** ̾õκ */
	MaxDepth = 8,
	/** ¨ͤο */
	CheckmateSpecialDepth = 127,
	/** »롼פο */
	HistorySpecialDepth = 126,
      };
      enum { FirstThreat = 6, SecondThreat = 2 };
      enum MoveType { 
	UNKNOWN, KING_ESCAPE, CAPTURE, PROMOTE, CHECK,
	ESCAPE, ATTACK, OTHER, 
      };
    };

    struct QuiescenceFlags
    {
      unsigned int data;
      enum Flags { 
	/** 
	 * ؼȽŤʤʤ褦ˡʳλؼϼPTYPE
	 * Ťʤʤ褦ꤹɬפ
	 */
	KING_ESCAPE = KING, PROMOTE = PTYPE_EMPTY, CHECK = PTYPE_EDGE,
	DROP = PTYPE_SIZE, ALL_ESCAPE, MAJOR_PIECE_ATTACK, KNIGHT_ATTACK,
	GOLDSILVER_ATTACK,
	ADVANCE_BISHOP, KING8_ATTACK, ESCAPE_FROM_LAST_MOVE,
	PINNED_ATTACK, UTILIZE_PROMOTED, BREAK_THREATMATE, KING_WALK,
	//
	LAST_FLAG,
      };      
      QuiescenceFlags() : data(0)
      {
      }
    private:
      bool isGenerated(int type) const
      {
	return data & (1u << type);
      }
      void recordGeneration(int type)
      {
	data |= (1u << type);
      }
    public:
      bool isGenerated(Ptype ptype) const
      {
	return isGenerated(static_cast<int>(ptype));
      }
      void recordGeneration(Ptype ptype)
      {
	recordGeneration(static_cast<int>(ptype));
      }
      bool isGenerated(Flags f) const
      {
	return isGenerated(static_cast<int>(f));
      }
      void recordGeneration(Flags f)
      {
	recordGeneration(static_cast<int>(f));
      }
      void clear() 
      {
	data = 0;
      }
    };

    /**
     * QuiescenceSearch ǥѥͭϤʻؼ
     */
    struct QuiescenceThreat
    {
      int value;
      Move move;
      explicit QuiescenceThreat(int v=0, Move m=Move::INVALID()) 
	: value(v), move(m)
      {
      }
    };    
      
    /**
     * QuiescenceSearch ɽ¸ǡ
     */
    class QuiescenceRecord
    {
    public:
      enum StaticValueType { UNKOWN, UPPER_BOUND, EXACT };
      static const char *toString(StaticValueType);
    private:
      int upper_bound, lower_bound;
      vector<Move> move_cache;
      QuiescenceFlags flags;
      Move best_move;
      int static_value;
      /** static_value ؤζ */
      QuiescenceThreat threat1, threat2;
      mutable Position8 sendoffs;
      mutable King8Info king8info;
      /** upper_bound, lower_bound, static_value õ */
      char upper_depth, lower_depth, static_value_depth;
      char static_value_type;
      mutable bool sendoffs_initialized, king8info_initialized;
      bool visited;
    public:
      DualThreatmateState threatmate;
      /**  oracle  */
      AttackOracleAges attack_oracle, threatmate_oracle;
      
      /**
       * Ȥʤ.
       * CAVEAT: -1 ȲĹ depth -1ˤʤäˤϤޤ
       */
      enum { InitialDepth = -128, };
      QuiescenceRecord () 
	: best_move(Move::INVALID()), king8info(0),
	  upper_depth(InitialDepth), lower_depth(InitialDepth),
	  static_value_depth(InitialDepth), 
	  static_value_type(UNKOWN), sendoffs_initialized(false), 
	  king8info_initialized(false), visited(false)
      {
      }

      template <Player Turn>
      const Position8& sendOffPosition(const NumEffectState& state) const
      {
	assert(Turn == state.getTurn());
	if (! sendoffs_initialized)
	{
	  const Position king_position = state.getKingPosition(alt(Turn));
	  effect_util::SendOffPosition::find<Turn>(state, king_position, 
						   sendoffs);
	  sendoffs_initialized = true;
	}
	return sendoffs;
      }
      const Position8& 
      sendOffPosition(Player turn, const NumEffectState& state) const
      {
	if (turn == BLACK)
	  return sendOffPosition<BLACK>(state);
	else
	  return sendOffPosition<WHITE>(state);
      }
      const King8Info& 
      king8Info(const NumEffectState& state) const
      {
	if (! king8info_initialized)
	{
	  king8info = King8Info::make(state.getTurn(), state);
	  king8info_initialized = true;
	}
	return king8info;
      }

      void setVisited(bool value=true) { 
	assert(! (visited && value));
	visited = value;
      }
      bool isVisited() const { return visited; }
      void clear()
      {
	move_cache.clear();
	flags.clear();
	upper_depth = lower_depth = static_value_depth = InitialDepth;
      }
      void setStaticValue(StaticValueType type, int value, int depth,
			  const QuiescenceThreat& t1=QuiescenceThreat(),
			  const QuiescenceThreat& t2=QuiescenceThreat())
      {
	assert((depth <= QSearchTraits::MaxDepth)
	       || (depth == QSearchTraits::CheckmateSpecialDepth));
	static_value_type = type;
	static_value = value;
	static_value_depth = depth;
	threat1 = t1;
	threat2 = t2;
      }
      void setLowerBound(int depth, int bound, Move best_move)
      {
	assert((depth <= QSearchTraits::MaxDepth)
	       || (depth == QSearchTraits::CheckmateSpecialDepth));
	if (depth >= lower_depth)
	{
	  lower_depth = depth;
	  lower_bound = bound;
	  this->best_move = best_move;
	}
      }
      void setUpperBound(int depth, int bound)
      {
	assert((depth <= QSearchTraits::MaxDepth)
	       || (depth == QSearchTraits::CheckmateSpecialDepth));
	if (depth >= upper_depth)
	{
	  upper_depth = depth;
	  upper_bound = bound;
	}
      }
      void setHistoryValue(int value)
      {
	lower_bound = upper_bound = value;
	lower_depth = upper_depth = QSearchTraits::HistorySpecialDepth;
      }
      void setHistoryValue(Move best_move, int value)
      {
	this->best_move = best_move;
	setHistoryValue(value);
      }
      void reserveMoves(size_t new_size)
      {
	move_cache.reserve(move_cache.size()+new_size);
      }
    private:
      void addMovesReserved(const MoveVector& new_moves)
      {
	move_cache.insert(move_cache.end(), 
			  new_moves.begin(), new_moves.end());
      }
      void addMoves(const MoveVector& new_moves)
      {
	reserveMoves(new_moves.size());
	addMovesReserved(new_moves);
      }
    public:
      void addKillerMoves(const MoveVector& new_moves)
      {
	addMoves(new_moves);
      }
      void addKillerMovesReserved(const MoveVector& new_moves)
      {
	assert(move_cache.capacity() >= move_cache.size() + new_moves.size());
	addMovesReserved(new_moves);
      }
      template <class MOVE_TYPE>
      void addMoves(MOVE_TYPE type, const MoveVector& new_moves)
      {
	assert(! flags.isGenerated(type));
	flags.recordGeneration(type);
	addMoves(new_moves);
      }

      StaticValueType staticValueType() const { 
	return static_cast<StaticValueType>(static_value_type); 
      }
      bool hasStaticValue() const { return staticValueType() != UNKOWN; }
      int staticValue() const { assert(hasStaticValue()); return static_value; }
      int staticValueDepth() const { return static_value_depth; }
      int upperDepth() const { return upper_depth; }
      int lowerDepth() const { return lower_depth; }
      int upperBound() const { return upper_bound; }
      int lowerBound() const { return lower_bound; }
      const Move bestMove() const { return best_move; }
      const vector<Move>& moves() const { return move_cache; }
      QuiescenceFlags cacheFlags() const { return flags; }

      void dump(std::ostream&) const;
    };
    
  } // namespace search
} // namespace osl

#endif /* _QUIESCENCERECORD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
