/* searchBase.h
 */
#ifndef _SEARCHBASE_H
#define _SEARCHBASE_H

#include "osl/search/searchEvalBase.h"
#include "osl/search/searchRecorder.h"
#include "osl/search/searchTable.h"
#include "osl/search/realizationProbability.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/searchWindow.h"
#include "osl/hash/hashCollision.h"
#include "osl/hash/hashKey.h"
#include "osl/state/numEffectState.h"
#include <cassert>
namespace osl
{
  namespace search
  {
    class SimpleHashTable;

    /**
     * MTDF  SearchFramework ˶̤
     * ʪΥ᥽å.
     */
    template<typename Eval, typename MoveGenerator, 
	     typename Table, typename Recorder, typename Probabilities>
    struct SearchBase :
      protected SearchEvalBase<Eval>
    {
      // types
      typedef Eval eval_t;
      typedef SearchEvalBase<eval_t> base_t;
    
      typedef MoveGenerator MoveGenerator_t; // ̾ΤĤɤ?
      typedef Probabilities Probabilities_t;
    protected:
      Recorder& recorder;
      Table *table;		// acquaintance
    public:
      SearchBase(Recorder& r, Table *t) : recorder(r), table(t)
      {
	assert(table);
      }
      virtual ~SearchBase() {}
      virtual bool abort(Move) const { return false; }
      /** 
       * ơ֥λؼΤ
       *
       * ĤǤ˼¸ΨΥå⤹
       * HashKey conflict table 鸣 move Ǥ뤳Ȥ
       * conflict  turn ΰ㤤 isAlmostValidMove ǤϥåǤʤ
       * Ȥ
       */
      bool validTableMove(const NumEffectState& state,
			  const MoveLogProb& move, int limit) const
      {
	if ((limit < move.getLogProb()) || (! move.validMove()))
	  return false;
	if (move.getMove().isPass())
	  return false;		// TODO: ξʬϼ줽ʵ
	const bool valid
	  = ((move.getMove().player() != state.getTurn())
	     || state.isAlmostValidMove<false>(move.getMove()));
	if (! valid)
	{
	  // 褳ƤϤʤ
	  recorder.recordInvalidMoveInTable(state, move, limit);
	  throw hash::HashCollision();
	}
	return valid;
      }

      void recordLowerBound(Player P,
			    SimpleHashRecord *record, int limit,
			    const SearchMove& best_move, int val) const
      {
	assert(limit >= 0);
	recorder.tableStoreLowerBound(P, best_move.moveLogProb(), val, limit);
	if (best_move.validMove())
	{
	  // ɽκѤ򤷤Ƥ硤best_move.record  null ξ礢
	  if (! (isWinValue(P, val) || isWinValue(alt(P), val)
		 || (val == drawValue()) || best_move.getMove().isPass()
		 || table->minimumRecordLimit() >= 0))
	  {
	    if (! best_move.record)
	      abort(best_move.getMove());
	    assert(best_move.record || abort(best_move.getMove()));
	  }
	  MoveLogProb store_move = best_move.moveLogProb();
	  store_move.setLogProbAtLeast(Probabilities::TableMove);
	  record->setLowerBound(P, limit, 
				SearchMove(store_move, best_move.record), val);
	}
	else
	{
	  record->setLowerBound(P, limit, best_move, val);
	}
      }
      void recordUpperBound(Player P,
			    SimpleHashRecord *record, int limit,
			    const SearchMove& best_move, int val) const
      {
	assert(limit >= 0);
	recorder.tableStoreUpperBound(P, best_move.moveLogProb(), val, limit);

	if (best_move.validMove())
	{
	  if (! (isWinValue(P, val) || isWinValue(alt(P), val)
		 || (val == drawValue()) || best_move.getMove().isPass()
		 || table->minimumRecordLimit() >= 0))
	    assert(best_move.record || abort(best_move.getMove()));
	  MoveLogProb store_move = best_move.moveLogProb();
	  store_move.setLogProbAtLeast(Probabilities::TableMove);
	  record->setUpperBound(P, limit, 
				SearchMove(store_move, best_move.record), val);
	}
	else
	{
	  record->setUpperBound(P, limit, best_move, val);
	}
      }
    private:
      void recordCheckmateResult(Player P, SimpleHashRecord *record,
				 int val, Move move) const
      {
	assert(isWinValue(P,val) || isWinValue(alt(P),val));
	const int limit = SearchTable::CheckmateSpecialDepth;
	const MoveLogProb best_move(move, 101);
	recorder.tableStoreLowerBound(P, best_move, val, limit);
	recorder.tableStoreUpperBound(P, best_move, val, limit);
	record->setAbsoluteValue(best_move.getMove(), val,
				 SearchTable::CheckmateSpecialDepth);
      }
    public:
      /** ;θĤ */
      void recordWinByCheckmate(Player P, SimpleHashRecord *record,
				Move check_move) const
      {
	assert(record);
	recordCheckmateResult(P, record, winByCheckmate(P), check_move);
      }
      /** ;θĤ */
      void recordLoseByCheckmate(Player P, SimpleHashRecord *record) const
      {
	assert(record);
	recordCheckmateResult(P, record, winByCheckmate(alt(P)), Move::INVALID());
      }

      using base_t::isWinValue;
      using base_t::winByCheckmate;
      using base_t::winByFoul;
      using base_t::winByLoop;
      using base_t::minusInfty;
      using base_t::drawValue;
    
    };
  } // namespace search
} // namespace osl


#endif /* _SEARCHBASE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
