#ifndef _SEARCH_FRAMEWORK_TCC
#define _SEARCH_FRAMEWORK_TCC

#include "osl/search/searchFramework.h"
#include "osl/search/searchTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/search/quiescenceSearch.h"
#include "osl/search/shouldPromoteCut.h"
#include "osl/search/searchMoveVector.h"
#include "osl/search/dominanceCheck.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/quiescenceWindow.h"
#include "osl/search/quiescenceWindow.tcc"
#include "osl/checkmate/immediateCheckmate.h"
#include "osl/move_classifier/pawnDropCheckmate.h"
#include "osl/move_classifier/moveAdaptor.h"
#include "osl/category/categoryListUtil.h"
#include "osl/record/csa.h"
#include "osl/container/moveStack.h"
#include "osl/container/moveLogProbSet.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/effect_util/effectUtil.h"
#include "osl/effect_util/additionalEffect.h"
#include <iostream>

#define search_assert(x, m) assert((x) || abort(m))
#define GUIDE_BY_QSEARCH_LIMITED

/* ------------------------------------------------------------------------- */
/*	Constructors							     */
/* ------------------------------------------------------------------------- */

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
SearchFramework(const HashEffectState& s, checkmate_t& c, 
		Table *t, Recorder& r)
  : base_t(r, t), SearchState(s, c), eval(s)
{
}
  
template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
~SearchFramework()
{
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
bool osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
abort(Move best_move) const
{
  return SearchState::abort(best_move);
}

/* ------------------------------------------------------------------------- */
/*	Methods								     */
/* ------------------------------------------------------------------------- */


namespace osl
{
  namespace search
  {
    namespace framework
    {
      /**
       * category ˼
       * examineMoves Ƥ֤helper
       * @param Policy examineMoves Ǽ¹Ԥ륳
       */
      template <Player P, class Policy>
      struct CategoryHelper
      {
	const Policy& policy;
	SearchMove *best_move;
	int *max_val;
	SearchMoveSet *tried;
	CategoryHelper(const Policy& h, SearchMove *b, int *m,
		       SearchMoveSet *t=0)
	  : policy(h), best_move(b), max_val(m), tried(t)
	{
	}
	template <bool isKingEscape>
	bool operator()(const char *category_name, MoveLogProbVector& moves)
	{
	  assert(tried);
	  return (*policy.searcher).template examineMoves<P,Policy,isKingEscape>
	    (category_name, moves, policy, *best_move, *max_val, *tried);
	}
      };
    } // namespace framework
  } // namespace search
} // namespace osl

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Helper>
int osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
searchWithMove(const SearchMove& search_move, Helper& helper)
{
  const Move move = search_move.getMove();
  const HashKey new_hash = currentHash().newHashWithMove(move);
  assert(P == move.player());

  if (move.isPass())
    pass_count.inc(P);

  // ǧ
  if (! pass_count.loopByBothPass())
  {
    const Sennichite next_sennichite
      = repetition_counter.isAlmostSennichite(new_hash);
    if (next_sennichite.isDraw())
      return base_t::drawValue();
    if (next_sennichite.hasWinner())
      return base_t::winByFoul(next_sennichite.winner());
    assert(next_sennichite.isNormal());
  }

  if (! move.isPass())
  {
    // ͥ۴طǧ
    const DominanceCheck::Result has_dominance
      = DominanceCheck::detect(repetition_counter.history(), new_hash);
    if (has_dominance == DominanceCheck::LOSE)
      return base_t::winByLoop(alt(P));
    if (has_dominance == DominanceCheck::WIN)
      return base_t::winByLoop(P);
    // Ϣ³Τ
    if (move.capturePtype() == PTYPE_EMPTY)
    {
      const int sacrifice_count = countSacrificeCheck2(curDepth());
      if (sacrifice_count == 2)
      {
	// 3ܤϻؤʤ
	const Position to = move.to();
	int offence = state().countEffect(P, to) + (move.isDrop() ? 1 : 0);
	const int deffense = state().hasEffectBy(alt(P), to); // max1
	if (offence <= deffense)
	  offence += AdditionalEffect::count2(state(), to, P);
	if (offence <= deffense)
	{
	  return base_t::winByLoop(alt(P));
	}
      }
    }
  }

  // õ
  int result;
  helper.setResultAddress(&result);
  base_t::recorder.addNodeCount();
  
  const eval_t old_eval = eval;
  ApplyMoveOfTurn::doMove(eval, state(), move);
  doUndoMoveOrPass<P,Helper>(new_hash, move, &search_move.record, helper);
  eval = old_eval;

  if (move.isPass())
    pass_count.dec(P);

  // ͳǧ
  if (/*base_t::isWinValue(P, result) && */(! move.isPass())
      && move_classifier::MoveAdaptor<move_classifier::PawnDropCheckmate<P> >
      ::isMember(state(), move))
    return base_t::winByFoul(alt(P));

  return result;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player Turn, class Policy, bool isKingEscape>
bool osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *category_name, const MoveLogProbVector& moves, 
	     const Policy& policy,
	     SearchMove& best_move, int& max_val, SearchMoveSet& triedMoves)
{
  assert(Turn == state().getTurn());
  base_t::recorder.newCategory(category_name, curLimit());
  bool try_next = true;
  for (MoveLogProbVector::const_iterator p=moves.begin(); p!=moves.end(); p++)
  {
    SearchMove *next_move = 0;
    if ((! isKingEscape) && (! p->getMove().isPass())
	&& ShouldPromoteCut::canIgnoreAndNotDrop<Turn>(p->getMove()))
      continue;
    if (! isKingEscape)
    {
      next_move = triedMoves.assignIfBetter(*p, ReSearchLimitMargin);
    }
    else
    {
      // ꤫ƨμʤȾԤ
      next_move = triedMoves.assignIfBetter(*p, 0);
    }
    if (! next_move)
      continue;
    if (try_next)
    {
      try_next = (! policy.searchWithMove(*next_move, max_val, best_move));
      assert(try_next || (! base_t::isWinValue(alt(Turn),max_val))
	     || abort(best_move.getMove()));
      // cut ǤĤλؼ triedMoves ˵Ͽʤȼ
      // ˬǤϤޤΤ return ʤ
    }
  }
  return try_next;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player Turn, class Policy>
inline 
bool osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
examineMoves(const char *category_name, const SearchMoveVector& moves, 
	     const Policy& policy, SearchMove& best_move, int& max_val)
{
  assert(Turn == state().getTurn());
  base_t::recorder.newCategory(category_name, curLimit());
  for (SearchMoveVector::const_iterator p=moves.begin(); p!=moves.end(); p++)
  {
    if (policy.searchWithMove(**p, max_val, best_move))
    {
      assert(! base_t::isWinValue(alt(Turn),max_val));
      return false;
    }
  }
  return true;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Policy>
int osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
searchAllMoves(SimpleHashRecord *& record_memorize,
	       const Policy& policy, SearchMove& best_move)
{
#ifndef NDEBUG
  checkPointSearchAllMoves();
#endif
  assert(P == state().getTurn());
  search_assert(policy.window().isConsistent(P), lastMove());
#ifndef NDEBUG
  const int MaxLimit = 4096;
#endif
  assert(curLimit() < MaxLimit);
  assert(curLimit() >= 0);

  if (pass_count.loopByBothPass())
    return quiescenceValue<P,Policy>(policy);

  int max_val=base_t::minusInfty(P);
  const Player Opponent = PlayerTraits<P>::opponent;
  // ʬ֤β => ľμ꤬ˡ => 
  if (EffectUtil::isKingInCheck(Opponent, state())) {
    max_val = base_t::minusInfty(alt(P));
    return max_val;
  }
  // ɽԲǽξˤδؿѤrecord
  boost::scoped_ptr<SimpleHashRecord> record_if_unavailable;
  // δؿѤrecord
  SimpleHashRecord *record = record_memorize;
  if (! record)
    record = record_memorize
      = base_t::table->allocate(currentHash(), curLimit());
  if (! record)
  {
    record_if_unavailable.reset(new SimpleHashRecord());
    record = record_if_unavailable.get(); // memorize ƤϤʤ
  }
  assert(hasLastRecord(1));
  const SimpleHashRecord *parent = lastRecord(1);
  SearchMove best_move_in_table;
  record->is_king_in_check = EffectUtil::isKingInCheck(P, state());
  // ơ֥ˤͤĴ٤
  if (record_memorize)
  {
    const TableHit hit = policy.isOutOfWindow(P, *record, curLimit(),
					      max_val, base_t::recorder);
    if (hit == LOWER_HIT)
      return max_val;
    Move checkmate_move=Move::INVALID();
    if ((!record->is_king_in_check)
	&& record->checkmateNodesLeft(1)
	&& (checkmate::ImmediateCheckmate::hasCheckmateMove<P>
	    (state(), checkmate_move)))
    {
      max_val= base_t::winByCheckmate(P);
      base_t::recordWinByCheckmate(P, record, checkmate_move);
      return max_val;
    }
#ifndef DONT_USE_CHECKMATE
    assert(record);
    // try simulation or simple checkmate search
    int additional_limit = 0;		// simulation only
    if (parent && parent->threatmate().mayHaveCheckmate(alt(P)))
    {
      additional_limit = record->checkmateNodesLeft(100);
    }
    base_t::recorder.gotoCheckmateSearch(state(), additional_limit);
    const bool win = isWinningState<P>(additional_limit, checkmate_move, 
				       record->qrecord.attack_oracle);
    base_t::recorder.backFromCheckmateSearch();
    if (win)
    {      
      assert(checkmate_move.isValid());
      max_val= base_t::winByCheckmate(P);
      base_t::recordWinByCheckmate(P, record, checkmate_move);
      setKillerMove(checkmate_move);
      return max_val;
    }
#endif
    if (hit == UPPER_HIT) 
      return max_val;
    assert(hit == NO_HIT);
    assert(max_val == base_t::minusInfty(P));
    best_move_in_table = record->bestMove();
  }

  search_assert(policy.window().isConsistent(P), lastMove());

#ifndef DONT_USE_CHECKMATE
  // ͤͤ
  if ((! record->is_king_in_check)
      && (! (record && record->threatmate().isThreatmate(P)))
      && tryThreatmate())
  {
    int threatmate_limit = std::max(4500-curDepth()*1000, 100);
    if (record_memorize)
    {
      threatmate_limit = record->threatmateNodesLeft(threatmate_limit);
    }
    else
    {
      threatmate_limit /= 2;		// for safety against multiple calls
    }
    Move threatmate_move;
    base_t::recorder.gotoCheckmateSearch(state(), threatmate_limit);
    const bool threatmate
      = isThreatmateState<P>(threatmate_limit, threatmate_move, 
			     record->qrecord.threatmate_oracle);
    base_t::recorder.backFromCheckmateSearch();
    if (threatmate)
    {
      record->threatmate().setThreatmate(P, threatmate_move);
#ifndef NDEBUG
      if (base_t::table->isVerbose() && curDepth() < 3)
      {
	std::cerr << "threatmate by " << record::csa::show(threatmate_move)
		  << " ";
	history().dump(curDepth());
      }
#endif
    }
  }
#endif

  /** ;Ϥ root β礭ˡޤ꿼oracle  */
  int checkmate_limit = 1;
  best_move = SearchMove();	// invalidate
  // õ ThreatmateState 
  record->threatmate().update(P, (parent ? &(parent->threatmate()) : 0), 
			      EffectUtil::isKingInCheck(P, state()));

  // ơ֥ˤ
  if (base_t::validTableMove(state(), best_move_in_table.moveLogProb(), 
			     curLimit()))
  {
    MoveLogProb ml=best_move_in_table.moveLogProb();
    assert(ml.getLogProb() >= Probabilities_t::TableMove);
    if (Policy::isBestMoveExtension)
      ml.setLogProbAtMost(Probabilities_t::ReSearch);
    SearchMove trying_move(ml, best_move_in_table.record);
    SearchMoveVector moves;
    moves.push_back(&trying_move);
    const bool try_next = examineMoves<P,Policy>
      ("table", moves, policy, best_move, max_val);
    if (! try_next)
    {
      assert(best_move.validMove());
      assert(! base_t::isWinValue(alt(P),max_val));
      goto register_table;
    }
  }
  // QuiescenceSearch best move
  if (curDepth() <= 1)
  {
    quiescenceValue<P>(base_t::winThreshold(alt(P)), base_t::winThreshold(P));
  }
  else
  {
#if defined(GUIDE_BY_QSEARCH_FULL)
    // full window
    quiescenceValue<P>(base_t::winThreshold(alt(P)), base_t::winThreshold(P), 4);
#elif defined(GUIDE_BY_QSEARCH_LIMITED)
    // limited window
    const int pawn_value2 = Eval::captureValue(newPtypeO(alt(P),PAWN));
    const int current = policy.window().beta();
    int alpha = policy.window().alpha();
    int beta = policy.window().beta();
    if (EvalTraits<P>::notLessThan(alpha, beta))
    {
      const std::pair<int,int> alpha_beta
	= NullWindowToQWindow::halfPawn(current, pawn_value2*4);
      alpha = eval::max(P, FixedEval::winThreshold(alt(P)), 
			alpha_beta.first);
      beta = eval::min(P, FixedEval::winThreshold(P), 
		       alpha_beta.second);
    }
    quiescenceValue<P>(alpha, beta);
#endif
  }
  // killer move 
  if ((! record->is_king_in_check)	// TODO: Ϥ
      && (curLimit() >= Probabilities_t::KillerMove))
  {
    // killer move
    MoveLogProbVector moves;
    MoveVector killer_moves;
    getKillerMoves(killer_moves);
    for (MoveVector::const_iterator p=killer_moves.begin(); 
	 p!=killer_moves.end(); ++p)
    {
      assert(p->player() == state().getTurn());
      assert(p->player() == P);
      moves.push_back(MoveLogProb(*p, Probabilities_t::KillerMove));
    }
    // QuiescenceSearch best move
    if (record->qrecord.bestMove().isNormal())
    {
      const MoveLogProb quiescence_move(record->qrecord.bestMove(),
					Probabilities_t::KillerMove);
      moves.push_back(quiescence_move);
    }
    if (! moves.empty())
    {
      const bool try_next 
	= examineMoves<P,Policy,false>
	("killer move", moves, policy, best_move, max_val,record->moves());
      if (! try_next)
      {
	assert(best_move.validMove());
	assert(! base_t::isWinValue(alt(P),max_val));
	goto register_table;
      }
    }
  }

  testTimeOut();

#ifndef DONT_USE_CHECKMATE
  if (root_limit >= 500)
  {
    int depth = curDepth();
    if (record_memorize && (depth <= 1))
      checkmate_limit = checkmate::limitToCheckCount(3500);
    else if (record_memorize && (depth == 2))
      checkmate_limit = 1000;
    else if (record_memorize && (depth == 3))
      checkmate_limit = 200;
    else if (((root_limit - curLimit()) <= 600) || (depth <= 5))
    {
      assert(static_cast<unsigned int>(curLimit()) < 4096);
      checkmate_limit = checkmate::limitToCheckCount(curLimit());
    }
    if (record->threatmate().mayHaveCheckmate(alt(P))
	|| (parent && parent->threatmate().maybeThreatmate(alt(P))))
      checkmate_limit += std::max(100, checkmate_limit);
    if (root_limit < 600)
      checkmate_limit /= 2;
  }
  if (record_memorize)
  {
    // ;ϤٿȤƤ
    checkmate_limit = record->checkmateNodesLeft(checkmate_limit);
    if (checkmate_limit <= 0)
      goto checkmate_done;
  }
  else
  {
    checkmate_limit /= 2;		// for safety against multiple calls
  }
  
  // ͤޤȤͤ 
  {
    Move checkmate_move=Move::INVALID();
    base_t::recorder.gotoCheckmateSearch(state(), checkmate_limit);
    const bool win = isWinningState<P>
      (checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
    base_t::recorder.backFromCheckmateSearch();
    if (win)
    {
      assert(checkmate_move.isValid());
      max_val= base_t::winByCheckmate(P);
      base_t::recordWinByCheckmate(P, record, checkmate_move);
      setKillerMove(checkmate_move);
      return max_val;
    }
  }
checkmate_done:
#endif
  search_assert(policy.window().isConsistent(P), lastMove());
  if (curLimit() < 200)
    goto move_generation_failure;
    
  // ˥ơ֥Ͽ줿
  if (record_memorize && (! record->moves().hasOnlyPass()))
  {
    SearchMoveVector moves;
    assert(record->moves().isValidAll());
    for (SearchMoveSet::const_iterator p=record->moves().begin(); 
	 p!=record->moves().end(); ++p)
    {
      if (p->getLogProb() <= curLimit())
	moves.push_back(&*p);
    }
    // ɾͽ¤٤
    //
    const int default_value 
      = ((policy.window().alpha()+policy.window().beta())/2 
	 + Eval::captureValue(newPtypeO(P,PAWN)));
    moves.sortByValue<P>(default_value);
    const bool try_next
      = examineMoves<P,Policy>("recorded", moves, policy, best_move, max_val);
    if (! try_next)
    {
      assert(best_move.validMove());
      assert((! base_t::isWinValue(alt(P),max_val)));
      goto register_table;
    }
  }
  search_assert(policy.window().isConsistent(P), lastMove());
  // ƥ˻
  {
    const category::CategoryEnv env(&state(), curLimit(), &history(),
				    record, curDepth(), &historyTable());
    bool cut_occurred;
    category::CategoryFlags& category_cache = 
      record->prepareGenerate(curLimit(), ReSearchLimitMargin);
    cut_occurred = CategoryListUtil::forEachCategory
      (MoveGenerator(), env,
       framework::CategoryHelper<P,Policy>
       (policy, &best_move, &max_val, &record->moves()), category_cache);
    if (! best_move.validMove())
    {
      assert(! cut_occurred);
      search_assert(policy.window().isConsistent(P), lastMove());
      goto move_generation_failure;
    }
  }
  if (record->moves().hasOnlyPass())
  {
    // goto quiescence search if tried move is only null move
    //  !policy.window().isConsistent(P) ξ礢
    goto move_generation_failure;
  }
register_table:
  assert(best_move.validMove());
  assert(eval::isConsistentValue(max_val));

  if (base_t::isWinValue(alt(P),max_val))
  {
    // ;ʳalt(P)ξ => ζ̤ɬ
    // ⤦١ͤޤȤͤ 
    {
      Move checkmate_move=Move::INVALID();
      checkmate_limit += 100;
      if (record_memorize)
	checkmate_limit = record->checkmateNodesLeft(checkmate_limit*2);
      base_t::recorder.gotoCheckmateSearch(state(), checkmate_limit);
      const bool win = isWinningState<P>
	(checkmate_limit, checkmate_move, record->qrecord.attack_oracle);
      base_t::recorder.backFromCheckmateSearch();
      if (win)
      {
	assert(checkmate_move.isValid());
	max_val= base_t::winByCheckmate(P);
	base_t::recordWinByCheckmate(P, record, checkmate_move);
	setKillerMove(checkmate_move);
	return max_val;
      }
    }
    // TODO: ľ꤬(䤱)Ϣ³ä
    // ɬǤϤʤƵͰɤǽϤ
    // դˡޤȤʲä礤Ƥɤ
    max_val = brinkmatePenalty(P, curLimit()) + eval.value();
  }
  else
  {
    setKillerMove(best_move.getMove());
  }
  if (record_memorize)
    policy.recordToTable(P, record_memorize, curLimit(), best_move, max_val);
  return max_val;
move_generation_failure:
  // Ǥʤä
  // limit  200 ̤äȡʾä move generator  limit ʲμǤʤä礳
  max_val = quiescenceValue<P,Policy>(policy);
  if (record_memorize)
  {
    // TODO: 礤λؼϿƤȼiteration?
    policy.recordToTable(P, record_memorize, curLimit(), SearchMove(), max_val);
  }
  assert(eval::isConsistentValue(max_val));
  return max_val;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P>
int osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
quiescenceValue(int alpha, int beta, int depth)
{
  testTimeOut();
  typedef QuiescenceSearch<eval_t> qsearcher_t;
  qsearcher_t qs(static_cast<SearchStateCore&>(*this), *base_t::table);
  Move last_move = lastMove();
  if (last_move.isInvalid())
    last_move = Move::PASS(alt(P));
  int result;
#ifdef DONT_USE_QSEARCH_PROBECUT
  const bool probecut_suitable = false;
#else
  const Position my_king = state().getKingPosition(P);
  const bool probecut_suitable 
    = (! last_move.isNormal()) 
    || (! Neighboring8Direct::hasEffect(state(), last_move.ptypeO(), 
					last_move.to(), my_king));
#endif
  if (probecut_suitable)
  {
    if (depth >= 8)
      result = qs.template searchProbCut<P>(alpha, beta, eval, last_move);
    else
      result = qs.template search<P>(alpha, beta, eval, last_move, depth);
  }
  else
  {
    result = qs.template search<P>(alpha, beta, eval, last_move, depth);
  }
  base_t::recorder.addQuiescenceCount(qs.nodeCount());
  return result;
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
template <osl::Player P, class Policy>
inline
int osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
quiescenceValue(const Policy& policy)
{
  typedef typename Policy::window_t window_t;
  const std::pair<int,int> alpha_beta
    = (policy.window().isConsistent(P) 
       ? QuiescenceWindow<window_t>::make(P, policy.window(), eval)
       : QuiescenceWindow<NullWindow>::make(P, NullWindow(policy.window().beta()), eval));
  return quiescenceValue<P>(alpha_beta.first, alpha_beta.second);
}

namespace osl
{
  namespace search
  {
    template <Player P, class Searcher>
    struct NextQuiescenceFramework
    {
      Searcher *searcher;
      const int alpha, beta;
      int result;
      void operator()(Position)
      {
	result = 
	  searcher->template quiescenceValue<P>(alpha, beta);
      }
    };
  }
}

template <class Eval,typename MoveGenerator, typename Table, typename Recorder, typename Probabilities>
bool osl::search::SearchFramework<Eval,MoveGenerator,Table,Recorder,Probabilities>::
isReasonableMove(Move move, int pawn_sacrifice)
{
  const Player turn = state().getTurn();
  assert(move.player() == turn);
  const int alpha = base_t::winThreshold(alt(turn));
  const int beta  = base_t::winThreshold(turn);
  const int current = (turn == BLACK) 
    ? quiescenceValue<BLACK>(alpha, beta)
    : quiescenceValue<WHITE>(alpha, beta);
  const int pawn_value = Eval::captureValue(newPtypeO(turn, PAWN))*pawn_sacrifice;
  
  const HashKey new_hash = currentHash().newHashWithMove(move);
  int result;
  SimpleHashRecord *record = 0;
  if (turn == BLACK)
  {
    NextQuiescenceFramework<WHITE,SearchFramework> helper = {this, beta, alpha};
    doUndoMoveOrPass<BLACK>(new_hash, move, &record, helper);
    result = helper.result;
  }
  else
  {
    NextQuiescenceFramework<BLACK,SearchFramework> helper = {this, beta, alpha};
    doUndoMoveOrPass<WHITE>(new_hash, move, &record, helper);
    result = helper.result;
  }
  return eval::betterThan(turn, result, current + pawn_value);
}

/* ------------------------------------------------------------------------- */

#endif /* _SEARCH_FRAMEWORK_TCC */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
