/* searchWindow.h
 */
#ifndef _SEARCHWINDOW_H
#define _SEARCHWINDOW_H

#include "osl/search/simpleHashRecord.h"
#include "osl/eval/evalTraits.h"
namespace osl
{
  namespace search
  {
    enum TableHit { NO_HIT=0, LOWER_HIT, UPPER_HIT };
    struct AlphaBetaWindow
    {
      int alpha_value, beta_value;
      explicit AlphaBetaWindow(int a, int b) : alpha_value(a), beta_value(b)
      {
	assert(a % 2);
	assert(b % 2);
      }
      int& alpha() { return alpha_value; }
      int& beta()  { return beta_value; }
      int alpha() const { return alpha_value; }
      int beta()  const { return beta_value; }
      const AlphaBetaWindow flipPlayer() const
      {
	return AlphaBetaWindow(beta(), alpha());
      }
      bool isConsistent(Player P) const
      {
	return eval::notLessThan(P, beta(), alpha());
      }
      void dump() const;
    };

    template <Player P, class EvalBase>
    struct AlphaBetaWindowUtil
    {
      typedef typename EvalBase::eval_t eval_t;
      /**
       * NullWindow ξȰä window 򶹤뤳Ȥ롥
       */
      template <class Recorder>
      static
      TableHit isOutOfWindow(const SimpleHashRecord& record, int limit, 
			     AlphaBetaWindow& w, int& val,
			     const Recorder& recorder) 
      {
	int table_value;
	if (record.template hasGreaterLowerBound<P>(limit, w.alpha(), 
						    table_value))
	{
	  assert(eval::isConsistentValue(table_value));
	  w.alpha() = table_value + EvalTraits<P>::delta;
	  if (EvalTraits<P>::betterThan(table_value, w.beta()))
	  {
	    recorder.tableHitLowerBound(P, table_value, w.beta(), limit);
	    val = table_value;
	    return LOWER_HIT;
	  }
	} 
	if (record.template hasLesserUpperBound<P>(limit, w.beta(), table_value))
	{
	  assert(eval::isConsistentValue(table_value));
	  w.beta() = table_value - EvalTraits<P>::delta;
	  if (EvalTraits<P>::betterThan(w.alpha(), table_value))
	  {
	    recorder.tableHitUpperBound(P, table_value, w.alpha(), limit);
	    val = table_value;
	    return UPPER_HIT;
	  }
	}
	return NO_HIT;
      }
    };
    

    struct NullWindow
    {
      int value;

      explicit NullWindow(int v) :value(v)
      {
      }
      int& alpha() { return value; }
      int& beta()  { return value; }
      int alpha() const { return value; }
      int beta()  const { return value; }
      bool isConsistent(Player) const { return true; }
      void dump() const;
    };

    template <Player P, class EvalBase, bool best_move_extension>
    struct NullWindowUtil
    {
      typedef typename EvalBase::eval_t eval_t;
      /** 
       * Record ٤ cut Ǥ뤫ɤȽꤹ.
       * @return cut Ǥ뤫ɤ
       * @param val cut  upper/lower bound 
       * @param best_move_extension ξ硤window ۤƤƤ
       *   ʤ꿼ɤ̤Ǥʤ cut ʤƱõ
       *   줫ɤγΨ⤯Ȥ뤳ȤǱĹ˻Ѥ뤿
       */
      template <class Recorder>
      static
      TableHit isOutOfWindow(const SimpleHashRecord& record, int limit, 
			     NullWindow w,
			     int& val, const Recorder& recorder)
      {
	const int lookUpLimit = (best_move_extension ? limit + 200 : limit);
	int table_value;
	if (record.template hasGreaterLowerBound<P>
	    (lookUpLimit, w.beta(), table_value))
	{
	  assert(eval::isConsistentValue(table_value));
	  recorder.tableHitLowerBound(P, table_value, w.beta(), limit);

	  val = table_value;
	  return LOWER_HIT;
	} 
	if (record.template hasLesserUpperBound<P>
	    (lookUpLimit, w.alpha(), table_value))
	{
	  assert(eval::isConsistentValue(table_value));
	  recorder.tableHitUpperBound(P, table_value, w.alpha(), limit);

	  val = table_value;
	  return UPPER_HIT;
	}
	return NO_HIT;
      }

    };
    
  } // namespace search
} // namespace osl


#endif /* _SEARCHWINDOW_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
