/* simpleHashRecord.h
 */
#ifndef _SEARCH_SIMPLEHASHRECORD_H
#define _SEARCH_SIMPLEHASHRECORD_H

#include "osl/search/quiescenceRecord.h"
#include "osl/search/searchTable.h"
#include "osl/search/searchMoveSet.h"
#include "osl/search/searchMove.h"
#include "osl/checkmate/oracleAges.h"
#include "osl/eval/evalTraits.h"
#include "osl/category/categoryFlags.h"
#include <cassert>

namespace osl
{
  namespace search
  {
    /**
     * SimpleHashTable ˵Ͽ륨ȥ
     */
    class SimpleHashRecord
    {
      signed short upper_limit;
      signed short lower_limit;
      int upper_bound;
      int lower_bound;
      int checkmate_nodes;
      SearchMove best_move;
      /** ޤǤ */
      SearchMoveSet move_cache;
      int threatmate_nodes;
      /** moves ä˻Ȥäƥ */
      category::CategoryFlags used_categories;
      /** moves 򹹿Ȥlimit */
      signed short limit_of_moves;
    public:
      /** 礤ǻȤǡ */
      QuiescenceRecord qrecord;
      bool is_king_in_check;

      SimpleHashRecord() 
	: upper_limit(-1), lower_limit(-1), checkmate_nodes(0), 
	  threatmate_nodes(0), limit_of_moves(-1), is_king_in_check(false)
      {
      }
      /**
       * @param max recordǻȤäɤΡɿ
       * @return ;˻ȤΡɿ֤
       */
      int checkmateNodesLeft(int max)
      {
	if (max > checkmate_nodes)
	{
	  const int left = max - checkmate_nodes;
	  checkmate_nodes = max;
	  return left;
	}
	return 0;
      }
      /**
       * @param max recordǻȤäɤΡɿ
       * @return ͤǧε;˻ȤΡɿ֤
       */
      int threatmateNodesLeft(int max)
      {
	if (max > threatmate_nodes)
	{
	  const int left = max - threatmate_nodes;
	  threatmate_nodes = max;
	  return left;
	}
	return 0;
      }
      category::CategoryFlags& prepareGenerate(int new_limit, int margin)
      {
	if (limit_of_moves + margin < new_limit)
	{
	  used_categories.clear();
	  limit_of_moves = new_limit;
	}
	return used_categories;
      }

      /**
       * ͤ񤭹. ;кꡥ
       */
      void setAbsoluteValue(Move best_move, int value, int limit)
      {
	lower_limit = limit;
	lower_bound = value;
	upper_limit = limit;
	upper_bound = value;
	this->best_move = SearchMove(MoveLogProb(best_move, value));
      }

      /**
       * lowerBound .
       * - ͥ
       * - Ʊʤɤͥ
       */
      void setLowerBound(Player P, int limit, const SearchMove& best_move, 
			 int value)
      {
	assert((value % 2) == 0);
	assert(limit >= 0);
	if (limit < lower_limit)
	  return;
	if (best_move.validMove()
	    && ((! this->best_move.moveLogProb().validMove())
		|| (limit > lower_limit)
		|| eval::betterThan(P, value, lower_bound)))
	{
	  this->best_move= best_move;
	}
	lower_limit = limit;
	lower_bound = value;
	if (hasUpperBound(0))
	  makeConsistent<true>(P);
      }

      /**
       * upperBound .
       * - ͥ
       * - Ʊʤ鰭ͥ
       */
      void setUpperBound(Player P, int limit, const SearchMove& best_move, 
			 int value)
      {
	assert((value % 2) == 0);
	assert(limit >= 0);
	if (limit < upper_limit)
	  return;
	if (best_move.validMove()
	    && (! this->best_move.moveLogProb().validMove()))
	{
	  // best move upper boundǤϴŪˤϹʤ
	  this->best_move= best_move;
	}
	upper_limit = limit;
	upper_bound = value;
	if (hasLowerBound(0))
	  makeConsistent<false>(P);
      }
    private:
      /**
       * ¤Ȳ¤̷⤬äĴ.
       * Ʊä夫褿ͥ褷ʤȡơõη̤ϿǤʤ
       * @param PreferLowerBound Ʊä lower bound ͥ褹
       */
      template <bool PreferLowerBound>
      void makeConsistent(Player P)
      {
	assert(hasLowerBound(0) && hasUpperBound(0));
	if (! eval::betterThan(P, lower_bound, upper_bound))
	  return;
	if ((upper_limit < lower_limit)
	    || (PreferLowerBound && (upper_limit == lower_limit)))
	{
	  upper_bound = lower_bound;
	  assert((upper_bound % 2) == 0);
	}
	else
	{
	  lower_bound = upper_bound;
	  assert((lower_bound % 2) == 0);
	}
      }
    public:
      /** upperBound ϿȤlimit */
      int upperLimit() const { return upper_limit; }
      /** ֤Υץ쥤䤫ߤɾͤξ */
      int upperBound() const { return upper_bound; }
      /** lowerBound ϿȤlimit */
      int lowerLimit() const { return lower_limit; }
      /** ֤Υץ쥤䤫ߤɾͤβ */
      int lowerBound() const { return lower_bound; }
      /** ޤǤ˵;õΡɿ */
      int checkmateNodes() const { return checkmate_nodes; }
      const SearchMove& bestMove() const { return best_move; }
      SearchMoveSet& moves() { return move_cache; }
      const SearchMoveSet& moves() const { return move_cache; }

      bool hasUpperBound(int limit) const
      {
	return (upperLimit() >= limit);
      }
      bool hasLowerBound(int limit) const
      {
	return (lowerLimit() >= limit);
      }
      bool hasAbsoluteUpperBound(Player p, int limit) const
      {
	return (p == BLACK) ? hasUpperBound(limit) : hasLowerBound(limit);
      }
      bool hasAbsoluteLowerBound(Player p, int limit) const
      {
	return (p == BLACK) ? hasLowerBound(limit) : hasUpperBound(limit);
      }
      int absoluteUpperBound(Player p) const
      {
	return (p == BLACK) ? upperBound() : lowerBound();
      }
      int absoluteLowerBound(Player p) const
      {
	return (p == BLACK) ? lowerBound() : upperBound();
      }
      void resetValue()
      {
	lower_limit = -1;
	upper_limit = -1;
      }
      /** high fail 뤫 */
      template <Player P>
      bool hasGreaterLowerBound(int limit, int threshold, int& val) const
      {
	if ((lowerLimit() < limit)
	    || (EvalTraits<P>::betterThan(threshold, lowerBound())))
	  return false;
	val = lowerBound();
	return true;
      }
      /** low fail 뤫 */
      template <Player P>
      bool hasLesserUpperBound(int limit, int threshold, int& val) const
      {
	if ((upperLimit() < limit)
	    || (EvalTraits<P>::betterThan(upperBound(), threshold)))
	  return false;
	val = upperBound();
	return true;
      }

      const DualThreatmateState& threatmate() const { return qrecord.threatmate; }
      DualThreatmateState& threatmate() { return qrecord.threatmate; }

      void dump(std::ostream&) const;

      /** SimpleHashTable::migrate  */
      void copyFrom(const SimpleHashRecord&);
      /** SimpleHashTable::migrate  */
      void fixBestMove();
    };
  } // namespace search

  using search::SimpleHashRecord;
} // namespace osl


#endif /* _SEARCH_SIMPLEHASHRECORD_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
