/* simpleState.h
 */
#ifndef _SIMPLE_STATE_H
#define _SIMPLE_STATE_H

#include "osl/misc/loki.h"
#include "osl/direction.h"
#include "osl/boardTable.h"
#include "osl/ptype.h"
#include "osl/ptypeTraits.h"
#include "osl/piece.h"
#include "osl/container/pieceMask.h"
#include "osl/effectContent.h"
#include "osl/move.h"
#include "osl/player.h"
#include "osl/handicap.h"
#include "osl/misc/carray.h"
#include "osl/effect_action/pieceFilter.h"
#include "osl/apply_move/applyDoUndoXMove.h"
#include "osl/apply_move/doUndoMoveLockForward.h"
#include "osl/ptypeTable.h"

#include <iosfwd>

namespace osl
{
  namespace state
  {
    class SimpleState;
    std::ostream& operator<<(std::ostream& os,const SimpleState& state);
    bool operator==(const SimpleState& st1,const SimpleState& st2);

    class SimpleState
    {
      /**
       * ˵Ȥʤ functor
       */
      class NotEmpty{
      public:
	bool operator()(Position) const{ return false; }
      };
    public:
      /** Ѿ饹 ĤФ˻Ѥ */
      typedef SimpleState simple_board_t;
      typedef SimpleState simple_state_t;
      typedef SimpleState effect_state_t;
    private:
      friend std::ostream& operator<<(std::ostream& os,const SimpleState& state);
      friend bool operator==(const SimpleState& st1,const SimpleState& st2);
      typedef SimpleState state_t;
    public:
      static const bool hasPawnMask=false;
    protected:
      PieceMask used_mask;
      CArray<PieceMask,2> stand_mask;
      CArray<Piece,Position::SIZE> board;
      /**
       * ƤpieceϿƤ
       */
      CArray<Piece,Piece::SIZE> pieces;

      /** . ĹŪˤϾä⤷ʤʬɬ */
      Player turn;

      friend class osl::apply_move::ApplyDoUndoSimpleMove<BLACK,SimpleState>;
      friend class osl::apply_move::ApplyDoUndoSimpleMove<WHITE,SimpleState>;
      friend class osl::apply_move::ApplyDoUndoCaptureMove<BLACK,SimpleState>;
      friend class osl::apply_move::ApplyDoUndoCaptureMove<WHITE,SimpleState>;
      friend class osl::apply_move::ApplyDoUndoDropMove<BLACK,SimpleState>;
      friend class osl::apply_move::ApplyDoUndoDropMove<WHITE,SimpleState>;
      friend class osl::apply_move::DoUndoMoveLockSimple<SimpleState>;
      friend class osl::apply_move::DoUndoMoveLockCapture<SimpleState>;
      friend class osl::apply_move::DoUndoMoveLockDrop<SimpleState>;
    public:
      // ˴ؤ
      explicit SimpleState();
      explicit SimpleState(Handicap h);
      // publicѾˤϡvirtual destructor롥
      virtual ~SimpleState();
      /** ̤ξ֤˽ */
      void init();
      /** ϥǥ˱֤˽ */
      void init(Handicap h);
      const Piece SimpleState::getPieceOf(int num) const{
	return pieces[num];
      }
      void setPieceOf(int num,Piece p) {
	pieces[num]=p;
      }
      template<Player P>
      const Piece getKingPiece() const{
	return getPieceOf(KingTraits<P>::index);
      }
      const Piece getKingPiece(Player P) const{
	assert(isValid(P));
	if (P==BLACK)
	  return getKingPiece<BLACK>();
	else
	  return getKingPiece<WHITE>();
      }
      template<Player P>
      Position getKingPosition() const{
	return getKingPiece<P>().position();
      }
      Position getKingPosition(Player player) const{
	assert(isValid(player));
	if (player==BLACK)
	  return getKingPosition<BLACK>();
	else
	  return getKingPosition<WHITE>();
      }

      void setBoard(Position pos,Piece piece)
      {
	board[pos.index()]=piece;
      }
    protected:
      PieceMask& standMask(Player p) {
	return stand_mask[playerToIndex(p)];
      }
    public:
      const PieceMask& standMask(Player p) const {
	return stand_mask[playerToIndex(p)];
      }
      const PieceMask& usedMask() const {return used_mask;}
      bool isOffBoard(int num) const{
	return standMask(BLACK).test(num) 
	  || standMask(WHITE).test(num);
      }
    
      template<Player P>
      bool isPawnMaskSet(int x) const{
	Position pos(x,(P==BLACK ? 2 : 1));
	Piece p1=getPieceAt(pos);
	Piece p2=getPieceAt(pos+DirectionTraits<D>::blackOffset());
	Piece p3=getPieceAt(pos+DirectionTraits<D>::blackOffset()*2);
	Piece p4=getPieceAt(pos+DirectionTraits<D>::blackOffset()*3);
	Piece p5=getPieceAt(pos+DirectionTraits<D>::blackOffset()*4);
	Piece p6=getPieceAt(pos+DirectionTraits<D>::blackOffset()*5);
	Piece p7=getPieceAt(pos+DirectionTraits<D>::blackOffset()*6);
	Piece p8=getPieceAt(pos+DirectionTraits<D>::blackOffset()*7);
	return (p1.ptypeOMask(newPtypeO(P,PAWN))
		& p2.ptypeOMask(newPtypeO(P,PAWN))
		& p3.ptypeOMask(newPtypeO(P,PAWN))
		& p4.ptypeOMask(newPtypeO(P,PAWN))
		& p5.ptypeOMask(newPtypeO(P,PAWN))
		& p6.ptypeOMask(newPtypeO(P,PAWN))
		& p7.ptypeOMask(newPtypeO(P,PAWN))
		& p8.ptypeOMask(newPtypeO(P,PAWN)))==0;
      }
      bool isPawnMaskSet(Player player, int x)const {
	if (player==BLACK)
	  return isPawnMaskSet<BLACK>(x);
	else
	  return isPawnMaskSet<WHITE>(x);
      }


      void setPiece(Player player,Position pos,Ptype ptype);
      void setPieceAll(Player player);

      /**
       * @param pos  isOnboard Position 12˵(8˵+Ϥ)
       * ! isOnBoard(pos) ξ PIECE_EDGE ֤
       */
      const Piece getPieceAt(Position pos) const { return board[pos.index()];}
      const Piece getPieceOnBoard(Position pos) const
      {
	assert(pos.isOnBoard());
	return getPieceAt(pos);
      }

      bool isOnBoard(int num) const {
	return getPieceOf(num).isOnBoard();
      }
      template<Player P,Ptype T>
      bool hasPieceOnStand() const {
	return (standMask(P).getMask(PtypeFuns<T>::indexNum) 
		& mask_t::makeDirect(PtypeFuns<T>::indexMask)).any();
      }
      bool hasPieceOnStand(Player player,Ptype ptype) const{
#if OSL_WORDSIZE == 64
	/** 
	 * Ptype_Table.getIndex(ptype)constant 0֤褦ѹΤ
	 * ʬפΤϤ
	 */
	return (standMask(player).getMask(0)
		& Ptype_Table.getMaskLow(ptype)).any();
#elif OSL_WORDSIZE == 32
	return (standMask(player).getMask(Ptype_Table.getIndex(ptype))
		& Ptype_Table.getMaskLow(ptype)).any();
#endif
      }
      /**
       * 
       */
      int countPiecesOnStand(Player pl,Ptype ptype) const{
	return (standMask(pl).getMask(Ptype_Table.getIndex(ptype))
		& Ptype_Table.getMaskLow(ptype)).countBit();
      }
      /**
       * diffˤPiece . 
       * @return ׳ʤPTYPE_EDGE
       */
      Piece nextPiece(Position cur, Offset diff) const
      {
	assert(! diff.zero());
	cur += diff;
	while (getPieceAt(cur) == Piece::EMPTY())
	  cur += diff;
	return getPieceAt(cur);
      }
    
      void setTurn(Player player) {
	turn=player;
      }
      Player getTurn() const{
	return turn;
      }
      /**
       * ֤ѹ
       */
      void changeTurn() {
	osl::changeTurn(turn);
      }

      // check
      bool isConsistent(bool showError=true) const;
      /** 顼ɽ򤹤뤫ɤtemplateѥ᡼ˤ® */
      template <bool showError>
      bool isAlmostValidMove(Move move) const;
      /**
       * ˡ꤫ɤñ˸롥̤˰¸åΤߡ
       * 롼ؤʤǤǽϡisValidMove Ѥ롥
       *
       * ̤˰¸븡Ǥ⡤̤ȴ䲦ɤǤ뤫
       * ꡤͤɤϸʤ
       */
      bool isAlmostValidMove(Move move,bool showError=true) const;
      /**
       * ˡ꤫ɤ̩˸롥isAlmostValidMove θޤࡥ
       */
      bool isValidMove(Move move,bool showError=true) const;

      /**
       * ̰ʳʬȿ§Υå
       *
       * static ˤǤ뤬ޤǤΤǸ
       */
      bool isValidMoveByRule(Move move,bool showError) const;

      /**
       * @param from - ޥΰ
       * @param to - ޥΰ
       * @param offset - fromtoؤshort offset
       * fromtoִطˤȤ
       * ǡ֤򤫤å
       * @param pieceExistsAtTo - toɬ𤬤 (toǤư)
       */
      bool isEmptyBetween(Position from, Position to,Offset offset,bool pieceExistsAtTo=false) const
      {
	assert(from.isOnBoard());
	assert(! offset.zero());
	assert(offset==Board_Table.getShortOffset(Offset32(to,from)));
	Position pos=from+offset;
	for (; getPieceAt(pos).isEmpty(); pos+=offset) {
	  if (!pieceExistsAtTo && pos==to) 
	    return true;
	}
	return pos==to;
      
      }
      /**
       * @param from - ޥΰ
       * @param to - ޥΰ
       * fromtoִطˤȤ
       * ǡ֤򤫤å
       */
      bool isEmptyBetween(Position from, Position to,bool noSpaceAtTo=false) const{
	assert(from.isOnBoard());
	Offset offset=Board_Table.getShortOffset(Offset32(to,from));
	assert(! offset.zero());
	return isEmptyBetween(from,to,offset,noSpaceAtTo);
      }
      /**
       * attackerptypeoζ𤬤Ȳꤷtarget뤫ɤ
       *  stateupdateʤǳΤ.
       * targetPosition϶Ǥɤ
       * ׾ζưƤƤȤϤˡ
       * ʬȤαäʤȸʤ뤳Ȥ
       */
      bool hasEffectFromTo(PtypeO ptypeo,Position attacker,
			   Position target) const
      {
	Offset32 offset32=Offset32(target,attacker);
	EffectContent effect=Ptype_Table.getEffect(ptypeo,offset32);
	if (! effect.hasEffect()) 
	  return false;
	if (effect.hasUnblockableEffect())
	  return true;
	assert(Board_Table.getShortOffset(offset32) == effect.offset());
	return isEmptyBetween(attacker,target,effect.offset());
      }
      /**
       * attackerPiecetargetPositionĤɤ.
       * targetPosition϶Ǥɤ
       * @param attackerPiece
       * @param targetPosition
       */
      bool hasEffectTo(Piece attackerPiece,Position targetPosition) const{
	assert(attackerPiece.isPiece() && targetPosition.isOnBoard());
	return hasEffectFromTo(attackerPiece.ptypeO(),
			       attackerPiece.position(),
			       targetPosition);
      }

      /**
       * μ꤬꤫ɤȽꤹ.
       *
       * State(e.g., NumEffectState)ξϡ
       * ñhasEffectByȤɤ.
       *
       * Ĥζˤ벦ξattckPieceˤζ
       * ʣζˤ벦ξPiece::EMPTY()
       * fromoffboardλ̵false֤
       * @param P(template) - ߤμ(ä, movealt(P))
       * @param setPiece(template) - attackPieceͤ뤫ɤ
       * safeMove::isSafeMoveȶͭǤȤ¿
       */
      template<Player P,bool setPieceP>
      bool lastMoveIsOpenCheck(Position from,Position kingPosition,Piece &attackPiece) const{
	if (from.isPieceStand()) 
	  return false;
	Offset offset=Board_Table.getShortOffset(Offset32(from,kingPosition));
	if (offset.zero()
	    || !isEmptyBetween(kingPosition,from,offset)) 
	  return false;
	Piece p;
	Position pos;
	for (pos=from+offset;(p=getPieceAt(pos)).isEmpty();pos+=offset)
	  ;
	if (p.isOnBoardByOwner<PlayerTraits<P>::opponent>() 
	    && Ptype_Table.getEffect(p.ptypeO(),pos,kingPosition).hasEffect())
	  {
	    if (setPieceP) attackPiece=p;
	    return true;
	  }
	return false;
      }

      /**
       * μ꤬꤫ɤ.
       * Ĥζˤ벦ξattck_pieceˤζ
       * ʣζˤ벦ξPiece::EMPTY()
       * @param P(template) - ߤμ(ä, movealt(P))
       */
      template<Player P>
      bool lastMoveIsCheck(Move last_move, Piece& attack_piece) const{
	assert(last_move.isNormal());
	assert(last_move.player() == alt(P));
	attack_piece = Piece::EMPTY();
	const Position king_position = getPieceOf(KingTraits<P>::index).position();
	const Position last_to   = last_move.to();
	const Position last_from = last_move.from();
	const Piece last_piece=getPieceAt(last_to);
	if (hasEffectTo(last_piece, king_position)) {
	  if (! lastMoveIsOpenCheck<P,false>(last_move.from(), king_position,
					     attack_piece))
	    attack_piece=last_piece;
	  return true;
	}
	return lastMoveIsOpenCheck<P,true>(last_from,king_position,attack_piece);
      }

      // ֤
      //
      /** T  isBasic ǤʤƤư unpromote(T) η̤Ʊ */ 
      template<Player P,Ptype T,typename F>
      void forEachOnBoard(F & func) const {
	const int first = PtypeTraits<T>::indexMin;
	const int last  = PtypeTraits<T>::indexLimit;
	for (int num=first; num<last; ++num)
	  {
	    Piece p=getPieceOf(num);
	    if (p.isOnBoardByOwner<P>()) {
	      func(p);
	    }
	  }
      }
      /**
       *
       */
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position, Action&, Int2Type<false>) const {}
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position piecePosition,Action & action,Int2Type<true>) const {
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset();
	action.template doAction<P>(getPieceAt(piecePosition),piecePosition+offset);
      }

      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceDir(Position piecePosition,Action & action) const {
	forEachEffectOfPieceDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
      }
      /**
       *
       */
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position, Action&, Int2Type<false>) const {}
      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position piecePosition,Action & action,Int2Type<true>) const {
	Piece piece=getPieceAt(piecePosition);
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset();
	assert(offset.intValue() != 35);
	Position pos=piecePosition+offset;
	for (;getPieceAt(pos).isEmpty();pos+=offset)
	  action.template doAction<P>(piece,pos);
	action.template doAction<P>(piece,pos);
      }

      template<Player P,Ptype Type,class Action,Direction Dir>
      void forEachEffectOfPieceLongDir(Position piecePosition,Action & action) const {
	forEachEffectOfPieceLongDir<P,Type,Action,Dir>(piecePosition,action,Int2Type<(PtypeTraits<Type>::moveMask & DirectionTraits<Dir>::mask)!=0>());
      }
      /**
       * piecePositionˤˤä뤹٤Ƥposition
       * (ޤ)ˤĤ
       * action¹Ԥ
       */
      template<Player P,Ptype Type,class Action>
      void forEachEffectOfPiece(Position piecePosition,Action & action) const;
      /**
       * pieceˤä뤹٤Ƥposition(ޤˤĤ
       * action¹Ԥ
       */
      template<class Action>
      void forEachEffectOfPiece(Piece piece,Action & action) const;

      /**
       * @param P(template) - ¦
       * @param pos - оݤΥޥ
       * @param action - ¹Ԥ
       */
      template<Player P,class Action,bool checkDone,class PositionCheck>
      void forEachEffectWithCheck(Position pos,Action & action,PositionCheck const& positionCheck) const;
    
      template<Player P,class Action>
      void forEachEffect(Position pos,Action & action) const {
	forEachEffectWithCheck<P,Action,false,NotEmpty>(pos,action,NotEmpty());
      }
      /**
       * @param P(template) - ¦
       * @param target - ĤɸΥޥ
       */
      template<Player P>
      const Piece hasEffectPiece(Position target) const;
      /**
       * @param pl - ¦
       * @param target - ĤɸΥޥ
       */
      const Piece hasEffectPiece(Player pl,Position target) const;

      template<Player P>
      bool hasEffectBy(Position pos) const{
	return hasEffectPiece<P>(pos) != Piece::EMPTY();
      }
      bool hasEffectBy(Player player,Position pos) const{
	if (player==BLACK)
	  return hasEffectBy<BLACK>(pos);
	else
	  return hasEffectBy<WHITE>(pos);
      }
      /**
       * 
       */
      template<Player P>
      bool hasEffectByWithRemove(Position target,Position removed) const;
    
      /**
       * @param P(template) - ĤƤ¦Υץ쥤
       * @param target - Ĥ줿
       * @param attackerPiece - multiple attackξPiece::EMPTY()
       *        ǤʤʤĤƤ֤
       */
      template<Player P>
      bool hasEffectBy(Position target,Piece& attackerPiece) const;

      /**
       * ٤ƤɤȤǸƤ
       */
      template<Player P>
      bool hasMultipleEffectBy(Position target) const;
      bool hasMultipleEffectBy(Player player,Position target) const;
    
      template<Player P,class Action>
      void forEachEffectNotBy(Position pos,Piece p,Action & action) const {
	typedef PieceFilter<Action> action_t;
	action_t notByAction(action,p);
	forEachEffect<P,action_t>(pos,notByAction);
      }
      /** 
       * @param P(template) - ¦
       * @param piece - ¦ζ, ζоݳȤ
       * @param target - оݤΥޥ
       */
      template<Player P>
      bool hasEffectNotBy(Piece piece,Position target) const;

      /** 
       * @param player - ¦
       * @param piece - ¦ζ
       * @param target - оݤΥޥ
       */
      bool hasEffectNotBy(Player player,Piece piece,Position target) const {
	if (player==BLACK)
	  return hasEffectNotBy<BLACK>(piece,target);
	else
	  return hasEffectNotBy<WHITE>(piece,target);
      }


      template<Player P,class Action>
      void forEachEffectPtypeDirection(Position pos,Action & action,Direction dir,Ptype ptype,Int2Type<false> canMove) const{}

      template<Player P,class Action>
      void forEachEffectPtypeDirection(Position pos,Action & action,Direction dir,Ptype ptype,Int2Type<true> canMove) const{
	Position pos1=pos-Board_Table.template getOffset<P>(dir);
	Piece p=getPieceAt(pos1);
	PtypeO ptypeo=p.ptypeO();
	if (Ptype_Table.getShortMoveMask(P,ptypeo,dir)!=0 && unpromote(getPtype(ptypeo))==ptype) {
	  action.template doAction<P>(p,pos);
	}
      }

      template<Player P,Ptype Type,Direction Dir,class Action>
      void forEachEffectPtypeDirection(Position pos,Action & action) const{
	BOOST_STATIC_ASSERT((PtypeTraits<Type>::isBasic));
	forEachEffectPtypeDirection<P,Action>(pos,action,Dir,Type,Int2Type<PtypeDirectionTraits<Type,Dir>::canMove || PtypeDirectionTraits<PtypeFuns<Type>::promotePtype,Dir>::canMove>());
      }

      template<Player P,Ptype Type,class Action>
      void forEachEffectPtype(Position pos,Action & action) const;
      /**
       * ޥˤDirectionǤĹ뤫ɤ
       */
      template<Direction Dir,Player P>
      bool hasEffectDir(Position to) const{
	BOOST_STATIC_ASSERT((DirectionTraits<Dir>::isLong));
	const Offset offset=DirectionPlayerTraits<Dir,P>::offset();
	Piece p;
	for (Position pos=to-offset;(p=getPieceAt(pos)).isEmpty();pos-=offset)
	  ;
	if (p.isOnBoardByOwner<P>() 
	    && (Ptype_Table.getMoveMask(p.ptype())&DirectionTraits<Dir>::mask)!=0)
	  return true;
	return false;
      }


      /** dump: ʬ cerr ɽ롣abort ʤɤ˥ǥХå˻Ȥ */
      bool dump() const;
      void doSimpleMove(Position from, Position to, int promoteMask);
      void doDropMove(Position to,Ptype ptype);
      void doCaptureMove(Position from, Position to, Piece target,int promoteMask);
      /**
       * from ɽ줿Piecenew_ownerλˤ̤.
       */
      const SimpleState emulateCapture(Piece from, Player new_owner) const;

      /**
       * from to  ptypeλϤ̤.
       */
      const SimpleState emulateHandPiece(Player from, Player to, Ptype ptype) const;
    };  
  } // namespace state
  using state::SimpleState;

  namespace apply_move
  {
    template<Player P>
    struct ApplyDoUndoSimpleMove<P,SimpleState>
    {
      static void prologue(SimpleState& s, 
			   Position from, Position to, int promoteMask,
			   Piece& oldPiece, int& num)
      {
	oldPiece=s.getPieceAt(from);
	Piece newPiece=oldPiece.promoteWithMask(promoteMask);
	newPiece+=(to-from);
	num=oldPiece.number();
	s.setPieceOf(num,newPiece);
	s.setBoard(to,newPiece);
	s.setBoard(from,Piece::EMPTY());
      }
      static void epilogue(SimpleState& s, Position from, Position to,
			   Piece oldPiece, int num)
      {
	// ʬȤθ򳰤
	s.setPieceOf(num,oldPiece);
	s.setBoard(from,oldPiece);
	s.setBoard(to,Piece::EMPTY());
      }
      template <typename F>
      static void doUndoSimpleMove(SimpleState& s, 
				   Position from, Position to, int promoteMask,F& func);
    };
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoSimpleMove<P,SimpleState>::
    doUndoSimpleMove(SimpleState& s, 
		     Position from, Position to, int promoteMask,F& func)
    {
      Piece oldPiece;
      int num;
      prologue(s, from, to, promoteMask, oldPiece, num);
      s.changeTurn();
      func(to);
      s.changeTurn();
      epilogue(s, from, to, oldPiece, num);
    }

    template<Player P>
    struct ApplyDoUndoDropMove<P,SimpleState>
    {
      typedef SimpleState state_t;

      /**
       * OSL_WORDSIZE == 64λϡnumIndex, numLow
       * standMask¸ɤ
       */
      static void prologue(SimpleState& s, Ptype ptype, Position to,
			   Piece& oldPiece, int& num, int& numIndex, int& numLow)
      {
#if OSL_WORDSIZE == 64
	numIndex=0;
	const mask_t ownMochigoma=
	  s.standMask(P).getMask(0) & Ptype_Table.getMaskLow(ptype);
	assert(ownMochigoma.any() || (s.dump(), 0));
	num=numLow=ownMochigoma.bsf();
#elif OSL_WORDSIZE == 32
	numIndex=Ptype_Table.getIndex(ptype);
	const mask_t ownMochigoma=
	  s.standMask(P).getMask(numIndex) & Ptype_Table.getMaskLow(ptype);
	assert(ownMochigoma.any() || (s.dump(), 0));
	numLow=ownMochigoma.bsf();
	num=numLow|(numIndex<<5);
#endif
	oldPiece=s.getPieceOf(num);
	Piece p=oldPiece;
	p += to-Position::STAND();

	s.setPieceOf(num,p);
	s.setBoard(to,p);
	s.standMask(P).xorMask(numIndex,PieceMask::numToMask(numLow));
      }

      static void epilogue(SimpleState& s, Position to,
			   Piece oldPiece, int num,
			   int numIndex, int numLow)
      {
	s.standMask(P).xorMask(numIndex,PieceMask::numToMask(numLow));
	s.setBoard(to,Piece::EMPTY());
	s.setPieceOf(num,oldPiece);
      }
      template <typename F>
      static void doUndoDropMove(SimpleState& s, 
				 Position to, Ptype ptype,F& func);
    };
  
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoDropMove<P,SimpleState>::
    doUndoDropMove(SimpleState& s, 
		   Position to, Ptype ptype,F& func)
    {
      Piece oldPiece;
      int num, numIndex, numLow;
      prologue(s, ptype, to, oldPiece, num, numIndex, numLow);
      s.changeTurn();
      func(to);
      s.changeTurn();
      epilogue(s, to, oldPiece, num, numIndex, numLow);
    }

    template<Player P>
    struct ApplyDoUndoCaptureMove<P,SimpleState>
    {
      typedef SimpleState state_t;
      template <typename F>
      static void doUndoCaptureMove(state_t& s,
				    Position from,Position to, Piece p1, int promoteMask,F& func);

      static
      void prologue(SimpleState& s, 
		    Position from, Position to, Piece target, int promoteMask,
		    Ptype& capturePtype, Piece& oldPiece, int& num0, 
		    int& num1, int& num1Index, mask_t& num1Mask)
      {
	capturePtype=target.ptype();
	num1=target.number();
	num1Index=PieceMask::numToIndex(num1);
	num1Mask=PieceMask::numToMask(num1);

	s.standMask(P).xorMask(num1Index,num1Mask);
	oldPiece=s.getPieceAt(from);
	Piece newPiece=oldPiece.promoteWithMask(promoteMask);
	newPiece+=(to-from);
	num0=oldPiece.number();
	s.setPieceOf(num0,newPiece);
	s.setPieceOf(num1,target.captured());
	s.setBoard(to,newPiece);
	s.setBoard(from,Piece::EMPTY());
      }

      static 
      void epilogue(SimpleState& s, Position from, Position to, Piece target, 
		    Piece oldPiece, int num0, 
		    int num1, int num1Index, mask_t num1Mask)
      {
	s.standMask(P).xorMask(num1Index,num1Mask);
	s.setPieceOf(num0,oldPiece);
	s.setPieceOf(num1,target);
	s.setBoard(from,oldPiece);
	s.setBoard(to,target);
      }
    };
  
  
    template<Player P>
    template <typename F>
    void ApplyDoUndoCaptureMove<P,SimpleState>::
    doUndoCaptureMove(state_t& s,
		      Position from,Position to, Piece target, int promoteMask,F& func)
    {
      Piece oldPiece;
      int num0, num1;
      int num1Index;
      mask_t num1Mask;
      Ptype capturePtype;
      prologue(s, from, to, target, promoteMask,
	       capturePtype, oldPiece, num0, num1, num1Index, num1Mask);
      s.changeTurn();
      func(to);
      s.changeTurn();

      epilogue(s, from, to, target, oldPiece, num0, num1, num1Index, num1Mask);
    }

    template<> class DoUndoMoveLockSimple<SimpleState>;
    template<> class DoUndoMoveLockDrop<SimpleState>;
    template<> class DoUndoMoveLockCapture<SimpleState>;
  } // namespace apply_move
} // namespace osl

#endif /* _SIMPLE_STATE_H */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
