/* categoryMoveVector.cc
 */
#include "osl/category/analyzer/categoryMoveVector.h"
#include "osl/category/analyzer/probOfEachCategory.h"
#include "osl/category/categoryListUtil.h"
#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryUtil.h"

namespace osl
{
  namespace category
  {
    namespace analyzer
    {
      template void CategoryMoveVector::
      generateAndSet<StandardMoveGenerator>(const CategoryEnv&);
      template void CategoryMoveVector::
      generateAndSetDetailed<StandardMoveGenerator>(const CategoryEnv&);
    } // namespace analyzer
  } // namespace category
} // namespace osl

struct osl::category::analyzer::CategoryMoveVector::Helper
{
  CategoryMoveVector *out;
  template <bool is_king_escape>
  bool operator()(const char *category_name,
		  const MoveLogProbVector& moves)
  {
    if (! moves.empty())
    {
      out->push_front(CategoryMoves(moves, category_name));
      // is_king_escape ǻؼ꤬ФʹߤΥƥ
      return ! is_king_escape;	
    }
    return true;
  }
};

osl::category::analyzer::
CategoryMoves::CategoryMoves(const MoveLogProbVector& v, const std::string& n)
  : moves(v), category(n)
{
}

osl::category::analyzer::
CategoryMoveVector::CategoryMoveVector()
{
}

osl::category::analyzer::
CategoryMoveVector::~CategoryMoveVector()
{
}

struct osl::category::analyzer::CategoryMoveVector::DetailFinder
  : public osl::category::analyzer::ProbOfEachCategory::CallBack
{
  const int expected;
  std::string match;
  explicit DetailFinder(int e) : expected(e)
  {
  }
  void found(const char *name, int prob)
  {
    if (prob == expected)
      match = name;
  }
  void notFound()
  {
  }
};

template <class Categories>
void osl::category::analyzer::
CategoryMoveVector::generateAndSetDetailed(const CategoryEnv& env)
{
  // NOTE: inefficient
  CategoryMoveVector rough;
  rough.generateAndSet<Categories>(env);
  for (const_iterator p=rough.begin(); p!=rough.end(); ++p)
  {
    const std::string category = p->category;
    if (category == "Pass")	// FIXME
    {
      push_front(CategoryMoves(p->moves, category));
      continue;
    }
    for (MoveLogProbVector::const_iterator q=p->moves.begin();
	 q!=p->moves.end(); ++q)
    {
      DetailFinder callback(q->getLogProb());
      ProbOfEachCategory analyzer(env, q->getMove(), &callback);
      analyzer.analyze(StandardMoveGenerator());
      
      const std::string detailed = (category == callback.match) 
	? category : category + callback.match;
      iterator target = end();
      for (iterator r=begin(); r!=end(); ++r)
      {
	if (r->category == detailed)
	{
	  target = r;
	  break;
	}
      }
      if (target == end())
      {
	push_front(CategoryMoves(MoveLogProbVector(), detailed));
	target = begin();
      }
      target->moves.push_back(*q);
    }
  }
  reverse();
}

template <class Categories>
void osl::category::analyzer::
CategoryMoveVector::generateAndSet(const CategoryEnv& env)
{
  clear();
  Helper out = { this };
  CategoryListUtil::forEachCategory(Categories(), env, out);
  reverse();
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
