/* checkHashRecord.cc
 */
#include "osl/checkmate/checkHashRecord.h"
#include "osl/checkmate/sameBoardList.h"
// TODO: PAWN_CHECKMATE_SENSITIVE, DELAY_SACRIFICE ΤʤΤʬΥ
#include "osl/checkmate/checkMoveGenerator.h"
#ifdef CHECKMATE_DEBUG
#  include "osl/stat/ratio.h"
#endif

// TODO: 4 餤Ȥɤ餬ɤ?
#define MAX_CHECK_EXTRA_LIMIT_PROOF 1
#define MAX_CHECK_EXTRA_LIMIT_DISPROOF 1

// #define TWINLIST_STAT

#ifdef TWINLIST_STAT
#  include "osl/stat/histogram.h"
#endif
#include <iostream>

/** ȶ̤ͥ۴ط򵭲/ԵͤǤʤŤ */
#define PROOF_NUMBER_PROPAGATE_BY_DOMINANCE

static const unsigned int check_extra_limit_max = 8u;

namespace osl
{
  namespace checkmate
  {
    // explicit instantiation
    template void CheckHashRecord::propagateCheckmateRecursive<BLACK>();
    template void CheckHashRecord::propagateCheckmateRecursive<WHITE>();
    template void CheckHashRecord::propagateNoCheckmateRecursive<BLACK>();
    template void CheckHashRecord::propagateNoCheckmateRecursive<WHITE>();

    // ۤ褦allocator Ĵɬ
    BOOST_STATIC_ASSERT(sizeof(CheckHashRecord) + sizeof(int)*6 
			+ sizeof(void*) <= 192);
  }
}

bool osl::checkmate::CheckHashRecord::
isConsistent() const
{
  assert(! proofDisproof().isLoopDetection());
  static int last_type = UNSET;
  if (proof_pieces_type != UNSET)
  {
    if (! proof_disproof.isFinal())
    {
      std::cerr << "record->proof_pieces_type " << last_type
		<< " => " << (int)proof_pieces_type << "\n";
    }
    check_assert(proof_disproof.isFinal());
  }
  last_type = proof_pieces_type;
  return true;
}

osl::checkmate::CheckHashRecord::
~CheckHashRecord()
{
#if 0
  // final_by_dominance  dangling pointer ˤʤäƤǽ뤿
  // äƤϤʤ
  if (final_by_dominance)
    assert(final_by_dominance->final_by_dominance == 0);
#endif
  assert(isConsistent());

#ifdef TWINLIST_STAT
  static stat::Histogram t_hist(1,300,0,true);
  t_hist.add(twins.size());
#endif
}

unsigned int osl::checkmate::CheckHashRecord::
selectBestAttackMove(const PathEncoding& path,
		     const TwinTable& table,
		     unsigned int& childBestProofNumber,
		     unsigned int& childSecondBestProofNumber,
		     unsigned int& currentDisproofNumber,
		     unsigned int& childBestDisproofNumber,
		     ProofDisproof& bestResultInSolved,
		     CheckMove *& bestChild)
{
#ifdef DELAY_SACRIFICE
  if ((! filter.isTarget(MoveFlags::Sacrifice))
      && (getProof(proofDisproof()) > 8))	// ο208դ꤬
    filter.addTarget(MoveFlags::Sacrifice);
#endif
  while (true)
  {
    const unsigned int result = 
      selectBestAttackMoveMain(path, table, childBestProofNumber,
			       childSecondBestProofNumber,
			       currentDisproofNumber,
			       childBestDisproofNumber, 
			       bestResultInSolved, bestChild);
    if (! bestChild)
    {
#ifdef DELAY_SACRIFICE
      if (! filter.isTarget(MoveFlags::Sacrifice))
      {
	filter.addTarget(MoveFlags::Sacrifice);
	continue;
      }
#endif
      if ((this->bestResultInSolved == ProofDisproof::PawnCheckmate())
	  && (! filter.isTarget(MoveFlags::NoPromote)))
      {
	filter.addTarget(MoveFlags::NoPromote);
	continue;
      }
#ifdef CHECK_DELAY_UPWARDS
      if (! filter.isTarget(MoveFlags::Upward))
      {
	filter.addTarget(MoveFlags::Upward);
	useMaxInsteadOfSum = true;
	continue;
      }
#endif
    }
    else
    {
      if (bestChild->record)
	check_assert(! bestChild->findLoop(path, table));
    }
    return result;
  }
}

unsigned int osl::checkmate::CheckHashRecord::
selectBestAttackMoveMain(const PathEncoding& path,
			 const TwinTable& table,
			 unsigned int& childBestProofNumber,
			 unsigned int& childSecondBestProofNumber,
			 unsigned int& currentDisproofNumber,
			 unsigned int& childBestDisproofNumber,
			 ProofDisproof& bestResultInSolved,
			 CheckMove*& bestChild)
{
  childBestProofNumber=ProofDisproof::BigProofNumber;
  childSecondBestProofNumber=ProofDisproof::BigProofNumber;
  currentDisproofNumber=0;
  childBestDisproofNumber=ProofDisproof::BigProofNumber;
  bestResultInSolved = this->bestResultInSolved;
  bestChild = 0;
  unsigned int sum_proofs = 0;

  // TODO: targetUpwards template ˤʬɤ
  const bool targetUpwards = filter.isTarget(MoveFlags::Upward);
  for (CheckMoveList::iterator p=moves.begin(); p!=moves.end(); ++p)
  {
#ifdef CHECK_DELAY_UPWARDS
    if ((! targetUpwards) && (p->flags.isSet(MoveFlags::Upward)))
    {
      check_assert(p->record);
      const ProofDisproof& pdp = p->record->proofDisproof();
      if (pdp.proof()==0)
      {
	if (! pdp.isPawnDropFoul(p->move))
	{
	  bestChild = &*p;
	  return 0;
	}
	else
	{
	  addToSolvedInAttack(*p, ProofDisproof::PawnCheckmate());
	}
      }
      continue;
    }
#endif
    if (! filter.isTarget(p->flags))
      continue;

    unsigned int child_proof, child_disproof;
    if (p->record)
    {
#ifdef PROOF_NUMBER_PROPAGATE_BY_DOMINANCE
      if (p->record->sameBoards && (! p->record->proofDisproof().isFinal())
	  && (! p->record->isVisited) && (! p->findLoop(path, table)))
	p->record->sameBoards->updateSlow<false>(p->move.player(), *p->record, 
						 PathEncoding(path, p->move));
#endif      
      const TwinEntry *loopDetected = p->findLoop(path, table);
      if (loopDetected)
      {
#ifdef PAWN_CHECKMATE_SENSITIVE
	if ((this->bestResultInSolved != ProofDisproof::PawnCheckmate())
	    && loopDetected->move.record
	    && (loopDetected->move.record->bestResultInSolved
		== ProofDisproof::PawnCheckmate()))
	{
	  this->bestResultInSolved 
	    = this->bestResultInSolved.betterForAttack(ProofDisproof::PawnCheckmate());
	}
#endif
#ifdef CHECK_ABSORB_LOOP
	if (loopDetected->loopTo == this)
	{
	  p->flags.set(MoveFlags::Solved);
	  this->bestResultInSolved = betterForAttack(this->bestResultInSolved, NoCheckmate);
	  continue;
	}
#endif
	bestResultInSolved 
	  = bestResultInSolved.betterForAttack(ProofDisproof::LoopDetection());
	continue;
      }
      if (p->record->isVisited)
      {
	if (p->record->proofDisproof() == ProofDisproof::PawnCheckmate())
	  this->bestResultInSolved 
	    = this->bestResultInSolved.betterForAttack(ProofDisproof::PawnCheckmate());
	bestResultInSolved = bestResultInSolved.betterForAttack(ProofDisproof::LoopDetection());
	continue;
      }
      if (targetUpwards && (distance > p->record->distance))
	distance = p->record->distance;

      const ProofDisproof& child_pdp = p->record->proofDisproof();
      child_proof=child_pdp.proof();
      child_disproof=child_pdp.disproof();
      if (child_proof == 0)
      {
	if (! child_pdp.isPawnDropFoul(p->move))
	{
	  bestChild = &*p;
	  return 0;
	}
	else
	{
	  addToSolvedInAttack(*p, ProofDisproof::PawnCheckmate());
	  continue;
	}
      }
      else if (child_disproof == 0)
      {
	check_assert(! child_pdp.isLoopDetection());
	addToSolvedInAttack(*p, child_pdp);
	continue;
      }
#ifdef CHECK_DELAY_UPWARDS
      else if ((!targetUpwards) && (p->record->distance <= distance))
      {
	p->flags.set(MoveFlags::Upward);
	continue;
      }
#endif
    }
    else
    {
      child_proof = p->h_proof;
      child_disproof = p->h_disproof;
    }
    
    check_assert(child_proof != 0);
    check_assert(child_disproof != 0);

    p->addCost(child_proof,child_disproof);
    sum_proofs += p->h_proof;

    if (child_proof <= childSecondBestProofNumber)
    {
      if ((child_proof < childBestProofNumber)
	  || ((child_proof == childBestProofNumber)
	      && (child_disproof < childBestDisproofNumber)))
      {
	bestChild=&*p;
	childSecondBestProofNumber=childBestProofNumber;
	childBestProofNumber=child_proof;
	childBestDisproofNumber=child_disproof;
      }
      else
      {
	childSecondBestProofNumber=child_proof;
      }
      currentDisproofNumber 
	= add(child_disproof, currentDisproofNumber);
      continue;
    }
    currentDisproofNumber 
      = add(child_disproof, currentDisproofNumber);
  }

  bestResultInSolved
    = bestResultInSolved.betterForAttack(this->bestResultInSolved);
#ifdef CHECK_EXTRA_LIMIT_PROOF
  // (ʬ줬)øʿ
  const unsigned int proofAverage = 
    std::min((size_t)MAX_CHECK_EXTRA_LIMIT_PROOF,
	     1 + (sum_proofs / (moves.size()+1))); // guard against div by 0 
  return std::min(check_extra_limit_max, proofAverage);
#else
  return 0;
#endif
}


unsigned int osl::checkmate::CheckHashRecord::
selectBestDefenseMove(const PathEncoding& path,
		      const TwinTable& table,
		      unsigned int& childBestDisproofNumber,
		      unsigned int& childSecondBestDisproofNumber,
		      unsigned int& currentProofNumber,
		      unsigned int& childBestProofNumber,
		      ProofDisproof& bestResultInSolved,
		      CheckMove *&bestChild)
{
  // TODO: targetUpwards template ˤʬɤ
  const bool targetUpwards = filter.isTarget(MoveFlags::Upward);
  childBestDisproofNumber=ProofDisproof::BigProofNumber;
  childSecondBestDisproofNumber=ProofDisproof::BigProofNumber;
  currentProofNumber=0;
  childBestProofNumber=ProofDisproof::BigProofNumber;
  bestChild = 0;
  bestResultInSolved = this->bestResultInSolved;
  unsigned int sum_disproofs = 0;

  for (CheckMoveList::iterator p=moves.begin(); p!=moves.end(); ++p)
  {
#ifdef CHECK_DELAY_UPWARDS
    if ((! targetUpwards) && (p->flags.isSet(MoveFlags::Upward)))
    {
      check_assert(p->record);
      const ProofDisproof& pdp = p->record->proofDisproof();
      if ((pdp.disproof()==0) || p->record->isVisited // Ե
	  || p->findLoop(path, table))
      {
	bestChild = &*p;
	return 0;
      }
      continue;
    }
#endif
    if (! filter.isTarget(p->flags))
      continue;

    unsigned int child_proof, child_disproof;
    if (p->record)
    {
#ifdef PROOF_NUMBER_PROPAGATE_BY_DOMINANCE
      if (p->record->sameBoards && (! p->record->proofDisproof().isFinal())
	  && (! p->record->isVisited) && (! p->findLoop(path, table)))
      {
	p->record->sameBoards->updateSlow<true>(alt(p->move.player()), *p->record, 
						PathEncoding(path, p->move));
      }
#endif      
      if (targetUpwards && (distance > p->record->distance))
	distance = p->record->distance;

      const ProofDisproof& child_pdp = p->record->proofDisproof();
      child_proof=child_pdp.proof();
      child_disproof=child_pdp.disproof();

      if ((child_disproof==0) || p->record->isVisited // Ե
	  || p->findLoop(path, table))
      {
	bestChild = &*p;
	return 0;
      }
      else if (child_proof == 0) // 
      {
	addToSolvedInDefense(*p, child_pdp);
	continue;
      }
#ifdef CHECK_DELAY_UPWARDS
      else if ((! targetUpwards) && (p->record->distance <= distance))
      {
	p->flags.set(MoveFlags::Upward);
	continue;
      }
#endif
    }
    else
    {
      child_proof = p->h_proof;
      child_disproof = p->h_disproof;
    }
    assert(child_proof);
    assert(child_disproof);
    
    p->addCost(child_proof, child_disproof);
    sum_disproofs += p->h_disproof;

    if ((child_disproof < childBestDisproofNumber)
	|| ((child_disproof == childBestDisproofNumber)
	    && (child_proof < childBestProofNumber)))
    {
      bestChild = &*p;

      childSecondBestDisproofNumber=childBestDisproofNumber;
      childBestDisproofNumber=child_disproof;
      childBestProofNumber=child_proof;
      currentProofNumber = add(currentProofNumber, child_proof);
    }
    else if (child_disproof < childSecondBestDisproofNumber)
    {
      childSecondBestDisproofNumber=child_disproof;
      currentProofNumber = add(currentProofNumber, child_proof);
    }
    else 
    {
      currentProofNumber = add(currentProofNumber, child_proof);
    }
  }

  bestResultInSolved 
    = bestResultInSolved.betterForDefense(this->bestResultInSolved);
#ifdef CHECK_EXTRA_LIMIT_DISPROOF
  const unsigned int disproofAverage
    = std::min((size_t)MAX_CHECK_EXTRA_LIMIT_DISPROOF, 
	       1 + (sum_disproofs / (moves.size() +1))); // guard against div by 0 
  return std::min(check_extra_limit_max, disproofAverage);
#else
  return 0;
#endif
}

template <osl::Player Attacker>
void osl::checkmate::CheckHashRecord::propagateCheckmateRecursive()
{
  assert(proof_pieces_type == PROOF);
#ifdef CHECKMATE_DEBUG
  static stat::Ratio ratio("propagateCheckmate");
  if ((! moves.empty()) && (moves.begin()->move.player() == Attacker))
  {
    // or-node
    check_assert(hasBestMove());
  }
#endif
  check_assert(final_by_dominance == 0);
  for (SameBoardList::iterator p=sameBoards->begin(); p!=sameBoards->end(); ++p)
  {
    if (p->stand(BLACK) == black_stand)
      continue;
    const PieceStand& attack_stand 
      = (Attacker == BLACK) ? p->black_stand : p->white_stand;
    if (attack_stand.template hasMoreThan<BLACK>(proofPieces()))
    {
#ifdef CHECKMATE_DEBUG
      ratio.add(p->proof());
#endif    
      check_assert(p->disproof());
      if (p->proof())
      {
	check_assert(p->final_by_dominance == 0);
	p->setProofPieces(proofPieces());
	p->setFinalByDominance(this);
	p->bestMove = bestMove;
	p->proof_disproof = proofDisproof();
	check_assert(p->isConsistent());
      }
    }
  }
}

template <osl::Player Attacker>
void osl::checkmate::CheckHashRecord::propagateNoCheckmateRecursive()
{
#ifdef CHECKMATE_DEBUG
  static stat::Ratio ratio("propagateNoCheckmate");
#endif
  check_assert(final_by_dominance == 0);
  check_assert(hasDisproofPieces());
  check_assert(! proofDisproof().isLoopDetection());
  for (SameBoardList::iterator p=sameBoards->begin(); p!=sameBoards->end(); ++p)
  {
    if (p->stand(BLACK) == black_stand)
      continue;
    const PieceStand& defense_stand 
      = (Attacker == BLACK) ? p->white_stand : p->black_stand;
    if (defense_stand.template hasMoreThan<BLACK>(disproofPieces()))
    {
#ifdef CHECKMATE_DEBUG
      ratio.add(p->disproof());
#endif    
      check_assert(p->proof()
		   || (p->dump(), dump(), std::cerr << Attacker << "\n", 0));
      if (p->disproof())
      {
	check_assert((p->final_by_dominance == 0) || (p->dump(), 0));
	p->setDisproofPieces(disproofPieces());
	p->setFinalByDominance(this);
	p->bestMove = bestMove;
	p->proof_disproof = proofDisproof();
	p->twins.clear();
      }
    }
  }
}

namespace osl
{
  namespace checkmate
  {
    namespace
    {
      void printMoves(std::ostream& os, const CheckMoveList& moves)
      {
	if (moves.empty())
	  return;
	os << "moves size " << moves.size() << "\n";
	for (CheckMoveList::const_iterator p=moves.begin();
	     p!=moves.end(); ++p)
	{
	  os << "moves " << p->flags << " " << p->move;
	  if (p->record)
	  {
	    os << " " << p->record->proofDisproof()
	       << " " << p->record << " parent " << p->record->parent;
	  }
	  else
	  {
	    os << " h " << p->h_proof << " " << p->h_disproof
	       << " c " << p->cost_proof << " " << p->cost_disproof
	       << "\n";
	  }
	  os << "\n";
	}
      }
      void printTwins(std::ostream& os, const TwinList& twins)
      {
	if (twins.empty())
	  return;
	os << "twins size " << twins.size() << "\n";
	for (TwinList::const_iterator p=twins.begin(); p!=twins.end(); ++p)
	{
	  os << "twins " << p->path << " " << p->move.move 
	     << " loopTo " << p->loopTo;
	  if (p->move.record)
	  {
	    os << " " << p->move.record->proofDisproof()
	       << " " << p->move.record
	       << " #next-twins " << p->move.record->twins.size();
	  }
	  os << "\n";
	}
      }
    }
  } // checkmate
} // namespace osl

void osl::checkmate::CheckHashRecord::dump(int dump_depth) const
{
  dump(std::cerr, dump_depth);
}

void osl::checkmate::CheckHashRecord::dump(std::ostream& os, int dump_depth) const
{
  if (dump_depth <= 0)
    return;
  os << "+ CheckHashRecord::dump(" << dump_depth << ") "
     << this << " distance " << distance << " ";
  os << proofDisproof() << " " << filter;
  if (proofDisproof().isFinal())
    os << " " << bestResultInSolved;
  os << "\n"
     << " black " << stand(BLACK) << " white " << stand(WHITE) << "\n"
     << " bestMove " << &*(bestMove);
  if (bestMove)
    os << " " << bestMove->move << " " << bestMove->record;
  if (hasProofPieces())
    os << " pp " << proofPieces();
  else if (hasDisproofPieces())
    os << " dp " << disproofPieces();
  os << "\n";
  if (final_by_dominance)
  {
    os << "--- final_by_dominance " << final_by_dominance << "\n";
    final_by_dominance->dump(os, 1);
    os << "+++ final_by_dominance\n";
  }
  if (parent)
  {
    os << "parent " << parent << " " << parent->proofDisproof();
    os << " " << parent->filter;
    if (parent->bestMove)
      os << " " << parent->bestMove->move << " " << parent->bestMove->record
	 << parent->bestMove->flags;
    if (parent->final_by_dominance)
      os << " f " << final_by_dominance;
    os << "\n";
  }
  printMoves(os, moves);
  printTwins(os, twins);
  if (bestMove)
  {
    os << "depth " << dump_depth << " " << bestMove->record << "\n";
    if ((dump_depth > 0) && (bestMove->record))
      bestMove->record->dump(os, dump_depth-1);
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
