/* oraclePool.cc
 */
#include "osl/checkmate/oraclePool.h"
#include "osl/checkmate/checkHashRecord.h"
#include "osl/state/simpleState.h"
#include "osl/hash/hashKey.h"
#include "osl/centering3x3.h"
#include "osl/stl/hash_map.h"
#include "osl/stl/vector.h"
#include <algorithm>

namespace osl
{
  namespace checkmate
  {
    // TODO: list ϻȤ®Ϥ
    typedef vector<const CheckHashRecord*> vector_t;
    typedef hash_map<HashKey,vector_t> hash_map_t;

    const unsigned int max_oracles_for_key_soft_limit = 128;
    const unsigned int max_oracles_for_key_hard_limit = 65534;
  }
} // namespace osl

/** 
 * Ȥꤢ8˵ȽꤷƤߤ.
 * 25˵ʤɤ¾θ
 */
struct osl::checkmate::OraclePool::Table 
{
  hash_map_t table;
  const Player defender;
  explicit Table(Player attacker) : defender(alt(attacker))
  {
  }
  template <Direction DIR>
  static void addKey(HashKey& key, const SimpleState& state, Position target)
  {
    const Offset offset = Board_Table.getOffsetForBlack(DIR);
    target += offset;	// 8 ˵ƻʤ֤ˤѴ
    const Piece piece = state.getPieceOnBoard(target);
    hash::Hash_Gen_Table.addHashKey(key, target, piece.ptypeO());
  }
  const HashKey makeKey(const SimpleState& state) const
  {
    const Position target_king=state.getKingPosition(defender);
    const Position center = Centering3x3::adjustCenter(target_king);

    HashKey key;
    hash::Hash_Gen_Table.addHashKey(key, center, 
			    state.getPieceOnBoard(center).ptypeO());
    addKey<UL>(key, state, center);
    addKey<U> (key, state, center);
    addKey<UR>(key, state, center);
    addKey<L> (key, state, center);
    addKey<R> (key, state, center);
    addKey<DL>(key, state, center);
    addKey<D> (key, state, center);
    addKey<DR>(key, state, center);
    return key;
  }
  void addOracle(const SimpleState& state, const CheckHashRecord* oracle)
  {
    const HashKey key = makeKey(state);
    vector_t& v = table[key];
    if (v.size() >= max_oracles_for_key_soft_limit)
    {
#ifdef CHECKMATE_DEBUG
      oracle->dump();
#endif
      return;
    }
    v.push_back(oracle);
  }
  template <bool is_attack>
  const CheckHashRecord * find(const SimpleState& state, PieceStand stand,
			       unsigned short& age) const;
  size_t keySize() const
  {
    return table.size();
  }
  size_t totalSize() const
  {
    size_t result = 0;
    for (hash_map_t::const_iterator p=table.begin(); p!=table.end(); ++p)
    {
      result += p->second.size();
    }
    return result;
  }
};

template <bool is_attack>
const osl::checkmate::CheckHashRecord *
osl::checkmate::OraclePool::
Table::find(const SimpleState& state, PieceStand stand,
	    unsigned short& age) const
{
  const HashKey key = makeKey(state);
  hash_map_t::const_iterator pos = table.find(key);
  if (pos == table.end())
    return 0;

  const vector_t& v = pos->second;
  assert(v.size() <= max_oracles_for_key_hard_limit);
  while (age<v.size())
  {
    const CheckHashRecord *record = v[age++];
    if (is_attack)
    {
      if (stand.hasMoreThan<BLACK>(record->proofPieces()))
	return record;
    }
    else
    {
      if (stand.hasMoreThan<BLACK>(record->disproofPieces()))
	return record;
    }
  }
  return 0;
}
/* ------------------------------------------------------------------------- */

osl::checkmate::
OraclePool::OraclePool(Player attacker)
  : proof_oracles(new Table(attacker)), disproof_oracles(new Table(attacker))
{
}
  
osl::checkmate::
OraclePool::~OraclePool()
{
}
  
void osl::checkmate::
OraclePool::addProofOracle(const SimpleState& state, 
			   const CheckHashRecord* oracle)
{
  assert(oracle->hasProofPieces());
  proof_oracles->addOracle(state, oracle);
}
void osl::checkmate::
OraclePool::addDisproofOracle(const SimpleState& state, 
			      const CheckHashRecord* oracle)
{
  assert(oracle->proofDisproof().isCheckmateFail()
	 || (! oracle->twins.empty()));
  if (oracle->hasDisproofPieces()) // TODO: loop ǤȸΨ
    disproof_oracles->addOracle(state, oracle);
}
  
const osl::checkmate::CheckHashRecord *
osl::checkmate::
OraclePool::findProofOracle(const SimpleState& state, PieceStand black_stand,
			    unsigned short& age) const
{
  const PieceStand attack_stand 
    = (proof_oracles->defender == WHITE) ? black_stand : PieceStand(WHITE, state);
  return proof_oracles->find<true>(state, attack_stand, age);
}

const osl::checkmate::CheckHashRecord *
osl::checkmate::
OraclePool::findDisproofOracle(const SimpleState& state, PieceStand black_stand,
			       unsigned short& age) const
{
  const PieceStand defense_stand 
    = (disproof_oracles->defender == BLACK) ? black_stand : PieceStand(WHITE, state);
  return disproof_oracles->find<false>(state, defense_stand, age);
}

size_t osl::checkmate::
OraclePool::totalSize() const
{
  return proof_oracles->totalSize() + disproof_oracles->totalSize();
}

size_t osl::checkmate::
OraclePool::keySize() const
{
  return proof_oracles->keySize() + disproof_oracles->keySize();
}
  
/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
