/**
 * attackKing.cc
 */
#include "osl/eval/endgame/attackKing.h"
#include "osl/eval/pieceEval.h"
#include "osl/misc/carray.h"

// YSSɾͤ򿿻
// http://www32.ocn.ne.jp/~yss/book.html#SEC3
//
// ζ̤Фζа֤ˤ硣ؼˤĤƤоΤǤ롣
//        0   1   2   3   4   5   6   7   8 (صΥ)
// 
//   +8   50  50  50  50  50  50  50  50  50
//   +7   50  50  50  50  50  50  50  50  50
//   +6   62  60  58  52  50  50  50  50  50
//   +5   80  78  72  67  55  51  50  50  50
//   +4  100  99  95  87  78  69  50  50  50
//   +3  140 130 110 100  95  75  54  50  50
//   +2  170 160 142 114  98  80  62  55  50  <--- ʿȤ롣
//   +1  170 165 150 121  94  78  58  52  50
//    0  170 145 137 115  91  75  57  50  50  <--- 濴(0 0)˲Ȥ
//   -1  132 132 129 102  84  71  51  50  50
//   -2  100  97  95  85  70  62  50  50  50
//   -3   90  85  80  68  60  53  50  50  50
//   -4   70  66  62  55  52  50  50  50  50
//   -5   54  53  51  50  50  50  50  50  50
//   -6   50  50  50  50  50  50  50  50  50
//   -7   50  50  50  50  50  50  50  50  50
//   -8   50  50  50  50  50  50  50  50  50

namespace osl
{
  namespace
  {
    const CArray<int,19*9> yss_bonus = {{
      // ˡϰʲ뤿ᡤ+9  -9 ޤǼ
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
         62,  60,  58,  52,  50,  50,  50,  50,  50, 
         80,  78,  72,  67,  55,  51,  50,  50,  50, 
        100,  99,  95,  87,  78,  69,  50,  50,  50, 
        140, 130, 110, 100,  95,  75,  54,  50,  50, 
        170, 160, 142, 114,  98,  80,  62,  55,  50,
        170, 165, 150, 121,  94,  78,  58,  52,  50, 
        170, 145, 137, 115,  91,  75,  57,  50,  50,
        132, 132, 129, 102,  84,  71,  51,  50,  50, 
        100,  97,  95,  85,  70,  62,  50,  50,  50, 
         90,  85,  80,  68,  60,  53,  50,  50,  50, 
         70,  66,  62,  55,  52,  50,  50,  50,  50, 
         54,  53,  51,  50,  50,  50,  50,  50,  50, 
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
         50,  50,  50,  50,  50,  50,  50,  50,  50,
         50,  50,  50,  50,  50,  50,  50,  50,  50, 
    }};
    inline int multiply(int value, int bonus)
    {
      // yss Ǥϻղòͤ2٤뤿1.2ǳ
      return (value * bonus / 120) & (~1); 
    }
    inline void adhoc_adjust(int& value, double scale)
    {
      value = static_cast<int>(value * scale);
      value &= (~1);
    }
  } // anonymous namespace
} // namespace osl



osl::eval::endgame::AttackKing::
Table::Table()
{
  data.fill(0);
  for (int king_x=1; king_x<=9; ++king_x)
  {
    for (int king_y=1; king_y<=9; ++king_y)
    {
      const Position king(king_x, king_y);
      for (int p=PTYPE_PIECE_MIN; p<=PTYPE_MAX; ++p)
      {
	const Ptype ptype = static_cast<Ptype>(p);
	assert(isPiece(ptype));
	const Ptype basic_type = unpromote(ptype);

	const int ptype_value
	  = Ptype_Eval_Table.value(newPtypeO(BLACK,ptype));
	// 
	const int stand_bonus 
	  = (isMajorBasic(basic_type) 
	     ? Ptype_Eval_Table.value(newPtypeO(BLACK,PAWN))
	     : 0);
	valueOf(king, BLACK, Position::STAND(), ptype) = -ptype_value 
	  - stand_bonus;
	valueOf(king, WHITE, Position::STAND(), ptype) = +ptype_value
	  + stand_bonus;
	// ׾
	for (int attack_x=1; attack_x<=9; ++attack_x)
	{
	  const int relative_x = abs(king_x - attack_x);
	  for (int attack_y=1; attack_y<=9; ++attack_y)
	  {
	    const Position attacker(attack_x, attack_y);
	    if (basic_type == KING)
	    {
	      // ̤ϤΤޤ
	      valueOf(king, BLACK, attacker, ptype) = -ptype_value;
	      valueOf(king, WHITE, attacker, ptype) = ptype_value;
	      continue;
	    }

	    const int relative_y_black_king 
	      = (((ptype == KNIGHT)||(ptype == LANCE))
		 ? king_y - attack_y - 1 // ˹ϰʱ󤯤ɾ
		 : king_y - attack_y);
	    const int relative_y_white_king 
	      = (((ptype == KNIGHT)||(ptype == LANCE))
		 ? attack_y - king_y - 1 // ˹ϰʱ󤯤ɾ
		 : attack_y - king_y);

	    const int bonus_black_king
	      = yss_bonus[relative_x + (-relative_y_black_king+9)*9];
	    const int bonus_white_king
	      = yss_bonus[relative_x + (-relative_y_white_king+9)*9];

	    const int black_attack =  multiply(ptype_value,bonus_white_king);
	    const int white_attack = -multiply(ptype_value,bonus_black_king);

	    if (isMajorBasic(basic_type))
	    {
	      //  max 
	      valueOf(king, BLACK, attacker, ptype) 
		= EvalTraits<WHITE>::max(-ptype_value, white_attack);
	      valueOf(king, WHITE, attacker, ptype) 
		= EvalTraits<BLACK>::max(ptype_value, black_attack);
	      continue;
	    }
	    valueOf(king, BLACK, attacker, ptype) = white_attack;
	    valueOf(king, WHITE, attacker, ptype) = black_attack;
	  }
	}
      }
    }
  }
  for (int king_x=1; king_x<=9; ++king_x)
  {
    for (int king_y=1; king_y<=9; ++king_y)
    {
      const Position king(king_x, king_y);
      // 1ĤưʤƱ
      for (int x=1; x<=9; ++x)
      {
	const Position b(x,2);
	valueOf(king, WHITE, b, LANCE) = valueOf(king, WHITE, b, PAWN);
	const Position w(x,8);
	valueOf(king, BLACK, w, LANCE) = valueOf(king, BLACK, w, PAWN);
      }
      // 1ܤζϲͤȾʬ
      for (int x=1; x<=9; ++x)
      {
	const Position b(x,1);
	adhoc_adjust(valueOf(king, WHITE, b, GOLD),    0.5);
	adhoc_adjust(valueOf(king, WHITE, b, PSILVER), 0.5);
	adhoc_adjust(valueOf(king, WHITE, b, PKNIGHT), 0.5);
	adhoc_adjust(valueOf(king, WHITE, b, PLANCE),  0.5);
	adhoc_adjust(valueOf(king, WHITE, b, PPAWN),   0.5);

	const Position w(x,9);
	adhoc_adjust(valueOf(king, BLACK, w, GOLD),    0.5);
	adhoc_adjust(valueOf(king, BLACK, w, PSILVER), 0.5);
	adhoc_adjust(valueOf(king, BLACK, w, PKNIGHT), 0.5);
	adhoc_adjust(valueOf(king, BLACK, w, PLANCE),  0.5);
	adhoc_adjust(valueOf(king, BLACK, w, PPAWN),   0.5);
      }
    }
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
