/**
 * defenseKing.cc
 */
#include "osl/eval/endgame/defenseKing.h"
#include "osl/eval/pieceEval.h"
#include "osl/misc/carray.h"

// YSSɾͤ򿿻
// http://www32.ocn.ne.jp/~yss/book.html#SEC3
//
//         0   1   2   3   4   5   6   7   8 (صΥ)
//   +8   50  50  50  50  50  50  50  50  50
//   +7   56  53  50  50  50  50  50  50  50
//   +6   64  61  55  50  50  50  50  50  50
//   +5   79  77  70  65  54  51  50  50  50
//   +4  100  99  95  87  74  58  50  50  50
//   +3  116 117 101  95  88  67  54  50  50
//   +2  131 129 124 114  90  71  59  51  50
//   +1  137 138 132 116  96  76  61  53  50
//    0  142 142 136 118  98  79  64  52  50  <--- 濴
//   -1  132 132 129 109  95  75  60  51  50
//   -2  121 120 105  97  84  66  54  50  50
//   -3   95  93  89  75  68  58  51  50  50
//   -4   79  76  69  60  53  50  50  50  50
//   -5   64  61  55  51  50  50  50  50  50
//   -6   56  52  50  50  50  50  50  50  50
//   -7   50  50  50  50  50  50  50  50  50
//   -8   50  50  50  50  50  50  50  50  50

namespace osl
{
  namespace
  {
    const CArray<int,19*9> yss_bonus = {{
      // ˡϰʲ뤿ᡤ+9  -9 ޤǼ
      50,  50,  50,  50,  50,  50,  50,  50,  50,
      50,  50,  50,  50,  50,  50,  50,  50,  50,
      56,  53,  50,  50,  50,  50,  50,  50,  50,
      64,  61,  55,  50,  50,  50,  50,  50,  50,
      79,  77,  70,  65,  54,  51,  50,  50,  50,
     100,  99,  95,  87,  74,  58,  50,  50,  50,
     116, 117, 101,  95,  88,  67,  54,  50,  50,
     131, 129, 124, 114,  90,  71,  59,  51,  50,
     137, 138, 132, 116,  96,  76,  61,  53,  50,
     142, 142, 136, 118,  98,  79,  64,  52,  50,
     132, 132, 129, 109,  95,  75,  60,  51,  50,
     121, 120, 105,  97,  84,  66,  54,  50,  50,
      95,  93,  89,  75,  68,  58,  51,  50,  50,
      79,  76,  69,  60,  53,  50,  50,  50,  50,
      64,  61,  55,  51,  50,  50,  50,  50,  50,
      56,  52,  50,  50,  50,  50,  50,  50,  50,
      50,  50,  50,  50,  50,  50,  50,  50,  50,
      50,  50,  50,  50,  50,  50,  50,  50,  50,
      50,  50,  50,  50,  50,  50,  50,  50,  50,
    }};
    /**  */
    const CArray<int,9> yss_king_bonus = {{
      0, 0, 0, 150, 450, 900,1300,1550,1600
    }};
    inline int multiply(int value, int bonus)
    {
      // http://www32.ocn.ne.jp/~yss/book.html#SEC3
      // > ɵ2005ǯ0122
      // > ̤ζ᤯Ϲ⤹Τǡߤϡܤ3ʬ1ͤޤǲƤޤ
      if (bonus > 100)
	bonus = (bonus - 100)/3 + 100;
      // yss Ǥϻղòͤ2٤뤿1.2ǳ
      return (value * bonus / 120) & (~1);
    }
  } // anonymous namespace
} // namespace osl



osl::eval::endgame::DefenseKing::
Table::Table()
{
  data.fill(0);
  for (int king_x=1; king_x<=9; ++king_x)
  {
    for (int king_y=1; king_y<=9; ++king_y)
    {
      const Position king(king_x, king_y);
      for (int p=PTYPE_PIECE_MIN; p<=PTYPE_MAX; ++p)
      {
	const Ptype ptype = static_cast<Ptype>(p);
	assert(isPiece(ptype));
	const Ptype basic_type = unpromote(ptype);

	const int ptype_value
	  = Ptype_Eval_Table.value(newPtypeO(BLACK,ptype));
	// 
	const int stand_bonus
	  = (isMajorBasic(basic_type) 
	     ? Ptype_Eval_Table.value(newPtypeO(BLACK,PAWN)) 
	     : 0);
	valueOf(king, BLACK, Position::STAND(), ptype) = ptype_value
	  + stand_bonus;
	valueOf(king, WHITE, Position::STAND(), ptype) = -ptype_value 
	  - stand_bonus;
	// ׾
	if (basic_type == KING)
	{
	  // ̤̤θ
	  valueOf(king, BLACK, king, ptype) 
	    = ptype_value + yss_king_bonus[9-king_y];
	  valueOf(king, WHITE, king, ptype) 
	    = -ptype_value - yss_king_bonus[king_y-1];
	  continue;
	}
	      
	for (int defense_x=1; defense_x<=9; ++defense_x)
	{
	  const bool near_edge 
	    = (((defense_x < 3) && (defense_x < king_x))
	       || ((defense_x > 7) && (defense_x > king_x)));
	  const int relative_x = abs(king_x - defense_x);
	  for (int defense_y=1; defense_y<=9; ++defense_y)
	  {
	    const Position defender(defense_x, defense_y);
	    if (isMajorBasic(basic_type))
	    {
	      // ϤȤꤢΤޤ 
	      valueOf(king, BLACK, defender, ptype) = ptype_value;
	      valueOf(king, WHITE, defender, ptype) = -ptype_value;
	      continue;
	    }

	    const int relative_y_black_king 
	      = (((ptype == KNIGHT)||(ptype == LANCE))
		 ? king_y - defense_y + 1 // ˹ϰʱ󤯤ɾ
		 : king_y - defense_y);
	    const int relative_y_white_king 
	      = (((ptype == KNIGHT)||(ptype == LANCE))
		 ? defense_y - king_y + 1 // ˹ϰʱ󤯤ɾ
		 : defense_y - king_y);

	    int bonus_black_king
	      = yss_bonus[relative_x + (-relative_y_black_king+9)*9];
	    int bonus_white_king
	      = yss_bonus[relative_x + (-relative_y_white_king+9)*9];
	    // üˤϥܡʥ򤢤ʤ
	    if (near_edge)
	    {
	      bonus_black_king = std::min(100, bonus_black_king);
	      bonus_white_king = std::min(100, bonus_white_king);
	    }

	    const int black_defense =  multiply(ptype_value,bonus_black_king);
	    const int white_defense = -multiply(ptype_value,bonus_white_king);

	    valueOf(king, BLACK, defender, ptype) = black_defense;
	    valueOf(king, WHITE, defender, ptype) = white_defense;
	  }
	}
      }
    }
  }
  // 1ĤưʤƱ
  for (int king_x=1; king_x<=9; ++king_x)
  {
    for (int king_y=1; king_y<=9; ++king_y)
    {
      const Position king(king_x, king_y);
      for (int x=1; x<=9; ++x)
      {
	const Position b(x,2);
	valueOf(king, BLACK, b, LANCE) = valueOf(king, BLACK, b, PAWN);
	const Position w(x,8);
	valueOf(king, WHITE, w, LANCE) = valueOf(king, WHITE, w, PAWN);
      }
    }
  }
  // ̤ü˴ȥܡʥ
  const int pawn_value = Ptype_Eval_Table.value(newPtypeO(BLACK,PAWN));
  valueOf(Position(9,8), BLACK, Position(9,8), KING) += pawn_value*3;
  valueOf(Position(8,8), BLACK, Position(8,8), KING) += pawn_value*3;
  valueOf(Position(7,8), BLACK, Position(7,8), KING) += pawn_value*3;
  valueOf(Position(1,8), BLACK, Position(1,8), KING) += pawn_value*3;
  valueOf(Position(2,8), BLACK, Position(2,8), KING) += pawn_value*3;
  valueOf(Position(3,8), BLACK, Position(3,8), KING) += pawn_value*3;

  valueOf(Position(9,9), BLACK, Position(9,9), KING) += pawn_value*3;
  valueOf(Position(8,9), BLACK, Position(8,9), KING) += pawn_value*3;
  valueOf(Position(7,9), BLACK, Position(7,9), KING) += pawn_value*2;
  valueOf(Position(1,9), BLACK, Position(1,9), KING) += pawn_value*3;
  valueOf(Position(2,9), BLACK, Position(2,9), KING) += pawn_value*3;
  valueOf(Position(3,9), BLACK, Position(3,9), KING) += pawn_value*2;

  valueOf(Position(9,2), WHITE, Position(9,2), KING) -= pawn_value*3;
  valueOf(Position(8,2), WHITE, Position(8,2), KING) -= pawn_value*3;
  valueOf(Position(7,2), WHITE, Position(7,2), KING) -= pawn_value*3;
  valueOf(Position(1,2), WHITE, Position(1,2), KING) -= pawn_value*3;
  valueOf(Position(2,2), WHITE, Position(2,2), KING) -= pawn_value*3;
  valueOf(Position(3,2), WHITE, Position(3,2), KING) -= pawn_value*3;

  valueOf(Position(9,1), WHITE, Position(9,1), KING) -= pawn_value*3;
  valueOf(Position(8,1), WHITE, Position(8,1), KING) -= pawn_value*3;
  valueOf(Position(7,1), WHITE, Position(7,1), KING) -= pawn_value*2;
  valueOf(Position(1,1), WHITE, Position(1,1), KING) -= pawn_value*3;
  valueOf(Position(2,1), WHITE, Position(2,1), KING) -= pawn_value*3;
  valueOf(Position(3,1), WHITE, Position(3,1), KING) -= pawn_value*2;

}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
