/* bookPlayer.cc
 */
#include "osl/game_playing/bookPlayer.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/openingBookTracer.h"
#include "osl/container/moveStack.h"

osl::game_playing::
BookPlayer::
BookPlayer(OpeningBookTracer *b, ComputerPlayer *s)
  : book(b), searcher(s)
{
}

osl::game_playing::
BookPlayer::~BookPlayer()
{
}

osl::game_playing::ComputerPlayer* osl::game_playing::
BookPlayer::clone() const
{
  return new BookPlayer(book->clone(), searcher->clone());
}

void osl::game_playing::
BookPlayer::pushMove(Move m)
{
  book->update(m);
  searcher->pushMove(m);
}
void osl::game_playing::
BookPlayer::popMove()
{
  book->popMove();
  searcher->popMove();
}

const osl::Move osl::game_playing::
BookPlayer::selectBestMove(const GameState& state, int seconds, int byoyomi)
{
  if (! book->isOutOfBook())
  {
    const Move best_move = book->selectMove();
    if (best_move.isNormal()
	&& (! state.isIllegal(best_move)))
    {
      return best_move;
    }
  }
  return searcher->selectBestMove(state, seconds, byoyomi);
}

void osl::game_playing::
BookPlayer::allowSpeculativeSearch(bool value)
{
  ComputerPlayer::allowSpeculativeSearch(value);
  searcher->allowSpeculativeSearch(value);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
