/* csaClient.cc
 */
#include "osl/game_playing/csaClient.h"
#include "osl/game_playing/gnuShogiClient.h"
#include "osl/game_playing/gameState.h"
#include "osl/game_playing/csaLogger.h"
#include "osl/game_playing/csaStopwatch.h"
#include "osl/record/csa.h"
#include "osl/record/csaIOError.h"
#include "osl/sennichite.h"

#include <iostream>

osl::game_playing::
CsaClient::CsaClient(ComputerPlayer *black, ComputerPlayer *white,
		     CsaLogger *l, std::istream& is, std::ostream& os)
  : CuiClient(black, white, l, is, os) {
  setComputerPlayer(WHITE, true);
}

osl::game_playing::
CsaClient::~CsaClient()
{
}

void osl::game_playing::
CsaClient::readAndProcessCommand()
{
  CsaStopwatch timer;
  std::string line;
  std::getline(is, line);
  const long op_think_time = timer.read();
  if (! is)
  {
    throw EndGame();
  }
  
  if (line == "%TORYO")
  {
    logger->resign(state->state().getTurn());
    throw EndGame();
  }

  // TODO: %MATTA, %CHUDAN
  try
  {
    const Move op_move=record::csa::strToMove(line, state->state());
    const GameState::MoveType illegal_move = state->isIllegal(op_move);
    if (illegal_move)
    {
      std::cerr << "illegal move: " << line << "\n";
      logger->inputError(line.c_str());
      if (illegal_move == GameState::PAWN_DROP_FOUL)
	logger->writeComment("pawn drop foul");
      else if (illegal_move == GameState::UNSAFE_KING)
	logger->writeComment("unsafe king");
      else if (illegal_move == GameState::OTHER_INVALID)
	logger->writeComment("other illegal move");
      return;
    }
    const Sennichite result = pushMove(op_move, op_think_time);
    if (! result.isNormal())
    {
      os << "%SENNICHITE" << std::endl;
      logger->endByRepetition(result);
      throw EndGame();
    }
  }
  catch (record::csa::CsaIOError&)
  {
    std::cerr << "bad input: " << line << "\n";
    logger->inputError(line.c_str());
  }
  return;
}

void osl::game_playing::
CsaClient::processComputerMove(Move best_move, int my_think_time)
{
  if ((! best_move.isNormal())
      || (state->isIllegal(best_move)))
  {
    if (best_move == Move::DeclareWin())
    {
      os << "%KACHI\n";
      logger->endByDeclaration(state->state().getTurn());
    }
    else
    {
      os << "%TORYO\n";
      logger->resign(state->state().getTurn());
    }
    throw EndGame();
  }

  os << record::csa::show(best_move)
     << std::endl << std::flush;

  assert(isComputer(state->state().getTurn()));

  const Sennichite result = pushMove(best_move, my_think_time);
  if (! result.isNormal())
  {
    logger->endByRepetition(result);
    throw EndGame();
  }
}

const osl::Move osl::game_playing::
CsaClient::computeMove(int& consumed)
{
  return GameManager::computeMove(consumed);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
