/* weightTracer.cc
 */
#include "osl/game_playing/weightTracer.h"
#include "osl/game_playing/openingBookTracer.h"
#include "osl/record/opening/openingBook.h"
#include "osl/stl/vector.h"
#include "osl/misc/random.h"
#include <iostream>

osl::game_playing::
WeightTracer::WeightTracer(WeightedBook& b, bool v)
  : book(b), state_index(b.getStartState()), turn(BLACK), verbose(v)
{
}

osl::game_playing::
WeightTracer::WeightTracer(const WeightTracer& copy)
  : book(copy.book), state_index(copy.state_index), turn(copy.turn), 
    verbose(copy.verbose), state_stack(copy.state_stack)
{
}

osl::game_playing::OpeningBookTracer* osl::game_playing::
WeightTracer::clone() const
{
  return new WeightTracer(*this);
}

void osl::game_playing::
WeightTracer::update(Move move)
{
  if (verbose)
    std::cerr << "WeightTracer " << this << " Move: " << move << std::endl;

  state_stack.push(state_index);
  assert(move.player() == turn);
  changeTurn(turn);

  if (! isOutOfBook())
  {
    const vector<record::opening::WMove>& moves = book.getMoves(state_index);
    for (size_t i=0; i<moves.size(); i++)
    {
      if(moves[i].getMove() == move)
      {
	state_index = moves[i].getStateIndex();
	if (verbose)
	  std::cerr << "book: " 
		    << state_stack.top() << "->" << state_index << "\n";
	return;
      }
    }
    if (verbose)
      std::cerr << "book: end" << "\n";
  }
  state_index = -1;
}

void osl::game_playing::
WeightTracer::popMove()
{
  state_index = state_stack.top();
  state_stack.pop();
  changeTurn(turn);
  if (verbose)
    std::cerr << "WeightTracer " << this << " pop: " << turn << std::endl;
}

bool osl::game_playing::
WeightTracer::isOutOfBook() const
{
  return state_index < 0; 
}

const osl::Move osl::game_playing::
WeightTracer::selectMove() const
{
  int maxIndex = -1;
  vector<record::opening::WMove> moves = book.getMoves(state_index);
  
  int random_value = time_seeded_random() % 10001;
  int weight = 0;

  if (verbose)
    std::cerr << "random number: " << random_value << std::endl;

  for (size_t index = 0; index < moves.size(); index++)
  {
    weight += moves[index].getWeight();
    if (random_value <= weight)
    {
      maxIndex = index;
      break;
    }
  }

  if (maxIndex >= 0)
    return moves[maxIndex].getMove();
  return Move::INVALID();
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
