/* alarm.cc
 */
#include "osl/misc/alarm.h"
#include <stdexcept>
#include <iostream>
#include <cassert>
#include <signal.h>
#include <unistd.h>

volatile int * volatile osl::misc::Alarm::notify = 0;

void osl::misc::
Alarm::catchSignal(int signum)
{
  assert(SIGALRM == signum);
  if (notify)
    *notify = TIMEOUT;
  notify = 0;
}

bool osl::misc::
Alarm::set(int timeout, AlarmSwitch notify)
{
  assert(notify.notify);
  assert(timeout > 0);
  if (timeout <= 0)
    throw std::range_error("negative or zero seconds in Alarm::set");
  if (Alarm::notify)
  {
    *notify.notify = UNAVAILABLE;
    return false;
  }
  Alarm::notify = notify.notify;
  signal(SIGALRM, catchSignal);
  alarm(timeout);
  *notify.notify = WATCHING;
  return true;
}

void osl::misc::
Alarm::reset(AlarmSwitch notify)
{
  if (Alarm::notify && notify.notify != Alarm::notify)
  {
    std::cerr << "warning: Alarm::reset inconsistency\n";
    return;
  }
  Alarm::notify = 0;
  alarm(0);
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
