#include "osl/record/csa.h"
#include "osl/record/csaIOError.h"
#include "osl/state/simpleState.h"
#include "osl/pieceTable.h"
#include <iostream>
#include <stdexcept>
#include <cassert>
#include <string>

/* ------------------------------------------------------------------------- */

osl::Player osl::record::csa::
charToPlayer(char c)
{
  if(c=='+') 
    return BLACK;
  if(c=='-') 
    return WHITE;
  std::cerr << "oops " << (int)c << "\n";
  throw CsaIOError("not a csa PlayerCharacter "+std::string(1,c));
}

const osl::Position osl::record::csa::
strToPos(const std::string& s)
{
  int x=s.at(0)-'0';
  int y=s.at(1)-'0';
  if(x==0 && y==0) 
    return Position::STAND();
  return Position(x,y);
}

osl::Ptype osl::record::csa::
strToPtype(const std::string& s)
{
  for(int i=0;i<16;i++){
    if(s == Ptype_Table.getCsaName(static_cast<Ptype>(i))) 
      return static_cast<Ptype>(i);
  }
  throw CsaIOError("unknown std::string in csa::strToPtype "+s);
}

const osl::Move osl::record::csa::
strToMove(const std::string& s,const SimpleState& state)
{
  Player pl=csa::charToPlayer(s.at(0));
  Position fromPos=csa::strToPos(s.substr(1,2));
  Position toPos=csa::strToPos(s.substr(3,2));
  Ptype ptype=csa::strToPtype(s.substr(5,2));
  if(fromPos==Position::STAND()){
    if (isPromoted(ptype))
      throw CsaIOError("drop with promote ?! in csa::strToMove "+s);
    return Move(toPos,ptype,pl);
  }
  else{
    Piece p0=state.getPieceAt(fromPos);
    Piece p1=state.getPieceAt(toPos);
    Ptype capturePtype=p1.ptype();
    bool isPromote=(p0.ptype()!=ptype);
    if (! ((p0.ptype()==ptype)||(p0.ptype()==unpromote(ptype))))
      throw CsaIOError("bad promotion in csa::strToMove "+s);
    return Move(fromPos,toPos,ptype,
		capturePtype,isPromote,pl);
  }
}

/* ------------------------------------------------------------------------- */
const std::string osl::record::csa::
show(Player player)
{
  if (player==BLACK)
    return "+";
  else
    return "-";
}

const std::string osl::record::csa::
show(Move move)
{
  if (move == Move::DeclareWin())
    return "%KACHI";
  if (move.isInvalid())
    return "%TORYO";
  if (move.isPass())
    return "%PASS";		// FIXME: not in CSA protocol
  return show(move.player())
    + show(move.from())
    + show(move.to())
    + show(move.ptype());
}

const std::string osl::record::csa::
show(Position pos)
{
  if (pos.isPieceStand())
    return "00";
  const int x = pos.x();
  const int y = pos.y();
  std::string result = "XX";
  result[0] = x + '0';
  result[1] = y + '0';
  return result;
}

const std::string osl::record::csa::
show(Ptype ptype)
{
  return Ptype_Table.getCsaName(ptype);
}

const std::string osl::record::csa::
show(Piece piece)
{
  if (piece.isEdge())
    return "   ";
  if (piece.isEmpty())
    return " * ";

  assert(piece.isPiece() && isPiece(piece.ptype()));
  assert(unpromote(piece.ptype()) == Piece_Table.getPtypeOf(piece.number()));
  return show(piece.owner()) 
    + show(piece.ptype());
}

/* ------------------------------------------------------------------------- */

std::ostream& osl::csaShow(std::ostream& os,const Position pos)
{
  return os << record::csa::show(pos);
}

std::ostream& osl::csaShow(std::ostream& os, const Piece piece)
{
  return os << record::csa::show(piece);
}

std::ostream& osl::csaShow(std::ostream& os,const osl::Ptype ptype)
{
  return os << record::csa::show(ptype);
}

std::ostream& osl::csaShow(std::ostream& os,const Move move)
{
  return os << record::csa::show(move);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
