#include "osl/record/csaRecord.h"
#include "osl/record/csaIOError.h"
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <cassert>
#include <string>

/* ------------------------------------------------------------------------- */

namespace osl
{
  namespace record
  {
    namespace
    {
      static void csaParseLine(RecordVisitor& rv,const std::string& s)
      {
	Record *rec=rv.getRecord();
	SimpleState* state=rv.getState();
	//  std::cerr << s << std::endl;
	if (s.length()==0) 
	  return;
	switch(s.at(0)){
	case '\'': /* ȹ */
	  if (s.substr(1,2) == "* ")
	    rv.getLastMove()->setComment(s.substr(3));
	  return;
	case '$': /* ȹ */
	  return;
	case 'V': /* Сֹ */
	  rec->setVersion(s.substr(1));
	  return;
	case 'N': /* жɼ̾ */
	  switch(s.at(1)){
	  case '+':
	  case '-':
	    rec->setPlayer(csa::charToPlayer(s.at(1)),s.substr(2));
	    break;
	  default:
	    std::cerr << "Illegal csa line " << s << std::endl;
	    throw CsaIOError("illegal csa line "+s);
	  }
	  break;
	case 'P': /*  */
	  switch(s.at(1)){
	  case 'I': /* ʿ */
	    state->init(HIRATE);
	    break;
	  case '+': /* ζ */
	  case '-':{ /* ζ */
	    Player pl=csa::charToPlayer(s.at(1));
	    for(unsigned int i=2;i<=s.length()-4;i+=4){
	      Position pos=csa::strToPos(s.substr(i,2));
	      if(s.substr(i+2,2) == "AL"){
		state->setPieceAll(pl);
	      }
	      else{
		Ptype ptype=csa::strToPtype(s.substr(i+2,2));
		state->setPiece(pl,pos,ptype);
	      }
	    }
	    break;
	  }
	  default:
	    if(isdigit(s.at(1))){
	      int y=s.at(1)-'0';
	      for(unsigned int x=9,i=2;i<s.length();i+=3,x--){
		if(s.at(i)==' ') continue;
		Player pl=csa::charToPlayer(s.at(i));
		Position pos(x,y);
		Ptype ptype=csa::strToPtype(s.substr(i+1,2));
		state->setPiece(pl,pos,ptype);
	      }
	    }
	  }
	  break;
	case '+':
	case '-':{
	  Player pl=csa::charToPlayer(s.at(0));
	  if(s.length()==1){
	    state->setTurn(pl);
	    rec->setInitialState(*state);
	  }
	  else{ // actual moves
	    const Move m = csa::strToMove(s,*state);
	    if (! state->isValidMove(m))
	    {
	      std::cerr << "Illegal move " << m << std::endl;
	      throw CsaIOError("illegal move "+s);
	    }
	    rv.addMoveAndAdvance(m);
	    return;
	  }
	  break;
	}
	case 'T':
	  (rv.getLastMove())->setTime(atoi(s.c_str()+1));
	  return;
	case '%':
	  return;
	default:
	  throw CsaIOError("unknown character in csaParseLine "+s);
	}
      }
    } // anonymous namespace
  } // namespace record
} // namespace osl

osl::record::csa::
ICsaRecordStream::ICsaRecordStream(std::istream& is) : is(is)
{
  if (! is)
  {
    std::cerr << "ICsaRecordStream::ICsaRecordStream cannot read \n";
    abort();
  }
}
  
osl::record::csa::
ICsaRecordStream::~ICsaRecordStream(){}
  
void osl::record::csa::
ICsaRecordStream::load(Record* rec)
{
  //  rec->init();
  state.init();
  rv.setState(&state);
  rv.setRecord(rec);
  std::string line;
  while (std::getline(is, line)) 
  {
    //    std::cerr << "Read: " << s << std::endl;
    // quick hack for \r
    if ((! line.empty())
	&& (line[line.size()-1] == 13))
      line.erase(line.size()-1);
    csaParseLine(rv, line);
  }
}

osl::record::csa::
CsaFile::CsaFile(std::string fileName)
{
  std::ifstream ifs(fileName.c_str());
  if (! ifs)
  {
    const std::string msg = "CsaFile::CsaFile file cannot read ";
    std::cerr << msg << fileName << "\n";
    throw CsaIOError(msg + fileName);
  }
  ICsaRecordStream irs(ifs);
  irs.load(&rec);
}
const osl::record::Record& osl::record::csa::
CsaFile::getRecord() const
{
  return rec;
}

const osl::SimpleState& osl::record::csa::
CsaFile::getInitialState() const
{
  return rec.getInitialState();
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
