#include "osl/record/kisen.h"
#include "osl/record/csaIOError.h"
#include "osl/apply_move/applyMove.h"
#include "osl/pieceStand.h"
#include <iostream>

namespace osl
{
namespace record
{
  Position KisenUtils::convertPosition( int pos ){
    assert(1<=pos && pos<=0x51);
    int y=((pos-1)/9)+1, x=((pos-1)%9)+1;
    return Position(x,y);
  }

  Move KisenUtils::convertMove(SimpleState const& state,int c0,int c1){
    Move move;

    if(1<=c1 && c1<=0x51){
      Position from=convertPosition(c1),to;
      Piece fromPiece=state.getPieceAt(from);
      assert(fromPiece.owner()==state.getTurn() ||
	     (std::cerr << c1 << "," << from << "," << fromPiece << std::endl,0)
	     );
      bool isPromote=false;
      if(1<=c0 && c0<=0x51){
	to=convertPosition(c0);
      }
      else if(0x65<=c0 && c0<=0xb5){
	to=convertPosition(c0-0x64);
	isPromote=true;
      }
      else{
	abort();
      }
      Piece toPiece=state.getPieceAt(to);
      assert(toPiece.isEmpty() || toPiece.owner()==alt(state.getTurn()));
      Ptype ptype=fromPiece.ptype();
      if(isPromote)ptype=promote(ptype);
      const Ptype captured = toPiece.ptype();
      if (captured == KING)
	return Move::INVALID();
      move=Move(from,to,ptype,captured,isPromote,state.getTurn());
    }
    else{
      assert(0x65<=c1);
      assert(1<=c0&&c0<=0x51);
      Position to=convertPosition(c0);
      Ptype ptype=PTYPE_EMPTY;
      int piece_on_stand = c1;
      const Ptype ptypes[]={ROOK,BISHOP,GOLD,SILVER,KNIGHT,LANCE,PAWN};
      for(size_t i=0;i<sizeof(ptypes)/sizeof(Ptype);i++){
	int count=state.countPiecesOnStand(state.getTurn(),ptypes[i]);
	if(count>0){
	  if(piece_on_stand>0x64){
	    piece_on_stand-=count;
	    if(piece_on_stand<=0x64) ptype=ptypes[i];
	  }
	}
      }
      assert(ptype!=PTYPE_EMPTY ||
	     (std::cerr << state << to << " " << c1
	      << " " << piece_on_stand << std::endl, false));
      move=Move(to,ptype,state.getTurn());
    }
    assert(state.isValidMove(move,true) ||
	   (std::cerr << state << move << std::endl, false));
    return move;
  }


  KisenFile::KisenFile(const std::string& fileName) 
    :ifs(fileName.c_str()),initialState(HIRATE) 
  {
    if (! ifs)
      throw CsaIOError("KisenFile not found");
    ifs.seekg(0,std::ios::end);
    assert((ifs.tellg() % 512)==0);
    numberOfGames=ifs.tellg()/512;
  }

  const vector<Move> KisenFile::getMoves(size_t index)
  {
    assert(index<size());
    vector<Move> moves;
    //    std::cerr << "Game[" << index << "]" << std::endl;
    ifs.seekg(index*512,std::ios::beg);
    CArray<unsigned char, 512> cbuf;
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),512);
    SimpleState state(HIRATE);
    //
    Player turn=BLACK;
    for(size_t turnCount=0; 
	(turnCount*2 < cbuf.capacity())
	  && cbuf[turnCount*2]!=0 && cbuf[turnCount*2+1]!=0;
	turnCount++, turn=alt(turn)){
      if(turnCount==KisenFile::maxMoves || cbuf[  turnCount *2 ] == 0 || cbuf[ turnCount * 2 + 1 ] == 0 ){ break; }
      int c0=cbuf[turnCount*2], c1=cbuf[turnCount*2+1];
      const Move move=KisenUtils::convertMove(state,c0,c1);
      if (move.isInvalid())
	break;
      moves.push_back(move);
      ApplyMoveOfTurn::doMove(state, move);
      assert(state.isConsistent( true ) );
    }
    return moves;
  }
  KisenIpxFile::KisenIpxFile(const std::string& fileName) 
    :ifs(fileName.c_str()) 
  {
    if (! ifs)
      throw CsaIOError("KisenIpxFile not found");
    ifs.seekg(0,std::ios::end);
    assert((ifs.tellg() % 256)==0);
    numberOfGames=ifs.tellg()/256;
  }
  const std::string KisenIpxFile::getPlayer(size_t index,Player pl)
  {
    assert(index<size());
    vector<Move> moves;
    ifs.seekg(index*256,std::ios::beg);
    CArray<unsigned char, 256> cbuf;
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),256);
    int startIndex=0;
    if(pl==WHITE)startIndex=14;
    CArray<char,15> buf;
    buf[14]='\0';
    strncpy(&buf[0],reinterpret_cast<char *>(&cbuf[startIndex]),14);
    return std::string(&buf[0]);
  }
  unsigned int KisenIpxFile::getRating(size_t index,Player pl)
  {
    assert(index<size());
    vector<Move> moves;
    ifs.seekg(index*256,std::ios::beg);
    CArray<unsigned char, 256> cbuf;
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),256);
    int startIndex=0324;
    if(pl==WHITE)startIndex=0326;
    return cbuf[startIndex]+256*cbuf[startIndex+1];
  }
  const std::string KisenIpxFile::getTitle(size_t index,Player pl)
  {
    assert(index<size());
    vector<Move> moves;
    ifs.seekg(index*256,std::ios::beg);
    CArray<unsigned char, 256> cbuf;
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),256);
    int startIndex=28;
    if(pl==WHITE)startIndex+=8;
    CArray<char,9> buf;
    buf[8]='\0';
    strncpy(&buf[0],reinterpret_cast<const char*>(&cbuf[startIndex]),8);
    return std::string(&buf[0]);
  }

  KisenPlusFile::KisenPlusFile(const std::string& fileName) 
    :ifs(fileName.c_str()),initialState(HIRATE) 
  {
    if (! ifs)
      throw CsaIOError("KisenPlusFile not found");
    ifs.seekg(0,std::ios::end);
    assert((ifs.tellg() % 2048)==0);
    numberOfGames=ifs.tellg()/2048;
  }

  const vector<Move> KisenPlusFile::getMoves(size_t index)
  {
    vector<Move> moves;
    vector<int> times;
    getMoves(index, moves, times);
    return moves;
  }

  void KisenPlusFile::getMoves(size_t index,
			       vector<Move>& moves, vector<int>& times)
  {
    assert(index<size());
    //    std::cerr << "Game[" << index << "]" << std::endl;
    ifs.seekg(index*2048,std::ios::beg);
    CArray<unsigned char, 2048> cbuf;
    ifs.read(reinterpret_cast<char *>(&cbuf[0]),2048);
    SimpleState state(HIRATE);
    for (size_t i = 0; 
	 i < 2048 && cbuf[i]!=0 && cbuf[i+1]!=0;
	 i += 8)
    {
      int c0 = cbuf[i];
      int c1 = cbuf[i + 1];
      bool is_promote = false;
      Move move;

      if (c0 > 100)
      {
	is_promote = true;
	c0 = 256 - c0;
      }

      Position to(c0 % 10, c0 / 10);

      if (c1 < 10)
      {
	// drop
	move = Move(to,
		    PieceStand::order[c1 - 1],
		    state.getTurn());
      }
      else
      {
	Position from(c1 % 10, c1 / 10);
	Ptype type = state.getPieceAt(from).ptype();
	if (is_promote)
	  type = promote(type);
	move = Move(from, to,
		    type, state.getPieceAt(to).ptype(),
		    is_promote, state.getTurn());
      }
      moves.push_back(move);
      times.push_back(cbuf[i + 7] * 60 + cbuf[i + 6]);
      ApplyMoveOfTurn::doMove(state, move);
      assert(state.isConsistent( true ) );
    }
  }
} // namespace record
} // namespace osl

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
