/* dotAnalyzerProof.cc
 */
#include "osl/search/analyzer/dotAnalyzerProof.h"
#include "osl/search/analyzer/dotWriter.h"
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/hash/hashKey.h"
#include "osl/moveLogProb.h"
#include <iostream>
#include <sstream>
osl::search::analyzer::DotAnalyzerProof::
DotAnalyzerProof(const SimpleHashTable& t, std::ostream& o)
  : DotAnalyzer(t, o)
{
}

osl::search::analyzer::DotAnalyzerProof::
~DotAnalyzerProof()
{
}

void osl::search::analyzer::DotAnalyzerProof::
absoluteBound(unsigned int depth, bool is_lower, int bound, Player turn, 
	      int limit, const SimpleHashRecord *from,
	      const SimpleHashRecord *record, const MoveLogProb& move)
{
  if (! record) 
    return;
  const bool or_node = ((turn == BLACK) ^ is_lower); // parent is or node
  if (move.validMove()) 
  {
    limit -= move.getLogProb();
    if (limit < 0)
      return;
  }

  const SearchMoveSet& moves = record->moves();
  LogWriter::NodeType type = LogWriter::NORMAL;
  int new_bound = bound;
  if (is_lower) 
  {
    if (! (record->hasAbsoluteLowerBound(turn, limit)
	   && (record->absoluteLowerBound(turn) >= bound)))
    {
      if (moves.empty() || or_node)
	return;
      std::cerr << "warning: lower bound proof failure at " 
		<< depth << " " << record << "\n";
      type = LogWriter::ABNORMAL;
    }
    else
    {
      new_bound = record->absoluteLowerBound(turn);
    }
  }
  else
  {
    if (! (record->hasAbsoluteUpperBound(turn, limit)
	   && (record->absoluteUpperBound(turn)-1 <= bound)))
    {
      if (moves.empty() || or_node || (depth == 1))
	return;
      std::cerr << "warning: upper bound proof failure at " 
		<< depth << " " << record << "\n";
      type = LogWriter::ABNORMAL;
    }
    else
    {
      new_bound = record->absoluteUpperBound(turn)-1;
    }
  }
  
  if (! moves.empty())
  {
    {
      std::stringstream debug;
      debug << depth << " " << (or_node ? "or" : "and") 
	    << (is_lower ? " > " : " < ") << bound << " -> " << new_bound;
      writer->showComment(debug.str().c_str());
    }
    if (from) 
      writer->showArc(from, record, move, false);
    writer->showNode(turn, record, limit, type);
    if (type == LogWriter::ABNORMAL)
      return;
    bound = new_bound;
    const MoveLogProb best_move = record->bestMove().moveLogProb();
    for (SearchMoveSet::const_iterator p=moves.begin();
	 p!=moves.end(); ++p)
    {
      assert(turn == p->getMove().player());
      if (p->getMove() == best_move.getMove()) 
      {
	// assume best move meets bounding condition
	absoluteBound(depth+1, is_lower, bound, alt(turn), limit, record,
		      p->record, best_move);
      }
      else 
      {
	absoluteBound(depth+1, is_lower, bound, alt(turn), limit, record, 
		      p->record, p->moveLogProb());
      }
    }
  }
  else
  {
#ifdef WRITE_QUIESCENCE
    // fixme
    if (from) 
      writer->showArc(from, record, move, false);
    writer->showNodeQuiescence(turn, record, limit, false);
#endif
  }
}

void osl::search::analyzer::DotAnalyzerProof::
analyzeQuiescence(const HashKey& root)
{
  MoveLogProb dummy;
  analyze(root, 0);	// fixme
}

void osl::search::analyzer::DotAnalyzerProof::
analyze(const HashKey& key, int limit)
{
  const SimpleHashRecord *record = table.find(key);
  if (! record) 
    return;
  const Player turn = key.turn();
  const bool is_black_turn = (turn == BLACK);
  const int bound
    = (is_black_turn ? record->lowerBound() : record->upperBound());
  const int bound_minus_one = bound - 1;
  MoveLogProb dummy;
  absoluteBound(0, true,  (is_black_turn ? bound : bound_minus_one), 
		turn, limit, 0, record, dummy);
  absoluteBound(0, false, (is_black_turn ? bound_minus_one : bound), 
		turn, limit, 0, record, dummy);
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
