#include "osl/search/searchMoveVector.h"
#include "osl/search/searchMove.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/eval/evalTraits.h"
#include <algorithm>
#include <iostream>

std::ostream& osl::search::operator<<(std::ostream& os,
				      SearchMoveVector const& mv)
{
  os<< "SearchMoveVector" << std::endl;
  for (SearchMoveVector::const_iterator p=mv.begin(); p!=mv.end(); ++p)
  {
    os << (*p)->moveLogProb() << " " << (*p)->record << std::endl;
  }
  return os << std::endl;
}

namespace osl
{
  namespace search
  {
    template <bool prefer_less>
    struct LogProbCompare
    {
      bool operator()(const SearchMove *l, const SearchMove *r) const
      {
	if (prefer_less)
	  return l->getLogProb() < r->getLogProb();
	else
	  return l->getLogProb() > r->getLogProb();
      }
    };

    template <Player P>
    struct SearchValueCompare
    {
      const int default_value;
      explicit SearchValueCompare(int d) : default_value(d) {}
      int value(const SearchMove& move) const
      {
	const SimpleHashRecord *record = move.record;
	if (! record)
	  return default_value;
	if (record->hasUpperBound(0))
	{
	  const int upper = record->upperBound();
	  if (record->hasLowerBound(0))
	    return EvalTraits<P>::max(upper, record->lowerBound());
	  return upper;
	}
	return (record->hasLowerBound(0) ? record->lowerBound() : default_value);
      }
      bool operator()(const SearchMove *l, const SearchMove *r) const
      {
	if (l->record == 0 && r->record == 0)
	{
	  if (l->getLogProb() != r->getLogProb())
	    return l->getLogProb() < r->getLogProb();
	}
	const int l_value = value(*l);
	const int r_value = value(*r);
	return EvalTraits<P>::betterThan(l_value, r_value);
      }
    };

    template void SearchMoveVector::sortByValue<BLACK>(int);
    template void SearchMoveVector::sortByValue<WHITE>(int);
  }
}

void osl::search::
SearchMoveVector::sortByProbability()
{
  std::sort(begin(), end(), LogProbCompare<true>());
}
void osl::search::
SearchMoveVector::sortByProbabilityReverse()
{
  std::sort(begin(), end(), LogProbCompare<false>());
}

template <osl::Player P>
void osl::search::
SearchMoveVector::sortByValue(int default_value)
{
  std::sort(begin(), end(), SearchValueCompare<P>(default_value));
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
