/* simpleHashRecord.cc
 */
#include "osl/search/simpleHashRecord.h"
#include "osl/record/csa.h"
#include <iostream>

void osl::search::
SimpleHashRecord::dump(std::ostream& os) const
{
  os << "SimpleHashRecord " << this << "\n";
  os << "best move " << record::csa::show(best_move.getMove())
	    << " " << best_move.getLogProb()
	    << " " << best_move.record << "\t";
  os << "limit: l " << lower_limit << " u " << upper_limit << "\n";
  os << "in_check " << is_king_in_check << "\n";
  os << "checkmate read " << checkmate_nodes << "\t"
     << "threatmate read " << threatmate_nodes << "\n";
  os << "tried oracle " << qrecord.attack_oracle.proof 
     << " " << qrecord.attack_oracle.proof_last_move
     << " " << qrecord.attack_oracle.disproof << "\n";
  if (hasLowerBound(0))
    os <<  lowerBound();
  else 
    os << "*";
  os << " < ";
  if (hasUpperBound(0))
    os <<  upperBound();
  else 
    os << "*";
  os << "\n";
  os << "moves " << limit_of_moves << " " << used_categories << "\n";
  moves().dump(os);
  qrecord.dump(os);
}

void osl::search::
SimpleHashRecord::copyFrom(const SimpleHashRecord& src)
{
  *this = src;
  // clear limitations
  checkmate_nodes = std::min(checkmate_nodes, 1);
  threatmate_nodes = std::min(threatmate_nodes, 1);
  qrecord.attack_oracle = AttackOracleAges();
  qrecord.threatmate_oracle = AttackOracleAges();
  for (SearchMoveSet::iterator p=moves().begin(); p!=moves().end(); ++p)
  {
    p->record = 0;
  }
}


void osl::search::
SimpleHashRecord::fixBestMove()
{
  if (bestMove().record)
  {
    SearchMove *stored_move = moves().find(bestMove().getMove());
    best_move.record = stored_move->record;
  }
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
