/* simpleHashTable.cc
 */
#include "osl/search/simpleHashTable.h"
#include "osl/search/simpleHashRecord.h"
#include "osl/container/generalSimpleHashTable.tcc"
#include <iostream>

namespace osl
{
  template class container::GeneralSimpleHashTable <SimpleHashRecord>;
} // namespace osl
  
osl::search::SimpleHashTable::
SimpleHashTable(unsigned int capacity, int minimum_recordlimit, bool v) 
  : GeneralSimpleHashTable<SimpleHashRecord>(capacity),
    minimum_limit(minimum_recordlimit), verbose(v)
{
}

osl::search::SimpleHashTable::
~SimpleHashTable() 
{
  if (verbose)
  {
    std::cerr << "SimpleHashTable size: " << size() 
	      << ", cache hit " << table->numCacheHit
	      << ", table full " << table->numRecordAfterFull << "\n";
  }
}

void osl::search::SimpleHashTable::
setVerbose(bool v)
{
  verbose = v;
}

void osl::search::SimpleHashTable::
setMinimumRecordLimit(int new_limit)
{
  minimum_limit = new_limit;
}

int osl::search::SimpleHashTable::
minimumRecordLimit() const
{
  return minimum_limit;
}

osl::search::SimpleHashRecord * 
osl::search::SimpleHashTable::
allocate(const HashKey& key, int limit)
{
  if (limit < minimumRecordLimit())
    return find(key);
  return GeneralSimpleHashTable <SimpleHashRecord>::allocate (key);
}

bool osl::search::SimpleHashTable::
isVerbose() const
{
  return verbose;
}

osl::search::SimpleHashRecord* osl::search::SimpleHashTable::
migrate(const HashKey& root, int max_depth, int min_limit,
	SimpleHashTable& out) const
{
  const SimpleHashRecord *original = find(root);
  if (original == 0)
    return 0;
  const int target_limit
    = std::max(original->lowerLimit(), original->upperLimit());
  if (target_limit < min_limit)
    return 0;
  SimpleHashRecord *copy = out.find(root);
  if (copy)			// already recorded
    return copy;
  copy = out.allocate(root, target_limit);
  if (! copy)
    return 0;
  copy->copyFrom(*original);
  if (max_depth > 0)
  {
    for (SearchMoveSet::const_iterator p=copy->moves().begin();
	 p!=copy->moves().end(); ++p)
    {
      const HashKey next = root.newHashWithMove(p->getMove());
      p->record = migrate(next, max_depth - 1, min_limit, out);
    }
  }
  copy->fixBestMove();
  return copy;
}

bool osl::search::SimpleHashTable::
isConsistent() const
{
  for (Table::const_iterator p=table->begin(); p!=table->end(); ++p)
  {
    const HashKey& key = p->first;
    const SimpleHashRecord& record = p->second;
    for (SearchMoveSet::const_iterator q=record.moves().begin();
	 q!=record.moves().end(); ++q)
    {
      if (q->record == 0)
	continue;
      const HashKey next = key.newHashWithMove(q->getMove());
      if (q->record != find(next))
	return false;
    }
  }
  return true;
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
