// generate.cc
#include "osl/category/standardMoveGenerator.h"
#include "osl/category/categoryListUtil.h"
#include "osl/container/moveLogProbSet.h"
#include "osl/record/csaRecord.h"
#include "osl/apply_move/applyMove.h"
#include <iostream>
#include <cmath>

using namespace osl;
using namespace osl::category;

void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-d] [-l limit] csa_files\n"
       << endl;
  exit(1);
}

/** move generation limit */
int limit = 800;

void testFile(const char *file_name);

int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  extern char *optarg;
  extern int optind;
    
  char c;
  while ((c = getopt(argc, argv, "l:vh")) != EOF)
  {
    switch(c)
    {
    case 'l':	limit = atoi(optarg);
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (argc < 1))
    usage(program_name);

  for (int i=0; i<argc; ++i)
  {
    testFile(argv[i]);
  }
}

void testState(const CategoryEnv& env)
{
  std::cout << *env.state;
  MoveLogProbSet moves;
  CategoryListUtil::gatherAllMoves<StandardMoveGenerator,MoveLogProbSet>
    (env, moves);

  for (MoveLogProbSet::const_iterator p=moves.begin(); p!=moves.end(); ++p)
  {
    std::cout << p->getMove() << " " << p->getLogProb()
	      << " " << pow(0.5, (p->getLogProb()/100.0)) << "\n";
  }
}

void testFile(const char *filename)
{
  const Record record = CsaFile(filename).getRecord();
  NumEffectState state(record.getInitialState());
  const vector<osl::Move> moves=record.getMoves();

  MoveStack history;
  CategoryEnv env(&state, limit, &history);

  testState(env);
  for (size_t i=0; i<moves.size(); ++i)
  {
    const Move move = moves[i];
    assert(state.isValidMove(move));

    ApplyMoveOfTurn::doMove(state, move);
    history.push(move);

    testState(env);
  }
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
