/* pairedit.cc
 */

#include "osl/eval/ppair/piecePairRawEval.h"
#include <boost/scoped_ptr.hpp>
#include <iostream>
#include <cstdlib>
#include <unistd.h>

/**
 * @file
 * PiecePairEvalTable ˿ͰŪ˼ä
 *
 * ˡ̤μ5x5 üǤ4x4 ˤŨζ+100Ƥߤ
 * ˹϶̤겼Ǥϲʤ˾Ф
 * ˤȾʬ?
 * ĺȾʬ?
 *
 * 
 * http://www32.ocn.ne.jp/~yss/book.html#SEC3
 * ˡζ϶̤ºݤΰ֤⤦ʾˤȤƷ׻Ƥʷ˹ǤṲ̃飳ʾ夬Ȥʤ
 */
using namespace osl;
using namespace osl::eval;

void usage(const char *prog)
{
  using namespace std;
  cerr << "Usage: " << prog << " [-f read-pair-file-name] [-o write-pair-file-name] "
       << endl;
  exit(1);
}

bool verbose = false;

void adjust(PiecePairRawTable& table, 
	    Position pos1, PtypeO ptypeo1,
	    Position pos2, PtypeO ptypeo2,
	    int value)
{
  const unsigned int index1 = table.indexOf(pos1, ptypeo1);
  const unsigned int index2 = table.indexOf(pos2, ptypeo2);
  
  int val = table.valueOf(index1, index2);
  assert(val == table.valueOf(index2, index1));
  if (verbose)
    std::cerr << pos1 << ptypeo1 << " " << pos2 << ptypeo2
	      << " " << val << " => ";
  val += value;
  val = std::min(127, val);
  val = std::max(-127, val);
  if (verbose)
    std::cerr << val << "\n";
  table.valueOf(index1, index2) = val;
  table.valueOf(index2, index1) = val;
}

void adjustKingBonus(PiecePairRawTable& table, 
		     Position pos1, PtypeO ptypeo1, // king
		     Position pos2, PtypeO ptypeo2, // attacker
		     int bonus)
{
  assert(getPtype(ptypeo1) == KING);
  assert(getPtype(ptypeo2) != KING);
  assert(getOwner(ptypeo1) != getOwner(ptypeo2));
  assert((bonus > 0) ^ (getOwner(ptypeo1) == BLACK));

  adjust(table, pos1, ptypeo1, pos2, ptypeo2, bonus);
}

void adjustDual(PiecePairRawTable& table, 
		Position king, Position attacker, Ptype attackerType,
		int blackAttackBonus, int whiteAttackBonus)
{
  adjustKingBonus(table, king, newPtypeO(BLACK, KING),
	 attacker, newPtypeO(WHITE, attackerType),
	 whiteAttackBonus);
  adjustKingBonus(table, king, newPtypeO(WHITE, KING),
	 attacker, newPtypeO(BLACK, attackerType),
	 blackAttackBonus);
}

void addValue(Player player, PiecePairRawTable& table,
	      Position pos1, Ptype ptype1,
	      Position pos2, Ptype ptype2,
	      int bonus)
{
  const PtypeO ptypeo1 = newPtypeO(player, ptype1);
  const PtypeO ptypeo2 = newPtypeO(player, ptype2);
  adjust(table, pos1, ptypeo1, pos2, ptypeo2, bonus);
}

void addPenalty(Player player, PiecePairRawTable& table,
		Position pos1, Ptype ptype1,
		Position pos2, Ptype ptype2,
		int bonus)
{
  assert(eval::betterThan(player, 0, bonus));
  addValue(player, table, pos1, ptype1, pos2, ptype2, bonus);
}

void addBonus(Player player, PiecePairRawTable& table,
	      Position pos1, Ptype ptype1,
	      Position pos2, Ptype ptype2,
	      int bonus)
{
  assert(eval::betterThan(player, bonus, 0));
  addValue(player, table, pos1, ptype1, pos2, ptype2, bonus);
}

void addPenaltyDual(PiecePairRawTable& table, 
		    Position pos1, Ptype ptype1,
		    Position pos2, Ptype ptype2,
		    int black_bonus)
{
  assert(black_bonus < 0);
  addPenalty(BLACK, table, pos1, ptype1, pos2, ptype2,  black_bonus);
  addPenalty(WHITE, table, pos1, ptype1, pos2, ptype2, -black_bonus);
}

void addSelfPenaltyDual(PiecePairRawTable& table, 
			Position pos, Ptype ptype,
			int black_bonus)
{
  addPenaltyDual(table, pos, ptype, pos, ptype, black_bonus);
}


int main(int argc, char **argv)
{
  const char *program_name = argv[0];
  bool error_flag = false;
  const char *read_pairfilename = 0;
  const char *write_pairfilename = 0;
  
  extern char *optarg;
  extern int optind;
  char c;
  while ((c = getopt(argc, argv, "f:o:vh")) != EOF)
  {
    switch(c)
    {
    case 'f':	read_pairfilename = optarg;
      break;
    case 'o':	write_pairfilename = optarg;
      break;
    case 'v':	verbose = true;
      break;
    default:	error_flag = true;
    }
  }
  argc -= optind;
  argv += optind;

  if (error_flag || (! read_pairfilename) || (! write_pairfilename))
    usage(program_name);

  boost::scoped_ptr<PiecePairRawTable> table(new PiecePairRawTable);
  table->loadFromBinaryFile(read_pairfilename);

  // ̾5x5ϰ
  for (int kingX=1; kingX<=9; ++kingX)
  {
    for (int kingY=1; kingY<=9; ++kingY)
    {
      const Position king(kingX,kingY);
      for (int attacker_x = ((kingX==9) ? kingX-3 : kingX-2); 
	   attacker_x <= ((kingX==1) ? kingX+3 : kingX+2); ++attacker_x)
      {
	if ((attacker_x < 1) || (attacker_x > 9))
	  continue;
	for (int attacker_y = ((kingY==9) ? kingY-3 : kingY-2); 
	     attacker_y <= ((kingY==1) ? kingY+3 : kingY+2); ++attacker_y)
	{
	  if ((attacker_y < 1) || (attacker_y > 9))
	    continue;
	  const Position attacker(attacker_x,attacker_y);
	  if (king == attacker)
	    continue;

	  adjustDual(*table, king, attacker, PPAWN,   100, -100);
	  adjustDual(*table, king, attacker, PLANCE,  100, -100);
	  adjustDual(*table, king, attacker, PKNIGHT, 100, -100);
	  adjustDual(*table, king, attacker, PBISHOP, 100, -100);
	  adjustDual(*table, king, attacker, PROOK,   100, -100);
	  adjustDual(*table, king, attacker, SILVER,  100, -100);
	  adjustDual(*table, king, attacker, BISHOP,  100, -100);
	  adjustDual(*table, king, attacker, ROOK,    100, -100);

	  // 1ܰʳ
	  if (attacker_y != 1)
	  {
	    const int bonus = 
	      (attacker_y >= kingY) ? 100 : 50;
	    adjustKingBonus(*table, king, newPtypeO(WHITE, KING),
			    attacker, newPtypeO(BLACK, PSILVER),
			    bonus);
	    adjustKingBonus(*table, king, newPtypeO(WHITE, KING),
			    attacker, newPtypeO(BLACK, GOLD),
			    bonus);
	  }
	  if (attacker_y != 9)
	  {
	    const int bonus = 
	      (attacker_y <= kingY) ? -100 : -50;
	    adjustKingBonus(*table, king, newPtypeO(BLACK, KING),
			    attacker, newPtypeO(WHITE, PSILVER),
			    bonus);
	    adjustKingBonus(*table, king, newPtypeO(BLACK, KING),
			    attacker, newPtypeO(WHITE, GOLD),
			    bonus);
	  }
	  // ϲƱ󤫤
	  if (attacker_y >= kingY)
	  {
	    adjustKingBonus(*table, king, newPtypeO(WHITE, KING),
			    attacker, newPtypeO(BLACK, PAWN),
			    100);
	  }
	  if (attacker_y <= kingY)
	  {
	    adjustKingBonus(*table, king, newPtypeO(BLACK, KING),
			    attacker, newPtypeO(WHITE, PAWN),
			    -100);
	  }
	} // attacker_y
	// ˹ĹܤˤȤ
	for (int attacker_y = ((kingY==9) ? kingY-4 : kingY-3);
	     attacker_y <= ((kingY==1) ? kingY+4 : kingY+3); ++attacker_y)
	{
	  if ((attacker_y < 1) || (attacker_y > 9))
	    continue;
	  const Position attacker(attacker_x,attacker_y);
	  if (king == attacker)
	    continue;

	  // , 25,45,65,85 礭Τǽ
	  if (! ((attacker_y == 5)
		 && ((attacker_x == 2)
		     || (attacker_x == 4)
		     || (attacker_x == 6)
		     || (attacker_x == 8))))
	  {
	    if ((attacker_y > kingY) && (attacker_y > 2))
	    {
	      adjustKingBonus(*table, king, newPtypeO(WHITE, KING),
			      attacker, newPtypeO(BLACK, LANCE),
			      100);
	      adjustKingBonus(*table, king, newPtypeO(WHITE, KING),
			      attacker, newPtypeO(BLACK, KNIGHT),
			      100);
	    }
	    if ((attacker_y < kingY) && (attacker_y < 8))
	    {
	      adjustKingBonus(*table, king, newPtypeO(BLACK, KING),
			      attacker, newPtypeO(WHITE, LANCE),
			      -100);
	      adjustKingBonus(*table, king, newPtypeO(BLACK, KING),
			      attacker, newPtypeO(WHITE, KNIGHT),
			      -100);
	    }
	  }
	} // attacker_y, knight, lance
      }	// attacker_x
    } // kingY
  } // kingX
  
  for (int attacker_x=1; attacker_x<=9; ++attacker_x)
  {
    for (int attacker_y=1; attacker_y<=9; ++attacker_y)
    {
      const Position attacker(attacker_x,attacker_y);
      // üζϿ
      if ((attacker_x == 1) || (attacker_x == 9))
      {
	addSelfPenaltyDual(*table, attacker, KNIGHT,  -100);
	addSelfPenaltyDual(*table, attacker, SILVER,  -100);
	addSelfPenaltyDual(*table, attacker, GOLD  ,  -100);
	addSelfPenaltyDual(*table, attacker, PPAWN ,  -100);
	addSelfPenaltyDual(*table, attacker, PLANCE,  -100);
	addSelfPenaltyDual(*table, attacker, PKNIGHT, -100);
	addSelfPenaltyDual(*table, attacker, PSILVER, -100);
      }
    }
  }
  // ˤ
  addPenalty(BLACK, *table, Position(2,6), SILVER, Position(3,7), KNIGHT, -80);
  addPenalty(WHITE, *table, Position(8,4), SILVER, Position(7,3), KNIGHT, +80);
  
  // 귧ط
  addPenalty(BLACK, *table, Position(9,9), KING, Position(7,7), KNIGHT, -120);
  addPenalty(BLACK, *table, Position(9,9), KING, Position(9,7), KNIGHT, -120);
  addPenalty(BLACK, *table, Position(1,9), KING, Position(3,7), KNIGHT, -120);
  addPenalty(BLACK, *table, Position(1,9), KING, Position(1,7), KNIGHT, -120);
  addPenalty(WHITE, *table, Position(9,1), KING, Position(7,3), KNIGHT, +120);
  addPenalty(WHITE, *table, Position(9,1), KING, Position(9,3), KNIGHT, +120);
  addPenalty(WHITE, *table, Position(1,1), KING, Position(3,3), KNIGHT, +120);
  addPenalty(WHITE, *table, Position(1,1), KING, Position(1,3), KNIGHT, +120);

  // ̤Ƥ˹᤬褵
  addPenalty(BLACK, *table, Position(9,9), KING, Position(9,9), KING, -120);
  addBonus(BLACK, *table, Position(9,9), KING, Position(9,8), LANCE, 40);
  addBonus(BLACK, *table, Position(9,9), KING, Position(8,9), KNIGHT, 80);

  addPenalty(BLACK, *table, Position(1,9), KING, Position(1,9), KING, -120);
  addBonus(BLACK, *table, Position(1,9), KING, Position(1,8), LANCE, 40);
  addBonus(BLACK, *table, Position(1,9), KING, Position(2,9), KNIGHT, 80);

  addPenalty(WHITE, *table, Position(9,1), KING, Position(9,1), KING, +120);
  addBonus(WHITE, *table, Position(9,1), KING, Position(9,2), LANCE, -40);
  addBonus(WHITE, *table, Position(9,1), KING, Position(8,1), KNIGHT, -80);

  addPenalty(WHITE, *table, Position(1,1), KING, Position(1,1), KING, +120);
  addBonus(WHITE, *table, Position(1,1), KING, Position(1,2), LANCE, -40);
  addBonus(WHITE, *table, Position(1,1), KING, Position(2,1), KNIGHT, -80);

  // 5ܤη (줽)
  for (int x=2; x<=8; ++x)	// 1,9ϥڥʥƥ
  {
    // 
    const Position attacker(x,5);
    addSelfPenaltyDual(*table, attacker, KNIGHT, -100);

    // ⤬󤱤פ⡥
    adjust(*table, Position(x,1), newPtypeO(WHITE,PAWN), attacker, newPtypeO(BLACK,KNIGHT), 100);
    adjust(*table, Position(x,2), newPtypeO(WHITE,PAWN), attacker, newPtypeO(BLACK,KNIGHT), 100);
    adjust(*table, Position(x,9), newPtypeO(BLACK,PAWN), attacker, newPtypeO(WHITE,KNIGHT), -100);
    adjust(*table, Position(x,8), newPtypeO(BLACK,PAWN), attacker, newPtypeO(WHITE,KNIGHT), -100);
  }

  // Ũȼؤζ
  for (int promoted_x=1; promoted_x<=9; ++promoted_x)
  {
    for (int x=1; x<=9; ++x)
    {
      for (int promoted_y=1; promoted_y<=3; ++promoted_y)
      {
	const Position promoted(promoted_x, promoted_y);
	for (int y=1; y<=3; ++y)
	{
	  const Position unpromoted(x, y);
	  if (promoted == unpromoted)
	    continue;
	  adjust(*table, promoted, newPtypeO(BLACK,PROOK), 
		 unpromoted, newPtypeO(WHITE,ROOK), 100);
	  adjust(*table, promoted, newPtypeO(BLACK,PROOK), 
		 unpromoted, newPtypeO(WHITE,BISHOP), 100);
	  adjust(*table, promoted, newPtypeO(BLACK,PBISHOP), 
		 unpromoted, newPtypeO(WHITE,ROOK), 100);
	  adjust(*table, promoted, newPtypeO(BLACK,PBISHOP), 
		 unpromoted, newPtypeO(WHITE,BISHOP), 100);
	}
      }
      for (int promoted_y=7; promoted_y<=9; ++promoted_y)
      {
	const Position promoted(promoted_x, promoted_y);
	for (int y=7; y<=9; ++y)
	{
	  const Position unpromoted(x, y);
	  if (promoted == unpromoted)
	    continue;
	  adjust(*table, promoted, newPtypeO(WHITE,PROOK), 
		 unpromoted, newPtypeO(BLACK,ROOK), -100);
	  adjust(*table, promoted, newPtypeO(WHITE,PROOK), 
		 unpromoted, newPtypeO(BLACK,BISHOP), -100);
	  adjust(*table, promoted, newPtypeO(WHITE,PBISHOP), 
		 unpromoted, newPtypeO(BLACK,ROOK), -100);
	  adjust(*table, promoted, newPtypeO(WHITE,PBISHOP), 
		 unpromoted, newPtypeO(BLACK,BISHOP), -100);
	}
      }
    } // x
  }
  // 4,6ܤޤߥܡʥ
  for (int x=1; x<=9; ++x)
  {
    // 
    const Position black_position(x,4);
    const Position white_position(x,6);
    addBonus(BLACK, *table, black_position, PAWN, black_position, PAWN, 100);
    addBonus(WHITE, *table, white_position, PAWN, white_position, PAWN,-100);
  }
  
  table->writeInBinaryFile(write_pairfilename);
}


/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
