#include "osl/category/check.h"
#include "osl/category/categoryUtil.h"
#include "osl/record/csaString.h"
#include "osl/state/numEffectState.h"
#include "osl/container/moveLogProbVector.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <algorithm>
#include <iostream>

extern bool isShortTest;

namespace osl
{
  namespace category
  {

    class CheckTest : public CppUnit::TestFixture 
    {
      CPPUNIT_TEST_SUITE(CheckTest);
      CPPUNIT_TEST(testCount);
      CPPUNIT_TEST(testSendOff);
      CPPUNIT_TEST_SUITE_END();
    public:
      void testCount();
      void testSendOff();
    };
  }
}

CPPUNIT_TEST_SUITE_REGISTRATION(osl::category::CheckTest);

using namespace osl;
using namespace osl::category;

void CheckTest::testSendOff(){
  NumEffectState state(CsaString(
			 "P1 * -GI * +HI *  *  *  *  * \n"
			 "P2-OU *  *  *  *  *  *  *  * \n"
			 "P3 *  * +TO *  *  *  *  *  * \n"
			 "P4 *  *  * +KI *  *  *  *  * \n"
			 "P5 * +KI *  *  *  *  *  *  * \n"
			 "P6 *  *  *  *  *  *  *  *  * \n"
			 "P7 *  *  *  *  *  *  *  *  * \n"
			 "P8 *  *  *  *  *  *  * +GI+KY\n"
			 "P9 *  *  *  *  *  *  * +KE+OU\n"
			 "P-00AL\n"
			 "+\n").getInitialState());
  MoveLogProbVector moves;
  CategoryEnv env(&state, 1000, 0);
  CategoryUtil<Check>::generate(env,moves);
  moves.sortByProbability();
  
  const Move m83(Position(7,3), Position(8,3), PPAWN, PTYPE_EMPTY, false, BLACK);
  const Move m82(Position(7,3), Position(8,2), PPAWN, PTYPE_EMPTY, false, BLACK);
  int p83 = -1, p82 = -1;
  for (size_t i=0; i<moves.size(); ++i)
  {
    if (moves[i].getMove() == m83)
      p83 = moves[i].getLogProb();
    else if (moves[i].getMove() == m82)
      p82 = moves[i].getLogProb();
  }
  CPPUNIT_ASSERT(p83 > 0);
  CPPUNIT_ASSERT(p82 > 0);
  CPPUNIT_ASSERT(p83 < p82);
}

void CheckTest::testCount()
{
  NumEffectState neState(CsaString(
			   "P1+NY+TO *  *  *  * -OU-KE-KY\n"
			   "P2 *  *  *  *  * -GI-KI *  *\n"
			   "P3 * +RY *  * +UM * -KI-FU-FU\n"
			   "P4 *  * +FU-FU *  *  *  *  *\n"
			   "P5 *  * -KE * +FU *  * +FU *\n"
			   "P6-KE *  * +FU+GI-FU *  * +FU\n"
			   "P7 *  * -UM *  *  *  *  *  *\n"
			   "P8 *  *  *  *  *  *  *  *  * \n"
			   "P9 * +OU * -GI *  *  *  * -NG\n"
			   "P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
			   "P-00KI00KY00FU00FU\n"
			   "P-00AL\n"
			   "+\n").getInitialState());

  MoveLogProbVector moves;
  CategoryEnv env(&neState, 1000, 0);
  CategoryUtil<Check>::generate(env,moves);
  CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(8u), moves.size());
#if 0
  if (! isShortTest)
    std::cerr << moves;
#endif
// Drop(+,GOLD,Position(22))
// Move(+,PBISHOP,capture=SILVER,Position(53)->Position(42),promote=0)
// Drop(+,GOLD,Position(41))
// Drop(+,ROOK,Position(41))
// Drop(+,ROOK,Position(51))
// Drop(+,ROOK,Position(61))
// Drop(+,ROOK,Position(71))
// Drop(+,KNIGHT,Position(43))
  moves.clear();

  env.limit = 400;
  CategoryUtil<Check>::generate(env,moves);
  if (! isShortTest)
    std::cerr << moves;
  CPPUNIT_ASSERT_EQUAL(static_cast<size_t>(3u), moves.size());
// Drop(+,ROOK,Position(51)) ֤ʤǧƤǤ
// Drop(+,ROOK,Position(61))
// Drop(+,ROOK,Position(71))
}

#if 0
void CheckTest::testShow2()
{
  SimpleState bugState=CsaString(
"P1-KY-KE * -KI *  *  * -KE-KY\n"
"P2 * -OU-GI * -KI * -HI *  * \n"
"P3 * -FU *  * -FU-GI-KA-FU-FU\n"
"P4-FU * -FU-FU * -FU-FU *  * \n"
"P5 *  *  *  *  *  *  * +FU * \n"
"P6+FU * +FU * +FU+GI+FU * +FU\n"
"P7 * +FU * +FU * +FU *  *  * \n"
"P8 * +KA+OU * +KI+GI * +HI * \n"
"P9+KY+KE * +KI *  *  * +KE+KY\n"
"+\n").getInitialState();
  bugState.doMove(Move(Position(9,6),Position(9,5),PAWN,PTYPE_EMPTY,false,BLACK));
  bugState.doMove(Move(Position(9,4),Position(9,5),PAWN,PAWN,false,WHITE));
  bugState.doMove(Move(Position(8,8),Position(4,4),BISHOP,PAWN,false,BLACK));
  
  NumEffectState neState(bugState);

  MoveLogProbVector moves;
  CaptureCheckMoveGenerator action1;
  MoveStack history;
  history.push(Move(Position(4,4),PAWN,WHITE));
  action1.generate(neState,history,500, moves);
  // std::cerr << action1;

  const MoveLogProb expected1(Move(Position(4,3),Position(4,4),SILVER,BISHOP,false,WHITE),44);
  const MoveLogProb expected2(Move(Position(3,3),Position(4,4),BISHOP,BISHOP,false,WHITE),44);

  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected1));
  CPPUNIT_ASSERT(std::find(moves.begin(), moves.end(), expected2));
}
#endif
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
