#include "osl/apply_move/applyMove.h"
#include "osl/category/categoryEnv.h"
#include "osl/category/gold.h"
#include "osl/category/classify.h"
#include "osl/category/classify.tcc"
#include "osl/container/moveStack.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/state/numEffectState.h"
#include "osl/stl/vector.h"
#include "osl/record/csaString.h"
#include "osl/record/csa.h"
#include "osl/record/record.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>

class CategoryGoldDiagonalTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(CategoryGoldDiagonalTest);
  CPPUNIT_TEST(testGenerate);
  CPPUNIT_TEST(testPrepareForkGold);
  CPPUNIT_TEST_SUITE_END();
 public:
  void testGenerate();
  void testGenerateOne(const char *board, const char *move);

  void testPrepareForkGold();
};

CPPUNIT_TEST_SUITE_REGISTRATION(CategoryGoldDiagonalTest);
  
void CategoryGoldDiagonalTest::testGenerate()
{
  testGenerateOne("P1 *  *  * +GI *  *  *  *  * \n"
		  "P2 *  *  *  * -KI-OU *  *  * \n"
		  "P3-FU-FU-FU *  * -FU-FU-FU-FU\n"
		  "P4 *  *  *  *  *  *  *  *  * \n"
		  "P5 *  *  *  *  *  *  *  *  * \n"
		  "P6 *  *  *  *  *  *  *  *  * \n"
		  "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		  "P8 * +KA *  *  *  *  * +HI * \n"
		  "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		  "P+00GI00HI\n"
		  "P-00AL\n"
		  "+\n", "+0063GI");
  testGenerateOne("P1 *  *  *  *  *  *  *  *  * \n"
		  "P2 *  *  *  * -KI-OU *  *  * \n"
		  "P3-FU-FU-FU *  * -FU-FU-FU-FU\n"
		  "P4 *  *  *  *  *  *  *  *  * \n"
		  "P5 *  *  *  *  *  *  *  *  * \n"
		  "P6 *  *  *  *  *  *  *  *  * \n"
		  "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		  "P8 * +KA *  *  *  *  * +HI * \n"
		  "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		  "P+00GI00HI\n"
		  "P-00AL\n"
		  "+\n", "");
  testGenerateOne("P1 *  *  * +GI *  *  *  *  * \n"
		  "P2 *  *  *  * -KI-OU *  *  * \n"
		  "P3-FU-FU-FU *  * -FU-FU-FU-FU\n"
		  "P4 *  *  *  *  *  *  *  *  * \n"
		  "P5 *  *  *  *  *  *  *  *  * \n"
		  "P6 *  *  *  *  *  *  *  *  * \n"
		  "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		  "P8 * +KA *  *  *  *  * +HI * \n"
		  "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		  "P+00GI\n"
		  "P-00AL\n"
		  "+\n", "");
  testGenerateOne("P1 *  *  * +GI-OU *  *  *  * \n"
		  "P2 *  *  *  * -KI *  *  *  * \n"
		  "P3-FU-FU-FU *  * -FU-FU-FU-FU\n"
		  "P4 *  *  *  *  *  *  *  *  * \n"
		  "P5 *  *  *  *  *  *  *  *  * \n"
		  "P6 *  *  *  *  *  *  *  *  * \n"
		  "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		  "P8 * +KA *  *  *  *  * +HI * \n"
		  "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		  "P+00GI00KI\n"
		  "P-00AL\n"
		  "+\n", "+0063GI");

  testGenerateOne("P1-KY-KE-GI-OU-KI-KI-GI-KE-KY\n"
		  "P2 * -HI *  *  *  *  * -KA * \n"
		  "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		  "P4 *  *  *  *  *  *  *  *  * \n"
		  "P5 *  *  *  *  *  *  *  *  * \n"
		  "P6 *  *  *  *  *  *  *  *  * \n"
		  "P7+FU+FU+FU * +FU+FU+FU+FU+FU\n"
		  "P8 * +KA *  * +KI *  * +HI * \n"
		  "P9 *  *  *  * +OU-GI *  *  * \n"
		  "P-00GI00KI\n"
		  "P+00AL\n"
		  "-\n", "-0067GI");
}

void CategoryGoldDiagonalTest::
testGenerateOne(const char *board, const char *move)
{
  osl::record::csa::CsaString csa(board);
  osl::state::SimpleState initialState = csa.getInitialState();
  osl::state::NumEffectState state(initialState);
  osl::container::MoveStack history;

  osl::container::MoveLogProbVector generatedMoves;
  osl::category::Classify<osl::category::CategoryList<osl::category::LureGoldDiagonally, osl::category::NullCategory> > classify;
  osl::category::CategoryEnv env(&state, osl::category::CategoryTraits::maxProb(),
				 &history);
  classify.generate(env, generatedMoves);
  if (strcmp(move, "") == 0)
  {
    CPPUNIT_ASSERT_EQUAL(0, static_cast<int>(generatedMoves.size()));
  }
  else
  {
    CPPUNIT_ASSERT_EQUAL(1, static_cast<int>(generatedMoves.size()));
    CPPUNIT_ASSERT(osl::record::csa::strToMove(move, initialState) ==  generatedMoves[0].getMove());
  }
}

void CategoryGoldDiagonalTest::testPrepareForkGold()
{
  using namespace osl;
  using namespace osl::category;
  {
    NumEffectState state(CsaString(
			   "P1-KY+RY *  *  * -KI * -KE-KY\n"
			   "P2 *  *  *  *  * -KI-OU *  * \n"
			   "P3 *  *  *  * -FU-FU * -FU * \n"
			   "P4-FU *  *  * -GI * -FU * -FU\n"
			   "P5 * -FU+FU *  *  *  *  *  * \n"
			   "P6+FU *  * +FU+GI+FU *  * +FU\n"
			   "P7+KY * +KE+KI *  * +FU+FU * \n"
			   "P8 *  *  *  *  *  * +GI+OU * \n"
			   "P9 * -RY *  *  * +KI * +KE+KY\n"
			   "P+00KA00KE00FU00FU00FU00FU\n"
			   "P-00KA00GI\n"
			   "-\n").getInitialState());
    const Move m48gi(Position(4,8),SILVER,WHITE);
    CategoryEnv env(&state, 1000, 0);
    CPPUNIT_ASSERT(PrepareForkGold::isInCategory(env, m48gi));
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY+RY *  *  * -KI * -KE-KY\n"
			   "P2 *  *  *  *  * -KI-OU *  * \n"
			   "P3 *  *  *  * -FU-FU * -FU * \n"
			   "P4-FU *  *  * -GI * -FU * -FU\n"
			   "P5 * -FU+FU *  *  *  *  *  * \n"
			   "P6+FU *  * +FU * +FU *  * +FU\n"
			   "P7+KY * +KE+KI+GI * +FU+FU * \n"
			   "P8 *  *  *  *  *  * +GI+OU * \n"
			   "P9 * -RY *  *  * +KI * +KE+KY\n"
			   "P+00KA00KE00FU00FU00FU00FU\n"
			   "P-00KA00GI\n"
			   "-\n").getInitialState());
    const Move m48gi(Position(4,8),SILVER,WHITE);
    CategoryEnv env(&state, 1000, 0);
    CPPUNIT_ASSERT(! PrepareForkGold::isInCategory(env, m48gi));
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY+RY *  *  * -KI * -KE-KY\n"
			   "P2 *  *  *  *  * -KI-OU *  * \n"
			   "P3 *  *  *  * -FU-FU * -FU * \n"
			   "P4-FU *  *  * -GI * -FU * -FU\n"
			   "P5 * -FU+FU *  *  *  *  *  * \n"
			   "P6+FU *  * +FU+GI+FU *  * +FU\n"
			   "P7+KY * +KE+KI *  * +FU+FU * \n"
			   "P8 *  *  *  *  *  * +GI+OU * \n"
			   "P9 * -RY *  * +FU+KI * +KE+KY\n"
			   "P+00KA00KE00FU00FU00FU\n"
			   "P-00KA00GI\n"
			   "-\n").getInitialState());
    const Move m48gi(Position(4,8),SILVER,WHITE);
    CategoryEnv env(&state, 1000, 0);
    CPPUNIT_ASSERT(! PrepareForkGold::isInCategory(env, m48gi));
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-HI *  *  * -OU * -KE-KY\n"
			   "P2 *  *  * -KI *  * -KI *  * \n"
			   "P3 *  * -KE *  * -GI * -FU-FU\n"
			   "P4-FU-FU-FU-FU *  *  *  *  * \n"
			   "P5 *  *  *  *  * -GI-FU *  * \n"
			   "P6+FU * +FU+FU+FU *  * +FU+FU\n"
			   "P7 * +FU+GI+KI * +FU *  *  * \n"
			   "P8 * +OU+KI *  *  *  * +HI * \n"
			   "P9+KY+KE *  *  *  *  *  * +KY\n"
			   "P+00KA00KA00FU\n"
			   "P-00GI00KE00FU00FU\n"
			   "+\n").getInitialState());
    const Move m63ka(Position(6,3),BISHOP,BLACK);
    CategoryEnv env(&state, 1000, 0);
    CPPUNIT_ASSERT(PrepareForkGold::isInCategory(env, m63ka));
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-HI *  *  * -OU * -KE-KY\n"
			   "P2 *  *  * -KI *  * -KI *  * \n"
			   "P3 *  * -KE *  * -GI * -FU-FU\n"
			   "P4-FU-FU-FU-FU *  *  *  *  * \n"
			   "P5 *  *  *  *  * -GI-FU *  * \n"
			   "P6+FU * +FU+FU+FU *  * +FU+FU\n"
			   "P7 * +FU+GI+KI * +FU *  *  * \n"
			   "P8 * +OU+KI *  *  *  * +HI * \n"
			   "P9+KY+KE *  *  *  *  *  * +KY\n"
			   "P+00KA00FU\n"
			   "P-00KA00GI00KE00FU00FU\n"
			   "+\n").getInitialState());
    const Move m63ka(Position(6,3),BISHOP,BLACK);
    CategoryEnv env(&state, 1000, 0);
    CPPUNIT_ASSERT(! PrepareForkGold::isInCategory(env, m63ka));
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
