#include "osl/category/knight.h"
#include "osl/category/classify.h"
#include "osl/category/classify.tcc"
#include "osl/category/categoryUtil.h"
#include "osl/record/csaString.h"
#include "osl/record/record.h"
#include "osl/record/csaRecord.h"
#include "osl/state/numEffectState.h"
#include "osl/container/moveLogProbVector.h"
#include "osl/apply_move/applyMove.h"
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <algorithm>
#include <fstream>
#include <iostream>

class KnightTest : public CppUnit::TestFixture
{
  CPPUNIT_TEST_SUITE(KnightTest);
  CPPUNIT_TEST(testKnightDoubleAttack);
  CPPUNIT_TEST(testKnightToAttack);
  CPPUNIT_TEST(testJoiningKnights);
  CPPUNIT_TEST_SUITE_END();
public:
  void testKnightDoubleAttack();
  void testKnightToAttack();
  void testJoiningKnights();
private:
  template<class Category>
  void testGenerateOne(const char *board, const char *move);
};

CPPUNIT_TEST_SUITE_REGISTRATION(KnightTest);

template<class Category>
void KnightTest::
testGenerateOne(const char *board, const char *move)
{
  osl::record::csa::CsaString csa(board);
  osl::state::SimpleState initialState = csa.getInitialState();
  osl::state::NumEffectState state(initialState);
  osl::container::MoveStack history;

  osl::container::MoveLogProbVector generatedMoves;
  osl::category::Classify<osl::category::CategoryList<Category, osl::category::NullCategory> > classify;
  osl::category::CategoryEnv env(&state, osl::category::CategoryTraits::maxProb(),
				 &history);
  classify.generate(env, generatedMoves);
  if (strcmp(move, "") == 0)
  {
    CPPUNIT_ASSERT_EQUAL(0, static_cast<int>(generatedMoves.size()));
  }
  else
  {
    if (1 != static_cast<int>(generatedMoves.size()))
    {
      std::cerr << state << std::endl;
      for (int i = 0; i < static_cast<int>(generatedMoves.size()); i++)
	std::cerr << generatedMoves[i] << std::endl;
    }
    CPPUNIT_ASSERT_EQUAL(1, static_cast<int>(generatedMoves.size()));
    CPPUNIT_ASSERT(osl::record::csa::strToMove(move, initialState) ==  generatedMoves[0].getMove());
  }
}

void KnightTest::testKnightDoubleAttack()
{
  testGenerateOne<osl::category::KnightDoubleAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI-FU-GI * -FU\n"
     "P5 * -FU *  *  *  *  * +FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE00FU\n"
     "+", "+0046KE");
  testGenerateOne<osl::category::KnightDoubleAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI * -GI * -FU\n"
     "P5 * -FU *  *  * -FU * +FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE00FU\n"
     "+", "");
  testGenerateOne<osl::category::KnightDoubleAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI-GI-FU * \n"
     "P4-FU * -FU-FU-GI-FU-FU * -FU\n"
     "P5 * -FU *  *  *  *  * +FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE\n"
     "+", "");
}

void KnightTest::testKnightToAttack()
{
  testGenerateOne<osl::category::KnightToAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI-FU-GI * -FU\n"
     "P5 * -FU *  *  *  *  *  *  * \n"
     "P6+FU * +FU+FU+FU *  * +FU+FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE00FU\n"
     "+", "+0046KE");
  testGenerateOne<osl::category::KnightToAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI * -GI * -FU\n"
     "P5 * -FU *  *  * -FU * +FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE00FU\n"
     "+", "+0026KE");
  testGenerateOne<osl::category::KnightToAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI-FU-FU * -FU\n"
     "P5 * -FU *  *  *  * -GI+FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE\n"
     "+", "");
  testGenerateOne<osl::category::KnightToAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI *  * \n"
     "P3 *  *  *  * -FU-KI-OU-FU * \n"
     "P4-FU * -FU-FU-GI-FU-FU * -FU\n"
     "P5 * -FU *  *  *  * -GI+FU * \n"
     "P6+FU * +FU+FU *  *  *  * +FU\n"
     "P7 * +FU+GI+KI+FU+FU+FU *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE\n"
     "P-00KE\n"
     "+", "+0017KE");
  testGenerateOne<osl::category::KnightToAttack>
    ("P1-KY-KE *  *  *  *  *  * -KY\n"
     "P2 * -HI *  *  * -KA-KI-OU * \n"
     "P3 *  *  *  * -FU-KI * -FU * \n"
     "P4-FU * -FU-FU-GI-FU-FU * -FU\n"
     "P5 * -FU *  *  *  * -GI+FU * \n"
     "P6+FU * +FU+FU+FU *  *  * +FU\n"
     "P7 * +FU+GI+KI * +FU *  *  * \n"
     "P8 * +OU+KI *  *  *  * +HI * \n"
     "P9+KY+KE+KA *  *  *  *  * +KY\n"
     "P+00GI00KE00FU\n"
     "P-00KE\n"
     "-", "-0084KE");
}

void KnightTest::testJoiningKnights()
{
  testGenerateOne<osl::category::JoiningKnights>
    ("P1-KY-KE * -KI *  *  * +RY * \n"
     "P2 * -OU-GI * -KI *  *  *  * \n"
     "P3 * -FU-FU-FU-FU *  *  * -FU\n"
     "P4-FU *  *  * -GI-FU *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU+KE+FU * +FU *  *  *  * \n"
     "P7 * +FU * +FU+GI+FU *  * +FU\n"
     "P8 * +KA+OU * +KI *  *  *  * \n"
     "P9+KY+KE+GI+KI *  * -RY *  * \n"
     "P+00KE00KY00FU00FU\n"
     "P-00KA00KY00FU00FU\n"
     "+", "+0074KE");

  testGenerateOne<osl::category::JoiningKnights>
    ("P1-KY-KE * -KI *  *  * +RY * \n"
     "P2 * -OU-GI * -KI *  *  *  * \n"
     "P3 * -FU-FU-FU-FU *  *  * -FU\n"
     "P4-FU *  *  * -GI-FU *  *  * \n"
     "P5 *  *  *  *  *  *  *  *  * \n"
     "P6+FU * +FU * +FU *  *  *  * \n"
     "P7 * +FU * +FU+GI+FU *  * +FU\n"
     "P8 * +KA+OU * +KI *  *  *  * \n"
     "P9+KY+KE+GI+KI *  * -RY *  * \n"
     "P+00KE00KY00FU00FU\n"
     "P-00KA00KY00KE00FU00FU\n"
     "+", "");
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
