/* proofPieces.t.cc
 */
#include "osl/checkmate/dualCheckmateSearcher.h"
#include "osl/checkmate/proofPieces.h"
#include "osl/state/hashEffectState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include <iostream>

class ProofPiecesTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(ProofPiecesTest);
  CPPUNIT_TEST(testLeaf);
  CPPUNIT_TEST(testAttack);
  CPPUNIT_TEST(testDefense);
  CPPUNIT_TEST_SUITE_END();
public:
  void testLeaf();
  void testAttack();
  void testDefense();
};
CPPUNIT_TEST_SUITE_REGISTRATION(ProofPiecesTest);

using namespace osl;
using namespace osl::checkmate;
extern bool isShortTest;
const size_t limit = 400000;

void ProofPiecesTest::testLeaf() 
{
  {
    HashEffectState state(CsaString(
			    "P1 *  *  *  *  *  *  * -OU * \n"
			    "P2 *  *  *  *  *  *  * +KI * \n"
			    "P3 *  *  *  *  *  *  * +FU * \n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  *  *  *  *  *  * \n"
			    "P7 *  *  *  *  *  *  *  *  * \n"
			    "P8 *  *  *  *  *  *  *  *  * \n"
			    "P9 *  * +OU *  *  *  *  *  * \n"
			    "P+00KI\n"
			    "P-00AL\n"
			    "-\n").getInitialState());
    const PieceStand proof_pieces
      = ProofPieces::leaf(state, BLACK, state.getHash().blackStand());
    CPPUNIT_ASSERT_EQUAL(PieceStand(), proof_pieces);
  }
  {
    HashEffectState state(CsaString(
			    "P1 *  *  *  *  *  *  * -OU * \n"
			    "P2 *  *  *  *  *  *  * +KI * \n"
			    "P3 *  *  *  *  *  *  * +FU * \n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  *  *  *  *  *  * \n"
			    "P7 *  *  *  *  *  *  *  *  * \n"
			    "P8 *  *  *  *  *  *  *  *  * \n"
			    "P9 *  * +OU *  *  *  *  *  * \n"
			    "P+00KI00KI00KI\n"
			    "P-00AL\n"
			    "-\n").getInitialState());
    const PieceStand proof_pieces 
      = ProofPieces::leaf(state, BLACK, state.getHash().blackStand());
    CPPUNIT_ASSERT_EQUAL(PieceStand(), proof_pieces);
  }
  {
    // ǽʾ
    HashEffectState state(CsaString(
			    "P1 *  *  *  *  * +HI * -OU * \n"
			    "P2 *  *  *  *  *  *  *  *  * \n"
			    "P3 *  *  *  *  *  *  * +KI * \n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  *  *  *  *  *  * \n"
			    "P7 *  *  *  *  *  *  *  *  * \n"
			    "P8 *  *  *  *  *  *  *  *  * \n"
			    "P9 *  * +OU *  *  *  *  *  * \n"
			    "P+00KI00KI00KI\n"
			    "P-00AL\n"
			    "-\n").getInitialState());
    const PieceStand proof_pieces 
      = ProofPieces::leaf(state, BLACK, state.getHash().blackStand());
    PieceStand gold;
    gold.add(GOLD, 3);
    CPPUNIT_ASSERT_EQUAL(gold, proof_pieces);
  }
}

void ProofPiecesTest::testAttack()
{
  HashEffectState state(CsaString(
			  "P1 *  *  *  *  *  * -KY * -KY\n"
			  "P2 *  *  *  *  *  *  * -OU * \n"
			  "P3 *  *  *  *  *  *  *  *  * \n"
			  "P4 *  *  *  *  *  *  * +FU * \n"
			  "P5 *  *  *  *  *  * +KE *  * \n"
			  "P6 *  *  *  *  *  *  *  *  * \n"
			  "P7 *  *  *  *  *  *  *  *  * \n"
			  "P8 *  *  *  *  *  *  *  *  * \n"
			  "P9 *  * +OU *  *  *  *  *  * \n"
			  "P+00KI\n"
			  "P-00AL\n"
			  "+\n").getInitialState());
  DualCheckmateSearcher<> checker(limit);
  checker.setVerbose(!isShortTest);
  Move check_move = Move::INVALID();
  const PathEncoding path(BLACK);
  AttackOracleAges oracle_age;
  const bool win = checker.isWinningState<BLACK>
    (10, state, state.getHash(), path, check_move, oracle_age);
  CPPUNIT_ASSERT_EQUAL(true, win);
  // +23{TO,NK} -21OU +22KI
  const CheckHashTable& table = checker.getTable(BLACK);
  const CheckHashRecord *record = table.find(state.getHash());
  PieceStand gold;
  gold.add(GOLD);
  CPPUNIT_ASSERT_EQUAL(gold, record->proofPieces());
}

void ProofPiecesTest::testDefense()
{
  AttackOracleAges oracle_age;
  {
    HashEffectState state(CsaString(
			    "P1 *  *  *  *  *  *  *  *  * \n"
			    "P2 *  *  *  *  *  *  * -OU * \n"
			    "P3 *  *  *  *  *  *  * +KI * \n"
			    "P4 *  *  *  *  *  *  * +FU * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  *  *  *  *  *  * \n"
			    "P7 *  *  *  *  *  *  *  *  * \n"
			    "P8 *  *  *  *  *  *  *  *  * \n"
			    "P9 *  * +OU *  *  *  *  *  * \n"
			    "P+00KI\n"
			    "P-00AL\n"
			    "-\n").getInitialState());
    DualCheckmateSearcher<> checker(limit);
    checker.setVerbose(!isShortTest);
    const PathEncoding path(WHITE);
    const bool checkmate
      = checker.isLosingState<WHITE>(100, state, state.getHash(), path);
    CPPUNIT_ASSERT_EQUAL(true, checkmate);

    const CheckHashTable& table = checker.getTable(BLACK);
    const CheckHashRecord *record = table.find(state.getHash());
    CPPUNIT_ASSERT(record);
    PieceStand gold;
    gold.add(GOLD);
    CPPUNIT_ASSERT_EQUAL(gold, record->proofPieces());
  }
  {
    HashEffectState state(CsaString(
			    "P1 *  *  *  * -KY-OU-KI *  * \n"
			    "P2 *  *  *  *  *  * +GI *  * \n"
			    "P3 *  *  *  *  * +KI *  *  * \n"
			    "P4 *  *  *  *  *  *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  *  *  *  *  *  * \n"
			    "P7 *  *  *  *  *  *  *  *  * \n"
			    "P8 *  *  *  *  *  *  *  *  * \n"
			    "P9 *  * +OU *  *  *  *  *  * \n"
			    "P+00KI00KE\n"
			    "P-00AL\n"
			    "-\n").getInitialState());
    DualCheckmateSearcher<> checker(limit);
    checker.setVerbose(!isShortTest);
    const PathEncoding path(WHITE);
    const bool checkmate
      = checker.isLosingState<WHITE>(100, state, state.getHash(), path);
    CPPUNIT_ASSERT_EQUAL(true, checkmate);

    const CheckHashTable& table = checker.getTable(BLACK);
    const CheckHashRecord *record = table.find(state.getHash());
    CPPUNIT_ASSERT(record);
    PieceStand gold_knight;
    gold_knight.add(GOLD);
    gold_knight.add(KNIGHT);
    CPPUNIT_ASSERT_EQUAL(gold_knight, record->proofPieces());
  }
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
