#include "osl/eval/endgame/attackKing.h"
#include "osl/eval/evalTraits.h"
#include "osl/eval/pieceEval.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <iostream>
#include <iomanip>
using namespace osl;
using namespace osl::eval;
using namespace osl::eval::endgame;

class EndgameAttackKingTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(EndgameAttackKingTest);
  // CPPUNIT_TEST(testShow);
  CPPUNIT_TEST(testStand);
  CPPUNIT_TEST(testPawnLance);
  CPPUNIT_TEST(testGold);
  CPPUNIT_TEST(testDirection);
  CPPUNIT_TEST_SUITE_END();
public:
  void testShow();
  void testStand();
  void testPawnLance();
  void testGold();
  void testDirection(Ptype ptype);
  void testDirection() {
    testDirection(GOLD);
    testDirection(LANCE);
  }
};

CPPUNIT_TEST_SUITE_REGISTRATION(EndgameAttackKingTest);

void EndgameAttackKingTest::testPawnLance()
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const Position p58(5,8);
  const Position p52(5,2);
  // 1ĤưʤƱ
  CPPUNIT_ASSERT_EQUAL(AttackKing::valueOf(black_king, newPtypeO(WHITE,PAWN), 
					   p58),
		       AttackKing::valueOf(black_king, newPtypeO(WHITE,LANCE), 
					   p58));
  CPPUNIT_ASSERT(AttackKing::valueOf(black_king, newPtypeO(WHITE,PAWN), p52)
		 > AttackKing::valueOf(black_king, newPtypeO(WHITE,LANCE), p52));
  CPPUNIT_ASSERT_EQUAL(AttackKing::valueOf(black_king, newPtypeO(BLACK,PAWN), 
					   p52),
		       AttackKing::valueOf(black_king, newPtypeO(BLACK,LANCE), 
					   p52));
  CPPUNIT_ASSERT(AttackKing::valueOf(black_king, newPtypeO(BLACK,PAWN), p58)
		<= AttackKing::valueOf(black_king, newPtypeO(BLACK,LANCE), p58));

  // ̤ǤƱ
  CPPUNIT_ASSERT_EQUAL(AttackKing::valueOf(white_king, newPtypeO(WHITE,PAWN), 
					   p58),
		       AttackKing::valueOf(white_king, newPtypeO(WHITE,LANCE), 
					   p58));
  CPPUNIT_ASSERT(AttackKing::valueOf(white_king, newPtypeO(WHITE,PAWN), p52)
		 >= AttackKing::valueOf(white_king, newPtypeO(WHITE,LANCE), p52));
  CPPUNIT_ASSERT_EQUAL(AttackKing::valueOf(white_king, newPtypeO(BLACK,PAWN), 
					   p52),
		       AttackKing::valueOf(white_king, newPtypeO(BLACK,LANCE), 
					   p52));
  CPPUNIT_ASSERT(AttackKing::valueOf(white_king, newPtypeO(BLACK,PAWN), p58)
		 < AttackKing::valueOf(white_king, newPtypeO(BLACK,LANCE), p58));
}

void EndgameAttackKingTest::testGold()
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const Position p51(5,1);
  const Position p52(5,2);
  const Position p58(5,8);
  const Position p59(5,9);
  // 1ܸ
  CPPUNIT_ASSERT(AttackKing::valueOf(black_king, newPtypeO(WHITE,GOLD), p58)
		 < AttackKing::valueOf(black_king, newPtypeO(WHITE,GOLD), p59));
  CPPUNIT_ASSERT(AttackKing::valueOf(white_king, newPtypeO(BLACK,GOLD), p52)
		 > AttackKing::valueOf(white_king, newPtypeO(BLACK,GOLD), p51));
}

void EndgameAttackKingTest::testDirection(Ptype ptype)
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const Position up(4,4);
  const Position down(4,6);

  // 夫ޤ⤤
  const int white_attack_front
    = AttackKing::valueOf(black_king, newPtypeO(WHITE,ptype), up);
  const int white_attack_back
    = AttackKing::valueOf(black_king, newPtypeO(WHITE,ptype), down);

  const int black_attack_front 
    = AttackKing::valueOf(white_king, newPtypeO(BLACK,ptype), down);
  const int black_attack_back
    = AttackKing::valueOf(white_king, newPtypeO(BLACK,ptype), up);

  CPPUNIT_ASSERT(betterThan(WHITE, white_attack_front, white_attack_back));
  CPPUNIT_ASSERT(betterThan(BLACK, black_attack_front, black_attack_back));

  // о
  CPPUNIT_ASSERT_EQUAL(black_attack_front, -white_attack_front);
  CPPUNIT_ASSERT_EQUAL(black_attack_back,  -white_attack_back);
}

void EndgameAttackKingTest::testStand() {
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const int black_attack
    = AttackKing::valueOf(white_king, newPtypeO(BLACK,GOLD), 
			  Position::STAND());
  CPPUNIT_ASSERT_EQUAL(Ptype_Eval_Table.value(GOLD), black_attack);

  const int white_attack
    = AttackKing::valueOf(black_king, newPtypeO(WHITE,GOLD), 
			  Position::STAND());
  CPPUNIT_ASSERT_EQUAL(-Ptype_Eval_Table.value(GOLD), white_attack);
}


void EndgameAttackKingTest::testShow()
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);
  // debug 
  for (int y=1; y<=9; ++y) 
  {
    for (int x=1; x<=9; ++x) 
    {
      std::cerr << std::setw(4) 
		<< AttackKing::valueOf(black_king, newPtypeO(WHITE,PAWN), 
				       Position(x,y));
    }
    std::cerr << "\n";
  }    
  std::cerr << "\n";
    
  for (int y=1; y<=9; ++y) 
  {
    for (int x=1; x<=9; ++x) 
    {
      std::cerr << std::setw(4) 
		<< AttackKing::valueOf(white_king, newPtypeO(BLACK,PAWN), 
				       Position(x,y));
    }
    std::cerr << "\n";
  }    
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
