#include "osl/eval/endgame/defenseKing.h"
#include "osl/eval/evalTraits.h"
#include "osl/eval/pieceEval.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;
using namespace osl::eval;
using namespace osl::eval::endgame;

class EndgameDefenseKingTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(EndgameDefenseKingTest);
  CPPUNIT_TEST(testDirection);
  CPPUNIT_TEST(testStand);
  CPPUNIT_TEST(testKing);
  CPPUNIT_TEST(testPawnLance);
  CPPUNIT_TEST_SUITE_END();
public:
  void testStand();
  void directionTest(Ptype ptype, bool up_is_better);
  void testDirection() {
    // Ͼ⤤
    directionTest(GOLD, true);
    // ˹ϲ
    directionTest(KNIGHT, false);
  }
  void testKing();
  void testPawnLance();
};

CPPUNIT_TEST_SUITE_REGISTRATION(EndgameDefenseKingTest);

void EndgameDefenseKingTest::testPawnLance()
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const Position p58(5,8);
  const Position p52(5,2);
  // 1ĤưʤƱ
  CPPUNIT_ASSERT_EQUAL(DefenseKing::valueOf(black_king, newPtypeO(WHITE,PAWN), 
					   p58),
		       DefenseKing::valueOf(black_king, newPtypeO(WHITE,LANCE), 
					   p58));
  CPPUNIT_ASSERT(DefenseKing::valueOf(black_king, newPtypeO(WHITE,PAWN), p52)
		 >= DefenseKing::valueOf(black_king, newPtypeO(WHITE,LANCE), p52));
  CPPUNIT_ASSERT_EQUAL(DefenseKing::valueOf(black_king, newPtypeO(BLACK,PAWN), 
					   p52),
		       DefenseKing::valueOf(black_king, newPtypeO(BLACK,LANCE), 
					   p52));
  CPPUNIT_ASSERT(DefenseKing::valueOf(black_king, newPtypeO(BLACK,PAWN), p58)
		< DefenseKing::valueOf(black_king, newPtypeO(BLACK,LANCE), p58));

  // ̤ǤƱ
  CPPUNIT_ASSERT_EQUAL(DefenseKing::valueOf(white_king, newPtypeO(WHITE,PAWN), 
					   p58),
		       DefenseKing::valueOf(white_king, newPtypeO(WHITE,LANCE), 
					   p58));
  CPPUNIT_ASSERT(DefenseKing::valueOf(white_king, newPtypeO(WHITE,PAWN), p52)
		 > DefenseKing::valueOf(white_king, newPtypeO(WHITE,LANCE), p52));
  CPPUNIT_ASSERT_EQUAL(DefenseKing::valueOf(white_king, newPtypeO(BLACK,PAWN), 
					   p52),
		       DefenseKing::valueOf(white_king, newPtypeO(BLACK,LANCE), 
					   p52));
  CPPUNIT_ASSERT(DefenseKing::valueOf(white_king, newPtypeO(BLACK,PAWN), p58)
		 <= DefenseKing::valueOf(white_king, newPtypeO(BLACK,LANCE), p58));
}

void EndgameDefenseKingTest::testKing()
{
  const Position p11(1,1);
  const Position p19(1,9);
  const Piece black_king11 = Piece::makeKing(BLACK, p11);
  const Piece black_king19 = Piece::makeKing(BLACK, p19);
  CPPUNIT_ASSERT(EvalTraits<BLACK>::betterThan
		 (DefenseKing::valueOf(black_king11, black_king11),
		  DefenseKing::valueOf(black_king19, black_king19)));

  const Piece white_king11 = Piece::makeKing(WHITE, p11);
  const Piece white_king19 = Piece::makeKing(WHITE, p19);
  CPPUNIT_ASSERT(EvalTraits<WHITE>::betterThan
		 (DefenseKing::valueOf(white_king19, white_king19),
		  DefenseKing::valueOf(white_king11, white_king11)));
}

void EndgameDefenseKingTest::directionTest(Ptype ptype, bool up_is_better)
{
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const Position up(4,4);
  const Position down(4,6);

  const int black_defense_front 
    = DefenseKing::valueOf(black_king, newPtypeO(BLACK,ptype), up);
  const int black_defense_back
    = DefenseKing::valueOf(black_king, newPtypeO(BLACK,ptype), down);

  const int white_defense_front
    = DefenseKing::valueOf(white_king, newPtypeO(WHITE,ptype), down);
  const int white_defense_back
    = DefenseKing::valueOf(white_king, newPtypeO(WHITE,ptype), up);

  if (up_is_better)
  {
    CPPUNIT_ASSERT(betterThan(WHITE, white_defense_front, white_defense_back));
    CPPUNIT_ASSERT(betterThan(BLACK, black_defense_front, black_defense_back));
  }
  else
  {
    CPPUNIT_ASSERT(! betterThan(WHITE, white_defense_front, white_defense_back));
    CPPUNIT_ASSERT(! betterThan(BLACK, black_defense_front, black_defense_back));
  }

  // о
  CPPUNIT_ASSERT_EQUAL(black_defense_front, -white_defense_front);
  CPPUNIT_ASSERT_EQUAL(black_defense_back,  -white_defense_back);
}
  

void EndgameDefenseKingTest::testStand() {
  const Position king_position(5,5);
  const Piece black_king = Piece::makeKing(BLACK, king_position);
  const Piece white_king = Piece::makeKing(WHITE, king_position);

  const int black_defense
    = DefenseKing::valueOf(black_king, newPtypeO(BLACK,GOLD), 
			   Position::STAND());
  CPPUNIT_ASSERT_EQUAL(Ptype_Eval_Table.value(GOLD), black_defense);

  const int white_defense
    = DefenseKing::valueOf(white_king, newPtypeO(WHITE,GOLD), 
			   Position::STAND());
  CPPUNIT_ASSERT_EQUAL(-Ptype_Eval_Table.value(GOLD), white_defense);
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
