#include "../recursiveDoUndoMove.h"
#include "../doUndoMoveLockTest.h"
#include "osl/state/simpleState.h"
#include "osl/state/evalHashEffectState.h"
#include "osl/record/record.h"
#include "osl/record/csaString.h"
#include "osl/record/csaRecord.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>
#include <fstream>

class EvalTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(EvalTest);
  CPPUNIT_TEST(testShow);
  CPPUNIT_TEST(testConsistent);
  CPPUNIT_TEST(testSimpleEval);
  CPPUNIT_TEST(testDoUndoMove);
  CPPUNIT_TEST_SUITE_END();
public:
  void testShow();
  void testConsistent();
  void testSimpleEval();
  void testDoUndoMove();
};

CPPUNIT_TEST_SUITE_REGISTRATION(EvalTest);

using namespace osl;
using namespace osl::eval;
extern bool isShortTest;

void EvalTest::testShow()
{
  for(int i=PTYPEO_MIN;i<=PTYPEO_MAX;i++){
    PtypeO ptypeO=static_cast<PtypeO>(i);
    if (! isShortTest)
      std::cout << ptypeO << ",val=" << Ptype_Eval_Table.value(ptypeO);
    if(isPiece(ptypeO))
    {
      CPPUNIT_ASSERT(Ptype_Eval_Table.value(ptypeO)
		     *playerToMul(getOwner(ptypeO))>0);
      if (! isShortTest)
	std::cout << ", captureVal=" <<  Ptype_Eval_Table.captureValue(ptypeO) << std::endl;
      CPPUNIT_ASSERT(Ptype_Eval_Table.captureValue(ptypeO)
		     *playerToMul(getOwner(ptypeO))<0);
      if(isPromoted(ptypeO))
      {
      
	if (! isShortTest)
	  std::cout << ", promoteVal=" <<  Ptype_Eval_Table.promoteValue(ptypeO);
	CPPUNIT_ASSERT(Ptype_Eval_Table.promoteValue(ptypeO)
		       *playerToMul(getOwner(ptypeO))>0);
      }
    }
    if (! isShortTest)
      std::cout << std::endl;
  }
}

static void testFileConsistent(std::string const& fileName){
  //  std::cout << "testFileConsistent(" << fileName << ")" << std::endl;
  Record rec=CsaFile(fileName).getRecord();
  EvalHashEffectState eState(rec.getInitialState());
  vector<osl::Move> moves=rec.getMoves();

  DoUndoMoveLockTest<EvalHashEffectState> lockTest(eState, moves);
  lockTest.run();
}

void EvalTest::testConsistent(){
  std::ifstream ifs(CSAFILES "/FILES");
  CPPUNIT_ASSERT(ifs);
  std::string fileName;
  for(int i=0;i<100 && (ifs >> fileName) ; i++){
    if(fileName == "") 
      break;
    testFileConsistent(CSAFILES "/" + fileName);
  }
}

const int Gold = PtypeEvalTraits<GOLD>::val;
const int Rook = PtypeEvalTraits<ROOK>::val;
const int Pawn = PtypeEvalTraits<PAWN>::val;

void EvalTest::testSimpleEval()
{
  EvalHashEffectState state(CsaString(
			       "P1-KY-KE-GI-KI *  * -HI-KE * \n"
			       "P2 * -HI *  * -KA-OU *  *  * \n"
			       "P3-FU * -FU-FU-FU * -GI-FU-KY\n"
			       "P4 * +KE *  *  * -FU *  * -FU\n"
			       "P5 * -FU+KE+KA *  *  *  * +FU\n"
			       "P6 *  * +FU *  *  * +FU *  * \n"
			       "P7+FU+FU * +FU+FU+FU *  *  * \n"
			       "P8 * +GI+KI *  *  * +GI *  * \n"
			       "P9+KY *  *  * +OU+KI *  * +KY\n"
			       "P+00FU\n"
			       "P+00FU\n"
			       "P+00KI\n"
			       "+\n").getInitialState());
  const int ret=state.getVal();
  CPPUNIT_ASSERT_EQUAL((Gold+Pawn-Rook)*2, ret);
}


void EvalTest::testDoUndoMove()
{
  extern bool isShortTest;
  std::ifstream ifs(CSAFILES "/FILES");
  CPPUNIT_ASSERT(ifs);
  int i=0;
  int count=100;
  if (isShortTest)
    count=10;
  std::string fileName;
  while((ifs >> fileName) && (++i<count)) {
    if(fileName == "") 
      break;
    if (! isShortTest)
      std::cerr << fileName << " ";

    CsaFile file(CSAFILES "/"+fileName);
    Record rec=file.getRecord();
    EvalHashEffectState eState(rec.getInitialState());

    vector<osl::Move> moves=rec.getMoves();
  
    size_t depth=0;
    typedef EvalHashEffectState state_t;
    if (moves.begin()->player()==BLACK){
      RecursiveDoUndoMove<BLACK,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<BLACK>::doUndoMove(eState, *moves.begin(), func);
    }
    else{
      RecursiveDoUndoMove<WHITE,state_t>
	func(eState, &*moves.begin(), &*moves.end(), &depth);
      ApplyMove<WHITE>::doUndoMove(eState, *moves.begin(), func);
    }
    CPPUNIT_ASSERT_EQUAL(moves.size(), depth);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
