#include "osl/game_playing/mtdfPlayer.h"
#include "osl/game_playing/gameState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;
using namespace osl::game_playing;

class MtdfPlayerTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(MtdfPlayerTest);
  CPPUNIT_TEST(testResign);
  CPPUNIT_TEST(testWin);
  CPPUNIT_TEST(testIsReasonableMove);
  CPPUNIT_TEST(testClone);
  CPPUNIT_TEST(testDeclare);
  CPPUNIT_TEST_SUITE_END();
public:
  void testResign();
  void testWin();
  void testIsReasonableMove();
  void testClone();
  void testDeclare();
};

CPPUNIT_TEST_SUITE_REGISTRATION(MtdfPlayerTest);

const int seconds = 30;
void MtdfPlayerTest::testResign()
{
  const SimpleState state(CsaString(
			 "P1 *  *  *  * -OU *  *  *  * \n"
			 "P2 *  *  *  * +KI *  *  *  * \n"
			 "P3 *  *  *  * +FU *  *  *  * \n"
			 "P4 *  *  *  *  *  *  *  *  * \n"
			 "P5 *  *  *  *  *  *  *  *  * \n"
			 "P6 *  *  *  *  *  *  *  *  * \n"
			 "P7 *  *  *  *  *  *  *  *  * \n"
			 "P8 *  *  *  *  *  *  *  *  * \n"
			 "P9 *  *  *  * +OU *  *  *  * \n"
			 "P-00AL\n"
			 "-\n").getInitialState());
  const GameState gs(state);
  PlainMtdfPlayer player;
  const Move best_move = player.selectBestMove(gs, seconds, 0);
  CPPUNIT_ASSERT_EQUAL(Move::INVALID(), best_move);
}

void MtdfPlayerTest::testWin()
{
  const SimpleState state(CsaString(
			 "P1 *  *  *  * -OU *  *  *  * \n"
			 "P2 *  *  *  *  *  *  *  *  * \n"
			 "P3 *  *  *  * +FU *  *  *  * \n"
			 "P4 *  *  *  *  *  *  *  *  * \n"
			 "P5 *  *  *  *  *  *  *  *  * \n"
			 "P6 *  *  *  *  *  *  *  *  * \n"
			 "P7 *  *  *  *  *  *  *  *  * \n"
			 "P8 *  *  *  *  *  *  *  *  * \n"
			 "P9 *  *  *  * +OU *  *  *  * \n"
			 "P+00KI\n"
			 "P-00AL\n"
			 "+\n").getInitialState());
  const GameState gs(state);
  PlainMtdfPlayer player;
  const Move best_move = player.selectBestMove(gs, seconds, 0);
  const Move expected(Position(5,2), GOLD, BLACK);
  CPPUNIT_ASSERT_EQUAL(expected, best_move);
}

void MtdfPlayerTest::testIsReasonableMove()
{
  const SimpleState state(CsaString(
			    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
			    "P2 * -HI *  *  *  *  * -KA * \n"
			    "P3-FU-FU-FU-FU * -FU-FU-FU-FU\n"
			    "P4 *  *  *  * -FU *  *  *  * \n"
			    "P5 *  *  *  *  *  *  *  *  * \n"
			    "P6 *  *  *  * +FU *  *  *  * \n"
			    "P7+FU+FU+FU+FU * +FU+FU+FU+FU\n"
			    "P8 * +KA *  *  *  *  * +HI * \n"
			    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
			    "+\n").getInitialState());
  const GameState gs(state);
  PlainMtdfPlayer player;
  const Move m52to(Position(5,6),Position(5,5),PAWN,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT(! player.isReasonableMove(gs, m52to,1));
  const Move m48gi(Position(3,9),Position(4,8),SILVER,PTYPE_EMPTY,false,BLACK);
  CPPUNIT_ASSERT(player.isReasonableMove(gs, m48gi,1));
}

void MtdfPlayerTest::testClone()
{
  PlainMtdfPlayer player;
  ComputerPlayer *clone = player.clone();
  PlainMtdfPlayer *pclone = dynamic_cast<PlainMtdfPlayer*>(clone);
  CPPUNIT_ASSERT(pclone);
  CPPUNIT_ASSERT(player.getConfig() == pclone->getConfig());
}

void MtdfPlayerTest::testDeclare()
{
  // ƥȶ̤ enter_king/enterKing.t.cc 
  {
    const NumEffectState state(CsaString(
				 "P1 *  * +RY *  * +KI * -KE * \n"
				 "P2+OU+TO+UM+UM+TO * -KI-OU * \n"
				 "P3-FU+FU+FU+FU+FU-GI-GI *  * \n"
				 "P4 *  *  *  * -FU-FU-FU-FU * \n"
				 "P5 *  * -RY-NY-GI *  *  *  * \n"
				 "P6 *  *  *  *  * +FU+FU+FU * \n"
				 "P7 *  *  *  *  * +KI+KE * -NY\n"
				 "P8 *  *  *  *  * +KI+GI *  * \n"
				 "P9 *  *  *  * -NY *  *  *  * \n"
				 "P+00FU00FU00FU00FU00KY00KE00KE\n"
				 "+\n").getInitialState());
    const GameState gs(state);
    PlainMtdfPlayer player;
    const Move result = player.selectBestMove(gs, 0, 5);
    CPPUNIT_ASSERT_EQUAL(Move::DeclareWin(), result);
  }
  {
    const NumEffectState state(CsaString(
				 "P1 *  *  *  * +NY *  *  *  * \n"
				 "P2 *  *  *  *  * -KI-GI *  * \n"
				 "P3 *  *  *  *  * -KI-KE * +NY\n"
				 "P4 *  *  *  *  * -FU-FU-FU * \n"
				 "P5 *  * +RY+NY+GI *  *  *  * \n"
				 "P6 *  *  *  * +FU+FU+FU+FU * \n"
				 "P7+FU-FU-FU-FU-FU+GI+GI *  * \n"
				 "P8-OU-TO-UM-UM-TO * +KI+OU * \n"
				 "P9 *  * -RY *  * -KI * +KE * \n"
				 "P-00FU00FU00FU00FU00KY00KE00KE\n"
				 "-\n").getInitialState());
    const GameState gs(state);
    PlainMtdfPlayer player;
    const Move result = player.selectBestMove(gs, 0, 5);
    CPPUNIT_ASSERT_EQUAL(Move::DeclareWin(), result);
  }
  // ä­ʤ꡼
  {
    const NumEffectState state(CsaString(
				 "P1 *  * +RY *  * +KI * -KE * \n"
				 "P2+OU+TO+UM+UM+TO * -KI-OU * \n"
				 "P3-FU+FU+FU+FU * -GI-GI *  * \n"
				 "P4 *  *  *  * -FU-FU-FU-FU * \n"
				 "P5 *  * -RY-NY-GI *  *  *  * \n"
				 "P6 *  *  *  *  * +FU+FU+FU * \n"
				 "P7 *  *  *  *  * +KI+KE * -NY\n"
				 "P8 *  *  *  *  * +KI+GI *  * \n"
				 "P9 *  *  *  * -NY *  *  *  * \n"
				 "P+00FU00FU00FU00FU00KY00KE00KE\n"
				 "P-00FU\n"
				 "+\n").getInitialState());
    const GameState gs(state);
    PlainMtdfPlayer player;
    player.setTableLimit(0, 2000);
    const Move result = player.selectBestMove(gs, 0, 5);
    CPPUNIT_ASSERT(Move::DeclareWin() != result);
  }
  {
    const NumEffectState state(CsaString(
				 "P1 *  *  *  * +NY *  *  *  * \n"
				 "P2 *  *  *  *  * -KI-GI *  * \n"
				 "P3 *  *  *  *  * -KI-KE * +NY\n"
				 "P4 *  *  *  *  * -FU-FU-FU * \n"
				 "P5 *  * +RY+NY+GI *  *  *  * \n"
				 "P6 *  *  *  * +FU+FU+FU+FU * \n"
				 "P7+FU-FU-FU-FU * +GI+GI *  * \n"
				 "P8-OU-TO-UM-UM-TO * +KI+OU * \n"
				 "P9 *  * -RY *  * -KI * +KE * \n"
				 "P+00FU\n"
				 "P-00FU00FU00FU00FU00KY00KE00KE\n"
				 "-\n").getInitialState());
    const GameState gs(state);
    PlainMtdfPlayer player;
    player.setTableLimit(0, 2000);
    const Move result = player.selectBestMove(gs, 0, 5);
    CPPUNIT_ASSERT(Move::DeclareWin() != result);
  }
}

// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
