#include "osl/record/csaRecord.h"
#include "osl/record/csa.h"
#include "osl/stl/vector.h"
#include "osl/record/csaString.h"
#include "osl/state/simpleState.h"
#include "osl/apply_move/doUndoMoveLock.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

#include <fstream>

class CsaTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE( CsaTest );
  CPPUNIT_TEST( testShow );
  CPPUNIT_TEST( testSpeed );
  CPPUNIT_TEST( testSpeedGame );
  CPPUNIT_TEST_SUITE_END();
 private:
  osl::Record rec;
  osl::SimpleState state;
  osl::vector<osl::Move> moves;

 public:
  void setUp();
  void testShow();
  void testLoad();
  void testSpeed();
  void testSpeedGame();
};

CPPUNIT_TEST_SUITE_REGISTRATION(CsaTest);

using namespace osl;
using namespace osl::record;

void CsaTest::setUp(){
  rec=CsaFile(OSL_DATADIR "../../osl/test/kcc2.csa").getRecord();
  state=rec.getInitialState();
  moves=rec.getMoves();
}

void CsaTest::testShow(){
}

void CsaTest::testLoad(){
  //  std::cout << rec << std::endl;
  // st/d::cout << irs.getState() << std::endl;
}

struct posToInt {
  int operator()(){
    return 1;
  }
};


inline void myStateTest(SimpleState& state){
  // std::cout << state << std::endl;
  {
    const Position pos0=csa::strToPos("55"),pos1=csa::strToPos("54");
    const Move move=Move(pos0,pos1,
			    PAWN,PTYPE_EMPTY,false,BLACK);
    const SimpleState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
  {
    const Position pos1=csa::strToPos("99");
    const Move move=Move(pos1,PAWN,BLACK);
    const SimpleState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
  {
    const Position pos0=csa::strToPos("53"),pos1=csa::strToPos("42");
    const Move move=Move(pos0,pos1,
			    PBISHOP,SILVER,false,BLACK);
    const SimpleState stateCopy(state);
    {
      DoUndoMoveLock lock(state, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state, move);
    }
  }
}

void CsaTest::testSpeed(){
  Record rec=CsaString(
"P1+NY+TO *  *  *  * -OU-KE-KY\n"
"P2 *  *  *  *  * -GI-KI *  *\n"
"P3 * +RY *  * +UM * -KI-FU-FU\n"
"P4 *  * +FU-FU *  *  *  *  *\n"
"P5 *  * -KE * +FU *  * +FU *\n"
"P6-KE *  * +FU+GI-FU *  * +FU\n"
"P7 *  * -UM *  *  *  *  *  *\n"
"P8 *  *  *  *  *  *  *  *  * \n"
"P9 * +OU * -GI *  *  *  * -NG\n"
"P+00HI00KI00KE00KY00FU00FU00FU00FU00FU00FU\n"
"P-00KI00KY00FU00FU\n"
"P-00AL\n"
"+\n").getRecord();
  SimpleState state0=rec.getInitialState();
  //  std::cout << state0;
  // std::cout << "BitXmask" << std::endl;
  {
    const Position pos0=csa::strToPos("55"),pos1=csa::strToPos("54");
    const Move move=Move(pos0,pos1,
			    PAWN,PTYPE_EMPTY,false,BLACK);
    const SimpleState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
  {
    const Position pos1=csa::strToPos("99");
    const Move move=Move(pos1,PAWN,BLACK);
    const SimpleState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
  {
    const Position pos0=csa::strToPos("53"),pos1=csa::strToPos("42");
    const Move move=Move(pos0,pos1,
			     PBISHOP,SILVER,false,BLACK);
    const SimpleState stateCopy(state0);
    {
      DoUndoMoveLock lock(state0, move);
    }
    CPPUNIT_ASSERT_EQUAL(stateCopy, state0);

    for (int i=0;i<10000;i++)
    {
      DoUndoMoveLock lock(state0, move);
    }
  }
}

void CsaTest::testSpeedGame(){
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
