#include "mtdfdebug.h"
#include "problems.h"
#include "osl/search/searchRecorder.h"
#include "osl/record/csaString.h"
#include "osl/category/fWMoveGen.h"
#include "osl/state/hashEffectState.h"
#include "osl/moveLogProb.h"
#include "osl/state/simpleState.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;

class MtdfTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(MtdfTest);
  CPPUNIT_TEST(testWinValue);
  CPPUNIT_TEST(testConstruct);
  CPPUNIT_TEST(testMoveLogProb);
  CPPUNIT_TEST(testProblems1);
  CPPUNIT_TEST(testProblems2);
  CPPUNIT_TEST(testProblems3);
  CPPUNIT_TEST_SUITE_END();

  void testProblems(SearchRecorder& recorder,
		    const Problem *problems, int numProblems, int depth);
public:
  MtdfTest();
  void testConstruct();
  void testMoveLogProb();
  void testWinValue();
  void testProblems1();
  void testProblems2();
  void testProblems3();
  bool verbose() const
  {
    extern bool isShortTest;
    return ! isShortTest;
  }
};

CPPUNIT_TEST_SUITE_REGISTRATION(MtdfTest);

const Move move61rook = Move(Position::STAND(),Position(6,1),ROOK,PTYPE_EMPTY,false,BLACK);
const Move move5364uma = Move(Position(5,3),Position(6,4),PBISHOP,PAWN,false,BLACK);

MtdfTest::MtdfTest() 
{
}

void MtdfTest::testConstruct()
{
  osl::SearchRecorder recorder("MtdfTest.testShow");
  HashEffectState eState((SimpleState(HIRATE)));

  SimpleHashTable nullTable(0,0,verbose());
  Mtdf::checkmate_t checkmate;
  Mtdf mtdf_searcher(eState,checkmate,nullTable,recorder);
}

void MtdfTest::testMoveLogProb()
{
  const Move m8273ryu
    = Move(Position(8,2),Position(7,3),PROOK,PTYPE_EMPTY,false,BLACK);
  const Move m82fu = Move(Position(8,2),PAWN,BLACK);
  MoveLogProb moveLogProb1(m8273ryu,45);
  MoveLogProb moveLogProb2(m82fu,60);
  CPPUNIT_ASSERT_EQUAL(m8273ryu, moveLogProb1.getMove());
  CPPUNIT_ASSERT_EQUAL(45, moveLogProb1.getLogProb());
  CPPUNIT_ASSERT_EQUAL(m82fu, moveLogProb2.getMove());
  CPPUNIT_ASSERT_EQUAL(60, moveLogProb2.getLogProb());
}

void MtdfTest::testProblems(SearchRecorder& recorder,
			    const Problem *problems, int numProblems,
			    int depth)
{
  for (int i=0; i<numProblems; ++i)
  {
    HashEffectState state((CsaString(problems[i].state).getInitialState()));
    SimpleHashTable nullTable(0,0,verbose());
    Mtdf::checkmate_t checkmate;
    Mtdf mtdf_searcher(state, checkmate, nullTable, recorder);

    recorder.stream() << "\nfile " << i << " limit " << depth << "\n";
    recorder.stream() << state;
    mtdf_searcher.setInitialGuess();

    const Move best_move = mtdf_searcher.computeBestMove(depth);
    CPPUNIT_ASSERT(problems[i].acceptable(best_move));
  }
}

void MtdfTest::testProblems1()
{
  osl::SearchRecorder recorder("MtdfTest.testProblems500");
  recorder.setLogMargin(100);
  testProblems(recorder, problems1, numProblems1, 500);
}
void MtdfTest::testProblems2()
{
  osl::SearchRecorder recorder("MtdfTest.testProblems1000");
  recorder.setLogMargin(100);
  testProblems(recorder, problems1, numProblems1, 1000);
}
void MtdfTest::testProblems3()
{
  osl::SearchRecorder recorder("MtdfTest.testProblems1500");
  recorder.setLogMargin(100);
  testProblems(recorder, problems1, numProblems1, 1500);
  recorder.setLogMargin(100);
  testProblems(recorder, problems3, numProblems3, 2000);
}


void MtdfTest::testWinValue()
{
  CPPUNIT_ASSERT(Mtdf::isWinValue(BLACK, Mtdf::winByCheckmate(BLACK)));
  CPPUNIT_ASSERT(Mtdf::isWinValue(WHITE, Mtdf::winByCheckmate(WHITE)));
  CPPUNIT_ASSERT(! Mtdf::isWinValue(BLACK, Mtdf::winByCheckmate(WHITE)));
  CPPUNIT_ASSERT(! Mtdf::isWinValue(WHITE, Mtdf::winByCheckmate(BLACK)));
}

/* ------------------------------------------------------------------------- */
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
