#include "osl/search/quiescenceGenerator.h"
#include "osl/state/numEffectState.h"
#include "osl/record/csaString.h"

#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>

using namespace osl;
using namespace osl::search;

class QuiescenceGeneratorTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE(QuiescenceGeneratorTest);
  CPPUNIT_TEST(testOpenCheck);
  CPPUNIT_TEST(testCheck);
  CPPUNIT_TEST(testEscape);
  CPPUNIT_TEST(testEscapeFromLastMove);
  CPPUNIT_TEST(testEscapeByMoveOnly);
  CPPUNIT_TEST(testPromote);
  CPPUNIT_TEST(testEscapeKing);
  CPPUNIT_TEST(testEscapeKingInTakeBack);
  CPPUNIT_TEST(testAdvanceBishop);
  CPPUNIT_TEST(testAttackKing8);
  CPPUNIT_TEST(testAttackMajorPiece);
  CPPUNIT_TEST(testAttackGoldWithPawn);
  CPPUNIT_TEST(testAddLance);
  CPPUNIT_TEST(testUtilizePromoted);
  CPPUNIT_TEST(testBreakThreatmate);
  CPPUNIT_TEST_SUITE_END();
public:
  void testPromote();
  void testEscape();
  void testEscapeFromLastMove();
  void testEscapeKing();
  void testEscapeKingInTakeBack();
  void testEscapeByMoveOnly();
  void testOpenCheck();
  void testCheck();
  void testAdvanceBishop();
  void testAttackKing8();
  void testAttackMajorPiece();
  void testAttackGoldWithPawn();
  void testAddLance();
  void testUtilizePromoted();
  void testBreakThreatmate();
};

CPPUNIT_TEST_SUITE_REGISTRATION(QuiescenceGeneratorTest);

void QuiescenceGeneratorTest::testPromote()
{
  {
    NumEffectState state(CsaString(
			   "P1+HI *  *  *  *  * -KI-KE-OU\n"
			   "P2 *  *  *  *  *  * -KI-GI-KY\n"
			   "P3 *  *  *  *  *  * -FU-FU-FU\n"
			   "P4 *  *  *  *  *  *  *  *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 *  *  *  *  *  *  *  *  * \n"
			   "P7-FU-GI * +GI *  *  *  *  * \n"
			   "P8 * -FU *  *  *  *  *  *  * \n"
			   "P9+OU *  *  *  *  *  *  *  * \n"
			   "P+00AL\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::promote(state, moves);
    // 98check ϾʤΤpromoteƤʤɤޤʤ
    CPPUNIT_ASSERT_EQUAL((size_t)3, moves.size());     
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY *  *  *  *  *  *  * -KY\n"
			   "P2 * -HI *  *  * -GI-KI-OU * \n"
			   "P3 *  * -KE * -KA-KI-KE-FU+FU\n"
			   "P4 * -GI-FU-FU-FU-FU *  *  * \n"
			   "P5-FU-FU *  *  *  * -FU+FU+HI\n"
			   "P6 *  * +FU+FU+FU+FU *  *  * \n"
			   "P7+FU+FU+GI+KI+GI * +KE *  * \n"
			   "P8 * +OU+KI *  *  *  *  *  * \n"
			   "P9+KY+KE *  *  *  *  *  * +KY\n"
			   "P+00FU\n"
			   "P-00KA00FU\n"
			   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::promote(state, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size()); // 12TO
  }
}

void QuiescenceGeneratorTest::testEscapeFromLastMove()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  * -OU * -KE-KY\n"
			   "P2 * -HI *  *  *  * -KI *  * \n"
			   "P3-FU *  * -FU * -KI-GI-FU-FU\n"
			   "P4 *  *  *  * -FU-FU-FU *  * \n"
			   "P5 *  *  *  *  *  *  * +FU * \n"
			   "P6 * -KA * +FU+FU+FU+FU *  * \n"
			   "P7+FU+FU * +KI+KA+GI+KE * +FU\n"
			   "P8 *  * +KI *  *  *  * +HI * \n"
			   "P9+KY+KE * +OU *  *  *  * +KY\n"
			   "P+00GI00FU\n"
			   "P-00GI00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    Move last_move(Position(8,7),PAWN,BLACK);
    QuiescenceGenerator<WHITE>::escapeFromLastMove(state, last_move, moves);
    // ƨ
    CPPUNIT_ASSERT_EQUAL((size_t)6, moves.size());     
  }
  {
    NumEffectState state(CsaString(
			   "P1+TO *  *  *  * -OU * -KE-KY\n"
			   "P2 *  *  *  *  *  * -KI-KA * \n"
			   "P3-FU * -KE-FU * -KI-GI-FU-FU\n"
			   "P4 *  * -FU-GI * -FU-FU *  * \n"
			   "P5 * -HI *  *  *  *  * +FU * \n"
			   "P6 * +KY+FU+FU * +FU+FU *  * \n"
			   "P7+FU * +GI+KI * +GI *  * +FU\n"
			   "P8 *  * +OU+KA *  *  * +HI * \n"
			   "P9+KY+KE *  *  *  *  * +KE+KY\n"
			   "P+00FU\n"
			   "P-00KI00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    Move last_move(Position(8,6),LANCE,BLACK);
    QuiescenceGenerator<WHITE>::escapeFromLastMove(state, last_move, moves);
    // ƨ
    CPPUNIT_ASSERT_EQUAL((size_t)2, moves.size()); // 95,55
  }
  {
    NumEffectState state(CsaString(
           "P1-KY-KE * -HI * -OU *  * -KY\n"
           "P2 *  *  *  * -KI-GI-KI *  * \n"
           "P3-FU * -FU *  * -FU-KE-FU-FU\n"
           "P4 * -FU *  * -HI *  *  *  * \n"
           "P5 *  *  *  * +KA *  *  *  * \n"
           "P6 *  * +FU+FU *  *  *  *  * \n"
           "P7+FU+FU+GI+KI * +FU * +FU+FU\n"
           "P8 *  * +KI *  * +GI *  *  * \n"
           "P9+KY+KE *  * +OU *  * +KE+KY\n"
           "P+00KA00GI00FU\n"
           "P-00FU00FU00FU00FU\n"
	   "+\n").getInitialState());
    MoveVector moves;
    const Move last_move(Position(5,4),ROOK,WHITE);
    QuiescenceGenerator<BLACK>::escapeFromLastMove(state, last_move, moves);
    const Move m56fu(Position(5,6),PAWN,BLACK);
    CPPUNIT_ASSERT(moves.isMember(m56fu)); 
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  * -KI * -KE-KY\n"
			   "P2 *  * -HI * -KI * -GI+KA * \n"
			   "P3-FU *  * -FU * -FU *  *  * \n"
			   "P4 *  *  *  * -FU * -FU-OU-FU\n"
			   "P5 * -FU-GI+FU *  *  * -FU * \n"
			   "P6 *  *  *  *  * +FU+FU * +FU\n"
			   "P7+FU+FU+GI * +FU+KI+KE *  * \n"
			   "P8 *  *  * +HI *  * +GI *  * \n"
			   "P9+KY+KE *  *  * +KI+OU * +KY\n"
			   "P-00KA00FU00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Move last_move(Position(2,2),BISHOP,BLACK);
    QuiescenceGenerator<WHITE>::escapeFromLastMove(state, last_move, moves);
    const Move m33ka(Position(3,3),BISHOP,WHITE);
    CPPUNIT_ASSERT(moves.isMember(m33ka)); 
  }
}

void QuiescenceGeneratorTest::testEscapeKingInTakeBack()
{
  {
    NumEffectState state(CsaString(
			   "P1 *  *  *  *  *  * -KI-KE-OU\n"
			   "P2 *  *  *  *  *  * -KI-GI-KY\n"
			   "P3 *  *  *  *  *  * -FU-FU-FU\n"
			   "P4 *  *  *  *  *  *  *  *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 *  *  *  *  *  *  *  *  * \n"
			   "P7 * -GI *  *  *  *  *  *  * \n"
			   "P8-TO-FU *  *  *  *  *  *  * \n"
			   "P9+OU *  *  *  *  *  *  *  * \n"
			   "P+00AL\n"
			   "+\n").getInitialState());
    MoveVector moves;
    const bool has_safe_move 
      = QuiescenceGenerator<BLACK>::escapeKingInTakeBack(state, moves, false);
    CPPUNIT_ASSERT_EQUAL((size_t)0, moves.size());     
    CPPUNIT_ASSERT_EQUAL(false, has_safe_move);     
  }
}

void QuiescenceGeneratorTest::testAdvanceBishop()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY *  * -KI *  *  *  *  * \n"
			   "P2 * -OU-GI *  * -KA-HI * -KY\n"
			   "P3 * -FU * -KI-KE-GI * -FU * \n"
			   "P4-FU * -FU-KE-FU * -FU * -FU\n"
			   "P5 *  *  * -FU * +GI * +FU * \n"
			   "P6+FU+FU+FU * +FU+FU+FU *  * \n"
			   "P7 * +GI+KI+FU+KA *  *  * +FU\n"
			   "P8 *  * +KI *  * +KE * +HI * \n"
			   "P9+KY+OU *  *  *  *  *  * +KY\n"
			   "P+00KE00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::advanceBishop(state, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)2, moves.size());     // 15Ѥˤ
  }
}

void QuiescenceGeneratorTest::testOpenCheck()
{
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI * -KI-GI-KE-KY\n"
		    "P2 *  *  *  * -FU *  * -KA * \n"
		    "P3-FU-FU-FU-FU-OU-FU-FU-FU-FU\n"
		    "P4 *  *  *  *  *  *  *  *  * \n"
		    "P5 *  *  * +KY *  * -HI *  * \n"
		    "P6 *  *  * +GI+KI+KE *  *  * \n"
		    "P7+FU+FU+FU+FU+HI+FU+FU+FU+FU\n"
		    "P8 * +KA *  * +FU *  *  *  * \n"
		    "P9+KY+KE+GI+KI+OU *  *  *  * \n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::check(state, moves);
    // 츫Ǥ̤ƨƤ֤֤Ȥ
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size()); // +5645KI
  }
}

void QuiescenceGeneratorTest::testCheck()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE * -KI *  *  *  * +RY\n"
			   "P2 * -OU *  * -GI *  * +NG-KY\n"
			   "P3 * -FU *  *  *  *  * -FU * \n"
			   "P4-FU *  *  * -FU * -FU * -FU\n"
			   "P5 * +FU *  *  *  *  * +FU * \n"
			   "P6+FU * -FU+KE+FU * +FU *  * \n"
			   "P7 *  *  * +KI *  *  * +KE+FU\n"
			   "P8 *  *  * +KI * -UM *  *  * \n"
			   "P9+KY+OU+KI *  *  *  *  * -RY\n"
			   "P+00GI00KY\n"
			   "P-00KA00GI00FU00FU00FU00FU00FU00KE\n"
			   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::check(state, moves);
    // 74ˤϻؤƤ
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size());
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  *  *  *  * -KY\n"
			   "P2 * -HI * -GI * -OU *  *  * \n"
			   "P3-FU * -FU-FU * -KI *  * -FU\n"
			   "P4 * -FU *  * -FU * -FU-FU * \n"
			   "P5 *  *  *  *  * +HI *  *  * \n"
			   "P6+FU * +FU *  *  * +FU+FU * \n"
			   "P7 * +FU * +FU+FU *  *  * +FU\n"
			   "P8 * +KA+KI+GI *  *  *  *  * \n"
			   "P9+KY+KE * +OU * +KI *  * -UM\n"
			   "P+00KI00KE\n"
			   "P-00GI00GI00KE00KY00FU00FU\n"
			   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::check(state, moves);
    // 43ФsendoffpositionΤǡ
    // 33Τ褦¿˽ʲɤ
    CPPUNIT_ASSERT_EQUAL((size_t)3, moves.size());
  }
  {
    NumEffectState state(CsaString(
			   "P1-OU *  *  *  * -HI * -KE-KY\n"
			   "P2-KY-KI-KI *  *  *  *  *  * \n"
			   "P3 * -GI *  * -FU *  *  * -FU\n"
			   "P4 * -FU-FU *  * -KA+FU-FU * \n"
			   "P5-FU+FU *  *  * -FU *  * +FU\n"
			   "P6 *  * +FU *  *  *  *  *  * \n"
			   "P7+FU *  * +KI+FU+FU+GI *  * \n"
			   "P8+KY+GI *  *  *  *  * +HI * \n"
			   "P9+OU+KE+KI *  *  *  *  * +KY\n"
			   "P+00KA00GI00KE00FU00FU\n"
			   "P-00KE00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::check(state, moves); // 87
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size());
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY *  *  *  *  *  *  *  * \n"
			   "P2-OU-KI+RY *  *  *  *  *  * \n"
			   "P3-FU-FU *  *  *  *  *  *  * \n"
			   "P4 *  *  *  *  *  *  *  *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 *  *  *  *  *  *  *  *  * \n"
			   "P7 *  *  *  *  *  * +FU+FU+FU\n"
			   "P8 *  *  *  *  *  * +KI+GI+KY\n"
			   "P9 *  *  *  *  *  * +KI+KE+OU\n"
			   "P+00GI\n"
			   "P-00AL\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::check(state, moves, true); // 81GI
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size());
  }
}

void QuiescenceGeneratorTest::testEscape()
{
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
		    "P2 * -HI *  *  *  *  * -KA * \n"
		    "P3-FU-FU-FU-FU * -FU-FU-FU-FU\n"
		    "P4 *  *  *  * -FU *  *  *  * \n"
		    "P5 *  *  *  * +KA *  *  *  * \n"
		    "P6 *  *  *  *  *  *  *  *  * \n"
		    "P7+FU+FU+FU+FU+FU+FU+FU+FU+FU\n"
		    "P8 *  *  *  *  *  *  * +HI * \n"
		    "P9+KY+KE+GI+KI+OU+KI+GI+KE+KY\n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeAll(state, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)2, moves.size()); // 46, 66
  }
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI-OU-KI-GI-KE * \n"
		    "P2 * -HI *  *  *  *  * -KA * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  *  * -KY *  *  *  * \n"
		    "P5 *  *  *  *  *  *  *  *  * \n"
		    "P6 *  *  * +FU * +FU *  *  * \n"
		    "P7+FU+FU+FU * +KA * +FU+FU+FU\n"
		    "P8 *  *  *  * +KY *  * +HI * \n"
		    "P9+KY+KE+GI+KI+OU+KI+GI+KE * \n"
		    "P+00FU\n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeAll(state, moves);
    // ̵̤ʹ򤷤ʤ 48, 68
    CPPUNIT_ASSERT_EQUAL((size_t)2, moves.size());
  }
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI-OU-KI-GI-KE * \n"
		    "P2 * -HI *  *  *  *  * -KA * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  *  * -KY *  *  *  * \n"
		    "P5 *  *  *  *  *  *  *  *  * \n"
		    "P6 *  *  * +FU * +FU *  *  * \n"
		    "P7+FU+FU+FU+GI+KA * +FU+FU+FU\n"
		    "P8 *  *  *  * +KY *  * +HI * \n"
		    "P9+KY+KE+KI * +OU+KI+GI+KE * \n"
		    "P+00FU\n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeAll(state, moves);
    // ̵̤ʹ򤷤ʤ 48, 68, +56FU
    CPPUNIT_ASSERT_EQUAL((size_t)3, moves.size());
  }
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
		    "P2 *  *  *  *  *  *  * -KA * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  *  *  *  *  *  *  * \n"
		    "P5-HI *  *  *  *  *  *  *  * \n"
		    "P6 *  *  *  *  *  *  *  *  * \n"
		    "P7 * +FU+FU+FU+FU+FU+FU+FU+FU\n"
		    "P8 * +KY+KE *  *  *  * +HI * \n"
		    "P9+KA * +KI+GI+OU+KI+GI+KE+KY\n"
		    "P-00FU\n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeAll(state, moves);
    // ƨʤϤĤ
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size()); // +7989KI
  }
  {
    NumEffectState state(CsaString(
           "P1-KY-KE * -HI * -OU *  * -KY\n"
           "P2 *  *  *  * -KI-GI-KI *  * \n"
           "P3-FU * -FU *  * -FU-KE-FU-FU\n"
           "P4 * -FU *  * -HI *  *  *  * \n"
           "P5 *  *  *  * +KA *  *  *  * \n"
           "P6 *  * +FU+FU *  *  *  *  * \n"
           "P7+FU+FU+GI+KI * +FU * +FU+FU\n"
           "P8 *  * +KI *  * +GI *  *  * \n"
           "P9+KY+KE *  * +OU *  * +KE+KY\n"
           "P+00KA00GI00FU\n"
           "P-00FU00FU00FU00FU\n"
	   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeAll(state, moves);
    const Move m56fu(Position(5,6),PAWN,BLACK);
    CPPUNIT_ASSERT(moves.isMember(m56fu)); 
  }
}

void QuiescenceGeneratorTest::testEscapeKing()
{
  {
    NumEffectState state(CsaString(
		    "P1-KY-KE-GI-KI-OU-KI-GI-KE-KY\n"
		    "P2 *  *  *  *  *  *  * -KA * \n"
		    "P3-FU-FU-FU-FU-FU-FU-FU-FU-FU\n"
		    "P4 *  *  *  *  *  *  *  *  * \n"
		    "P5 * -HI *  *  *  *  * +OU * \n"
		    "P6 *  *  *  *  * +FU *  *  * \n"
		    "P7+FU+FU+FU+FU+FU * +FU+FU+FU\n"
		    "P8 * +KA *  *  *  *  * +HI * \n"
		    "P9 * +KE+GI+KI * +KI+GI+KE+KY\n"
		    "P+00KY\n"
		    "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::escapeKing(state, moves);
      // ػ: king 36,26,16, lance 35, 45
    CPPUNIT_ASSERT_EQUAL((size_t)5, moves.size()); 
  }
}

void QuiescenceGeneratorTest::testAttackKing8()
{
  {
    NumEffectState state(CsaString(
			   "P1+UM *  *  *  *  * -OU * -KY\n"
			   "P2 *  * -HI-KI * -GI-KI *  * \n"
			   "P3 *  *  * -FU-GI-FU *  * -KE\n"
			   "P4-FU *  *  * -FU * -FU+HI-FU\n"
			   "P5 * +KE * +FU * +FU *  *  * \n"
			   "P6+FU *  * +KI+FU * +GI * +FU\n"
			   "P7+OU+FU-TO *  *  * +KE *  * \n"
			   "P8 *  *  *  *  *  *  *  *  * \n"
			   "P9+KY *  *  *  *  *  *  * +KY\n"
			   "P+00KA00GI00KE00KY00FU00FU\n"
			   "P-00KI00FU00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::attackKing8(state, moves);
    // 8688⡤95⤯餤?
    // ˤäƤ76TO Ȥ 76KI Ȥ⤷礦ʤ?
    CPPUNIT_ASSERT_EQUAL((size_t)3, moves.size()); 
  }
}

void QuiescenceGeneratorTest::testAttackMajorPiece()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  *  * -OU * -KY\n"
			   "P2-HI *  *  *  *  * -KI *  * \n"
			   "P3-FU *  * -FU * -KI * -FU-FU\n"
			   "P4 * -FU *  *  *  * +FU *  * \n"
			   "P5 *  * +GI+FU+FU-KE+KA+FU-KA\n"
			   "P6 *  *  * +KI *  *  *  *  * \n"
			   "P7+FU+FU * +KI *  *  *  * +FU\n"
			   "P8 *  *  *  *  * +GI+HI *  * \n"
			   "P9+KY+KE * +OU *  *  * +KE+KY\n"
			   "P+00GI00FU00FU00FU00FU00FU\n"
			   "P-00GI00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::attackMajorPiece(state, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)6, moves.size());
// Drop(-,PAWN,Position(36)) 뤫ɤ̯
// Drop(-,PAWN,Position(37))
// Drop(-,SILVER,Position(27))
// Drop(-,SILVER,Position(49))
// Drop(-,SILVER,Position(44))
// Drop(-,SILVER,Position(26))
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE+KA *  *  *  *  *  * \n"
			   "P2-OU * -GI *  * +TO *  *  * \n"
			   "P3 * -FU-FU-KI *  * -KE * -FU\n"
			   "P4-FU *  *  *  *  *  *  *  * \n"
			   "P5 *  *  * -KI * -FU *  *  * \n"
			   "P6+FU * +FU-FU-FU *  *  * +FU\n"
			   "P7 * +FU+KI *  * +FU+GI-FU * \n"
			   "P8 *  * +OU *  *  *  *  * +HI\n"
			   "P9+KY+KE * +KI+RY *  *  * +KY\n"
			   "P+00KY00FU00FU\n"
			   "P-00KA00GI00GI00KE00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<WHITE>::attackMajorPiece(state, moves);
    const Move m62gi(Position(6,2), SILVER, WHITE);
    CPPUNIT_ASSERT(moves.isMember(m62gi));
  }
  {
    NumEffectState state(CsaString(
			   "P1-OU-KE *  *  *  *  *  * -KY\n"
			   "P2-KY-GI-GI * -KI *  *  * -HI\n"
			   "P3-FU * -FU * -FU *  * -FU * \n"
			   "P4 * +FU *  *  * +UM *  * -FU\n"
			   "P5 *  *  * -FU *  *  *  *  * \n"
			   "P6+FU * +FU *  *  *  * +FU * \n"
			   "P7 * +GI+GI+FU * -TO+FU * +FU\n"
			   "P8 * +OU+KI+KI *  *  *  *  * \n"
			   "P9+KY+KE+HI *  *  *  * +KE-UM\n"
			   "P+00KI00FU00FU00FU\n"
			   "P-00KE00KY00FU\n"
			   "+\n").getInitialState());
    // ƨ
    MoveVector moves;
    QuiescenceGenerator<BLACK>::attackMajorPiece(state, moves);
    CPPUNIT_ASSERT(moves.empty());
  }
  {
    NumEffectState state(CsaString(
			   "P1-OU-KE *  *  *  * +UM * -KY\n"
			   "P2-KY-GI-GI * -KI *  *  * -HI\n"
			   "P3-FU * -FU * -FU *  * -FU-FU\n"
			   "P4 * +FU *  *  *  *  *  *  * \n"
			   "P5 *  *  * -FU *  *  *  *  * \n"
			   "P6+FU * +FU *  *  *  * +FU * \n"
			   "P7 * +GI+GI+FU * -TO+FU * +FU\n"
			   "P8 * +OU+KI+KI *  *  *  *  * \n"
			   "P9+KY+KE+HI *  *  *  * +KE-UM\n"
			   "P+00KI00FU00FU00FU\n"
			   "P-00KE00KY00FU\n"
			   "+\n").getInitialState());
    // α
    MoveVector moves;
    QuiescenceGenerator<BLACK>::attackMajorPiece(state, moves);
    const Move m22ki(Position(2,2), GOLD, BLACK);
    CPPUNIT_ASSERT(moves.isMember(m22ki));
  }
}

void QuiescenceGeneratorTest::testAttackGoldWithPawn()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  * -OU *  * -KY\n"
			   "P2 * -HI * -GI * -KA-KI *  * \n"
			   "P3-FU * -FU-FU * -KI *  * -FU\n"
			   "P4 * -FU *  * -FU * -FU-FU * \n"
			   "P5 *  *  *  *  * -KE *  *  * \n"
			   "P6 *  * +FU *  *  * +FU+FU * \n"
			   "P7+FU+FU * +FU+FU *  *  * +FU\n"
			   "P8 * +KA+KI+GI * +HI *  *  * \n"
			   "P9+KY+KE * +OU * +KI *  * +KY\n"
			   "P+00GI00FU\n"
			   "P-00GI00KE00FU\n"
			   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::attackGoldWithPawn(state, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size());
  }
}

void QuiescenceGeneratorTest::testAddLance()
{
  {
    NumEffectState state(CsaString(
			   "P1 *  *  *  *  *  * -GI-OU-KY\n"
			   "P2 *  *  *  *  * -KA-KI-GI * \n"
			   "P3 *  *  *  *  *  * -KE-FU-FU\n"
			   "P4+RY * +KI * -FU-FU+FU *  * \n"
			   "P5 *  *  * -FU *  *  *  * +FU\n"
			   "P6 * +FU *  * +FU+FU *  *  * \n"
			   "P7+KE * -UM * +GI+GI *  *  * \n"
			   "P8 *  *  *  *  * +KI+OU *  * \n"
			   "P9+KY *  *  *  *  *  * +HI+KY\n"
			   "P+00KE00KE00KY00FU00FU00FU\n"
			   "P-00KI00FU00FU00FU00FU00FU\n"
			   "+\n").getInitialState());
    MoveVector moves;
    QuiescenceGenerator<BLACK>::capture<KNIGHT,false>(state, moves,Piece::EMPTY());
    // 33to
    // 35,36,37kyo
    CPPUNIT_ASSERT_EQUAL((size_t)4, moves.size());
  }
}

void QuiescenceGeneratorTest::testUtilizePromoted()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  *  * -KI-KE-OU\n"
			   "P2 *  *  *  * -HI *  * -GI-KY\n"
			   "P3-FU * +NK-FU * -FU-KA-FU-FU\n"
			   "P4 * -FU * -KI-FU * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6+FU * +HI+FU *  *  *  *  * \n"
			   "P7 * +FU * +GI+FU+FU+FU+FU+FU\n"
			   "P8 *  *  *  * +KI * +OU *  * \n"
			   "P9+KY *  *  *  * +KI+GI+KE+KY\n"
			   "P+00GI00FU00FU\n"
			   "P-00KA\n"
			   "+\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(7,3));
    QuiescenceGenerator<BLACK>::utilizePromoted(state, target, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)5, moves.size());
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  *  *  * -KI-KE-OU\n"
			   "P2 *  *  *  * -HI *  * -GI-KY\n"
			   "P3-FU * +NK-FU * -FU-KA-FU-FU\n"
			   "P4 * -FU * -KI-FU * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6+FU+HI * +FU *  *  *  *  * \n"
			   "P7 * +FU * +GI+FU+FU+FU+FU+FU\n"
			   "P8 *  *  *  * +KI * +OU *  * \n"
			   "P9+KY *  *  *  * +KI+GI+KE+KY\n"
			   "P+00GI00FU00FU\n"
			   "P-00KA\n"
			   "+\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(7,3));
    QuiescenceGenerator<BLACK>::utilizePromoted(state, target, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)1, moves.size());
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY *  *  *  *  * -KI-KE-OU\n"
			   "P2 *  *  *  * -HI *  * -GI-KY\n"
			   "P3-FU-KE+NK-FU * -FU-KA-FU-FU\n"
			   "P4 * -FU * -KI-FU * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6+FU * +HI+FU *  *  *  *  * \n"
			   "P7 * +FU * +GI+FU+FU+FU+FU+FU\n"
			   "P8 *  *  *  * +KI * +OU *  * \n"
			   "P9+KY *  *  *  * +KI+GI+KE+KY\n"
			   "P+00GI00FU00FU\n"
			   "P-00KA\n"
			   "+\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(7,3));
    QuiescenceGenerator<BLACK>::utilizePromoted(state, target, moves);
    CPPUNIT_ASSERT_EQUAL((size_t)2, moves.size());
  }
}

void QuiescenceGeneratorTest::testEscapeByMoveOnly()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  * -OU-KI * -KE-KA\n"
			   "P2 *  *  * -GI-KI * -GI *  * \n"
			   "P3-FU * -FU-FU * -FU * -FU-FU\n"
			   "P4 *  *  *  * -FU * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 * -HI+FU+FU *  *  *  *  * \n"
			   "P7+FU+KY * +KI+FU+FU+FU+FU+FU\n"
			   "P8 * +GI+KI *  *  *  * +HI * \n"
			   "P9+KY+KE * +OU *  * +GI+KE+KY\n"
			   "P+00FU\n"
			   "P-00KA00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(8,6));
    const bool has_safe_move
      = QuiescenceGenerator<WHITE>::escapeByMoveOnly(state, target, moves);
    // ˲äƤ
    CPPUNIT_ASSERT(! has_safe_move);
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE *  * -OU-KI * -KE-KA\n"
			   "P2 *  *  * -GI-KI * -GI *  * \n"
			   "P3-FU * -FU-FU * -FU * -FU-FU\n"
			   "P4 *  *  *  * -FU * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 * -HI+FU+FU *  *  *  *  * \n"
			   "P7+FU+FU * +KI+FU+FU+FU+FU+FU\n"
			   "P8 * +GI+KI *  *  *  * +HI * \n"
			   "P9+KY+KE * +OU *  * +GI+KE+KY\n"
			   "P+00KY\n"
			   "P-00KA00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(8,6));
    const bool has_safe_move
      = QuiescenceGenerator<WHITE>::escapeByMoveOnly(state, target, moves);
    // ˲
    CPPUNIT_ASSERT(has_safe_move);
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE * -KI *  * -KI-KE-OU\n"
			   "P2 *  *  * -GI *  *  * -GI-KY\n"
			   "P3-FU-HI-FU-FU-FU-FU-UM-FU-FU\n"
			   "P4 * +FU *  *  *  * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 *  * +FU * +FU *  *  *  * \n"
			   "P7+FU * +KE+FU * +FU+FU+FU+FU\n"
			   "P8 * +HI * +GI+KI * +GI+OU * \n"
			   "P9+KY *  *  *  * +KI * +KE+KY\n"
			   "P+00FU\n"
			   "P-00KA\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(8,3));
    const bool has_safe_move
      = QuiescenceGenerator<WHITE>::escapeByMoveOnly(state, target, moves);
    // ˲äƤɷ⤵
    CPPUNIT_ASSERT(! has_safe_move);
  }
  {
    NumEffectState state(CsaString(
			   "P1-KY-KE * -KI *  * -KI-KE-OU\n"
			   "P2 *  *  * -GI *  *  * -GI-KY\n"
			   "P3-FU-HI *  * -FU-FU-UM-FU-FU\n"
			   "P4 * +FU-FU-FU *  * -FU *  * \n"
			   "P5 *  *  *  *  *  *  *  *  * \n"
			   "P6 *  * +FU * +FU *  *  *  * \n"
			   "P7+FU * +KE+FU * +FU+FU+FU+FU\n"
			   "P8 * +HI * +GI+KI * +GI+OU * \n"
			   "P9+KY *  *  *  * +KI * +KE+KY\n"
			   "P+00FU\n"
			   "P-00KA\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Piece target = state.getPieceOnBoard(Position(8,3));
    const bool has_safe_move
      = QuiescenceGenerator<WHITE>::escapeByMoveOnly(state, target, moves);
    // ƨ
    CPPUNIT_ASSERT(has_safe_move);
  }
}

void QuiescenceGeneratorTest::testBreakThreatmate()
{
  {
    NumEffectState state(CsaString(
			   "P1-KY-HI *  *  *  * -GI-OU-KY\n"
			   "P2 *  *  *  * +FU *  *  *  * \n"
			   "P3 *  *  *  * -KI *  * +KI * \n"
			   "P4 * -FU-FU * -GI * +KE+UM * \n"
			   "P5-FU *  * +FU-FU-KE * -FU-FU\n"
			   "P6 *  * +FU *  *  *  *  *  * \n"
			   "P7+FU+FU+KI * -TO *  *  *  * \n"
			   "P8 * +OU+KI *  *  * +FU * +HI\n"
			   "P9+KY+KE *  * -UM *  *  *  * \n"
			   "P+00KE\n"
			   "P-00GI00GI00KY00FU00FU00FU00FU00FU\n"
			   "-\n").getInitialState());
    MoveVector moves;
    const Move m33ke(Position(3,3), KNIGHT, BLACK);
    QuiescenceGenerator<WHITE>::breakThreatmate(state, m33ke, moves);
    CPPUNIT_ASSERT(! moves.empty());
  }
}


// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
