#include "../recursiveDoUndoMove.h"
#include "../doUndoMoveLockTest.h"
#include "osl/apply_move/applyMove.h"
#include "osl/record/csaRecord.h"
#include "osl/record/csaString.h"
#include "osl/state/simpleState.tcc"

#include "osl/container/pieceVector.h"
#include "osl/effect_action/storePiece.h"
#include "osl/apply_move/doUndoMoveLock.h"

#include <iomanip>
#include <fstream>
#include <vector>
#include <cppunit/TestCase.h>
#include <cppunit/extensions/HelperMacros.h>


class SimpleStateTest : public CppUnit::TestFixture 
{
  CPPUNIT_TEST_SUITE( SimpleStateTest );
  CPPUNIT_TEST(testEmulateCapture);
  CPPUNIT_TEST(testEmulateHandPiece);
  CPPUNIT_TEST_SUITE_END();
public:
  void testEmulateCapture();
  void testEmulateHandPiece();
};


CPPUNIT_TEST_SUITE_REGISTRATION(SimpleStateTest);
using namespace osl;
using namespace osl::state;

void SimpleStateTest::testEmulateCapture()
{
  SimpleState state(CsaString(
			      "P1+NY * +TO *  *  * -OU-KE-KY\n"
			      "P2 *  *  *  *  * -GI-KI *  * \n"
			      "P3 * +RY *  * +UM * -KI-FU-FU\n"
			      "P4 *  * +FU-FU *  * -FU *  * \n"
			      "P5 *  * -KE * +FU * +KY+FU * \n"
			      "P6-KE *  * +FU+GI-FU *  * +FU\n"
			      "P7 *  * -UM *  *  *  *  *  * \n"
			      "P8 *  *  *  *  *  *  *  *  * \n"
			      "P9 * +OU * -GI *  *  *  * -NG\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00KE\n"
			      "P+00KI\n"
			      "P-00KI\n"
			      "P-00KY\n"
			      "P+00HI\n"
			      "+\n"
			      ).getInitialState());
  {
    SimpleState state1=state.emulateCapture(state.getPieceAt(Position(5,3)),WHITE);
    CPPUNIT_ASSERT(state1.isConsistent());
    SimpleState state2(CsaString(
				 "P1+NY * +TO *  *  * -OU-KE-KY\n"
				 "P2 *  *  *  *  * -GI-KI *  * \n"
				 "P3 * +RY *  *  *  * -KI-FU-FU\n"
				 "P4 *  * +FU-FU *  * -FU *  * \n"
				 "P5 *  * -KE * +FU * +KY+FU * \n"
				 "P6-KE *  * +FU+GI-FU *  * +FU\n"
				 "P7 *  * -UM *  *  *  *  *  * \n"
				 "P8 *  *  *  *  *  *  *  *  * \n"
				 "P9 * +OU * -GI *  *  *  * -NG\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00KE\n"
				 "P+00KI\n"
				 "P-00KI\n"
				 "P-00KY\n"
				 "P-00KA\n"
				 "P+00HI\n"
				 "+\n"
				 ).getInitialState());

    CPPUNIT_ASSERT_EQUAL(state1,state2);
  }
}

void SimpleStateTest::testEmulateHandPiece()
{
  SimpleState state(CsaString(
			      "P1+NY * +TO *  *  * -OU-KE-KY\n"
			      "P2 *  *  *  *  * -GI-KI *  * \n"
			      "P3 * +RY *  * +UM * -KI-FU-FU\n"
			      "P4 *  * +FU-FU *  * -FU *  * \n"
			      "P5 *  * -KE * +FU * +KY+FU * \n"
			      "P6-KE *  * +FU+GI-FU *  * +FU\n"
			      "P7 *  * -UM *  *  *  *  *  * \n"
			      "P8 *  *  *  *  *  *  *  *  * \n"
			      "P9 * +OU * -GI *  *  *  * -NG\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00FU\n"
			      "P+00KE\n"
			      "P+00KI\n"
			      "P-00KI\n"
			      "P-00KY\n"
			      "P+00HI\n"
			      "+\n"
			      ).getInitialState());
  {
    SimpleState state1=state.emulateHandPiece(BLACK,WHITE,ROOK);
    CPPUNIT_ASSERT(state1.isConsistent());
    SimpleState state2(CsaString(
				 "P1+NY * +TO *  *  * -OU-KE-KY\n"
				 "P2 *  *  *  *  * -GI-KI *  * \n"
				 "P3 * +RY *  * +UM * -KI-FU-FU\n"
				 "P4 *  * +FU-FU *  * -FU *  * \n"
				 "P5 *  * -KE * +FU * +KY+FU * \n"
				 "P6-KE *  * +FU+GI-FU *  * +FU\n"
				 "P7 *  * -UM *  *  *  *  *  * \n"
				 "P8 *  *  *  *  *  *  *  *  * \n"
				 "P9 * +OU * -GI *  *  *  * -NG\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00FU\n"
				 "P+00KE\n"
				 "P+00KI\n"
				 "P-00KI\n"
				 "P-00KY\n"
				 "P-00HI\n"
				 "+\n"
				 ).getInitialState());

    CPPUNIT_ASSERT_EQUAL(state1,state2);
  }
}
// ;;; Local Variables:
// ;;; mode:c++
// ;;; c-basic-offset:2
// ;;; End:
